#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>

#include "dynamic-lard.h"

#define LOG2(x) log(x)/log(2.0)
#define DIRECTORY_MESSAGE_HDR_SIZE 128;

#define MESSAGE_DELIMITER " \0"
#define MESSAGE_HOLE "0x0"

DynamicLard::DynamicLard(   ServerToStringFunc *toString ) {
    depth   = 0;
    size    = 0;
    num_servers = 0;
    servers = NULL;

    this->toString      = toString;
    this->name  = NULL;

        // set up network structure
    dmessage.hdrsize = DIRECTORY_MESSAGE_HDR_SIZE;
    dmessage.msgsize = 512; 
    dmessage.msg     = NULL;
    equality_check   = NULL;
    refreshDirectoryMessage();
}

DynamicLard::~DynamicLard( ) {
    if ( servers )      NestFree( servers );
    if ( dmessage.msg ) NestFree( dmessage.msg );
    if ( this->name )   NestFree( this->name );
}

void
DynamicLard::setServerToString( ServerToStringFunc *toString ) {
    this->toString = toString;
    assert( this->toString );
}

void
DynamicLard::setServerEqualityFunction( serversAreSameFunction *fp ) {
    equality_check = fp;
}

void
DynamicLard::setName( const char *name ) {
    this->name = NestStrdup( name );
}

const char *
DynamicLard::getName( ) {
    return name;
}

int
DynamicLard::numServers( ) {
    return num_servers;
}

int
DynamicLard::getSize( ) {
    return size;
}

bool
DynamicLard::addHelper( void *server, void *helper ) {
    
    int serverIndex = findIndex( server ); 
    unsigned helperIndex; 
    int generation = 1;

        // first check for error
    if ( serverIndex < 0 ) return false;

    do {
        helperIndex = findDescendant( serverIndex, generation++ ); 
    } while( helperIndex < size && servers[helperIndex] != NULL );

    debug( "placing helper for %d at %d\n", serverIndex, helperIndex );
    addServerAtIndex( helper, helperIndex );
    return true;
}

void
DynamicLard::addServerAtIndex( void *server, unsigned index ) {
    while( index >= size ) grow();
    servers[index] = server;
    if ( server != NULL ) {
        num_servers++;
        refreshDirectoryMessage();
    }
}

void
DynamicLard::addServer( void *server ) {
        // first look for an open slot
    unsigned index;
    for( index = 0; index < size; index++ ) {
        if ( servers[index] == NULL ) {
            break;
        }
    }
    addServerAtIndex( server, index );
    debug( "new server 0x%x indexed at slot %d\n", server, index );
}

void 
DynamicLard::debug( char *fmt, ... )
{
	va_list args;
	va_start(args,fmt);

	if(getenv("LARD_DEBUG")) {
		//vfprintf(stderr,"extend_server debug: ",args);
		vfprintf(stderr,fmt,args);
		fflush(stderr);
	}

	va_end(args);
}

void
DynamicLard::dump( FILE * out, ServerToStringFunc *foreachserver ) {
    for( unsigned int i = 0; i < size; i++ ) {
        fprintf( out, "\t%d %s\n", 
                i, 
                ( servers[i] ? foreachserver( servers[i] ) : "0x0" ) );
    }
}

int
DynamicLard::findIndex( void *server ) {
        // just move through the tree until you find the server
    unsigned index;
    for( index = 0; index < size; index++ ) {
        if ( equality_check!=NULL && equality_check(server, servers[index]) ) {
            return index;
        } else if ( servers[index] == server ) {
            return index;
        }
    }
    return -1;  // not found
}

void
DynamicLard::grow( ) {

    unsigned newsize = ( size == 0 ) ? 1 : size * 2;
    void **newservers = (void **)NestCalloc((sizeof(void *)) * newsize);

    for( unsigned i = 0; i < size; i++ ) {
        newservers[i] = servers[i];
    }
    if ( servers != NULL ) NestFree( servers );
    servers = newservers;

    size = newsize;
    depth = (unsigned)ceil( LOG2( (double)size ) );
    debug( "grew to size %u, depth %u.\n", size, depth );
}

int
DynamicLard::findAncestor( int index ) {

    int mask_bits = 0;
    int mask, next_pow;

    next_pow = (int)pow( 2.0, mask_bits );
    do {
        mask = next_pow - 1;
        next_pow *= 2;
        mask_bits++;
    } while( next_pow - 1 < index );
    
    int ancestor = mask & index;
    debug( "Index %d & %d yields ancestor %d.\n", index, mask, ancestor );
    return ancestor;
}

int
DynamicLard::findDescendant( int index, int generation ) {
    int mask;

        // find the first mask above this index
    mask = 1;
    while( mask < index ) {
        mask *= 2;
    }
        // now skip generations
    for( int i = 1; i < generation; i++ ) {
        mask *= 2;
    }
        // here it is
    return mask | index;
}

bool
DynamicLard::replaceServer( void *server, void *replacement ) {
	int index = findIndex( server );
	if ( index >= 0 ) {
		servers[index] = NULL;
        num_servers--;
        addServerAtIndex( replacement, index );
		debug( "Replaced 0x%x with 0x%x\n", server, replacement );
		return true;
	} else {
		debug( "Unable to replace 0x%x with 0x%x\n", server, replacement );
		return false;
	}
}

bool
DynamicLard::removeServer( void *server, bool clean ) {

    int index = findIndex( server );
    if ( index < 0 ) {
        debug( "Tried to remove 0x%x but couldn't find.\n", server );
        return false;
    }

        // remove this server from the directory
    server = servers[index];
    if ( clean ) NestFree( server );
    servers[index] = NULL;
    num_servers--;
    assert( num_servers >= 0 );

        // special case zero.  we always need a root of the tree
    if( index == 0 && num_servers > 0 ) {
        index = size - 1;
        while( servers[index] == NULL ) index--;
        debug( "Lost server at index 0, replacing with %d.\n", index );
        assert( index > 0 );
        assert( servers[index] );
        servers[0] = servers[index];
        servers[index] = NULL;
    }

    refreshDirectoryMessage();
    return true;
}

void *
DynamicLard::getServer( const char *file ) {

    if ( size == 0 || num_servers == 0 ) {
        debug( "Empty lard, returning NULL\n" );
        return NULL;
    }
    
    int index = 0;

    unsigned hash = (unsigned)hash_one_at_a_time( file, strlen( file ) ); 
	debug( "Current depth: %d, %s hashes 0x%x", depth, file, hash );
	for( int i = 1; i <= 5; i++ ) {
		debug( "(%d bits %d),", i, least_significant_bits( hash, i ) );
	}
	debug( "\n" );

        // look for ancestor if no server at this index
    index = least_significant_bits( hash, depth );
    while( servers[index] == NULL && index > 0 ) {
        index = findAncestor( index );
    }

    /*
        // we will always promote someone to index 0, so we should never
        // fail to find an ancestor
        // if no ancestor, just find the lowest server to absorb the load 
    if ( servers[index] == NULL ) {
        debug( "Could not find an ancestor, will just find anyone.\n" );
        index = size - 1;
        while( servers[index] == NULL ) index--;
    }
    */
    
        // if we make it here, we'd better have someone
    assert( servers[index] );
    debug( "Found server (index %d, 0x%x) for %s (hash 0x%x)\n", 
                index, servers[index], file, hash );
    return servers[index];
}

ssize_t
writen( int fd, const void *vptr, ssize_t n ) {
    ssize_t nleft;
    ssize_t nwritten;
    const char *ptr;

    ptr = (char *)vptr;
    nleft = n;

    while ( nleft > 0 ) {
        if ( ( nwritten = write( fd, ptr, nleft ) ) < 0 ) {
            if ( errno == EINTR ) continue;
            else return -1;
        }
        nleft -= nwritten;
        ptr += nwritten;
    }

    return n;
}


// returns -1 on error
ssize_t
readn( int fd, void *vptr, ssize_t n ) {
    ssize_t nleft;
    ssize_t nread;
    char *ptr;

    ptr = (char *) vptr;
    nleft = n;
    while( nleft > 0 ) {
        if ( ( nread = read( fd, ptr, nleft ) ) < 0 ) {
            if ( errno == EINTR ) continue;
            return -1;
        } else if ( nread == 0 ) {
            break;        // EOF
        }
        nleft -= nread;
        ptr   += nread;
    }

    return( n - nleft );
}


int
DynamicLard::receive( int sd, StringToServerFunc *stringToServer, void *args ) {

        // first read the total size and scan out remaining size and name
    char buffer[dmessage.hdrsize];
    debug( "Receiving hrd of size %d...\n", dmessage.hdrsize );
    ssize_t n = readn( sd, buffer, dmessage.hdrsize );
    if ( n != dmessage.hdrsize ) {
        debug( "Readn from %d of %d got %d: %s\n",
                sd, dmessage.hdrsize, n, strerror( errno ) );
        return -1;
    }
    buffer[dmessage.hdrsize-1] = '\0';

        // set up the name array and scan into it
        // make it one extra byte big for safety
    if ( name == NULL ) {
        name = (char*)NestCalloc( dmessage.hdrsize + 1 );
    } else {
        NestMemset( name, 0, dmessage.hdrsize + 1);
    }
    int msgsize;
    if ( sscanf( buffer, "%d %s", &msgsize, name ) != 2 ) {
        debug( "Received header %s, but couldn't parse out size...\n", buffer );
        return -1;
    }

    debug( "Recvd hrd of size %d: %s, now reading %d\n", 
            dmessage.hdrsize, buffer, msgsize );
    char *message = (char *)NestCalloc( msgsize + 1 );
    n = readn( sd, message, msgsize );
    if ( n != msgsize ) {
        debug( "Readn from %d of %d got %d: %s\n",
                sd, msgsize, n, strerror(errno));
        return -1;
    }
    message[msgsize] = '\0';

    debug( "Received directory message: %s\n", message );

        // now tokenize it and add each server to our index after wiping out
        // old index
    clear();    // TODO - this could leak.  
                // Need to pass a free message to receive?

    char *server = strtok( message, MESSAGE_DELIMITER );
    unsigned index = 0;
    while( server ) {
        if ( index >= size ) grow();
        debug( "Adding server %s at index %d.\n", server, index );
        void *s = ( strcmp( MESSAGE_HOLE, server ) 
                        ? stringToServer( server, args ) 
                        : NULL );
        addServerAtIndex( s, index );
        server = strtok( NULL, MESSAGE_DELIMITER );
        if ( ++index >= size ) grow();
    }
    return 1;
}

void
DynamicLard::clear( bool clean ) {
        // doesn't clean up completely.
    if ( clean ) {
        for( unsigned index = 0; index < size; index++ ) {
            if ( servers[index] ) NestFree( servers[index] ); 
        }
    }
    NestMemset( servers, 0, (sizeof(void *)) * size);
    num_servers = 0;
}

int
DynamicLard::send( int sd ) {
    if ( ! toString ) {
        fprintf(stderr,"Trying to send but did not supply a toString func!\n");
        return -1;
    }
    //debug( "Sending %d bytes to %d.\n", dmessage.offset, sd );
    return writen( sd, dmessage.msg, dmessage.offset );
}

bool
DynamicLard::refreshDirectoryMessage( bool grow ) {

        // if owner of this class didn't provide a toString func
        // upon creation, then they are not interested in sending over network
    if ( ! toString ) return false;
    
        // grow if necessary
    if( grow || dmessage.msg == NULL ) {

            // with high failure rates, proxys segfault on this free.
            // they seemed to be confused about the size of the directory
            // and were in this function recursively at a depth of 1.
            // so I've made it so only the coordinators maintain the directory
            // message.  Which certainly doesn't fix the bug, but maybe it
            // avoids it.
        if( dmessage.msg ) NestFree( dmessage.msg );

        dmessage.msgsize *= 2;
        dmessage.msg = (char*)NestCalloc(dmessage.msgsize);
        debug( "Growing directorystring to %d.\n", dmessage.msgsize );
    }
    dmessage.offset = dmessage.hdrsize;

        // write in each server name
    for( unsigned i = 0; i < size; i++ ) {
        const char *servername = (  servers[i] 
                                    ? toString( servers[i] ) 
                                    : MESSAGE_HOLE );
        NestSnprintf(   &(dmessage.msg[dmessage.offset]), 
                    dmessage.msgsize - dmessage.offset,
                    "%s%s", 
                    servername, MESSAGE_DELIMITER ); 
        dmessage.offset += strlen( servername ) + strlen(MESSAGE_DELIMITER);
    }
    
        // just recurse one time if we need to grow
    if ( dmessage.offset >= dmessage.msgsize ) {
        return refreshDirectoryMessage( true );
    }

        // now just add the length to the front
        // for fun and debugging clear out the whole header with spaces
    NestMemset( dmessage.msg, ' ', dmessage.hdrsize );
    int ret = NestSnprintf( dmessage.msg, 
                            dmessage.hdrsize,
                            "%d %s",
                            dmessage.offset - dmessage.hdrsize,
                            ( ( name ) ? name : MESSAGE_HOLE ) );
    dmessage.msg[dmessage.offset] = '\0';
    assert( ret > 0 );
    
    debug( "Refreshed dir msg: %s\n", dmessage.msg );
    return true;
}


/* UTILITY FUNCTIONS provided free from http://burtleburtle.net/bob/hash/doobs.html */

/*
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() was built out of 36 single-cycle latency instructions in a
  structure that could supported 2x parallelism, like so:
      a -= b;
      a -= c; x = (c>>13);
      b -= c; a ^= x;
      b -= a; x = (a<<8);
      c -= a; b ^= x;
      c -= b; x = (b>>13);
      ...
  Unfortunately, superscalar Pentiums and Sparcs can't take advantage
  of that parallelism.  They've also turned some of those single-cycle
  latency instructions into multi-cycle latency instructions.  Still,
  this is the fastest good hash I could find.  There were about 2^^68
  to choose from.  I only looked at a billion or so.
--------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}


int
most_significant_bits( unsigned hash, int bits ) {
    return ( bits == 0 ) ? 0 : hash >> ( (8 * sizeof(unsigned)) - bits );
}

int
least_significant_bits( unsigned hash, int bits ) {
    return hash & ( ( 1 << (bits) ) - 1 );
    /*
       // this is the same thing, maybe just a couple of insts slower
	unsigned mask = (unsigned)pow( (double)2, (double)bits ) - 1;
	return hash & mask;
    */
}


int hash_additive(const char *key, int len, int prime)
{
  int hash, i;
  for (hash=len, i=0; i<len; ++i)
    hash += key[i];
  return (hash % prime);
}

int hash_rotating(const char *key, int len, int prime)
{
  int hash, i;
  for (hash=len, i=0; i<len; ++i)
    hash = (hash<<4)^(hash>>28)^key[i];
  return (hash % prime);
}

int hash_one_at_a_time(const char *key, int len)
{
  int   hash, i;
  for (hash=0, i=0; i<len; ++i)
  {
    hash += key[i];
    hash += (hash << 10);
    hash ^= (hash >> 6);
  }
  hash += (hash << 3);
  hash ^= (hash >> 11);
  hash += (hash << 15);

    // what is mask?
  //return (hash & mask);
  return hash;
}

