/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_common.h"
#include "nest_speak.h"

#include "dynamic-lard.h"

#include <sys/utsname.h>
#include <map>
#include <vector>

    // time between any regular maintenance (in millisecs)
#define POLLING_INTERVAL 30000
#define THRASH_THRESHOLD 7

typedef struct Proxy {
    unsigned long last_hits;
    unsigned long last_capac_misses;
    unsigned long last_comp_misses;
    int turns_thrashing;
    int turns_thrashing_since_reset;
    struct timeval arrival;
    bool initial;
    char *name;
};

    // helper function 
int Poll();
char *DnsLookup( const struct sockaddr_in *addr, unsigned addr_sz );
void SubmitProxys( );
int SubmitProxy(int q, int proxysize );
int StartVacateFork();
int checkThrashing( Proxy **thrashiest );
void acceptNewApplication( int application_socket );
void acceptNewProxy( int proxy_socket );
int checkLostProxys( fd_set *readys );
int checkLostApplications( fd_set *readys );
void getInitialList();
Proxy * newProxy( NestConnection csd, char *name );
void resetFds( fd_set *readfds );
int sizeBackups( int mttf, int mttr, int numProxys ); 

    // some global stuff 
map<int, Proxy *, less<int> > proxys;
map<int, Proxy *, less<int> > backups;
vector<int> applications;

DynamicLard *dlard;

void pushNewDirectory();

NestConnection homec;   // connection back home
char *home;

fd_set writefds, proxy_readfds;
extern int admin_pid;   // needed by NestSelect
char *proxypasswd;
char *scriptsdir;
int threshold;
int maxfd;

bool adaptive = false;


int mttf = 0, mttr = 0;     // mean time to failure and recovery

    // how many proxys to try to maintain
int desiredBackups = 0;
int stableSize = 1;     // need at least one proxy
bool specifiedStable = false;
int proxysInQueue = 0;
int waitingToSettle = 0;

int proxyStorageSize = 10;

char *submitscript = "proxy_submit.pl";
char *vacatescript = "vacate_job.pl";
char *Logfile      = "proxy_monitor.out";

bool dealWithThrashing( );

void
Usage( char *p, int line ) {
    fprintf( stderr, "Usage (line %d): %s\n"
                     "\t-d debugflags\n"
                     "\t-i poll interval (secs)\n"
                     "\t-f mttf (secs)\n"
                     "\t-r mttr (secs)\n"
                     "\t-h homeserver\n"
                     "\t-o output file\n"
                     "\t-a (turns on adaptive)\n"
                     //"\t-p passwd\n"
                     "\t-s scripts dir\n"
                     "\t-z proxysize (MB)\n"
                     "\t-g initial number of proxys\n"
                     "\t-t threshold\n"
                     "",
            line, p );
    exit( 1 );
}


FILE *out = NULL;

void
initLog( char *Logfile ) {
    out = fopen( Logfile, "w" );
}
 
void 
log( char *fmt, ... ) {
    va_list args;     
    va_start(args,fmt);  
    vfprintf(stdout,fmt,args);         
    vfprintf( out, fmt, args );
    fflush( out );     
    fflush( stdout );
    va_end(args);
}

char *
proxyToString( void *server ) {
    return ((Proxy *)server)->name;
}

int
main( int argc, char *argv[] ) {

    home = scriptsdir = NULL;

    proxysInQueue = 0;

    proxys.clear();
    backups.clear();
    applications.clear();

    int poll_interval = POLLING_INTERVAL; 
    threshold     = THRASH_THRESHOLD;

    specifiedStable = false;

    int c;
    char *prog = argv[0];
    int arg = 0;
    while( ( c = getopt( argc, argv, "ag:d:z:f:r:i:h:o:p:s:t:" )) != EOF ) {
        switch( c ) {
        case 'z':
            proxyStorageSize = atoi( optarg );
            break;
        case 'g':
            stableSize = atoi( optarg );
            specifiedStable = true;
            //log( "Stable size %d\n", stableSize );
            break;
        case 'a':
            adaptive = true;
            break;
        case 'd':
            add_nest_debug_level( optarg );
            break;
        case 'i':
            poll_interval = atoi( optarg ) * 1000;
            break;
        case 'f':
            mttf = atoi( optarg );
            break;
        case 'r':
            mttr = atoi( optarg );
            break;
        case 'h':
            home = NestStrdup( optarg );
            break;
        case 'o':
            Logfile = NestStrdup( optarg );
            break;
        case 'p':    
            printf( "The -p is deprecated\n" );
            Usage( prog, __LINE__ );
            //proxypasswd = NestStrdup( optarg );
            break;
        case 's':
            scriptsdir = NestStrdup( optarg );
            break;
        case 't':
            threshold = atoi( optarg );
            break;
        default:
            Usage( prog, __LINE__ );
        }
    }
    arg = optind;
    if( ! home || ! scriptsdir ) Usage( prog, __LINE__ );
    //if ( ! mttf || ! mttr ) Usage( prog, __LINE__ );
    //log( "ARGS: %s mttf %d mttr %d %d threshold\n", home, mttf, mttr, threshold );
    
    initLog( Logfile );

        // fork a little process to sit there and open up nodes for proxys
    StartVacateFork( );

        // set up directory
    dlard = new DynamicLard( proxyToString );
    dlard->setName( home );

        // ignore SIGPIPES
    NestSignal( SIGPIPE, SIG_IGN );

        // set up our fd sets
    fd_set r_readyfds, w_readyfds;
    FD_ZERO( &writefds );
    FD_ZERO( &proxy_readfds  );

    int proxy_socket = NestsetSocket( 
            AF_INET, 
            SOCK_STREAM, 
            0, 
            PROXY_MONITOR_PORT );
    FD_SET( proxy_socket, &proxy_readfds );
    maxfd = proxy_socket;
    int application_socket = NestsetSocket(
            AF_INET, 
            SOCK_STREAM, 
            0, 
            PROXY_APPLICATION_PORT );
    FD_SET( application_socket, &proxy_readfds );
    maxfd = NEST_MAX( application_socket, maxfd );


    admin_pid = NestGetpid();

    struct timeval last_poll;
    NestGettimeofday( &last_poll );

        // check to see if any are already running
    getInitialList();

    
        // main select loop here
    while( true ) {

            // first see if we need to submit any more proxys
        SubmitProxys(); // this may be no-op if we don't need to

            // print out our state
        nest_debug( NEST_DEBUG_PROXY, 
                "%d proxys, %d backups, %d stablesize, %d target backups, "
                "%d submitted.\n",
                proxys.size(), backups.size(), stableSize, desiredBackups,
                proxysInQueue );

        r_readyfds = proxy_readfds; 
        w_readyfds = writefds;
        bool changes = false;
        int dealtWith = 0;
        int readies = NestSelect(
                maxfd + 1, 
                &r_readyfds,
                &w_readyfds, 
                NULL,
                NestNextTimeout( poll_interval, &last_poll ) );
        if ( readies == 0 ) {
                // no-one ready, we must have timed out
            NestGettimeofday( &last_poll );
            Poll();
            bool changedDirectory = dealWithThrashing( );
            if ( changedDirectory ) pushNewDirectory();
            if ( waitingToSettle > 0 ) waitingToSettle--;
            continue;
        } else if ( readies < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, "select failed: %s\n", strerror( errno ) );
            continue;
        }

            // if we make it here, we have some socket activity

            // are we losing an old application?
        int lost = checkLostApplications( &r_readyfds );
        dealtWith += lost;

            // are we losing an old proxy? 
        lost = checkLostProxys( &r_readyfds );
        dealtWith += lost;
        if ( lost ) changes = true;
        
            // is it a new proxy?
        if ( FD_ISSET( proxy_socket, &r_readyfds ) ) {
            acceptNewProxy( proxy_socket );
            dealWithThrashing();    // maybe new proxy can help a thrasher
            dealtWith++;
            changes = true;
        }

            // see if we need to push new info to our apps
            // do this before check the new app, so we don't send new one
            // the directory twice
        if ( changes ) {
            pushNewDirectory();
        }

            // is it a new application?
        if ( FD_ISSET( application_socket, &r_readyfds ) ) {
            acceptNewApplication( application_socket );
            dealtWith++;
        }
        
            // this shouldn't happen, but somehow we sometimes
            // get our readfd set screwed up . . .
        if ( dealtWith != readies ) {
            nest_warn( __FILE__, __FUNCTION__, "Weirdness in proxy-monitor, trying to recover.\n" );
            FD_ZERO( &proxy_readfds  );
            FD_SET( proxy_socket, &proxy_readfds );
            FD_SET( application_socket, &proxy_readfds );
            resetFds( &proxy_readfds );
        }

    } // end of main service loop

        // we should never get here
    nest_die( "Big trouble in Little China.\n" );
    return ERROR_EXIT;
}

void
foreachproxy( char *proxy, void *args ) {
    NestConnection gc;
    NestReplyStatus status = NestOpenConnection( gc, proxy );
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "Could not open initial contact to report proxy %s: %s\n",
                proxy, NestErrorString( status ) );
    } else {
        Proxy *g = newProxy( gc, proxy );
        log( "%ld Initial proxy %s\n", time(NULL), g->name );
        if ( proxys.size() >= (unsigned)stableSize ) {
            backups[gc] = g;
        } else {
            proxys[gc] = g;
        }
        dlard->addServer( g );
    }
}

int 
sizeBackups( int mttf, int mttr, int numProxys ) {

    if ( mttr == 0 || mttf == 0 ) return 0;
    
    float num = mttr;
    float den = (float)mttf / numProxys;
    float frac = num / den;
    
    int size = (int)floor( frac );
    if ( frac - floor(frac) >= 0.5 ) size++;
    
    nest_debug( NEST_DEBUG_PROXY, "For %d mttf, %d mttr, %d proxys, "
            "returning %d desiredBackups\n", mttf, mttr, numProxys, size );
    return size;
}

bool
dealWithThrashing( ) {

    bool changedDirectory = false;
        
        // first see if any backups are thrashing, if so, promote them
    map<int, Proxy *, less<int> >::iterator i, j;
    j = backups.begin();
    while( j != backups.end() ) {
        i = j++;
        Proxy *g = (*i).second;
        if ( g->turns_thrashing_since_reset > threshold ) {
            log( "Promoting thrashing %s from backup to active.\n", g->name );
            proxys[(*i).first] = g;
            backups.erase( i );
        }
    }

        // now for any thrashing proxys, help by promoting any available backup 
    Proxy *worstThrasher;
    int thrashing;
    if ( adaptive ) {
        while( backups.size() && (thrashing = checkThrashing(&worstThrasher)) ){
            map<int, Proxy *, less<int> >::iterator b;
            b = backups.begin();
            dlard->removeServer( (*b).second );
            dlard->addHelper( worstThrasher, (*b).second );
            changedDirectory = true;
            proxys[(*b).first] = (*b).second;
            log( "Helping active thrasher %s with backup %s.\n", 
                    worstThrasher->name, (*b).second->name );
            worstThrasher->turns_thrashing_since_reset = 0;
            backups.erase( b );
        }
    }

    return changedDirectory;
}

void
openHomeConnection( ) {
    NestReplyStatus status = NestOpenConnection( homec, home );
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not open connection to %s: %s\n",
                home, NestErrorString( status ) );
    }
}

void
getInitialList( ) {
    openHomeConnection();
    NestReplyStatus status = NestProxyQuery( homec, foreachproxy, NULL );
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not query proxys from %s: %s\n",
                home, NestErrorString( status ) );
    }
    proxysInQueue = proxys.size() + backups.size();
}

void
sendDirectory( int csd ) {
    Proxy homeProxy;
    bool  usingHome = false;

        // special case, empty list, just send the main server
    if ( proxys.size() == 0 ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "Special case, no proxys, sending home\n" );
        homeProxy.name = home;
        assert( dlard->numServers() == 0 );
        dlard->addServer( &homeProxy );
        usingHome = true;
    }

        // go ahead and send list
    if ( dlard->send( csd ) < 0 ) {
        nest_debug( NEST_DEBUG_PROXY, "Probably lost app %d\n", csd );
    }

        // clean up after special case
    if ( usingHome ) {
        assert( proxys.size() == 0 );
        dlard->removeServer( &homeProxy );
    }
}

void
pushNewDirectory( ) {
    nest_debug( NEST_DEBUG_PROXY, "Sending revised dlard to all %d apps.\n", 
            applications.size() );
    vector<int>::iterator i;
    for( i = applications.begin(); i != applications.end(); i++ ) {
        sendDirectory( *i );
    }
}

void
acceptNewApplication( int application_socket ) {
    struct sockaddr_in address;
    unsigned addr_sz = sizeof( address );

    int csd =  NestAccept( application_socket, (sockaddr*)&address, &addr_sz );
    if ( csd > 0 ) {
        char *apploc = DnsLookup( &address, addr_sz );
        nest_debug( NEST_DEBUG_PROXY, "New app from %s.\n", apploc ); 
        log( "%ld New app from %s (%d).\n", time(NULL), apploc, csd );
        NestFree( apploc );
        applications.push_back( csd );
        FD_SET( csd, &proxy_readfds );
        maxfd = NEST_MAX( maxfd, csd );
        sendDirectory( csd );
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Failed to accept new app: %s\n", strerror( errno ) );
    }
}

void
acceptNewProxy( int proxy_socket ) {
    struct sockaddr_in address;
    unsigned addr_sz = sizeof( address );

    int csd =  NestAccept( proxy_socket, (sockaddr*)&address, &addr_sz );
    if ( csd > 0 ) {
            // accept proxy, and set him up
        char *proxyloc = DnsLookup( &address, addr_sz );
        Proxy *g = newProxy( csd, proxyloc );
        NestFree( proxyloc );

            // always make sure we have at least one real proxy
        if ( proxys.size() < (unsigned)stableSize ) {
            log( "%ld New proxy %s is active proxy\n", time(NULL), g->name );
            dlard->addServer( g );
            proxys[csd] = g;
        } else {
                // here if anyone is thrashing, add this one to help and
                // make this one be an active proxy
            log( "%ld New proxy %s to be backup\n", time(NULL), g->name );
            dlard->addServer( g );
            backups[csd] = g;
        }
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Failed to accept new proxy: %s\n", strerror( errno ) );
    }
}

// little helper routine to free the structure
void
freeProxy( Proxy *g, int csd ) {
    NestFree( g->name );
    if ( maxfd == csd ) maxfd--;
    FD_CLR( csd, &proxy_readfds );
    NestCloseConnection( csd );
    NestFree( g );
}

// little helper routine to populate the structure
Proxy *
newProxy( NestConnection csd, char *name ) {
    Proxy *g = (Proxy *)NestCalloc( sizeof( *g ) );
    g->name = NestStrdup( name );
    g->initial = true;
    maxfd = NEST_MAX(csd,maxfd);
    FD_SET( csd, &proxy_readfds );
    NestGettimeofday( &g->arrival );
    return g;
}

// returns by value how many are thrashing,
// returns by reference, who is the worstThrasher
// TODO should return who is mostLikely to thrash even if no thrashers . . .
int
checkThrashing( Proxy **worstThrasher ) {
    int worst     = 0;
    int thrashing = 0;
    int thrashing_since_reset = 0;
    map<int, Proxy *, less<int> >::iterator i;
    for( i = proxys.begin(); i != proxys.end(); i++ ) {
        if ( (*i).second->turns_thrashing_since_reset > threshold ) {
            nest_debug( NEST_DEBUG_PROXY, "%s is thrashing.\n", 
                    (*i).second->name );
            thrashing_since_reset++;
            thrashing++;
            if ( (*i).second->turns_thrashing_since_reset > worst ) {
                if ( worstThrasher != NULL ) *worstThrasher = (*i).second;
                worst = (*i).second->turns_thrashing_since_reset;
            }
        } else {
            if ( (*i).second->turns_thrashing > threshold ) thrashing++;
        }
    }
    if ( thrashing == 0 && proxys.size() > 0 
            && ! specifiedStable 
            && ! waitingToSettle ) 
    {
        stableSize = proxys.size();
        nest_debug( NEST_DEBUG_PROXY, "No thrashing, setting stable to %d.\n",
                stableSize );
    }
    return thrashing_since_reset;
}

void
resetFds( fd_set *readfds ) {
    nest_debug( NEST_DEBUG_PROXY, "Resetting fdset...\n" );
    map<int, Proxy *, less<int> >::iterator i;
    for( i = backups.begin(); i != backups.end(); i++ ) {
        FD_SET( (*i).first, readfds );
    }
    for( i = proxys.begin(); i != proxys.end(); i++ ) {
        FD_SET( (*i).first, readfds );
    }
    vector<int>::iterator j;
    for( j = applications.begin(); j != applications.end(); j++ ) {
        FD_SET( *j, readfds );
    }
}

int
checkLostApplications( fd_set *readys ) {
    int removals = 0;
    vector<int>::iterator j, i;
    j = applications.begin();
    while( j != applications.end() ) {
        i = j++;
        if( FD_ISSET( (*i), readys ) ) {
            log( "%d Lost application %d.\n", time(NULL), (*i) );
            FD_CLR( (*i), &proxy_readfds );
            NestClose( *i );
            if ( maxfd == *i ) maxfd--;
            j = applications.erase( i );
            removals++;
        }
    }
    return removals;
}

int
checkLostProxys( fd_set *readys ) {

    int removals = 0;
    map<int, Proxy *, less<int> >::iterator i, j;

        // first check the backups
    j = backups.begin();
    while( j != backups.end() ) {
        i = j++;
        if ( FD_ISSET( (*i).first, readys ) ) {
            log( "%d Lost backup proxy %s.\n", time(NULL), (*i).second->name );
            dlard->removeServer( (*i).second );
            freeProxy( (*i).second, (*i).first );
            backups.erase( i );
            removals++;
        }
    }

        // then check the actives, replace any lost actives with backups
    j = proxys.begin();
    while( j != proxys.end() ) {
        i = j++;
        if ( FD_ISSET( (*i).first, readys ) ) {
            if ( backups.size() > 0 ) {
                map<int, Proxy *, less<int> >::iterator k;
                k = backups.begin();
                Proxy *backup = (*k).second;
                log( "%ld Lost active proxy %s, replacing with backup %s.\n", 
                        time(NULL), (*i).second->name, backup->name );
                dlard->removeServer( backup );
                dlard->replaceServer( (*i).second, backup );
                proxys[(*k).first] = (*k).second;
                backups.erase( k );
            } else {
                log( "%ld Lost active proxy %s.\n",
                        time( NULL), (*i).second->name );
                dlard->removeServer( (*i).second );
            }
            freeProxy( (*i).second, (*i).first );
            proxys.erase( i );
            removals++;
        }
    }
    return removals;
}


static char *command = NULL;
static int commandlen = 0;
static char *silence;

void
growCommand( ) {
    if( command == NULL ) {
        commandlen = 4096;
    } else {
        commandlen *= 2;
        NestFree( command );
    }
    command = (char*)NestMalloc( commandlen );

    if ( get_nest_debug_level() & NEST_DEBUG_PROXY ) {
        silence = "";
    } else {
        silence = " > /dev/null";
    }
    
}

int
StartVacateFork( ) {

    if ( 1 ) {
        nest_debug( NEST_DEBUG_PROXY, "Run the fork thing yourself.\n" );
        return 0;
    }

    pid_t pid = fork();
    if ( pid < 0 ) {
        perror( "fork" );
        exit( 1 );
    } else if ( pid != 0 ) {
        nest_debug( NEST_DEBUG_PROXY, "Forked child %d to run vacate script.\n",
                pid );
        return 0;
    }

        // here we are the child
    if ( command == NULL ) growCommand();

    while( NestSnprintf( command, commandlen, "%s/%s %s",
                scriptsdir, vacatescript, silence ) < 0 )
    {
        growCommand();
    }

    nest_debug( NEST_DEBUG_PROXY, "Will try to run vacate script with %s.\n",
        command );
    int ret = system( command );
    if ( ret < 0 ) {
        perror( "system" );
    }
    nest_die( "This command %s is never supposed to return.\n" );
    return 0;
}

void
SubmitProxys( ) {
    int thrash = checkThrashing( NULL );
    desiredBackups = sizeBackups( mttf, mttr, stableSize );

        // always see if any proxys showed up who we weren't expecting.
    if ( (unsigned)proxysInQueue < (unsigned)proxys.size() + backups.size() ) {
        proxysInQueue = proxys.size() + backups.size();
    }

    int needed = stableSize + desiredBackups;
    if ( adaptive ) {
        needed = thrash + proxys.size() + backups.size();
        //needed += thrash;
        // special case 0
        if ( proxysInQueue == 0 ) needed = 1;
    }

    if ( needed > proxysInQueue ) {
        SubmitProxy( needed - proxysInQueue, proxyStorageSize );
        proxysInQueue += (needed - proxysInQueue) ;
        waitingToSettle = threshold * 2;
    }

    /*
        // special case 0
    if ( proxys.size() == 0 ) {
        assert( backups.size() == 0 );
        needed = stableSize + desiredBackups - backups.size();
    } else {
        needed = thrash + desiredBackups - backups.size();
    }
    needed -= proxysInQueue;
    
        // do it always and pass it our thoughts about pending so if we're out
        // of whack, it will fix it for us.  it should really kill some too maybe
//    if ( needed > 0 ) {
    nest_debug( NEST_DEBUG_PROXY, 
            "%d thrash, %d act, %d bu, %d desired, %d stable, %d pending\n"
            "\t=> %d needed\n",
            thrash, proxys.size(), backups.size(), desiredBackups,
            stableSize, proxysPending, needed );
    if ( needed > 0 ) {
        SubmitProxy( needed );
        proxysPending += needed;
    }
//    } else if ( needed < 0 ) {
    if ( needed < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "Oversubscribed.  More proxys than are needed.\n" );
    }
    */
}

int
SubmitProxy( int quant, int storagesize ) {
    int ret = 0;

    if ( command == NULL ) growCommand();
    while ( NestSnprintf( command, commandlen, 
                "%s/%s %s %s %d %d",
                scriptsdir, submitscript,
                home, NestUname(), storagesize, quant ) < 0 )
    {
        growCommand();
    }


    if ( quant > 0 ) {
        nest_debug( NEST_DEBUG_PROXY, "Submitting proxy with %s.\n", command );
        log( "%ld Submit proxy %d\n", time(NULL), quant );
    }

    ret = system( command );
    if ( ret < 0 ) {
        perror( "system" );
    } 
    return ret;
}

unsigned long
ClassAdLongValue( ClassAd *ad, char *attr ) {
    static ClassAdUnParser unp;
    ExprTree *tree = NULL;
    string buf;

    tree = ad->Lookup( attr );
    unp.Unparse( buf, tree );
    return ( tree ) ?  (unsigned long)atol( buf.c_str() ) : (unsigned long)-1;
}

void
updateProxy( Proxy *g, char *adstring ) {
    unsigned long misses, hits, compmisses, thismisses, thishits,thiscompmisses;
    ClassAdParser cap;
    ClassAd *ad = cap.ParseClassAd( adstring );
    misses      = ClassAdLongValue( ad, "Capac_Misses" );
    hits        = ClassAdLongValue( ad,  "Hits" );
    compmisses  = ClassAdLongValue( ad, "Comp_Misses" ); 

    if ( g->initial ) {
        thismisses = thishits = thiscompmisses = 0;
        g->initial = false;
    } else {
        thismisses = misses - g->last_capac_misses;
        thishits   = hits   - g->last_hits;
        thiscompmisses = compmisses - g->last_comp_misses;
    }
    if ( thismisses || thishits || thiscompmisses ) {
        if ( thismisses > 0 ) {
            g->turns_thrashing++;
            g->turns_thrashing_since_reset++;
        } else {
            g->turns_thrashing = 0;
            g->turns_thrashing_since_reset = 0;
        }
    }
    g->last_hits = hits;
    g->last_capac_misses = misses;
    g->last_comp_misses = compmisses;

    log(    "Proxy %s "
            "cap misses %ld (%ld), "
            "comp misses %ld (%ld), "
            "hits %ld (%ld).\n",
            g->name, 
            misses, thismisses, 
            compmisses, thiscompmisses,
            hits, thishits );
    if ( ad ) NestFree( ad );
}

int
myPoll( map<int, Proxy *, less<int> > *pool ) {
    map<int, Proxy *, less<int> >::iterator i;
    for( i = pool->begin(); i != pool->end(); i++ ) {
        char *ad = NULL;
        unsigned adlen;
        NestReplyStatus status = NestRequestAd( (*i).first, &ad, &adlen );
        if ( status == NEST_SUCCESS ) {
            updateProxy( (*i).second, ad );
            if ( ad )   delete( ad );
        } else {
            nest_debug( NEST_DEBUG_PROXY, "Probably lost proxy %s: %s\n",
                    (*i).second->name, NestErrorString( status ) );
        }
    }
    return 0;
}

int
Poll( ) {

        // first poll the main server, then poll the proxys
    char *adstring = NULL;
    unsigned adlen;
    NestReplyStatus status = NestRequestAd( homec, &adstring, &adlen );
    if ( status == NEST_SUCCESS ) {
        ClassAdParser cap;
        ClassAd *ad = cap.ParseClassAd( adstring );
        long kbs_out, kbs_in, concur, errors, transf, ts;
        ts      = ClassAdLongValue( ad, "NestTime" );
        kbs_out = ClassAdLongValue( ad, "Total_kilobytes_out" );
        kbs_in  = ClassAdLongValue( ad, "Total_kilobytes_in" );
        concur  = ClassAdLongValue( ad, "Current_concurrency" );
        errors  = ClassAdLongValue( ad, "Total_errors" );
        transf  = ClassAdLongValue( ad, "Total_transfers" );
            //( my $ts, my $t, my $k_o, my $k_i, my $con, my $e )
        log( "Home %s %ld %ld %ld %ld %ld %ld\n",
            home, ts, transf, kbs_out, kbs_in, concur, errors );
        if ( ad ) NestFree( ad );
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Could not retrieve ad from home %s: %s. Will reopen.\n",
                home, NestErrorString( status ) );
        openHomeConnection();
    }

        // poll each proxy
    nest_debug( NEST_DEBUG_PROXY, "Dumping directory:\n" );
    if ( get_nest_debug_level() & NEST_DEBUG_PROXY ) {
        dlard->dump( stderr, proxyToString ); 
    }
    nest_debug( NEST_DEBUG_PROXY, "Polling %d proxys and %d backups.\n", 
            proxys.size(), backups.size() );
    myPoll( &proxys );
    myPoll( &backups );

    return 0;
}

char *
DnsLookup( const struct sockaddr_in *from, unsigned addr_sz ) {
    struct hostent * mach;
    mach = gethostbyaddr( (char*)&(from->sin_addr), addr_sz, AF_INET );
    if (mach == NULL) {
        nest_warn( __FILE__, __FUNCTION__, "gethostbyaddr %s: %s\n",
                (char *)&(from->sin_addr), hstrerror (h_errno)  );
        return (char *)&(from->sin_addr); 
    } else {
        return NestStrdup( mach->h_name );
    }
}
    
