/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_client.h"
#include "nest_client_pool.h"

#define   lock_mutex()   lockMutex( __FUNCTION__, __LINE__ )
#define unlock_mutex() unlockMutex( __FUNCTION__, __LINE__ )


    // constructor
nestClientPool::nestClientPool( bool thread_safe ) {
    maxfd = 0;
    admissions = 0;
    total_sizes = 0;
    max_len = 0;

    clients.clear();
    if ( this->thread_safe = thread_safe ) {
        cond_var = (pthread_cond_t * ) NestMalloc( sizeof( pthread_cond_t  ) );
        mutex    = (pthread_mutex_t *) NestMalloc( sizeof( pthread_mutex_t ) );
        nest_pthread_cond_init ( cond_var, NULL );
        nest_pthread_mutex_init( mutex, NULL );
    } else {
        cond_var = NULL;
        mutex = NULL;
    }
}

nestClientPool::~nestClientPool( ) {
    if ( thread_safe ) {
        NestFree( cond_var );
        NestFree( mutex );
    }
    clients.clear();
}

bool
nestClientPool::empty( ) {
    return ( clients.size() == 0 );
}

ssize_t
nestClientPool::size( ) {
    return clients.size();
}

ssize_t
nestClientPool::clear( ) {
    ssize_t s = size();
    lock_mutex();
    clients.clear();
    unlock_mutex();
    return s;
}

bool
nestClientPool::isThreadSafe() {
    return thread_safe;
}

int
nestClientPool::addClient( nestProtocol* connection, sockaddr_in *address ) {

    nestClient *newclient = new nestClient( connection, address );
    return addClient( newclient );
}

bool
nestClientPool::contains( nestClient *client ) {
        // this used to just check the address.  That is clearly flawed.
        // Now it checks the id which is better.
    bool found = false;
    lock_mutex();
    for( CLIENT_ITERATOR itr = clients.begin(); itr != clients.end(); itr++ ) {
        if ( client == ( nestClient* ) *itr ) {
            found = true;
            break;
        }
    }
    unlock_mutex();
    return found;
}

int
nestClientPool::addClient( nestClient *newclient ) {
    
    assert( newclient != NULL );
    assert( ! contains( newclient ) ); 
    lock_mutex();
    clients.push_back( newclient );

        // maintain maxfd to allow select on idle clients
    if ( newclient->hasSocket() ) {
        int sd = newclient->getConnection()->getSocket();
        if ( sd > maxfd ) maxfd = sd;
    }
    int size = clients.size( );
    admissions++;
    total_sizes += size;
    //ave_len = ( ( ave_len * ( admissions - 1 ) ) + size ) / admissions; 
    max_len = NEST_MAX( max_len, size );
    signal();
    unlock_mutex();
    return size;
}

int
nestClientPool::getMaxfd( ) {
        // notice that we don't check whether maxfd can be decremented
        // when we lose a client, maybe we should . . .
    lock_mutex();
    int max = maxfd;
    unlock_mutex();
    return max;
}

int
nestClientPool::removeClient( nestClient *client ) {

    lock_mutex();
    clients.remove( client );
    int size = clients.size();
    unlock_mutex();
    return size;
}

nestClient *
nestClientPool::getFront( bool unused ) {
    lock_mutex();
    while ( clients.empty() ) {
        if ( thread_safe ) nest_pthread_cond_wait( cond_var, mutex );
        else return NULL;
    }
    nestClient *client = (nestClient *) clients.front();
    clients.pop_front();
    unlock_mutex();
    return client;
}

// this should not be called internally since it relocks
CLIENT_ITERATOR
nestClientPool::begin() {
    lock_mutex();
    CLIENT_ITERATOR itr = clients.begin();
    unlock_mutex();
    return itr;
}

// this should not be called internally since it relocks
CLIENT_ITERATOR
nestClientPool::end() {
    lock_mutex();
    CLIENT_ITERATOR itr = clients.end();
    unlock_mutex();
    return itr;
}

CLIENT_ITERATOR
nestClientPool::erase( CLIENT_ITERATOR itr ) {
    lock_mutex();
    CLIENT_ITERATOR next = clients.erase( itr );
    unlock_mutex();
    return next;
}

bool
nestClientPool::signal() {
    if ( thread_safe ) {
        nest_pthread_cond_signal( cond_var );
        return true;
    } else {
        return false;
    }
}

bool
nestClientPool::wait() {
    if ( thread_safe ) {
        nest_pthread_mutex_lock( mutex );
        return true;
    } else {
        return false;
    }
}

bool
nestClientPool::lockMutex( const char *func, int line ) {
    if ( thread_safe ) {
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "Acquiring client pool lock for 0x%x at %s:%d\n", 
                this, func, line );
        nest_pthread_mutex_lock( mutex );
        return true;
    } else {
        return false;
    }
}

bool
nestClientPool::unlockMutex( const char *func, int line ) {
    if ( thread_safe ) {
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "Releasing client pool lock for 0x%x at %s:%d\n", 
                this, func, line );
        nest_pthread_mutex_unlock( mutex );
        return true;
    } else {
        return false;
    }
}
