/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_common.h"
#include "nest_request.h"
#include "nest_pthread_util.h"


bool    NestEnableLots      = true;
char    *NestAdminUser      = "";
user_id_t   AdminUid;

const char    *NestSystemUser     = "system";
const char    *NestAnonymousUser  = "";

const char    *SystemPwdFile      = "/etc/passwd";
const char    *SystemGrpFile      = "/etc/group";
const char    *GridMapFile        = "/etc/grid-security/grid-mapfile";

const char    *MetaDir            = "./__NEST__";

const char    *ActualGroup        = "";
user_id_t   ActualGid;

pthread_mutex_t nest_common_mutex;
pthread_mutex_t nest_malloc_mutex;
pthread_mutex_t nest_strdup_mutex;

#ifdef USE_PTHREADS

void
initCommonMutex( ) {
    nest_pthread_mutex_init( &nest_common_mutex, NULL );
    nest_pthread_mutex_init( &nest_malloc_mutex, NULL );
    nest_pthread_mutex_init( &nest_strdup_mutex, NULL );
}

void
lockCommonMutex( pthread_mutex_t *mux, char *func, int line ) {
    nest_debug( NEST_DEBUG_CONCURRENCY, "%d acquiring 0x%x from %s:%d\n",
            NestGetpid(), mux, func, line );
    nest_pthread_mutex_lock( mux );
}

void
unlockCommonMutex( pthread_mutex_t *mux, char *func, int line ) {
    nest_debug( NEST_DEBUG_CONCURRENCY, "%d released 0x%x from %s:%d\n",
            NestGetpid(), mux, func, line );
    nest_pthread_mutex_unlock( mux );
}

#else

void initCommonMutex()                                      {}
void lockCommonMutex    (pthread_mutex_t *, char *, int )   {}
void unlockCommonMutex  (pthread_mutex_t *, char *, int )   {}

#endif

int admin_pid = 0;
void
NestcheckPid( const char *file, const char *function ) {
    nest_debug( NEST_DEBUG_ADMIN, "%s:%s - %d =?= %d\n",
            file, function, NestGetpid(), admin_pid );
    if ( NestGetpid() != admin_pid ) {
        nest_die( "Admin violoation in %s:%s.  %d != %d\n",
                file, function, NestGetpid(), admin_pid );
    }
}


    // User management utilities

int
Nestset_grid_mapfile( const char *fname ) {
    return NestOpen( fname, O_RDONLY );
}

grid_mapfile_entry  *
Nestget_grid_mapfile_entry(int grid_mapfile_fd) {
    static  grid_mapfile_entry  entry;
    if (grid_mapfile_fd < 0) {
        return NULL;
    }

    char    ch; //line[sizeof(grid_mapfile_entry)];
    bool    error;
    int     i;

    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ((ch != ' ') &&
            (ch != '\t') &&
            (ch != '\n')) {
            error = false;
            break;
        }
    }
    if (error || ch != '"') {
        return NULL;
    }

    i = 0;
    entry.identity[i++] = ch;
    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        entry.identity[i++] = ch;
        if (ch == '"') {
            error = false;
            break;
        }
    }
    if (error) {
        return NULL;
    }
    entry.identity[i] = 0;

    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ((ch != ' ') &&
            (ch != '\t') &&
            (ch != '\n')) {
            error = false;
            break;
        }
    }
    if (error) {
        return NULL;
    }

    i = 0;
    entry.name[i++] = ch;
    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ((ch == ' ') ||
            (ch == '\t') ||
            (ch == '\n')) {
            error = false;
            break;
        }
        entry.name[i++] = ch;
    }
    if (error) {
        return NULL;
    }
    entry.name[i] = 0;

    return &entry;
}

int
Nestinsert_grid_mapfile_entry( const char *fname, 
                               const char *identity,
                               const char *name ) {
    
    struct grid_mapfile_entry *user;
 
    int state = Nestset_grid_mapfile(fname);
    if ( state < 0) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Couldn't read system grid-mapfile at %s!\n",
                    fname );
        return state;
    }

    char entry[NEST_MAXPATHLEN];
    
    // An identity cannot be mapped to more than one user name
    while( (user = Nestget_grid_mapfile_entry(state)) != NULL) {
        if(!strcmp(user->identity, identity) && 
           strcmp(user->name, name))
            return -1;
        if(!strcmp(user->identity, identity) && 
           !strcmp(user->name, name))
            return 0;
    }   //  End-of-while
    Nestend_grid_mapfile(state);    

    state = NestOpen(fname, O_WRONLY);
    if ( state < 0) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Couldn't update system grid-mapfile at %s!\n",
                    fname );
        return state;
    }

    strcat(strcat(strcat(strcpy(entry, identity), " "), name), "\n");

    int size = strlen(entry), written;
    NestLseek(state,0,SEEK_END);
    written = NestWriten(state, entry, size);
    NestClose(state);
    if(written < size)
        return -1;
    else
        return 0;
}

int
Nestend_grid_mapfile(int grid_mapfile_fd) {
    if (grid_mapfile_fd >= 0) {
        NestClose(grid_mapfile_fd);
    }
    return 0;
}

struct passwd *
Nestnew_pwd_ent(){

    struct passwd    *ret_ptr = (struct passwd *) 
        NestMalloc(sizeof(struct passwd));

    ret_ptr->pw_name   = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_passwd = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_gecos  = (char *) NestMalloc (NEST_NAMELEN * 10);
    ret_ptr->pw_dir    = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_shell  = (char *) NestMalloc (NEST_NAMELEN);

    return ret_ptr;
}


void
Nestfree_pwd_ent(struct passwd *ptr) {

    NestFree(ptr->pw_name);
    NestFree(ptr->pw_passwd);
    NestFree(ptr->pw_gecos);
    NestFree(ptr->pw_dir);
    NestFree(ptr->pw_shell);

    NestFree(ptr);
}

void
Nestpw_assign(struct passwd *dst, struct passwd *src) {

    //memcpy((void *) dst, (void *) src, sizeof(struct passwd));

    dst->pw_uid = src->pw_uid;
    dst->pw_gid = src->pw_gid;
    strcpy(dst->pw_name, src->pw_name);
    strcpy(dst->pw_passwd, src->pw_passwd);
    strcpy(dst->pw_gecos, src->pw_gecos);
    strcpy(dst->pw_dir, src->pw_dir);
    strcpy(dst->pw_shell, src->pw_shell);

}


int
Nestfgetpwuid(user_id_t u, struct passwd *p, FILE *fp) {

    fpos_t    save_pos;
    struct passwd *tmp_p;

    if (!fp || !p){
        return -1;
    }

    if (fgetpos(fp, &save_pos) < 0) {
        return -1;
    }

    while ( (tmp_p = fgetpwent(fp)) != NULL){
        if ((user_id_t)u == (user_id_t)(tmp_p->pw_uid)) {
            break;
        }
    }

    fsetpos(fp, &save_pos);

    if (!tmp_p) {
        return -1;
    }

    Nestpw_assign(p, tmp_p);
    return 0;
}


int
Nestfgetpwnam(const char *user, struct passwd *p, FILE *fp) {

    fpos_t    save_pos;
    struct passwd *tmp_p;

    if (!user || !p || !fp){
        return -1;
    }

    if (fgetpos(fp, &save_pos) < 0) {
        return -1;
    }

    LOCK_MUTEX( &nest_common_mutex );
    while( (tmp_p = fgetpwent(fp)) != NULL){
        if(!strcmp(user, tmp_p->pw_name)) {
            break;
        }
    }
    UNLOCK_MUTEX( &nest_common_mutex );

    fsetpos(fp, &save_pos);

    if (!tmp_p) {
        return -1;
    }

    Nestpw_assign(p, tmp_p);
    return 0;
}


int
Nestmy_getpwuid(user_id_t u, struct passwd *p, const char *fname) {

    FILE    *p_fp = fopen(fname, "r");
    int     ret_val;

    if (p_fp == NULL) {
        return -1;
    }

    ret_val = Nestfgetpwuid(u, p, p_fp);

    fclose(p_fp);
    return(ret_val);
}

int
Nestmy_getpwnam(const char *u, struct passwd *p, const char *fname) {

    FILE    *p_fp = fopen(fname, "r");
    int     ret_val;

    if (p_fp == NULL) {
        return -1;
    }

    ret_val = Nestfgetpwnam(u, p, p_fp);

    fclose(p_fp);
    return(ret_val);
}


int
NestGet_uid(user_id_t *u, const char *uname, const char *fname) {
    struct passwd *pw;

    pw = Nestnew_pwd_ent();

    nest_debug( NEST_DEBUG_CLIENT, "Looking up uid for uname %s from %s.\n",
            uname, fname );

    if (Nestmy_getpwnam(uname, pw, fname) < 0) {
		nest_debug( NEST_DEBUG_CLIENT, "No passwd entry found for user '%s'\n",
					uname );
        Nestfree_pwd_ent(pw);
        return -1;
    }

	nest_debug( NEST_DEBUG_CLIENT, "UID for '%s' found %d\n",
				uname, pw->pw_uid );
    *u = pw->pw_uid;
    Nestfree_pwd_ent(pw);
    return 0;
}

size_t
NestStrlen( const char *s ) {
    return strlen( s );
}

int
NestSnprintf( char *buf, int len, const char *fmt, ... ) {

    va_list ap;
    int n;

    va_start(ap, fmt);
    n = vsnprintf (buf, len, fmt, ap);
    va_end(ap);

    /*
    nest_debug( NEST_DEBUG_SERVER,
                "NestSnprintf called - %d %s!\n",
                len,
                fmt );
    */

    if ((n < 0) || (n >= len)) {
        //nest_debug( NEST_DEBUG_SERVER, "NestSnprintf returns error!\n" );
        if (buf) {
            *buf = 0;
        }
        return -1;
    }

    /*
    nest_debug( NEST_DEBUG_SERVER,
                "NestSnprintf returns %s!\n",
                buf );
    */

    return n;
}

char*
NestMakeString(int initial_size, const char *fmt, ...) {
    /* Guess we need no more than 100 bytes. */
    int n, size = initial_size;
    char *p;
    va_list ap;
    if ((p = (char *) NestMalloc (size)) == NULL) {
        return NULL;
    }

    while (1) {
        /* Try to print in the allocated space. */
        va_start(ap, fmt);
        n = vsnprintf (p, size, fmt, ap);
        va_end(ap);
        /* If that worked, return the string. */
        if (n > -1 && n < size) {
            return p;
        }
        /* Else try again with more space. */
        if (n > -1) {       /* glibc 2.1 */
            size = n+1;     /* precisely what is needed */
        } else    {           /* glibc 2.0 */
            size *= 2;      /* twice the old size */
        }
        if ((p = (char *) NestRealloc (p, size)) == NULL) {
            return NULL;
        }
    }
}

int
NestGet_uname( char *uname, user_id_t u ) {
    return NestGet_uname( uname, u, SystemPwdFile );
}

int
NestGet_uname(char *uname, user_id_t u, const char *fname) {

    struct passwd *pw;

    pw = Nestnew_pwd_ent();

    nest_debug( NEST_DEBUG_CLIENT, "Looking up uname for uid %d from %s.\n",
            u, fname );
    if (Nestmy_getpwuid(u, pw, fname) < 0) {
        Nestfree_pwd_ent(pw);
        return -1;
    }

    strcpy(uname, pw->pw_name);
    Nestfree_pwd_ent(pw);
    return 0;
}


    //  Group management utilities

int
Nestgetgrpent(group_ent *grp_ent, FILE *fp) {

    char    c,
            tmp_buf[NEST_NAMELEN];

    int        fld = 1,
            i=0;

    c = getc(fp);

    while(!feof(fp)){
        switch(c) {
            case ',':
            case ':':
            case '\n':    tmp_buf[i++] = '\0';

                        switch(fld) {

                            case 1:    strncpy(grp_ent->grp_name, tmp_buf, i);
                                    break;

                            case 2: strncpy(grp_ent->grp_passwd, tmp_buf, i);
                                    break;

                            case 3: sscanf(tmp_buf,"%d",&i);
                                    grp_ent->grp_gid = i;
                                    break;

                            default: strncpy(grp_ent->grp_entries[fld - 4], 
                                             tmp_buf, i);
                        }

                        fld++;

                        if (fld > MAX_USR_ENTRIES + 3) {
                            nest_debug( NEST_DEBUG_GROUP_MNGMNT, 
                                         "Number of users in group %s > %d\n",
                                        grp_ent->grp_name,
                                        MAX_USR_ENTRIES);
                            return 0;
                        }

                        i = 0;

                        if ( c == '\n') {
                            strcpy(grp_ent->grp_entries[fld - 4], "");
                            strcpy(grp_ent->grp_entries[fld - 3], "");//Paranoia
                            return 0;
                        }
                        break;

            default:    tmp_buf[i++] = c;
                        if (i > NEST_NAMELEN) {
                            nest_debug( NEST_DEBUG_GROUP_MNGMNT, 
                                    "Name too big to accomodate!\n");
                            i--;
                        }
        }
        c = getc(fp);
    }
    return -1;
}


int
Nestputgrpent(group_ent *grp_ent, FILE *fp) {

     char    tmp_buf[NEST_NAMELEN];
    char    flag = 'f';

    sprintf(tmp_buf,"%lu",(unsigned long)grp_ent->grp_gid);

    fprintf(fp,"%s:%s:%s:", grp_ent->grp_name, grp_ent->grp_passwd, tmp_buf);
    fflush(fp);

    for(int i=0; i < MAX_USR_ENTRIES; i++) {
        if (!strcmp(grp_ent->grp_entries[i],"")) {
            if (flag == 't') {
                fseek(fp, -1, SEEK_CUR);
            }
            fprintf(fp,"\n");
            return 0;
        }
        flag = 't';
        fprintf(fp,"%s,",grp_ent->grp_entries[i]);
    }
    return 0;
}


int
Nestgetgrpnam(const char *grp_name, group_ent *grp_ent, const char *fname) {

    FILE    *tmp_fp = fopen(fname, "r");

    if (tmp_fp == NULL) {
        nest_perror("fopen");
        return -1;
    }

    while(Nestgetgrpent(grp_ent, tmp_fp) == 0) {
        if (!strcmp(grp_ent->grp_name, grp_name)) {
            fclose(tmp_fp);
            return 0;
        }
    }

    fclose(tmp_fp);
    return -1;
}

int
Nestgetgrpgid(user_id_t grp_gid, group_ent *grp_ent, const char *fname) {

    FILE    *tmp_fp = fopen(fname, "r");

    if (tmp_fp == NULL) {
        nest_perror("fopen");
        return -1;
    }

    while(Nestgetgrpent(grp_ent, tmp_fp) == 0) {
        if ((unsigned)grp_ent->grp_gid == (unsigned)grp_gid) {
            fclose(tmp_fp);
            return 0;
        }
    }

    fclose(tmp_fp);
    return -1;
}


/*
    Converts the "days:hours:minutes" into  double minutes.
*/
double
Nesttimestr_to_double(char *time) {
    char *tmp,
         cor_time[NEST_NAMELEN];
    int i=0,
        t[3];       //days:hours:minutes:seconds

    double result;

    tmp = time;
    *cor_time = 0;

    while (tmp && *tmp) {
        if (*tmp == ':') {
            i++;
        }
        tmp++;
    }

    if (i > 2) {
        return 0;
    }

    while (2-i) {
        strcat(cor_time,"0:");
        i++;
    }

    strcat(cor_time, time);
    sscanf(cor_time, "%d:%d:%d", t, t+1, t+2);

    result = (double) (DAYS(t[0]) + HOURS(t[1]) + MINUTES(t[2]));

    return result;
}



char *NEST_TRANSFER_NAMES[] = {
    "NestTime Request",
    "Client Send File",
    "Client Recv File",
    "File Delete Request",
    "NestBandwidth Statistics Request",
    "List Available Files Request",
    "Filesize Request",
    "Server Terminate Request",
    "Client Terminate Request",
    "Test Message",
    "Test Client to Server NestBandwidth",
    "Test Server to Client NestBandwidth",
    "Request Done"
};

int
NestAccept( int sockfd, struct sockaddr *cliaddr, socklen_t *addrlen ) {
    int accepted;

    while( true ) {
        accepted = accept( sockfd, cliaddr, addrlen );
        if ( accepted < 0 ) {
            if ( errno == EINTR ) {
                continue;
            } else {
                nest_perror( "accept" );
                exit( ERROR_EXIT );
            }
        } else {
            nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS, 
                    "Accepted %d on port %d\n",
                    accepted, sockfd );
            return accepted;
        }
    }
}

int
NestClose( int sockfd ) {
    int ret = close( sockfd );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "close of %d: %s\n", sockfd, strerror( errno ) );
    }
    nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS, 
            "Closed %d: %s\n", sockfd, 
            ( ( ret < 0 ) ? strerror( errno ) : "success" ) );
    return ret;
}

int
NestChangeUser( const char *name ) {
    struct passwd *p;

    LOCK_MUTEX( &nest_common_mutex );
    if( ( p = getpwnam( name ) ) == NULL ) return -1;
    UNLOCK_MUTEX( &nest_common_mutex );

    return setuid( p->pw_uid );
}

uid_t
NestGetuid() {
    return getuid();
}

gid_t
NestGetgid() {
    return getgid();
}

int
NestSeteuid( uid_t uid ) {
    return seteuid(uid);
}

int
NestSetegid( gid_t gid ) {
    return setegid(gid);
}

NestReplyStatus
NestConnect( int sockfd, struct sockaddr *serv_addr, int addrlen ) {
    if ( connect( sockfd, serv_addr, addrlen ) < 0 ) {
        nest_debug( NEST_DEBUG_PROXY, "Can't connect: %s.\n", strerror(errno) );
        return ErrnoToNest( errno );
    }
    return NEST_SUCCESS;
}

void
NestGetsockname( int sockfd, struct sockaddr *localaddr, socklen_t *addrlen ) {
    if ( getsockname( sockfd, localaddr, addrlen ) < 0 ) {
        nest_perror( "getsockname" );
        exit( ERROR_EXIT );
    }
}

pid_t
NestSetpgrp() {
    int ret = setpgrp();
    if ( ret < 0 ) {
        perror( "setpgrp" );
        assert( 0 );
    }
    return ret;
}

pid_t
NestGetpid( ) {
    return getpid();
}

void
NestListen( int sockfd, int backlog ) {
    if ( listen( sockfd, backlog ) < 0 ) {
        nest_perror( "listen" );
        exit( ERROR_EXIT );
    }
}

int
NestsetSocket( int domain, int type, int protocol, int port ) {

    static sockaddr_in sin;
    int sd;
    int on = 1;
    unsigned addr_sz = sizeof( sin );
    
    sd = NestSocket( domain, type, protocol );

    memset( &sin, 0, sizeof( sockaddr_in ) );
    sin.sin_family = domain; 
    sin.sin_addr.s_addr = htonl( INADDR_ANY );
    sin.sin_port = htons( (unsigned short) port );

        // maybe we need to set TCP buffer sizes ??
    NestSetsockopt( sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof( on ) );
    if ( NestBind( sd, (sockaddr *)&sin, sizeof( struct sockaddr_in ) ) != 0 ) {
        return -1;
    }
    NestGetsockname( sd, (sockaddr *)&sin, (socklen_t *) &addr_sz );
    NestListen( sd, BACKLOG );
    return sd;
}

int
NestBind( int sockfd, const struct sockaddr *addr, socklen_t addrlen ) {
    if ( bind( sockfd, addr, addrlen ) < 0 ) { 
        nest_perror( "bind" );
        return -1;    
    }
    return 0;
}

void
NestSetsockopt( int sockfd, int level, int name, void *val, socklen_t len ) {
    if ( setsockopt( sockfd, level, name, (char *)val, len ) < 0 ) {
        nest_perror( "setsockopt" );
    }
}

int
NestSocket( int family, int type, int protocol ) {
    int sockfd;
    if ( ( sockfd = socket( family, type, protocol ) ) < 0 ) { 
        nest_perror( "socket" );
        return -1;    
    }
    return sockfd;
}

// this function has been added for the test bandwidth protocols which
// just send and receive into memory buffers and don't go to disk
// insure was complaining however about uninit'ed mem reads 
void
NestBufferInit( char * buffer, unsigned long size ) {
    memset( buffer, 0, size );
}


// used by nest_get_put and by nest_client_exe
transferInfo *
NestmakeTransferInfo( ) {
    struct transferInfo *trans = (struct transferInfo *)
            NestCalloc( sizeof ( struct transferInfo ) );
    NestGettimeofday( &trans->starttime );
    memcpy(&trans->lastprogress, &trans->starttime, sizeof(struct timeval));
    return trans;
}

int
NestSpipe( int fd[2] ) {
    return( socketpair( AF_UNIX, SOCK_STREAM, 0, fd ) );
}

int 
NestDup( int filedes ) {
    int retval = dup( filedes );
    if ( retval < 0 ) {
        nest_perror( "dup" );
        assert( 0 );
        return ERROR_EXIT;
    }
    nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS,
            "Duped %d from %d.\n", retval, filedes );
    return retval;
}

int 
NestDup2( int filedes, int filedes2 ) {
    int retval = dup2( filedes, filedes2 );
    if ( retval < 0 ) {
        nest_perror( "dup2" );
        assert( 0 );
        return ERROR_EXIT;
    }
    nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS, 
            "Dup2ed %d from %d.\n", retval, filedes );
    return retval;
}

int
NestLogMessage( int fd, char * msg ) {

    struct timeval tv;
    char * now;
    
    NestGettimeofday( &tv );
    now = ctime( (time_t *) &tv.tv_sec );
    if ( NestWriten( fd, now, NestStrlen( now ) ) < 0 ) return ERROR_EXIT;
    if ( NestWriten( fd, "  ", 2 ) < 0 )             return ERROR_EXIT;
    if ( NestWriten( fd, msg, NestStrlen( msg ) ) < 0 ) return ERROR_EXIT;      

    return NEST_SUCCESS;
}

char *
NestFdGets( char *buffer, ssize_t bufsize, int fd ) {

    for( ssize_t i = 0; i < bufsize; i++ ) {
        if ( NestReadn( fd, (void *)&buffer[i], 1 ) < 0 ) return NULL;
        if ( buffer[i] == '\n' ) {
            if ( i + 1 < bufsize ) buffer[i+1] = '\0';
            return buffer; 
        }
    }
    return NULL;
}

NestReplyStatus
NestMergeFdsets( fd_set *dst, fd_set *src, int maxfd ) {

    int m = 0;
    for( int i = 0; i < maxfd; i++ ) {
        if ( NestFd_isset( i, src ) ) {
            NestFd_set( i, dst );
            m++;
        }
    }
    nest_debug( NEST_DEBUG_FDSETS, "Merged fd_sets (%d fds were merged)\n", m );
        
    return NEST_SUCCESS;
}

int
NestSelect( int numfds, fd_set * readfds, fd_set * writefds, 
        fd_set * exceptfds, struct timeval * timeout )
{
    CHECK_PID;
    int n = -1;
    while( n < 0 ) {
    if ( ( n = select( numfds, readfds, writefds, exceptfds, timeout ) ) < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "select: %s (maxfd %d)\n", 
                strerror( errno ), numfds );
        if ( errno == EINTR ) continue;
        int old_level = get_nest_debug_level();
        add_nest_debug_level( NEST_DEBUG_FDSETS ); 
        nest_debug( NEST_DEBUG_FDSETS, "select failed: %s (maxfd %d)\n",
                strerror( errno ), numfds );
        NestFd_print( NEST_DEBUG_FDSETS, readfds, numfds, true );
        set_nest_debug_level( old_level );
        assert( 0 ); 
        return -1;
    }
    }

    nest_debug( NEST_DEBUG_FDSETS, "Ready reads: " );
    NestFd_print( NEST_DEBUG_FDSETS, readfds, numfds );
    nest_debug( NEST_DEBUG_FDSETS, "\nReady writes: " );
    NestFd_print( NEST_DEBUG_FDSETS, writefds, numfds );

    return n;
}

// returns length of file, or -1 if filename does not refer to a valid file
long
NestFilelength( const char *filename ) {
    
    nest_stat buf;

    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISREG( buf.st_mode ) ) {
        return -1;
    }

    return buf.st_size;
}

bool
NestIsDir( const char *filename ) {
    nest_stat buf;

    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISDIR( buf.st_mode ) ) {
        return false;
    }

    return true;
}

bool
NestIsFile( const char *filename ) {
    nest_stat buf;

    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISREG( buf.st_mode ) ) {
        return false;
    }

    return true;
}

bool
NestIsLink( const char *filename ) {
    nest_stat buf;

    if ( ( lstat( filename, &buf ) ) < 0 || ! S_ISLNK( buf.st_mode ) ) {
        return false;
    }

    return true;
}

bool
NestReadLink( const char *path, char *linkto, int length )
{
    int result;
    result = readlink(path,linkto,length);
    if(result>0) {
        linkto[result] = 0;
        return true;
    } else {
        return false;
    }
}

struct hostent*
NestGethostbyname( const char *hostname ) {
    struct hostent *host = NULL;
    struct hostent *tmp  = NULL;
    LOCK_MUTEX( &nest_common_mutex );
    tmp = gethostbyname( hostname );
    if ( tmp == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "gethostbyname %s: %s\n", hostname, strerror( errno ) );
    } else {
        host = (struct hostent *)NestMalloc( sizeof( *tmp ) );
        NestMemcpy( host, tmp, sizeof ( *host ) );
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    return host;
}

#ifndef HAS_SCANDIR

// scandir code swiped off the web . . . 
/*******************************************************************************
*                                                                              *
*                                   Viewmol                                    *
*                                                                              *
*                              S C A N D I R . C                               *
*                                                                              *
*                 Copyright (c) Joerg-R. Hill, December 2000                   *
*                                                                              *
*******************************************************************************/

/* This function is only required for SunOS, all other supported OS
   have this function in their system library */

int scandir(const char *dir, struct dirent ***namelist,
            int (*select)(const struct dirent *),
            int (*compar)(const struct dirent **, const struct dirent **))
{
  DIR *d;
  struct dirent *entry;
  register int i=0;
  size_t entrysize;

  if ((d=opendir(dir)) == NULL)
     return(-1);

  *namelist=NULL;
  while ((entry=readdir(d)) != NULL)
  {
    if (select == NULL || (select != NULL && (*select)(entry)))
    {
      *namelist=(struct dirent **)NestRealloc((void *)(*namelist),
                 (size_t)((i+1)*sizeof(struct dirent *)));
        if (*namelist == NULL) return(-1);
        entrysize=sizeof(struct dirent)-sizeof(entry->d_name)+
            strlen(entry->d_name)+1;
        (*namelist)[i]=(struct dirent *)NestMalloc(entrysize);
        if ((*namelist)[i] == NULL) return(-1);
        memcpy((*namelist)[i], entry, entrysize);
        i++;
    }
  }
  if (closedir(d)) return(-1);
  if (i == 0) return(-1);
  if (compar != NULL) {
    int (*cmp_fn) (const void *, const void *);
    cmp_fn = (int (*) (const void *, const void *))cmp_fn;
    qsort((void *)(*namelist), (size_t)i, sizeof(struct dirent *), cmp_fn);
  }
    
  return(i);
}

int alphasort(const struct dirent **a, const struct dirent **b)
{
  return(strcmp((*a)->d_name, (*b)->d_name));
}


#endif


#ifdef HAS_SCM_RIGHTS

const char*
NestInet_ntop( int family, const void*addrptr, char*strptr, ssize_t len) {
    LOCK_MUTEX( &nest_common_mutex );
    const char * p;
    if ( ( p = inet_ntop( family, addrptr, strptr, len ) ) == NULL ) {
        nest_perror( "inet_ntop" );
        exit (ERROR_EXIT );
    }
    const char *ret = NestStrdup( p );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;    
}

// this functions receives a fd from the sd param and returns the fd 
int 
NestReceiveFd( int sd ) {
    
    struct iovec vector;    /* file name for the passed file - phony */
    struct msghdr msg;        /* full message */
    static struct cmsghdr *cmsg = NULL;
                            /* control message which includes the fd */
    int fd;

        // set up the iovec to read in the filename
    char filename[NEST_MESSAGESIZE];
    vector.iov_base = filename;
    vector.iov_len  = NEST_MESSAGESIZE;

        // the message we're expecting to receive
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &vector;
    msg.msg_iovlen = 1;

        // the control message
    if ( cmsg == NULL ) {
            // it will be NULL the first time the function is called
        cmsg = (struct cmsghdr *)
            NestMalloc( sizeof(struct cmsghdr) + sizeof(fd) );
        cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(fd);
    }

        // connect the message and the control message
    msg.msg_control = cmsg;
    msg.msg_controllen = cmsg->cmsg_len;
    
    if ( recvmsg( sd, &msg, 0 ) <= 0 ) {
        nest_perror( "recvmsg" );
        assert( 0 );
        return ERROR_EXIT;
    }

        // grab the file descriptor
    memcpy( &fd, CMSG_DATA( cmsg ), sizeof( fd ) );
    nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS, "Received fd %d\n",fd);
    return fd;
}

// this function passes the param fd to the socket sd
// returns OK or ERROR_EXIT on failure
NestReplyStatus
NestSendFd( int sd, int fd ) {
    struct iovec   vector;     /* some data to pass with fd */
    struct msghdr  msg;         /* the complete message */
    static struct cmsghdr *cmsg = NULL;
                            /* the control message which includes the fd */

        // send a bogus filename to make the code work
        // I can't figure out why this is necessary
    char filename[NEST_MESSAGESIZE];
    memset( filename, 0, NEST_MESSAGESIZE );
    NestSnprintf( filename, sizeof(filename), "%s", "foo" );
    vector.iov_base = filename;
    vector.iov_len  = NestStrlen( filename ) + 1;

        // assemble first part of message
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &vector;
    msg.msg_iovlen = 1;

        // assemble the control message
    if ( cmsg == NULL ) {
            // cmsg is NULL the first time the function is called
        cmsg = (struct cmsghdr *)NestMalloc(
                sizeof(struct cmsghdr) + sizeof(fd) );
        cmsg->cmsg_len   = sizeof( struct cmsghdr) + sizeof( fd );
        cmsg->cmsg_level = SOL_SOCKET;
        cmsg->cmsg_type  = SCM_RIGHTS;
    }

        // copy the file descriptor onto the end of the control message
    memcpy( CMSG_DATA( cmsg ), &fd, sizeof( fd ) );

        // connect the message and the control message
    msg.msg_control = cmsg;
    msg.msg_controllen = cmsg->cmsg_len;

    if ( sendmsg( sd, &msg, 0 ) != (int) vector.iov_len ) {
        nest_perror( "sendmsg" );
        return ErrnoToNest( errno );
    }

    return NEST_SUCCESS;
}

#else /* HAS_SCM_RIGHTS */

int
NestReceiveFd( int sd ) {

    struct strrecvfd recvfd;

    if( ioctl(sd, I_RECVFD, &recvfd) < 0)
        return -1;
    else
        return recvfd.fd;

}

NestReplyStatus
NestSendFd( int sd, int fd ) {

    if( ioctl(sd, I_SENDFD, fd) < 0){
        nest_perror( "ioctl" );
        return ErrnoToNest( errno );
    }

    return NEST_SUCCESS;
}

#endif /* HAS_SCM_RIGHTS */

/*
returns -1 on error
*/
ssize_t
NestWriten( int fd, const void *vptr, ssize_t n ) {
    ssize_t nleft;
    ssize_t nwritten;
    const char *ptr;

    /*  this is old stupidness from when I did test sends and receives
        // if the passed descriptor is negative, then NestWriten is a no-op
    if ( fd < 0 ) return n;
    */

    ptr = (char *)vptr;
    nleft = n;

    while ( nleft > 0 ) {
        if ( ( nwritten = NestWrite( fd, ptr, nleft ) ) < 0 ) {
            return -1;
        }
        nleft -= nwritten;
        ptr += nwritten;
    }

    return n;
}

ssize_t
NestWrite( int fd, const void *vptr, ssize_t n ) {
    ssize_t nwritten;

    if ( ( nwritten = write( fd, vptr, n ) ) < 0 ) {
        if ( errno == EINTR ) {
            return 0;
        } else { 
            return -1;
        }
    } 

    return nwritten;
}


// returns -1 on error
ssize_t
NestReadn( int fd, void *vptr, ssize_t n ) {
    ssize_t nleft;
    ssize_t nread;
    char *ptr;

         // if the passed descriptor is negative, then NestReadn is a no-op
    if ( fd < 0 ) return n;

    ptr = (char *) vptr;
    nleft = n;
    while( nleft > 0 ) {
        if ( ( nread = NestRead( fd, ptr, nleft ) ) < 0 ) {
            return -1;
        } else if ( nread == 0 ) {
                // is this right?  isn't it possible just no bytes right now?
            break;        // EOF    
        }

        nleft -= nread;
        ptr   += nread;
    }

    return( n - nleft );
}

ssize_t
NestRead( int fd, void *vptr, ssize_t n ) {
    ssize_t nread;

    while ( ( nread = read( fd, vptr, n ) ) < 0 ) {
        if ( errno == EINTR ) {
            continue;    
        } else { 
            return -1;
        }
    }

    return nread;
} 

int
NestChmod( const char *pathname, mode_t mode ) {
    int reply = chmod(pathname, mode);

    if (reply < 0) {
        nest_perror("chmod");
    }

    return reply;
}

int
NestFchmod( int fd, mode_t mode ) {
    int reply = fchmod(fd, mode);

    if (reply < 0) {
        nest_perror("fchmod");
    }

    return reply;
}

int
NestChown ( const char *pathname, uid_t uid, gid_t gid ) {
    int reply = chown(pathname, uid, gid);

    if (reply < 0) {
        nest_perror("chown");
    }

    return reply;
}

int
NestLchown ( const char *pathname, uid_t uid, gid_t gid ) {
    int reply = lchown(pathname, uid, gid);

    if (reply < 0) {
        nest_perror("lchown");
    }

    return reply;
}

int
NestFchown( int fd, uid_t uid, gid_t gid ) {
    int reply = fchown(fd, uid, gid);

    if (reply < 0) {
        nest_perror("fchown");
    }

    return reply;
}


/**
takes a sin and return the hostname
*/
char *
NestDns_lookup( const struct sockaddr_in *from ) {
    
        // turn this off
    //static char *deprecated = "dns lookup turned off";
    //return deprecated; 

    LOCK_MUTEX( &nest_common_mutex );
    struct hostent * mach;
    char *ret = NULL;
    mach = gethostbyaddr( (char *)&from->sin_addr,
        sizeof( struct in_addr ), AF_INET );
    if (mach == NULL) {
        const char *error_string = NULL;
        #ifdef HAS_HSTRERROR
            error_string = hstrerror( h_errno );
        #else
            error_string = strerror( h_errno );
        #endif
        nest_warn( __FILE__, __FUNCTION__, "gethostbyaddr %s: %s\n",
                (char *)&from->sin_addr, error_string );    
        ret = NestStrdup( (char *)&from->sin_addr );
    } else {
        ret = NestStrdup( mach->h_name );
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
}

char *
NestGetIP( struct in_addr addr ) {
    LOCK_MUTEX( &nest_common_mutex );
    char *ret = NestStrdup( inet_ntoa( addr ) );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
}

NestReplyStatus 
NestGetcwd( char *buffer, size_t size ) {
    LOCK_MUTEX( &nest_common_mutex );
    char *reply = getcwd(buffer, size);
    UNLOCK_MUTEX( &nest_common_mutex );

    if (reply == NULL) {
        nest_perror("getcwd");
        return ErrnoToNest( errno );
    } else {
        return NEST_SUCCESS;
    }
}


void
NestChdir( const char *pathname ) {
    if ( chdir( pathname ) < 0 ) {
        nest_die( "chdir to %s failed: %s\n", pathname, strerror(errno));
    }
}

int
Nestsafe_Chdir( const char *pathname ) {
    return chdir( pathname );
}

int
NestMkdir( const char *pathname, mode_t mode ) {
    int reply = mkdir(pathname, mode);

    if (reply < 0 && errno != EEXIST ) {
        nest_warn( __FILE__, __FUNCTION__,"mkdir of %s failed: %s\n", 
                pathname, strerror( errno ) );
    }

    return reply;
}

int
NestRmdir( const char *pathname ) {
    int reply = rmdir(pathname);

    if (reply < 0 && errno != ENOENT) {
        nest_perror("rmdir");
    }

    return reply;
}

int  
NestScandir( const char *path,
              struct dirent ***namelist,
              int (*select_fn) (const struct dirent *),
              int (*cmp_fn) (const struct dirent **, const struct dirent **)) {

    int (*actual_cmp_fn) (const void *, const void *);
    actual_cmp_fn = (int (*) (const void *, const void *))cmp_fn;

    return scandir( path, namelist, select_fn, actual_cmp_fn );
}

DIR  *
NestOpendir(const char *name) {
    return opendir( name );
}

struct dirent *
NestReaddir(DIR *dir) {
    return readdir( dir );
}


/**
* takes a pathname and keeps moving forward through it until it finds the
* final '/' and returns a ptr to the filename
*/
char* 
NestExtractFilename( char *pathname ) {
    char *ptr, *tmp;

    ptr = pathname;
    tmp = strchr( pathname, '/' );

    while( tmp != NULL ) {
        ptr = ++tmp;
        tmp = strchr( ptr, '/' );
    }    

    return ptr;
}


void
NestClearFileFlags( int fd, int flags ) {
    int val;
    if ( ( val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
    val &= ~(flags);
    if ( fcntl( fd, F_SETFL, val ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
}

void 
NestSetFileFlags( int fd, int flags ) {
    int val;
    if ( ( val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
    val |= flags;
    if ( fcntl( fd, F_SETFL, val ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
}

int
NestShutdown( int sd, int how ) {
    //CHECK_PID;
    int ret = shutdown( sd, how );
    if ( ret != 0 ) {
        if ( errno != ENOTCONN ) {
            nest_warn( __FILE__, __FUNCTION__, "shutdown of %d failed: %s\n", sd, strerror( errno ) );
        }
    } else {
        nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_ADMIN, 
        //nest_printf( 
                "NestShutdown %d: %s\n", sd, 
                ( (ret<0) ? strerror( errno ) : "success" ) );
    }
    return ret;
}

int
NestCreat( const char *pathname, mode_t mode ) {

    return creat( pathname, mode );
}

int
NestOpen( const char* pathname, int flags, mode_t mode ) {
    int fd = open( pathname, flags, mode );
    /*
    if ( fd < 0) {
        nest_warn( __FILE__, __FUNCTION__, "open %s: %s\n", pathname, strerror( errno ) );
    }
    */
    nest_debug( NEST_DEBUG_FDSETS | NEST_DEBUG_FD_LEAKS,
        "Opened %d for %s: %s\n", fd, pathname,
            ( ( fd < 0 ) ? strerror( errno ) : "success" ) );
    return fd;
}

off_t
NestLseek( int fd, off_t offset, int whence ) {
    off_t retval = lseek( fd, offset, whence );
    if ( retval == (off_t)-1 ) nest_perror( "lseek" );
    return retval;
}

int
NestRename( const char *oldpath, const char *newpath ) {
    int retval = rename(oldpath, newpath);

    if (retval < 0) {
        char out[3*NEST_MAXPATHLEN];
        NestSnprintf(out,
                     3*NEST_MAXPATHLEN,
                     "rename - %s %s",
                     oldpath,
                     newpath);
        out[3*NEST_MAXPATHLEN - 1] = 0;
        nest_perror(out);
    }

    return retval;
}

int
NestStat( const char *file, nest_stat *st ) {
    int retval = stat(file, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "stat of %s: %s\n",
                    file,
                    strerror( errno ) );
    }

    return retval;
}

int
NestFstat( int fd, nest_stat *st ) {
    int retval = fstat(fd, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "fstat %s\n",
                    strerror( errno ));
    }

    return retval;
}

int
NestLstat( const char *file, nest_stat *st ) {
    int retval = lstat(file, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "lstat of %s: %s\n",
                    file,
                    strerror( errno ) );
    }

    return retval;
}


int NestStatFS( const char *path, nest_fsstat *s ) 
{
#if defined(HAS_SYS_STATFS_H)
    return statfs(path,s);
#elif defined(HAS_SYS_STATVFS_H)
    return statvfs(path,s);
#else
#error Must have either statvfs or statfs.
#endif
}

int
NestSymlink( const char *oldpath, const char *newpath ) {

    return symlink( oldpath, newpath );
}

int
NestReadlink( const char *path, char *buf, size_t buf_len ) {

    int retval;

    if (!buf || !buf_len) {
        return -1;
    }

    retval = readlink( path, buf, buf_len );

    /*
        stupid readlink doesn't put a '\0' at the end of the string
        I hate readlink!!
    */

    if (retval < 0) {
        *buf = 0;
        return -1;
    }

    buf[retval] = 0;
    return 0;
}

int
NestTruncUnlink( char *file ) {

    /*
        first O_TRUNC the file and then unlink it

        the stupid quota system doesn't update the usage statistics,
        sometimes, weirdly, if you just do a unlink()

        this was actually found out accidentally :)
        and this O_TRUNC idea seems to solve problems!
    */

    int tmp_fd = NestOpen( file, O_WRONLY | O_TRUNC );

    if (tmp_fd >= 0) {
        NestClose(tmp_fd);
    }

    int retval = unlink(file);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "unlink %s: %s\n",
                    file,
                    strerror( errno ));
    }

    return retval;
}

int
NestUnlink( const char *file ) {

    int retval = unlink(file);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "unlink %s: %s\n",
                    file,
                    strerror( errno ));
    }

    return retval;
}

int
NestSync( ) {
    // Some sync()'s return a value, some void.
    // We'll just ignore them all, and assume success.
    // Ugly, yes, but it works.  -NRL
    sync();
    return 0;
}

int
NestGetFileflags( int transfer_type ) {
    if ( transfer_type == CLIENT_RECV_FILE ) {
        return READ_FLAGS;
    } else {
        assert( transfer_type == CLIENT_SEND_FILE );
        return WRITE_FLAGS;
    }
}

char *
NestGetUserName( char *user ) {
    static char *result = NULL;
    static bool init = false;
    LOCK_MUTEX( &nest_common_mutex );
    if ( ! init ) {
        init = true;
        struct passwd *pass = getpwuid( geteuid() );
        if(pass) {
            result = NestStrdup( pass->pw_name );
        } else {
            result = 0;
        }
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    if ( result ) {
        strcpy( user, result );
    }
    return result;
}

void
NestGettimespec( struct timespec *ts ) {
  struct timeval tv;
  NestGettimeofday( &tv );

  ts->tv_sec = tv.tv_sec;
  ts->tv_nsec = tv.tv_usec*1000;
}

void
NestGettimeofday( struct timeval *time ) {
    if ( gettimeofday( time, 0 ) < 0 ) {
        nest_perror( "gettimeofday" );
        assert( 0 );
    }
}

time_t
NestTime(time_t *tsecs) {
    time_t retval;
    if ((retval = time(tsecs)) == ((time_t)-1)) {
        nest_perror("time");
        assert( 0 );
    }
   return retval;
}

void *
NestMemset( void *dst, int c, size_t n ) {
    return memset( dst, c, n );
}

void *
NestMemcpy( void *dst, const void *src, size_t n ) {
    return memcpy( dst, src, n );
}

void
NestFreeNormal( void *ptr ) {
    assert( ptr != NULL );
    LOCK_MUTEX( &nest_malloc_mutex );
    free( ptr );
    UNLOCK_MUTEX( &nest_malloc_mutex );
}

void
NestFreeDebug( void *ptr, char *func, int line ) {
    nest_debug( NEST_DEBUG_MALLOC, "Freeing 0x%x at %s:%d\n", ptr, func, line );
    NestFreeNormal( ptr );
}

void *
NestCallocNormal(size_t size) {
    void *retptr = NestMalloc(size);
    memset( retptr, 0, size );
    return retptr;
}                                                        

void *
NestCallocDebug(size_t size, char *func, int line ) {
    void *ptr = NestCallocNormal( size );
    nest_debug( NEST_DEBUG_MALLOC, 
            "Calloced %ld bytes at 0x%x from %s:%d.\n", size, ptr, func, line);
    return ptr; 
}

void *
NestMallocDebug( size_t size, char *func, int line ) {
    void *vptr = NestMallocNormal( size );
    nest_debug( NEST_DEBUG_MALLOC, 
        "Mallocced %ld at 0x%x from %s:%d.\n", size, vptr, func, line );
    return vptr;
}

void *
NestMallocNormal(size_t size){
  assert(size > 0);
  LOCK_MUTEX( &nest_malloc_mutex );
  void *retptr = malloc(size);
  UNLOCK_MUTEX( &nest_malloc_mutex );
  if(retptr == NULL){
    nest_die( "failed to malloc %d bytes: %s\n", size, strerror( errno ) );
  }
  return retptr;
}

void *
NestReallocNormal(void *ptr, size_t size) {
    LOCK_MUTEX( &nest_malloc_mutex );
    void *retptr = realloc( ptr, size );
    UNLOCK_MUTEX( &nest_malloc_mutex );
    return retptr;
}

void *
NestReallocDebug( void *ptr, size_t size, char *func, int line ) {
    void *vptr = NestReallocNormal( ptr, size );
    nest_debug( NEST_DEBUG_MALLOC,
            "Reallocced 0x%x to grow 0x%x by %d %s:%d\n",
            vptr, ptr, size, func, line );
    return vptr;
}

char *
NestStrdupDebug( const char *src, char *func, int line ) {
    char *ret = NestStrdupNormal( src );
    nest_debug( NEST_DEBUG_MALLOC, 
        "Strduped %d at 0x%x from %s:%d.\n", 
            strlen(ret), (void*)ret, func, line );
    return ret;       
}

char *
NestStrdupNormal( const char *src ) {
    assert( src != NULL );
    LOCK_MUTEX( &nest_strdup_mutex );
    char *ret = strdup( src );
    UNLOCK_MUTEX( &nest_strdup_mutex );
    if ( ret == NULL ) {
        nest_die( "Couldn't strdup %s: %s\n", src, strerror( errno ) );
    }
    return ret;
}

float
NestBandwidth( long bytes, 
        struct timeval *start, struct timeval *end, const char *function ) {
    return NestBandwidth( bytes, NestTimediff( start, end, function ) );
}

float
NestBandwidth( long bytes, long micros ) {

    if ( bytes <= 0 || micros <= 0 ) {
        if ( bytes < 0 || micros < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, "Asked to compute negative bw: %ld bytes, %ld micros\n",
                bytes, micros );
            return -1;
        } else {
            return 0;
        }
    }

    /*
    float secs;
    float MBs;
    secs = (float) micros / ( THOUSAND * THOUSAND );
    MBs  = (float) bytes / ( 1024 * 1024 );

    return MBs / secs;  
    */
        // do one less division, compiler should do constant for us
    float constant = ( (float)THOUSAND * THOUSAND ) / ( 1024.0 * 1024.0 );
    float bw = ( (float)bytes / micros ) * constant;
    return bw;
}


/* returns microseconds between the interval */
long
NestTimediff( struct timeval *start, struct timeval *end, const char *function ) {

    long long micros = (long long)end->tv_sec * MILLION + end->tv_usec;
    micros -= ((long long)start->tv_sec * MILLION + start->tv_usec);
    /*
    long millis = micros / THOUSAND;

    if ( millis == 0 && micros != 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
        "Change NestTimediff to return micros:  %ld != %ld / 1000\n",
                millis, micros );
    }
    return ( millis );
    */
    long ret_micros = (long)micros;
    if ( ret_micros < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Returning negative timediff from %s: %ld\n"
                "\tstart %ld.%ld\n"
                "\tend   %ld.%ld\n", 
                function,
                ret_micros,
                start->tv_sec, start->tv_usec,
                end->tv_sec,   end->tv_usec );
    }
    return ret_micros; 
}

int 
NestSetenv( const char *name, const char *value, int overwrite ) {
#ifdef HAS_SETENV
    LOCK_MUTEX( &nest_common_mutex );
    int ret = setenv( name, value, overwrite );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
#else
    char command[NEST_MESSAGESIZE];
    NestSnprintf( command, NEST_MESSAGESIZE, "export %s=%s\n",
            name, value );
    if ( overwrite ) {
        nest_warn( __FILE__, __FUNCTION__, "Overwrite ignored in setenv\n" );
    }
    nest_warn( __FILE__, __FUNCTION__, "Not setting env %s -> %s", 
            name, value );
    return 1;
#endif
}


void 
NestSignal( int signum, void (*handler)(int)) {
    bool ok = true;
    if ( signal( signum, handler ) == SIG_ERR ) {
        ok = false;
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    if ( ! ok ) {
        nest_perror( "signal" );
        exit( ERROR_EXIT );
    }
}

void
NestNanosleep( unsigned long nanos )
{
#ifdef HAS_NANOSLEEP
    struct timespec w,r;
    int result;

    w.tv_sec = 0;
    w.tv_nsec = nanos;

    while(w.tv_nsec>BILLION) {
        w.tv_sec++;
        w.tv_nsec-=BILLION;
    }

    while(1) {
        result = nanosleep(&w,&r);
        if(result<0) {
            w = r;
        } else {
            break;
        }
    }
#else
    nest_warn( __FILE__, __FUNCTION__, "This platform does not have nanosleep.\n" );
#endif
}

// fdset utilities
int 
NestFd_print( int debug_level, fd_set *fds, int maxfd ) {
    return NestFd_print( debug_level, fds, maxfd, false );
}

int
NestFd_print( int debug_level, fd_set *fds, int maxfd, bool use_fstat ) {

    struct stat sbuf;

    debug_level |= NEST_DEBUG_FDSETS;

    for( int i = 0; i < maxfd; i++ ) {
        if ( NestFd_isset( i, fds ) ) {
            nest_debug( debug_level, "%d ", i );
            if ( use_fstat ) {
                int ret = fstat( i, &sbuf );
                if ( ret < 0 ) {
                    nest_warn( __FILE__, __FUNCTION__, "fstat of %d failed: %s\n", i, strerror(errno));
                }
            }
        }
    }
    nest_debug( debug_level, "\n" );
    return 0;
}

void
NestFd_set( int fd, fd_set *fds ) {
    //CHECK_PID;
    nest_debug( NEST_DEBUG_FDSETS, "Setting %d in fdset 0x%x\n", fd, fds );
    FD_SET( fd, fds );
}

void
NestFd_clear( int fd, fd_set *fds ) {
    CHECK_PID;
    nest_debug( NEST_DEBUG_FDSETS, "Clearing %d in fdset 0x%x\n", fd, fds );
    FD_CLR( fd, fds );
}

int
NestFd_isset( int fd, fd_set *fds ) {
    return( FD_ISSET( fd, fds ) );
}

int
NestAdd_id_to_handle(NestLotHandle   *handle,
                     rsv_id_t rsv_id) {
    if (!handle) {
        return -1;
    }
    handle->rsv_id = rsv_id;
    return 0;
}

int
NestGet_id_from_handle(NestLotHandle   *handle,
                       rsv_id_t *rsv_id) {

    if (!handle || !rsv_id) {
        if (rsv_id) {
            *rsv_id = 0;
        }
        return -1;
    }

    *rsv_id = handle->rsv_id;
    return 0;
}

int
NestHandle_to_string(NestLotHandle *handle,
                     char *str,
                     int len) {
    if (!handle || !str) {
        return -1;
    }

    rsv_id_t    tmp;

    if (NestGet_id_from_handle( handle, &tmp ) < 0) {
        return -1;
    }

    NestSnprintf( str, len, "%lu", tmp );
    str[len - 1] = 0;

    return 0;
}


int
NestString_to_handle(NestLotHandle *handle,
                     const char *str) {
    if (!handle || !str) {
        return -1;
    }

    rsv_id_t    tmp;
    sscanf( str, "%lu", &tmp );

    if (NestAdd_id_to_handle( handle, tmp ) < 0) {
        return -1;
    }

    return 0;
}

double
NestGetLotDuration(char *duration){
    char   *time;
    double days;
    double hours;
    double minutes;

    time    = NestStrdup(duration);
    days    = atoi( strtok(duration,":") );
    hours   = atoi( strtok(NULL,":") );
    minutes = atoi( strtok(NULL,":") );

    return days*24*60 + hours*60 + minutes;

}

double  
NestStrLotList_to_double(const char *lot_str){
    char    *str_lot_list = strdup(lot_str);
    double  lot_allocation = 0,temp_allocation;
    char    *token = strtok(str_lot_list, ":");

    
    while(token) {
        token = strtok(NULL, ":");
        if(token) {
            sscanf(token,"%lf", &temp_allocation);
            if(temp_allocation < 0) {
                lot_allocation = 0;
                break;
            }
            lot_allocation += temp_allocation;
        }
    }
    
    return lot_allocation;
}

int
NestFileInfo_to_string(NestFileInfo *file,
                       char *msg,
                       int len) {

    if (!msg || !len) {
        return -1;
    }

    if (!file) {
        *msg = 0;
        return 0;
    }

    return NestSnprintf(msg,
                    len,
                    "%s %ld %s",
                    file->name,
                    file->size,
                    file->LotList);
}

int
NestString_to_NestFileInfo(NestFileInfo *file,
                           const char *msg) {

    if (!file || !msg) {
        return -1;
    }

    char    tmp_size[NEST_NAMELEN];

    if (sscanf( msg,
                "%s %s",
                file->name,
                tmp_size) != 2) {
        return -1;
    }

    int src = NestStrlen(file->name) + NestStrlen(tmp_size) + 1,
        dest = 0;


    while ((msg[src] != '\n') && (src < NEST_MAXPATHLEN)) {
        file->LotList[dest++] = msg[src++];
    }

    file->LotList[dest] = 0;
    sscanf( tmp_size, "%ld", &file->size );

    return 3;
}


NestReplyStatus
NestAdd_ACL( NestACLList **result,
             const char *name,
             int is_user,
             nest_access_rights_t right ) {

    if (!result || !name) {
        return NEST_BAD_PARAMETERS;
    }
    NestACLList *cur;

    for( cur = *result; cur; cur = cur->next) {
        if (!strcmp(name, cur->acl.name) &&
            (is_user == cur->acl.is_user)) {
            cur->acl.rights |= right;
            return NEST_SUCCESS;
        }
    }

    cur = (NestACLList *) NestMalloc (sizeof(NestACLList));
    if (!cur) {
        return NEST_OUT_OF_RESOURCES;
    }

    strcpy(cur->acl.name, name);
    cur->acl.is_user = is_user;
    cur->acl.rights = right;

    cur->next = (*result);
    *result = cur;

    return NEST_SUCCESS;
}

/*
    Helper function to add an entry (user) to a list (members)
*/

NestReplyStatus
NestAdd_entry_to_list(string &new_members,
                      string members,
                      string user) {

    unsigned int i;
    char flag = 's';         //  s == start

    new_members = members;
    if (new_members.find(user)
        != string::npos) {
        return NEST_ALREADY_EXISTS;
    }

    for(i = 0; i < new_members.length(); i++) {
        if (new_members[i] == '{') {
            flag = 'o';  //  o == open braces
            continue;
        }
        if (new_members[i] != ' ' &&
            new_members[i] != '\t' &&
            new_members[i] != '\n' &&
            new_members[i] != '}') {
            if (flag == 'f') {
                continue;
            }
            if (flag == 'o') {
                flag = 'f';  //  f == found an alphabet after '{'
            }
            continue;
        }
        if (new_members[i] == '}') {
            break;
        }
    }

    string append_str;

    switch (flag) {
        case 's':   nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Improperly formed list!\n" );
                    return NEST_UNKNOWN_ERROR;
        case 'o':   /*  First member */
                    append_str = "\"" + user + "\"}";
                    break;
        case 'f':   /* Not the first member */
                    append_str = ",\"" + user + "\"}";
                    break;
    };

    new_members.replace(i, 1, append_str);
    return NEST_SUCCESS;
}

/*
    Helper function to delete an entry (user) from a list (members)
*/

NestReplyStatus
NestDelete_entry_from_list(string &new_members,
                           string members,
                           string user) {
    string tmp_user;
    unsigned int pos;

    new_members = members;
    if (new_members.find(user)
        == string::npos) {
        return NEST_USER_NOT_FOUND;
    }
    tmp_user = ",\"" + user + "\"";
    pos = new_members.find(tmp_user);
    if (pos == string::npos) {
        tmp_user = "\"" + user + "\",";
        pos = new_members.find(tmp_user);
        if (pos == string::npos) {
            tmp_user = "\"" + user + "\"";
            pos = new_members.find(tmp_user);
            if (pos == string::npos) {
                return NEST_UNKNOWN_ERROR;
            }
        }
    }

    new_members.erase(pos, tmp_user.length());
    return NEST_SUCCESS;
}

/*
Check to see if this path corresponds to a device.
If so, fill in that device name and return true.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
/*
int NestPathIsDevice( const char *path, char *device )
{
    struct fstab *f;

    setfsent();
    f = getfsfile(path);
    endfsent();

    if(f) {
        strncpy(device, f->fs_spec, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
*/
int 
NestPathIsDevice( const char *path, char *device ) {
    FILE *file;
    char mpath[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s",device,mpath)==2) {
            if(!strcmp(path,mpath)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

/*
Given a path name, try to determine the device containing it.
*/

int 
NestPathToDevice( const char *path, char *device ) {
    char cwd[NEST_MAXPATHLEN];

    strcpy(cwd,path);

    while (cwd[0]) {
        int i;

        if(NestPathIsDevice(cwd,device)) break;

        for(i=NestStrlen(cwd) - 1; i >= 0; i--) {
            if (cwd[i] == '/') {
                cwd[i] = 0;
                break;
            }
        }
    }

    if(cwd[0]) {
        return 1;
    } else {
        return NestPathIsDevice("/",device);
    }
}

/*
Return the pathname where this device is mounted.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
/*
int NestDeviceToPath( const char *device, char *path )
{
    struct fstab *f;

    setfsent();
    f = getfsspec(device);
    endfsent();

    if(f) {
        strncpy(path, f->fs_file, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
*/
int 
NestDeviceToPath( const char *device, char *path ) {
    FILE *file;
    char tempdev[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s",tempdev,path)==2) {
            if(!strcmp(device,tempdev)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

/*
Return the filesystem type of this device.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
/*
int NestFSType( const char *device, char *type )
{
    struct fstab *f;

    setfsent();
    f = getfsspec(device);
    endfsent();

    if(f) {
        strncpy(type, f->fs_vfstype, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
*/

int 
NestFSType( const char *device, char *type ) {
    FILE *file;
    char tempdev[NEST_MAXPATHLEN];
    char path[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s %s",tempdev,path,type)==3) {
            if(!strcmp(device,tempdev)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

long
NestElapsedTime( struct timeval *start ) {
    struct timeval now;
    NestGettimeofday( &now );
    return NestTimediff( start, &now, __FUNCTION__ );
}


struct timeval *
NestNextTimeout( int seconds, struct timeval *last ) {

    static struct timeval now;
    static struct timeval timeout;
    long micros;

    NestGettimeofday( &now );

        // how long since the last one and how long til the next one
    micros = NestTimediff( last, &now, __FUNCTION__ );
    if ( micros < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "%s.%s: negative micros: %ld (will set to zero).\n",
                __FILE__,
                __FUNCTION__,
                micros );
        micros = 0;
    }

    micros = ( seconds * 1000 ) - micros;
    nest_debug( NEST_DEBUG_SERVER, "%ld millis until next maintenance\n", 
            ( micros / THOUSAND ) );

    if ( micros <= 0 ) {
            // overdue already
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;
    } else {
            // compute remaining time by subtracting elapsed from max
        timeout.tv_sec = micros / ( THOUSAND * THOUSAND );
        timeout.tv_usec = micros % ( THOUSAND * THOUSAND );
    }

    return &timeout; 
}

char *
NestUname( ) {     
    static struct utsname *buf = NULL;
    if ( ! buf ) { //initialize
        buf = (struct utsname*)NestMalloc(sizeof(*buf));
        uname(buf);
    }
    return buf->nodename; 
}

void
NestParseLine( char *line, char ***argv, int *argc ) {
    char tmp_arg[NEST_MAXPATHLEN];
    int i,j;
    
    if(*argv){
        for(i=0; i<(*argc); i++)
            if((*argv)[i])
                NestFree((*argv)[i]);
        NestFree(*argv);
    }
    *argc=0;
 
    *argv = (char**) NestCalloc(NEST_MAXARGS*sizeof(char*));
    if(!*argv) return;
    
    for(i=0; i<NEST_MAXARGS; i++)
        (*argv)[i] = 0;
 
    i=0;
    bool quoted = false;
    for( (*argc)=0; (*argc)<NEST_MAXARGS; (*argc)++ ) {
        j=0;
        while(line[i] != 0 && line[i] != '\n' && 
              (line[i] == ' ' || line[i] == '\t'))
            i++;
        if(line[i] == '"'){
            i++;
            tmp_arg[j] = '"';
            j++;
            quoted = true;
        }
        if(quoted){
            while(line[i] != 0 && line[i] != '\n' && line[i] != '"'){
                tmp_arg[j] = line[i];
                i++;
                j++;
            }
            if(line[i] == '"') 
                i++;
        } else {
            while(line [i] != 0 && line[i] != '\n' && 
                  line[i] != ' ' && line[i] != '\t'){
                tmp_arg[j] = line[i];
                i++;
                j++;
            }
        }
        if(quoted){
            tmp_arg[j] = '"';
            j++;
        }
        tmp_arg[j]=0;
        quoted = false;
 
        if(!strcmp(tmp_arg,"")){
            (*argv)[*argc] = 0;
            break;
        }
        else{
            (*argv)[*argc] = (char*) NestMalloc(strlen(tmp_arg)+1);
            strcpy((*argv)[*argc],tmp_arg);
        }
    }
 
}

char *
NestGetModifiedIdentity( const char *identity )
{
    /* 
     * This function creates a physical file name from the
     * global identity which will be used for creating 
     * physical file names and as key in the classad 
     * collections. To make sure special chars like ' ','/',
     * '=','"',etc. in the identity are not mis-interpreted,
     * they are all converted to '_'. Change the following code
     * suitably to include more special chars.
     */

    if(!identity) {
        return NULL;
    }

    char *file_entry = (char*) NestMalloc( strlen(identity)+1 );
    int  i = 0;

    const char *cursor = identity;
    while(*cursor) {
        switch(*cursor) {
            case '"':
                cursor++;
                continue;                 
            case ' ':
            case '/':
            case '=':
                *(file_entry+i) = '_';
                break;
            default:
                *(file_entry+i) = *cursor;
                break;
        }
        cursor++;
        i++;
    }
    *(file_entry+i) = 0;

    return file_entry; 
}


char *
NestGetUserFromIdentity( const char *identity )
{
    if(!strcmp(identity,NEST_EVERYBODY) ||
       !strcmp(identity,NEST_NOBODY) ||
       !strcmp(identity,NEST_NO_GROUP))
        return (char*)identity;
            
    char *user = NULL;
    int  state = Nestset_grid_mapfile( GridMapFile );
    
    if ( state < 0) {
        nest_warn( __FILE__, __FUNCTION__,  
                    "Couldn't read grid-mapfile at %s: Can't lookup %s!\n",
                    GridMapFile,
                    identity );
        return user;
    }

    grid_mapfile_entry  *entry;
    while( (entry = Nestget_grid_mapfile_entry(state)) != NULL) {
        char *tmp_identity = NestGetModifiedIdentity( entry->identity );
        if( !strcmp(identity, tmp_identity) ) {
            user = entry->name;
            if (tmp_identity) {
				nest_debug( NEST_DEBUG_AUTH, "Matched mapfile user '%s' / '%s'\n",
							entry->identity, entry->name );
				NestFree( tmp_identity );
			}
            break;
        } else {
            if (tmp_identity) {
				nest_debug( NEST_DEBUG_AUTH, "Mapfile user '%s' / '%s' did not match\n",
							entry->identity, entry->name );
				NestFree( tmp_identity );
			}
        }
    }

    if ( !user ) {
        nest_warn( __FILE__, __FUNCTION__,
                "Unable to find a gridmap entry for %s.\n",
                identity );
    }
    nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_AUTH,
            "%s returning %s from %s.\n",
            __FUNCTION__, user, identity );
    Nestend_grid_mapfile( state );
    return user;
}

NestReplyStatus
NestCreateFullPath( const char *path )
{
	char *temp;
	char *delim;
	char oldchar;
	int result;

	temp = (char *) alloca(strlen(path)+1);
	strcpy(temp,path);

	delim = temp;

	while( (delim = strchr(delim,'/')) ) {

		if(delim==temp) {
			delim++;
			continue;
		}

		oldchar = *delim;
		*delim = 0;

		result = NestMkdir(temp,0700);
		if(result!=0) {
			if(errno==EEXIST) {
				/* no problem, keep going */
			} else {
				return ErrnoToNest(errno);
			}
		} else {
			/* ok, made it successfully */
		}

		*delim = oldchar;
		delim++;
	}

	/* Now, last chance */

	result = NestMkdir(temp,0700);
	if(result!=0) {
		if(errno==EEXIST) {
			return NEST_SUCCESS;
		} else {
			return ErrnoToNest(errno);
		}
	} else {
		return NEST_SUCCESS;
	}
}

void NestCollapsePath( const char *l, char *s )
{
	char *start = s;

	while(*l) {
		if(!strncmp(l,"//",2)) {
			l++;
		} else if(!strncmp(l,"/./",3)) {
			l+=2;
		} else if(!strncmp(l,"/..",3)) {
			l+=3;
			while( ((*s)!='/') && (s>start) ) {
				s--;
			}
		} else {
			*s++ = *l++;
		}
	}

	*s = 0;

	if(s==start) {
		strcpy(s,"/");
	}
}
