/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "nest_debug.h"
#include "nest_protocol.h"

bool daemonize = false;
nestProtocol *logfile = NULL;


int nest_debug_level = NEST_DEBUG_NEVER;
FILE * nest_debug_output = stderr;

#define DEBUG_DELIMITER ","

struct debug_s
{
    char *name;
    int  value;
};

struct debug_s debug_list[] = {
    { "pot",            NEST_DEBUG_POT },
    { "pop",            NEST_DEBUG_POP },
    { "nob",            NEST_DEBUG_NOB },
    { "concurrency",    NEST_DEBUG_CONCURRENCY },

    { "gftp",           NEST_DEBUG_GRIDFTP },
    { "chirp",          NEST_DEBUG_CHIRP },
    { "http",           NEST_DEBUG_HTTP },
    { "fs",             NEST_DEBUG_FS },
    { "nfs",            NEST_DEBUG_NFS },
    { "protocol",       NEST_DEBUG_PROTOCOL },

    { "user",           NEST_DEBUG_USER_MNGMNT },
    { "group",          NEST_DEBUG_GROUP_MNGMNT },
    { "account",        NEST_DEBUG_ACCT_MNGMNT },

    { "lots",           NEST_DEBUG_LOTS },
    { "bandwidth",      NEST_DEBUG_BW_RSV },
    { "rsv",            NEST_DEBUG_RSV },

    { "acl",            NEST_DEBUG_ACL },
    { "fdsets",         NEST_DEBUG_FDSETS },
    { "fdleaks",        NEST_DEBUG_FD_LEAKS },
    { "proxy",          NEST_DEBUG_PROXY },

    { "debug",          NEST_DEBUG_DEBUG },
    { "transfer",       NEST_DEBUG_TRANSFER },
    { "client",         NEST_DEBUG_CLIENT },
    { "storage",        NEST_DEBUG_STORAGE },
    { "server",         NEST_DEBUG_SERVER },
    { "auth",           NEST_DEBUG_AUTH },

    { "cachemodel",     NEST_DEBUG_CACHEMODEL },
    { "admin",          NEST_DEBUG_ADMIN },
    { "protpref",       NEST_DEBUG_PROTOCOL_SCHED },
    { "flock",          NEST_DEBUG_FLOCK },
    { "malloc",         NEST_DEBUG_MALLOC },
    { "fd_cache",       NEST_DEBUG_FD_CACHE },

    { "always",         NEST_DEBUG_ALWAYS },
    { "never",          NEST_DEBUG_NEVER }, 
    { 0, 0 }
};

NestReplyStatus
nest_debug_init( nestProtocol *file, bool daemon ) {
    logfile = file;
    daemonize = daemon;
    return NEST_SUCCESS;
}

long
get_nest_debug_level( ) {
    return nest_debug_level;
}

void
set_nest_debug_level( long level ) {
    nest_debug_level = level;
}

void
add_nest_debug_level( long new_level ) {
    if ( new_level == NEST_DEBUG_NEVER ) {
        nest_debug_level = NEST_DEBUG_NEVER;
    } else {
        nest_debug_level |= new_level;
    }
}

NestReplyStatus
add_nest_debug_level( const char *debug_values ) {
    char *debug_values_copy = NestStrdup(debug_values);
    if (debug_values_copy != NULL) {
        char *debug = strtok( debug_values_copy, DEBUG_DELIMITER );
        struct debug_s * debugs;
        
        while( debug ) {
            for( debugs = debug_list; debugs->name != NULL; debugs++ ) {
                if ( ! strcmp( debug, debugs->name ) ) {
                    nest_debug( NEST_DEBUG_DEBUG, "Setting debug for %s\n",
                                debugs->name );
                    add_nest_debug_level( debugs->value );
                    break;
                }
            }
            if ( debugs->name == NULL ) {
                return NEST_UNKNOWN_DEBUG_VALUE;
            }
            debug = strtok( NULL, DEBUG_DELIMITER );
        }
        NestFree(debug_values_copy);
    }

    return NEST_SUCCESS;
}

int
nest_print( const char *initial_string, const char *format, va_list args ) {
    char output[NEST_MESSAGESIZE];

    time_t ticks = time( NULL );
    if ( initial_string ) {
        NestSnprintf( output, NEST_MESSAGESIZE, "%s %ld:", 
                initial_string, (long)ticks );
    } else {
        NestSnprintf( output, NEST_MESSAGESIZE, "%ld:", (long)ticks );

    }

        // TODO: count the number of args in the format string
        // and make sure it is the same as the number of args in the va_list.
    
    int retval = vsnprintf(    &output[NestStrlen(output)], 
                            NEST_MESSAGESIZE - NestStrlen(output), 
                            format, 
                            args );
        // TODO: how do I check the return value of stupid vsnprintf

    if ( daemonize ) {
        syslog( LOG_NOTICE | LOG_USER, "%s", output );
    } else if ( logfile != NULL ) {
        logfile->Sendn( output, NestStrlen( output ) );
    } else {
        fprintf( nest_debug_output, "%s", output );
        fflush(  nest_debug_output );
    }
    return retval;
}

int 
nest_printf( const char *format, ... ) {
    va_list args;
    va_start( args, format );
    int ret = nest_print( NULL, format, args );
    va_end( args );
    return ret;
}


#ifdef DEBUG

bool 
check_nest_debug_status( long level ) {
    return (nest_debug_level & level);
}

int
nest_debug( long level, const char *format, ... ) {
    int ret = 0;
    if ( check_nest_debug_status( level ) ) {
        va_list args;
        va_start( args, format );
        ret = nest_print( NULL, format, args );
        va_end( args );
    }
    return ret;
}

#else

int 
nest_debug( long level, const char *format, ... ) {
    return 1;
}

#endif

int
nest_warn( const char *filename, const char *func_name, const char *format, ... ) {
    char output[NEST_MESSAGESIZE];
    int ret = 0;
    if ( ! getenv( "NEST_SILENCE" ) ) {
        va_list args;
        va_start( args, format );
        NestSnprintf(output, NEST_MESSAGESIZE, "NeST warning:%s:%s: ", filename, func_name );
        ret = nest_print( output, format, args );
        va_end( args );
    }
    return ret;
}

void
nest_die( const char *format, ... ) {
    va_list args;
    va_start( args, format );
    nest_print( "NeST Fatal Error: ", format, args );
    va_end( args );
    _exit( 1 );
}


void
nest_perror( const char *msg ) {
    nest_printf( "%s: %s\n", msg, strerror( errno ) );
}
