/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_file_catalog.h"

NestFileCatalog::NestFileCatalog() {
    this->root_dir = "/";
    this->file_catalog = NULL;
    return;
}

NestFileCatalog::~NestFileCatalog() {
    return;
}

bool
NestFileCatalog::is_file_present( const char *file ) {

    if (!file) {
        return false;
    }

    if (this->is_system_file( file )) {
        nest_stat   tmp;
        char    physical_file[NEST_MAXPATHLEN];

        if (NestSnprintf(physical_file,
                         NEST_MAXPATHLEN,
                         "%s",
                         file + NestStrlen(NestSystemUser) + 1) < 0) {
            return false;
        }

        if (NestLstat( physical_file, &tmp ) < 0) {
            return false;
        }

        return true;
    }

    string str_key = STORAGE_OBJECT_KEY_PREFIX;
    str_key = str_key + file;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(str_key);
    if (!file_class_ad) {
        return false;
    }

    return true;
}

bool
NestFileCatalog::is_system_file( const char *filename ) {

    if (!strncmp( filename, NestSystemUser, sizeof(NestSystemUser))) {
        return true;
    }

    return false;
}

NestReplyStatus
NestFileCatalog::get_physical_name( string &physical_name,
                                    string logical_name,
                                    string user,
                                    string group,
                                    string lot_handle ) {

    physical_name = this->root_dir +
                    logical_name;

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::add_entry(const char *logical_file,
                           const char *physical_file,
                           const char *logical_dir,
                           const char *user,
                           const char *group,
                           nest_obj_type_t file_type) {

    user_id_t   uid;
    int         tmp_fd;

    if (this->is_system_file( logical_file )) {
        switch (file_type) {
            case NEST_FILE:
                tmp_fd = NestOpen( physical_file, O_WRONLY | O_CREAT );

                if (tmp_fd < 0 ) {
                    nest_warn( __FILE__, __FUNCTION__, "create of %s failed: %s\n",
                                physical_file,
                                strerror( errno ) );
                    return ErrnoToNest(errno);
                }
                NestClose( tmp_fd );
                NestChmod( physical_file, FILE_MODE );
                if (NestGet_uid( &uid, user, SystemPwdFile ) < 0) {
                    uid = (user_id_t) -1;
                }
                NestChown( physical_file, uid, ActualGid );

                break;

            case NEST_DIR:
                if ( NestMkdir( physical_file, DIR_MODE ) < 0 ) {
                    nest_warn( __FILE__, __FUNCTION__, "mkdir of %s failed: %s\n",
                                physical_file,
                                strerror( errno ) );
                    return ErrnoToNest(errno);
                }
                NestChmod( physical_file, DIR_MODE );
                if (NestGet_uid( &uid, user, SystemPwdFile ) < 0) {
                    uid = (user_id_t) -1;
                }
                NestChown( physical_file, uid, ActualGid );

                break;

            case NEST_LINK:
                return NEST_NOT_YET_IMPLEMENTED;
        }

        return NEST_SUCCESS;
    }

    if (this->is_file_present(logical_file)) {
        return NEST_FILE_EXISTS;
    }

    string str_file_type;

    switch (file_type) {
        case NEST_FILE: str_file_type = "file";
                        break;
        case NEST_DIR:  str_file_type = "dir";
                        break;
        case NEST_LINK: str_file_type = "link";
                        break;
        default:        str_file_type = "unknown";
                        break;
    }

    string str_key,
           str_file_classad;

    str_key = STORAGE_OBJECT_KEY_PREFIX;
    str_key = str_key + logical_file;
    str_file_classad =  " ";
    str_file_classad = str_file_classad +
                        "[  Type = \"" + STORAGE_OBJECT_KEY_PREFIX + "\"; " +
                        "   OwnerUser = \"" + user + "\"; " +
                        "   OwnerGroup = \"" + group + "\"; " +
                        "   LogicalName = \"" + logical_file + "\"; " +
                        "   DirName = \"" + logical_dir + "\"; " +
                        "   PhysicalName = \"" + physical_file + "\"; " +
                        "   FileType = \"" + str_file_type + "\"; " +
                        "   LotHandle = {}; " +
                        "   AttachedLots = {}; " +
                        "   Size = 0; " +
                        "   TransferInProgress = false; " +

                        "   UserReadPerm = {\"" + user + "\"}; " +
                        "   UserWritePerm = {\"" + user + "\"}; " +
                        "   UserListPerm = {\"" + user + "\"}; " +
                        "   UserDeletePerm = {\"" + user + "\"}; " +
                        "   UserInsertPerm = {\"" + user + "\"}; " +
                        "   UserModifyACPerm = {\"" + user + "\"}; " +

                        "   GroupReadPerm = {}; " +
                        "   GroupWritePerm = {}; " +
                        "   GroupListPerm = {}; " +
                        "   GroupDeletePerm = {}; " +
                        "   GroupInsertPerm = {}; " +
                        "   GroupModifyACPerm = {} " +
                        " ]";


    ClassAd *file_class_ad = parser.ParseClassAd( str_file_classad, true );

    if (!file_class_ad) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "ParseClassAd failed - key = %s - CondorErrMsg = %s \n",
                    str_key.c_str(),
                    CondorErrMsg.c_str() );

        return NEST_CLASSAD_ERROR;
    }

    if (!this->file_catalog->AddClassAd( str_key, file_class_ad )) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "AddClassAd failed - key = %s - CondorErrMsg = %s \n",
                    str_key.c_str(),
                    CondorErrMsg.c_str() );

        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::add_entry(const char *logical_file,
                           const char *user,
                           nest_obj_type_t file_type) {

    if (this->is_file_present(logical_file)) {
        return NEST_FILE_EXISTS;
    }

    if (this->is_system_file( logical_file )) {
        char    physical_file[NEST_MAXPATHLEN];

        this->get_physical_name( physical_file,
                                 logical_file );

        return this->add_entry( logical_file,
                                physical_file,
                                NULL,
                                user,
                                NEST_NO_GROUP,
                                file_type );
    }

    /*
        Got to find the user's default group and default lot!
    */

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->file_catalog->GetClassAd(key);
    if (!user_class_ad) {
        // TODO:  user anonymous cannot create files even if they have
        // access permissions . . . 
        return NEST_USER_NOT_FOUND;
    }

    string  attr,
            default_group = "",
            default_lot_handle = "";

    attr = "DefaultGroup";
    user_class_ad->EvaluateAttrString( attr, default_group );
    if (default_group == "") {
        default_group = NEST_NO_GROUP;
    }

    attr = "DefaultLotHandle";
    user_class_ad->EvaluateAttrString( attr, default_lot_handle );

    string physical_name;
    if (this->get_physical_name( physical_name,
                                 logical_file,
                                 user,
                                 default_group,
                                 default_lot_handle ) != NEST_SUCCESS) {
        return NEST_FILE_EXISTS;
    }

    char    logical_dir[NEST_MAXPATHLEN];

    NestReplyStatus reply = this->get_parent_dir( logical_dir,
                                                  logical_file );
    NEST_REPLY_CHECK(reply);

    return this->add_entry( logical_file,
                            physical_name.c_str(),
                            logical_dir,
                            user,
                            default_group.c_str(),
                            file_type );
}


NestReplyStatus
NestFileCatalog::add_entry(const char *newpath_logical,
                           const char *oldpath,
                           const char *user,
                           nest_obj_type_t type) {

    if (type != NEST_LINK) {
        return NEST_BAD_PARAMETERS;
    }

    char    oldpath_logical[NEST_MAXPATHLEN];
    NestReplyStatus reply = this->get_logical_name( oldpath_logical,
                                                    oldpath,
                                                    user );
    NEST_REPLY_CHECK(reply);

    if (this->is_system_file( newpath_logical )) {
        if (!this->is_system_file( oldpath_logical )) {
            return NEST_NOT_YET_IMPLEMENTED;
        }

        char    old_physical[NEST_MAXPATHLEN],
                new_physical[NEST_MAXPATHLEN];

        this->get_physical_name( old_physical,
                                 oldpath_logical );

        this->get_physical_name( new_physical,
                                 newpath_logical );

        int result = NestSymlink( old_physical,
                                  new_physical );

        if (result < 0) {
            return ErrnoToNest(errno);
        }

        user_id_t   uid;
        if (NestGet_uid( &uid, user, SystemPwdFile ) < 0) {
            uid = (user_id_t) -1;
        }
        NestLchown( new_physical, uid, ActualGid );

        return NEST_SUCCESS;
    }

    if (this->is_file_present(newpath_logical)) {
        return NEST_FILE_EXISTS;
    }

    if (!this->is_file_present(oldpath_logical)) {
        return NEST_FILE_NOT_FOUND;
    }

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->file_catalog->GetClassAd(key);
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string  default_group;
    if (!user_class_ad->EvaluateAttrString( "DefaultGroup", default_group )) {
        default_group = NEST_NO_GROUP;
    }

    char    logical_dir[NEST_MAXPATHLEN];

    reply = this->get_parent_dir( logical_dir,
                                  newpath_logical );
    NEST_REPLY_CHECK(reply);

    return this->add_entry( newpath_logical,
                            oldpath,
                            logical_dir,
                            user,
                            default_group.c_str(),
                            NEST_LINK);
}

NestReplyStatus
NestFileCatalog::delete_entry(const char *file) {

    if (this->is_system_file( file )) {
        return NEST_SUCCESS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    ClassAd *file_ad = this->file_catalog->GetClassAd(key);

    if (!file_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    nest_obj_type_t     file_type;
    NestReplyStatus     reply;

    reply = this->get_type(&file_type, file);

    NEST_REPLY_CHECK(reply);

    if ((file_type == NEST_DIR) && (!this->is_dir_empty(file))) {
        return NEST_NOT_EMPTY;
    }

    if (!this->file_catalog->RemoveClassAd(key)) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_system_link_physical_name(char *physical_filename,
                                               const char *logical_filename) {

    nest_stat   stat_buf;

    if (!this->is_system_file( logical_filename )) {
        return NEST_FILE_NOT_FOUND;
    }

    if (NestSnprintf(physical_filename,
                     NEST_MAXPATHLEN,
                     "%s",
                     logical_filename + NestStrlen(NestSystemUser) + 1 ) < 0) {
        return NEST_FILENAME_TOO_LONG;
    }

    if (NestLstat( physical_filename, &stat_buf ) < 0) {
        return ErrnoToNest(errno);
    }

    if (S_ISLNK(stat_buf.st_mode)) {
        char    physical_link[NEST_MAXPATHLEN];

        strncpy( physical_link, physical_filename, NEST_MAXPATHLEN);

        if (NestReadlink( physical_link,
                        physical_filename,
                        NEST_MAXPATHLEN ) < 0) {
            return ErrnoToNest(errno);
        }
        return NEST_SUCCESS;
    }

    return NEST_FILE_NOT_FOUND;
}


NestReplyStatus
NestFileCatalog::get_physical_name(char *physical_filename,
                                   const char *logical_filename) {

    if (!physical_filename ||
        !logical_filename) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file( logical_filename )) {
        if (NestSnprintf(physical_filename,
                         NEST_MAXPATHLEN,
                         "%s",
                         logical_filename + NestStrlen(NestSystemUser) + 1 ) 
                < 0) 
        {
            return NEST_FILENAME_TOO_LONG;
        }

        return NEST_SUCCESS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + logical_filename;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string  attr,
            str_physical_name;
    attr = "PhysicalName";
    if (!file_class_ad->EvaluateAttrString(attr, str_physical_name)) {
        return NEST_FILE_NOT_FOUND;
    }

    if (NestSnprintf(physical_filename,
                     NEST_MAXPATHLEN,
                     "%s",
                     str_physical_name.c_str()) < 0) {
        *physical_filename = 0;
        return NEST_FILENAME_TOO_LONG;
    }
    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_logical_name( char *logical_filename,
                                   const char *physical_filename) {

    if (!physical_filename ||
        !logical_filename) {
        return NEST_BAD_PARAMETERS;
    }

    string constraint = "";
    constraint = constraint +
                    "other.PhysicalName == \"" + physical_filename + "\"";

    ExprTree    *constraint_tree = this->parser.ParseExpression(constraint);
    this->query.Query(this->file_catalog_view, constraint_tree);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return NEST_FILE_NOT_FOUND;
    }

    ClassAd *file_ad = this->file_catalog->GetClassAd( query_key );

    if (!file_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string  attr,
            str_logical_name;

    attr = "LogicalName";
    if (!file_ad->EvaluateAttrString(attr, str_logical_name)) {
        return NEST_FILE_NOT_FOUND;
    }

    strcpy(logical_filename, str_logical_name.c_str());
    return NEST_SUCCESS;
}

/*
    Except while returning NEST_BAD_PARAMETERS -
    logical_filename WILL have the relevant filename

    the return value can be optionally ignored, if people
    dont care about whether the file is present or not!
*/

NestReplyStatus
NestFileCatalog::get_logical_name(char *logical_filename,
                                  const char *filename,
                                  const char *user) {

    if (!logical_filename || !user) {
        return NEST_BAD_PARAMETERS;
    }

    if (!filename || !filename[0]) {
        NestSnprintf( logical_filename,
                      NEST_MAXPATHLEN,
                      "~%s/",
                      user );
        return NEST_SUCCESS;
    }

    if (filename[0] == '~') {
        /*
            hiya, somebody is trying to access somebodyelse's files
        */
        char    file_owner[NEST_NAMELEN];
        char    actual_file[NEST_MAXPATHLEN];

        file_owner[0] = actual_file[0] = 0;

        unsigned int i;
        for(i=1; i<NestStrlen(filename); i++) {
            if (filename[i] == '/') {
                file_owner[i-1] = 0;
                break;
            }
            file_owner[i-1] = filename[i];
        }
        file_owner[i-1]=0;

        strncpy( actual_file, filename + i, NEST_MAXPATHLEN );
        if (!actual_file[0]) {
            strcpy(actual_file, "/");
        }

        if (!strcmp(file_owner, "")) {
            strcpy(file_owner, user);
        }

        return this->get_logical_name(logical_filename,
                                      actual_file,
                                      file_owner);
    } else {
        NestSnprintf(logical_filename,
                     NEST_MAXPATHLEN,
                     "%s:%s",
                     user,
                     filename);
        logical_filename[NEST_MAXPATHLEN-2] = 0;
    }

    if (this->is_file_present(logical_filename)) {
        return NEST_SUCCESS;
    }

    /*
        This could be a NEST_DIR without a '/' at the end!
    */

    int len = NestStrlen(logical_filename);

    if (logical_filename[len-1] == '/') {
        return NEST_FILE_NOT_FOUND;
    }

    logical_filename[len] = '/';
    logical_filename[len+1] = 0;

    if (this->is_file_present(logical_filename)) {
        return NEST_SUCCESS;
    }

    logical_filename[len] = 0;
    return NEST_FILE_NOT_FOUND;
}

NestReplyStatus
NestFileCatalog::get_reverse_logical_name( char *dest,
                                           const char *src ) {

    int i = 0,
        j = 1;

    strcpy( dest, src );
    dest[0] = '~';

    while (src[i] != ':' && src[i]) {
        dest[j++] = src[i++];
    }

    if (!src[i]) {
        dest[j] = 0;
        return NEST_SUCCESS;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_owner_user( char *user,
                                 const char *file ) {

    if (!user || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file( file )) {

        nest_stat   stat_buf;
        char        physical_file[NEST_MAXPATHLEN];

        if (this->get_physical_name( physical_file,
                                     file ) != NEST_SUCCESS) {
            return NEST_FILE_NOT_FOUND;
        }

        if (NestLstat( physical_file, &stat_buf ) < 0) {
            return NEST_FILE_NOT_FOUND;
        }

        if (NestGet_uname( user, stat_buf.st_uid, SystemPwdFile ) < 0) {
            return NEST_USER_NOT_FOUND;
        }

        return NEST_SUCCESS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string  attr,
            str_user;
    attr = "OwnerUser";
    if (!file_class_ad->EvaluateAttrString(attr, str_user)) {
        return NEST_FILE_NOT_FOUND;
    }

    strcpy( user, str_user.c_str() );

    return NEST_SUCCESS;
}


NestReplyStatus
NestFileCatalog::get_type(nest_obj_type_t *type,
                          const char *file) {

    if (!type || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file( file )) {

        nest_stat   stat_buf;
        char        physical_file[NEST_MAXPATHLEN];

        if (this->get_physical_name( physical_file,
                                     file ) != NEST_SUCCESS) {
            return NEST_FILE_NOT_FOUND;
        }

        if (NestLstat( physical_file, &stat_buf ) < 0) {
            return NEST_FILE_NOT_FOUND;
        }

        if (S_ISREG( stat_buf.st_mode )) {
            *type = NEST_FILE;
            return NEST_SUCCESS;
        }

        if (S_ISDIR( stat_buf.st_mode )) {
            *type = NEST_DIR;
            return NEST_SUCCESS;
        }

        if (S_ISLNK( stat_buf.st_mode )) {
            *type = NEST_LINK;
            return NEST_SUCCESS;
        }

        return NEST_FILE_NOT_FOUND;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string  str_type;

    if (!file_class_ad->EvaluateAttrString("FileType", str_type)) {
        return NEST_FILE_NOT_FOUND;
    }

    if (!strcmp(str_type.c_str(), "file")) {
        *type = NEST_FILE;
        return NEST_SUCCESS;
    }

    if (!strcmp(str_type.c_str(), "dir")) {
        *type = NEST_DIR;
        return NEST_SUCCESS;
    }

    if (!strcmp(str_type.c_str(), "link")) {
        *type = NEST_LINK;
        return NEST_SUCCESS;
    }

    return NEST_FILE_NOT_FOUND;
}

NestReplyStatus
NestFileCatalog::get_size(off_t *result,
                          const char *file) {

    if (!result || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file( file )) {

        nest_stat   stat_buf;
        char        physical_file[NEST_MAXPATHLEN];

        if (this->get_physical_name( physical_file,
                                     file ) != NEST_SUCCESS) {
            return NEST_FILE_NOT_FOUND;
        }

        if (NestLstat( physical_file, &stat_buf ) < 0) {
            return NEST_FILE_NOT_FOUND;
        }

        *result = stat_buf.st_size;
        return NEST_SUCCESS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    double size;
    if (!file_class_ad->EvaluateAttrNumber("Size", size)) {
        return NEST_FILE_NOT_FOUND;
    }

    *result = (off_t) size;
    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::set_size(const char *file,
                          off_t size) {

    if (this->is_system_file(file)) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Calling set_size(%s) to a system file!\n",
                    file );
        return NEST_SUCCESS;
    }

    if (size < 0) {
        size = 0;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    char    str_size[NEST_NAMELEN];
    NestSnprintf(str_size, NEST_NAMELEN, "%ld", size);

    string modify_classad = "";
    modify_classad = modify_classad +
                    " [ Size = " + str_size + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad,
                                                true );

    if (!this->file_catalog->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::set_transfer_flag(const char *file,
                                   bool flag) {

    char    logical_filename[NEST_MAXPATHLEN];
    NestReplyStatus reply;

    if (this->is_system_file(file)) {

        char    physical_filename[NEST_MAXPATHLEN];

        reply = this->get_physical_name( physical_filename,
                                         file );

        NEST_REPLY_CHECK(reply);

        reply = this->get_logical_name( logical_filename,
                                        physical_filename );

        if (reply != NEST_SUCCESS) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Trying to set_transfer_flag to a "
                        "non-nest system file %s !\n",
                        file );

            return NEST_SUCCESS;
        }
    }
    else {
        strncpy( logical_filename, file, NEST_MAXPATHLEN );
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + logical_filename;

    const char *str_flag = flag ? "true" : "false";

    string modify_classad = "";
    modify_classad = modify_classad +
                    " [ TransferInProgress = " + str_flag + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad,
                                                true );

    if (!this->file_catalog->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}
        
NestReplyStatus 
NestFileCatalog::get_file_uniquifier(const char *physical_filename,
                                     NestFilehandle* uniquifier ) {
	if(!physical_filename) {
		return NEST_BAD_PARAMETERS;
	}

    nest_stat statentry;
    if ( NestLstat( physical_filename, &statentry ) < 0 ) {
        return ErrnoToNest( errno );
    }

    *uniquifier = (NestFilehandle)statentry.st_ino;
    return NEST_SUCCESS;     
}



NestReplyStatus
NestFileCatalog::get_file_classad(const char *file,
						          ClassAd** file_class_ad) {
	if(!file) {
		return NEST_BAD_PARAMETERS;
	}

	if(this->is_system_file(file)) { 
		return NEST_NOT_YET_IMPLEMENTED; 
	}

	string key = STORAGE_OBJECT_KEY_PREFIX;
	key = key + file;

	*file_class_ad = this->file_catalog->GetClassAd(key);
	if ( ! (*file_class_ad) ) {
		return NEST_FILE_NOT_FOUND;
	}

	return NEST_SUCCESS;
}


NestReplyStatus
NestFileCatalog::get_acl(const char *file,
                         const char *name,
                         int is_user,
                         nest_access_rights_t *res) {

    if (!file || !name || !res) {
        return NEST_BAD_PARAMETERS;
    }

    memset((void *) res, 0, sizeof(nest_access_rights_t));

    if (this->is_system_file(file)) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    ClassAd *file_class_ad = NULL;

	NestReplyStatus status = get_file_classad(file, &(file_class_ad) );

	if(status != NEST_SUCCESS) {
		return status;
	}

    string perms[6];
    perms[0] = "Read";
    perms[1] = "Write";
    perms[2] = "List";
    perms[3] = "Delete";
    perms[4] = "Insert";
    perms[5] = "ModifyAC";

    nest_access_rights_t    rights[6];
    rights[0] = NEST_READ_ACCESS;
    rights[1] = NEST_WRITE_ACCESS;
    rights[2] = NEST_LIST_ACCESS;
    rights[3] = NEST_DELETE_ACCESS;
    rights[4] = NEST_INSERT_ACCESS;
    rights[5] = NEST_MODIFYAC_ACCESS;

    string identity = is_user ? "User" : "Group";

    for(int permno=0; permno<6; permno++) {
        string attr = identity + perms[permno] + "Perm";
        ExprTree *perm_list = file_class_ad->Lookup( attr );

        if (!perm_list) {
            return NEST_CLASSAD_ERROR;
        }

        string str_perm_list;

        unparser.Unparse( str_perm_list, perm_list );
        nest_debug( NEST_DEBUG_ACL, "\t%s: '%s'\n",
                    perms[permno].c_str(), str_perm_list.c_str() );

        string str_find_param = "\"";
        str_find_param = str_find_param + name + "\"";

        if (str_perm_list.find(str_find_param)
            != string::npos) {
            *res |= rights[permno];
        }
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_all_acl(NestACLList **result,
                             const char *file,
                             int is_user) {

    // Verify that 'result' a vaild pointer
    if (!result || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file(file)) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    //*result = NULL;                // Force to known state
        // responsibility moved to the caller!

    // Build the ClassAd key
    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    // Find the ClassAd for this file
    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string perms[6];
    perms[0] = "Read";
    perms[1] = "Write";
    perms[2] = "List";
    perms[3] = "Delete";
    perms[4] = "Insert";
    perms[5] = "ModifyAC";

    nest_access_rights_t    rights[6];
    rights[0] = NEST_READ_ACCESS;
    rights[1] = NEST_WRITE_ACCESS;
    rights[2] = NEST_LIST_ACCESS;
    rights[3] = NEST_DELETE_ACCESS;
    rights[4] = NEST_INSERT_ACCESS;
    rights[5] = NEST_MODIFYAC_ACCESS;

    string              identity = is_user ? "User" : "Group";
    NestReplyStatus     error_reply;
    ExprTree            *perm_list;
    string              attr;
    ExprListIterator    perm_itr;
    Value               perm_val;
    char                perm_str[NEST_NAMELEN];

    // Loop through the set of available permissions
    for(int permno=0; permno<6; permno++) {
        attr = identity + perms[permno] + "Perm";

        // Pull the permissions list from the ClassAd
        perm_list = file_class_ad->Lookup( attr );
        if (!perm_list) {
            error_reply = NEST_CLASSAD_ERROR;
            goto clean_and_error_return;
        }

        // Initialize the iterator
        perm_itr.Initialize((ExprList *) perm_list);

        // Start the iterator; are there any values?
        if (! perm_itr.CurrentValue( perm_val )) {
            nest_debug( NEST_DEBUG_ACL, "\t%s: No Permissions\n",
                        perms[permno].c_str() );
            continue;            // None -- go check the next perm.
        } else {
            // Debugging code
            string str_perm_list;
            unparser.Unparse( str_perm_list, perm_list );
            nest_debug( NEST_DEBUG_ACL, "\t%s: '%s'\n",
                        perms[permno].c_str(), str_perm_list.c_str() );
        }

        // Loop through the rest of the values...
        do {

            // Is this a valid name?
            if (!perm_val.IsStringValue( perm_str, NEST_NAMELEN ) ||
                !strcmp(perm_str, "")) {
                continue;                    // No name -- skip it
            }

            // Update 'result' to reflect the permission
            error_reply = NestAdd_ACL( result,
                                       perm_str,
                                       is_user,
                                       rights[permno] );
            if (error_reply != NEST_SUCCESS) {
                goto clean_and_error_return;
            }

        } while ( perm_itr.NextValue( perm_val ) ) ;
    }

    return NEST_SUCCESS;

clean_and_error_return:
    NestACLList *tmp;

    while (*result) {
        tmp = (*result)->next;
        NestFree(*result);
        *result = tmp;
    }

    return error_reply;
}

NestReplyStatus
NestFileCatalog::set_acl(const char *file,
                         const char *name,
                         int is_user,
                         nest_access_rights_t arg) {

    if (!file || !name) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file(file)) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + file;

    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string perms[6];
    perms[0] = "Read";
    perms[1] = "Write";
    perms[2] = "List";
    perms[3] = "Delete";
    perms[4] = "Insert";
    perms[5] = "ModifyAC";

    nest_access_rights_t    rights[6];
    rights[0] = NEST_READ_ACCESS;
    rights[1] = NEST_WRITE_ACCESS;
    rights[2] = NEST_LIST_ACCESS;
    rights[3] = NEST_DELETE_ACCESS;
    rights[4] = NEST_INSERT_ACCESS;
    rights[5] = NEST_MODIFYAC_ACCESS;

    string identity = is_user ? "User" : "Group";

    for(int i=0; i<6; i++) {

        string attr = identity + perms[i] + "Perm";
        ExprTree *perm_list = file_class_ad->Lookup( attr );

        if (!perm_list) {
            return NEST_UNKNOWN_ERROR;
        }

        string str_perm_list,
               new_perm_list,
               str_name(name);

        unparser.Unparse( str_perm_list, perm_list );

        char need_to_modify = 'f';

        if (arg & rights[i]) {
            /*
                Need to give this user/group right[i]!
                Check to see if he already has that right!!
            */

            string str_find_param = "\"";
            str_find_param = str_find_param + name + "\"";

            if (str_perm_list.find(str_find_param)
                == string::npos) {
                if (NestAdd_entry_to_list(new_perm_list,
                                      str_perm_list,
                                      str_name)
                            != NEST_SUCCESS) {
                    return NEST_UNKNOWN_ERROR;
                }
                need_to_modify = 't';
            }
        }
        else {
            /*
                Need to revoke this user/group right[i]!
                Check to see if he already doesn't has that right!!
            */
            string str_find_param = "\"";
            str_find_param = str_find_param + name + "\"";

            if (str_perm_list.find(str_find_param)
                != string::npos) {
                if (NestDelete_entry_from_list(new_perm_list,
                                           str_perm_list,
                                           str_name)
                            != NEST_SUCCESS) {
                    return NEST_UNKNOWN_ERROR;
                }
                need_to_modify = 't';
            }
        }

        if (need_to_modify == 't') {
            string modify_classad = "";
            modify_classad = modify_classad +
                             " [ " + attr + " = " + new_perm_list + " ] ";
            modify_classad = " [ Updates = " + modify_classad + " ] ";

            ClassAd *mod_classad = parser.ParseClassAd( modify_classad,
                                                        true );

            if (!this->file_catalog->ModifyClassAd( key, mod_classad )) {
                return NEST_CLASSAD_ERROR;
            }
        }

    }   //  end-of-for

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_user_acl(const char *file,
                              const char *user,
                              nest_access_rights_t *res) {
    return this->get_acl(file, user, IS_USER, res);
}

NestReplyStatus
NestFileCatalog::set_user_acl(const char *file,
                              const char *user,
                              nest_access_rights_t arg) {
    return this->set_acl(file, user, IS_USER, arg);
}

NestReplyStatus
NestFileCatalog::get_group_acl(const char *file,
                               const char *group,
                               nest_access_rights_t *res) {
    return this->get_acl(file, group, IS_GROUP, res);
}

NestReplyStatus
NestFileCatalog::set_group_acl(const char *file,
                               const char *group,
                               nest_access_rights_t arg) {
    return this->set_acl(file, group, IS_GROUP, arg);
}

NestReplyStatus
NestFileCatalog::get_parent_dir( char *dest,
                                 const char *src ) {

    if (!dest || !src) {
        return NEST_BAD_PARAMETERS;
    }

    *dest = 0;

    int   cptr = NestStrlen(src) - 1;

    //  paranoia
    if (cptr < 0) {
        return NEST_BAD_PARAMETERS;
    }

    if (src[cptr] == '/') {
        cptr--;                 //  to skip the last '/' for directories
    }

    while ((cptr >= 0) &&
           (src[cptr] != '/')) {
        cptr--;
    }

    if (cptr < 0) {
        //  no '/' other than probably the last one!
        //      => no parent present except for some user's home-dir!

        cptr = NestStrlen(src) - 1;
        if ((src[cptr-1] == ':') &&
            (src[cptr] == '/')) {
            strcpy(dest, src);
            return NEST_SUCCESS;
        }
        return NEST_BAD_PARAMETERS;
    }

    strncpy(dest, src, (cptr+1));
    dest[cptr+1] = 0;

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::get_dest_file( char *dest,
                                const char *src_file,
								const char *user ) {
	NestReplyStatus reply;

    if (!dest || !src_file) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestSnprintf( dest,
                      NEST_MAXPATHLEN,
                      "%s",
                      src_file ) < 0) {
        return NEST_FILENAME_TOO_LONG;
    }

    if (dest[0] != '~') {

        /*
            the assumption is that, we will get
            logical "NeST" absolute paths
        */

        nest_debug( NEST_DEBUG_STORAGE,
                    "get_dest_file is passed an invalid filename %s\n",
                    src_file );

        return NEST_BAD_PARAMETERS;
    }

    int     i,
            index,
            limit = NestStrlen( dest );

    if (!strncmp( dest, "./", 2 )) {

		/* Shouldn't have reached here! 
		 * Don't know why this 'if stmt' remained here!!
	 	 */ 
		assert(0);

        for (i=2; i<limit; i++) {
            dest[i-2] = dest[i];
        }
        dest[i-2] = 0;
        limit -= 2;
    }

    for (index = 0; index < limit; index++) {
        if (((index + 2) < limit) &&
            dest[index] == '/' &&
            dest[index+1] == '.' &&
            dest[index+2] == '/') {
            for (i=index+3; i < limit; i++) {
                dest[i-2] = dest[i];
            }
            dest[i-2] = 0;
            limit -= 2;
            index--;
        }
        if (((index + 1) < limit) &&
            dest[index] == '/' &&
            dest[index+1] == '/') {
            for (i=index+2; i < limit; i++) {
                dest[i-1] = dest[i];
            }
            dest[i-1] = 0;
            limit--;
            index--;
        }
    }

    if (limit < 2) {
        return NEST_SUCCESS;
    }

    if (dest[limit-2] == '/' &&
        dest[limit-1] == '.') {
        dest[limit-1] = 0;
        limit--;
    }

    /*
        if it ends with a /.. add a '/' at the end

        I am skipping boundary checks for using one more byte - Sue me!
    */

    if (limit < 3) {
        return NEST_SUCCESS;
    }

    if (dest[limit-3] == '/' &&
        dest[limit-2] == '.' &&
        dest[limit-1] == '.') {
        dest[limit++] = '/';
        dest[limit] = 0;
    }

    for (index = 0; index < (limit - 3); index++) {
        if (dest[index] == '/' &&
            dest[index + 1] == '.' &&
            dest[index + 2] == '.' &&
            dest[index + 3] == '/') {

            int dst = index - 1,
                src = index + 3,
				ptr;
			char tmp_filename[NEST_MAXPATHLEN],
				 logical_filename[NEST_MAXPATHLEN],
				 phy_filename[NEST_MAXPATHLEN];	

			ptr = 0;
			while(ptr <= index){
				tmp_filename[ptr] = dest[ptr];
				ptr++;
			}
			tmp_filename[ptr] = 0;	
			reply = this->get_logical_name(logical_filename,
										   tmp_filename,
										   user);	
			reply = this->get_physical_name(phy_filename,
											logical_filename);
			
			NEST_REPLY_CHECK(reply);
			if(!this->is_access_allowed(logical_filename,
										user,
										NEST_LIST_ACCESS)){
				return NEST_NONPRIVELEDGED_USER;
			}
            while (dst >= 0 &&
                   dest[dst] != '/') {
                  dst--;
            }

            if (dst < 0) {
                /*
                    <no parent found> - oops!

                    since the filepath is assured to start with a '~'
                    just ignore this "../" as somebody is trying to climb
                    above once's home-dir

                    if you dont want NeST to emulate UNIX FS semantics
                    in this climbing-above-root behavior,

                    just have
                    "return NEST_BAD_PARAMETERS;"
                    inside this if-statement
                */
                dst = index;
            }

            index = dst - 1;

            while (src < limit) {
                dest[dst++] = dest[src++];
            }
            dest[dst]=0;
            limit = dst;
        }
    }

    return NEST_SUCCESS;
}

/*
    Replicated from NestIdentityManager::get_uid()
*/

NestReplyStatus
NestFileCatalog::get_uid(user_id_t *uid,
                         const char *user) {

    if (!uid || !user) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestGet_uid(uid, user, SystemPwdFile) < 0){
        return NEST_USER_NOT_FOUND;
    }

    return NEST_SUCCESS;
}

bool
NestFileCatalog::is_system_dir_access_allowed(const char *physical_dirname,
                                              user_id_t uid,
                                              user_id_t owner_uid,
                                              nest_access_rights_t flags) {

    int         fd;
    bool        result = true;
    int		success;
    user_id_t   my_id = NestGetuid();
    char        tmp_path[NEST_MAXPATHLEN];
    char        *cwd = NULL;

    NestSeteuid(uid);
    NestReplyStatus status;

    switch (flags) {
        case NEST_LIST_ACCESS:

            status = NestGetcwd(tmp_path, NEST_MAXPATHLEN);
            if ( status != NEST_SUCCESS ){
                result = false;
                break;
            }

            if (Nestsafe_Chdir(physical_dirname) < 0) {
                result = false;
                break;
            }

            Nestsafe_Chdir(tmp_path);
            break;

        case NEST_DELETE_ACCESS:
        case NEST_INSERT_ACCESS:

            NestSnprintf( tmp_path,
                          NEST_MAXPATHLEN,
                          "%s/.__NeST_proxy_file__",
                          physical_dirname );

            fd = NestCreat( tmp_path, FILE_MODE );
            if (fd < 0) {
                result = false;
                break;
            }

            NestClose(fd);
            success = NestUnlink( (const char *)tmp_path );

            if((flags == NEST_DELETE_ACCESS) && success < 0){
                result = false;
            }

            break;

        case NEST_MODIFYAC_ACCESS:
            //  NOT_YET_IMPLEMENTED
            result = false;
            break;
        case NEST_READ_ACCESS:
            nest_debug( NEST_DEBUG_ACL,
                        "READ_ACCESS should never be checked with"
                        " dir's permissions! - Disallowing "
                        " read ( %s ) \n",
                        physical_dirname );
            result = false;
            break;
        case NEST_WRITE_ACCESS:
            nest_debug( NEST_DEBUG_ACL,
                        "WRITE_ACCESS should never be checked with"
                        " dir's permissions! - Disallowing "
                        " write ( %s ) \n",
                        physical_dirname );
            result = false;
            break;
    }

    NestSeteuid(my_id);
    if ( cwd ) NestFree( cwd );
    return result;
}


bool
NestFileCatalog::is_system_file_access_allowed(const char *physical_filename,
                                               user_id_t uid,
                                               user_id_t owner_uid,
                                               nest_access_rights_t flags) {

    int open_flags = 0;

    switch (flags) {
        case NEST_READ_ACCESS:
            open_flags = O_RDONLY;
            break;

        case NEST_WRITE_ACCESS:
            open_flags = O_WRONLY;
            break;

        case (NEST_READ_ACCESS | NEST_WRITE_ACCESS):
            open_flags = O_RDWR;
            break;

        case NEST_DELETE_ACCESS:
            nest_debug( NEST_DEBUG_ACL,
                        "DELETE_ACCESS should never be checked with"
                        " file's permissions! - Disallowing "
                        " delete ( %s ) \n",
                        physical_filename );
            return false;
        
        case NEST_LIST_ACCESS:
            nest_debug( NEST_DEBUG_ACL,
                        "LIST_ACCESS should never be checked with"
                        " file's permissions! - Disallowing "
                        " list ( %s ) \n",
                        physical_filename );
            return false;

        case NEST_INSERT_ACCESS:
            nest_debug( NEST_DEBUG_ACL,
                        "INSERT_ACCESS should never be checked with"
                        " file's permissions! - Disallowing "
                        " insert ( %s ) \n",
                        physical_filename );
            return false;

        case NEST_MODIFYAC_ACCESS:
            //  NOT_YET_IMPLEMENTED
            return false;
    }

    user_id_t   my_id = NestGetuid();

    NestSeteuid(uid);
    int fd = NestOpen(physical_filename, open_flags);
    NestSeteuid(my_id);

    if (fd < 0) {
        nest_perror("acl open:");
        nest_debug( NEST_DEBUG_ACL,
                    "(%s,%d) %d\n",
                    physical_filename,
                    open_flags,
                    uid );
        return false;
    }

    close(fd);
    return true;
}


bool
NestFileCatalog::is_system_entry_access_allowed(const char *physical_filename,
                                                user_id_t uid,
                                                nest_access_rights_t flags) {

    nest_stat   stat_buf;

    if (NestLstat(physical_filename, &stat_buf) < 0) {
        return false;
    }

    if (S_ISDIR(stat_buf.st_mode)) {
        return is_system_dir_access_allowed(physical_filename,
                                            uid,
                                            stat_buf.st_uid,
                                            flags);
    }

    return is_system_file_access_allowed(physical_filename,
                                         uid,
                                         stat_buf.st_uid,
                                         flags);

/*

    //  Just leave this commented code out here
    //      a good place to start working at for any
    //      ~system file operation!

    // Checking this way will not work if any other filesystem
    // is being used. For eg. afs uses access control lists.
    // Should do the actual operation on some proxy file
    // to know whether it will succeed. This is implemented in
    // is_system_file_access_allowed and 
    // is_system_dir_access_allowed.
   
    switch (flags) {
        case NEST_READ_ACCESS:
        case NEST_LIST_ACCESS:
            if (uid == stat_buf.st_uid) {
                return (stat_buf.st_mode & S_IRUSR);
            }
            else {
                return (stat_buf.st_mode & S_IROTH);
            }

        case NEST_WRITE_ACCESS:
            if (uid == stat_buf.st_uid) {
                return (stat_buf.st_mode & S_IWUSR);
            }
            else {
                return (stat_buf.st_mode & S_IWOTH);
            }

        case NEST_DELETE_ACCESS:
            if (uid == stat_buf.st_uid) {
                return (stat_buf.st_mode & S_IXUSR);
            }
            else {
                return (stat_buf.st_mode & S_IXOTH);
            }

        case (NEST_READ_ACCESS | NEST_WRITE_ACCESS):
            if (uid == stat_buf.st_uid) {
                return ((stat_buf.st_mode & S_IRUSR) &&
                        (stat_buf.st_mode & S_IWUSR));
            }
            else {
                return ((stat_buf.st_mode & S_IROTH) &&
                        (stat_buf.st_mode & S_IWOTH));
            }

        case NEST_MODIFYAC_ACCESS:
            return (uid == stat_buf.st_uid);

        case NEST_INSERT_ACCESS:
            
    }
*/
}

bool
NestFileCatalog::is_system_access_allowed(const char *logical_filename,
                                          const char *user,
                                          nest_access_rights_t flags) {

    user_id_t       actual_uid;
    char            physical_filename[NEST_MAXPATHLEN];
    nest_obj_type_t filetype;
    char            *local_user;

    if (this->get_physical_name(physical_filename,
                                logical_filename) != NEST_SUCCESS) {
        return false;
    }

    if (this->get_type(&filetype, logical_filename) != NEST_SUCCESS) {
        return false;
    }

    if (NestAnonymousUser && (!strcmp( user, NestAnonymousUser ))) {
        return this->is_anonymous_access_allowed( physical_filename,
                                                  flags );
    }

    local_user = strdup(NestGetUserFromIdentity( user ));
    NestReplyStatus reply = this->get_uid(&actual_uid, local_user);
    if(local_user) NestFree(local_user);

    if (reply != NEST_SUCCESS) {
        return false;
    }
    
    return is_system_entry_access_allowed( physical_filename,
                                           actual_uid,
                                           flags );

}

/*
    only for ~system files
*/
bool
NestFileCatalog::is_anonymous_access_allowed( const char *physical_filename,
                                              nest_access_rights_t flags ) {

    nest_stat   stat_buf;
    char        dest_filename[NEST_MAXPATHLEN];

    if (NestLstat(physical_filename, &stat_buf) < 0) {
        return false;
    }

    switch (flags) {
        case NEST_READ_ACCESS:
        case NEST_LIST_ACCESS:
            return (stat_buf.st_mode & S_IROTH);

        case NEST_WRITE_ACCESS:
        case NEST_INSERT_ACCESS:
            /*
                takes care of hacks using ../s !!
            */
            if (this->get_dest_file( dest_filename,
                                     physical_filename,
									 NestAnonymousUser ) != NEST_SUCCESS) {
                return false;
            }

            return (strncmp( dest_filename, "/tmp/", 5 ) == 0);

        case NEST_DELETE_ACCESS:
        case (NEST_READ_ACCESS | NEST_WRITE_ACCESS):
        case NEST_MODIFYAC_ACCESS:
            return false;
    }

    return false;

}

/*
    only for NeST files
*/
bool
NestFileCatalog::is_access_allowed_for_entry(const char *file,
                                             const char *user,
                                             nest_access_rights_t flags) {

    nest_debug( NEST_DEBUG_ACL,
                "checking for entry %s - user %s - flags %d\n",
                file,
                user,
                flags );

    if (!file || !user) {
        return false;
    }

    nest_access_rights_t    file_acl;

    if (this->get_acl(file,
                      user,
                      IS_USER,
                      &file_acl) == NEST_SUCCESS) {

        if (file_acl & flags) {
            return true;
        }
    }

    if (strcmp( user, NEST_EVERYBODY )) {
        if (this->get_acl(file,
                          NEST_EVERYBODY,
                          IS_USER,
                          &file_acl) == NEST_SUCCESS) {

            if (file_acl & flags) {
                return true;
            }
        }
    }

    string  key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->file_catalog->GetClassAd(key);
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    ExprTree *group_list = user_class_ad->Lookup( "Groups" );
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;
    if (!list_itr.CurrentValue( val )) {
        return false;
    }

    string  group;
    do {
        if (val.IsStringValue( group )) {
            if (this->get_acl(file,
                              group.c_str(),
                              IS_GROUP,
                              &file_acl) == NEST_SUCCESS) {
                if (file_acl & flags) {
                    return true;
                }
            }
        }
    } while( list_itr.NextValue( val ) );

    return false;
}

bool
NestFileCatalog::is_access_allowed(const char *file,
                                   const char *user,
                                   nest_access_rights_t flags) {

    if (!file || !user) {
        return false;
    }

    nest_debug( NEST_DEBUG_ACL,
                "checking for %s - user %s - flags %d\n",
                file,
                user,
                flags );

    if (this->is_system_file( file )) {
        return this->is_system_access_allowed( file,
                                               user,
                                               flags );
    }

    if (NestAdminUser && !strcmp( user, NestAdminUser )) {
        return true;
    }

    if (NestAnonymousUser && !strcmp( user, NestAnonymousUser )) {
        return this->is_access_allowed( file,
                                        NEST_EVERYBODY,
                                        flags );
    }

    char    path_component[NEST_MAXPATHLEN];
    int     path_len;

    strncpy(path_component, file, NEST_MAXPATHLEN);
    path_len = NestStrlen(path_component) - 1;

    do {

        if (!this->is_access_allowed_for_entry(path_component,
                                               user,
                                               flags)) {
            return false;
        }

        //  go beyond the last '/', if any!

        path_len--;

        while ((path_len >= 0) && (path_component[path_len] != '/')) {
            path_len --;
        }

        if (path_len < 0) {
            break;
        }

        //  end the path after the '/'
        path_component[path_len + 1] = 0;

    } while (1);

    return true;
}

bool
NestFileCatalog::is_access_allowed_in_parent(const char *file,
                                             const char *user,
                                             nest_access_rights_t flags) {

    char  dest[NEST_MAXPATHLEN];

    NestReplyStatus  reply = this->get_parent_dir( dest,
                                                   file );
    if (reply != NEST_SUCCESS) {
        return false;
    }

    return this->is_access_allowed(dest, user, flags);
}

bool
NestFileCatalog::is_dir_empty( const char *logical_dir ) {

    if (!logical_dir) {
        return true;
    }

    if (this->is_system_file( logical_dir )) {
        char            physical_dir[NEST_MAXPATHLEN];
        struct dirent   **namelist;
        int num_entries;

        if (this->get_physical_name( physical_dir,
                                    logical_dir )
            != NEST_SUCCESS) {
            return false;
        }

        num_entries = NestScandir( physical_dir,
                                   &namelist,
                                   0,
                                   (int (*)(
                                       const struct dirent **,
                                       const struct dirent **)) alphasort );
    
        if (num_entries < 0) {
            return false;
        }

        for(int i=0; i<num_entries; i++) {
            if (namelist[i]->d_name) {
                NestFree(namelist[i]->d_name);
            }
            if (namelist[i]) {
                NestFree(namelist[i]);
            }
        }
        NestFree(namelist);

        /*
            empty iff "." and ".." are present
        */
        return (num_entries == 2);
    }

    string constraint = "";
    constraint = constraint +
                    "other.DirName == \"" + logical_dir + "\"";

    ExprTree    *constraint_tree = this->parser.ParseExpression(constraint);
    this->query.Query(this->file_catalog_view, constraint_tree);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return true;
    }

    ClassAd *file_ad = this->file_catalog->GetClassAd( query_key );

    if (!file_ad) {
        return true;
    }

    return false;
}

NestReplyStatus
NestFileCatalog::system_dir_list( NestFileInfo **result,
                                  const char *physical_dir ) {

    if (!physical_dir ||
        !result) {
        return NEST_BAD_PARAMETERS;
    }

    DIR *cwd;
    struct dirent *current;
    nest_stat statentry;
    NestFileInfo *info = NULL;
    NestFileInfo *head = NULL;
    char    physical_file[NEST_MAXPATHLEN];

    cwd = NestOpendir( physical_dir );
    if (cwd == NULL) {
        nest_perror("opendir");
        return NEST_FILE_NOT_FOUND;
    }

        // read each directory entry
    while( ( current = NestReaddir( cwd ) ) != NULL ) {

        if (strcmp( current->d_name, "."  ) == 0 ||
            strcmp( current->d_name, ".." ) == 0 ) {

            continue;    // ignore dot and dot-dot
        }

        NestSnprintf( physical_file,
                  NEST_MAXPATHLEN,
                  "%s/%s",
                  physical_dir,
                  current->d_name );

        if ( NestLstat( physical_file, &statentry ) < 0 ) {
            continue;
        }

        NestFileInfo *temp = info;
        info = (NestFileInfo*)NestMalloc(sizeof(NestFileInfo));

        strncpy( info->name, current->d_name, sizeof( info->name ) );
        info->size = statentry.st_size;
        strcpy( info->LotList, "" );

        info->next = NULL;
        info->classad = NULL;
        info->uniquifier = (NestFilehandle)statentry.st_ino;

        if ( temp != NULL ) {
            temp->next = info;
        }
        else {
            head = info;
        }
    }
    closedir( cwd );

    *result = head;
    return NEST_SUCCESS;
}


NestReplyStatus
NestFileCatalog::dir_list( NestFileInfo **result,
                           const char *logical_dir ) {

    if (!logical_dir ||
        !result) {
        return NEST_BAD_PARAMETERS;
    }

    if (this->is_system_file( logical_dir )) {
        char    physical_filename[NEST_MAXPATHLEN];

        NestReplyStatus reply = this->get_physical_name( physical_filename,
                                                         logical_dir );
        NEST_REPLY_CHECK(reply);

        return this->system_dir_list( result,
                                      physical_filename );
    }

    string constraint = "";
    constraint = constraint +
                    "other.DirName == \"" + logical_dir + "\"";

    ExprTree    *constraint_tree = this->parser.ParseExpression(constraint);
    this->query.Query(this->file_catalog_view, constraint_tree);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return NEST_CLASSAD_ERROR;
    }

    ClassAd         *file_ad;
    NestFileInfo    *tmp_node;
    char            tmp_name[NEST_MAXPATHLEN];

    *result = NULL;

    do {
        file_ad = this->file_catalog->GetClassAd( query_key );

        if (!file_ad) {
            continue;
        }

        tmp_node = (NestFileInfo *) NestMalloc (sizeof(NestFileInfo));
        if (!tmp_node) {
            continue;
        }

            // set type and name
        if (!file_ad->EvaluateAttrString( "LogicalName",
                                          tmp_name,
                                          NEST_MAXPATHLEN )) {
            NestFree(tmp_node);
            continue;
        }
        if (this->get_type( &(tmp_node->type), tmp_name ) != NEST_SUCCESS ) {
            NestFree( tmp_node );
            continue;
        }

        this->get_reverse_logical_name( tmp_node->name,
                                        tmp_name );

            // set size
        int     tmp_size;
        if (!file_ad->EvaluateAttrInt( "Size",
                                       tmp_size )) {
            NestFree(tmp_node);
            continue;
        }
        tmp_node->size = tmp_size > 0 ? tmp_size : 0;

        tmp_node->classad = file_ad;

            // set uniquifier
        if ( !file_ad->EvaluateAttrString( "PhysicalName",
                                            tmp_name,
                                            NEST_MAXPATHLEN ) ) {
            NestFree( tmp_node );
            continue;
        }
        this->get_file_uniquifier(tmp_name, &(tmp_node->uniquifier) );

        tmp_node->LotList[0] = 0;
        if (NestEnableLots) {

            ExprTree *lot_list = file_ad->Lookup( "LotHandle" );
            if (!lot_list) {
                NestFree(tmp_node->classad);
                NestFree(tmp_node);
                continue;
            }

            string str_lot_list = "";
            this->unparser.Unparse( str_lot_list, lot_list );
            strncpy(tmp_node->LotList,
                    str_lot_list.c_str(),
                    sizeof(tmp_node->LotList));

        }

        tmp_node->next = *result;
        *result = tmp_node;

    } while (this->query.Next(query_key));

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::attach_lot_to_file( const char *logical_file,
                                     NestLotHandle *handle ) {
    if(!logical_file || !handle) {
        return NEST_BAD_PARAMETERS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + logical_file;
        
    char str_lot_handle[NEST_NAMELEN];
    if(NestHandle_to_string( handle, str_lot_handle, NEST_NAMELEN ) < 0 ) {
        return NEST_LOT_INVALID;
    }
    
    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string attr = "AttachedLots";

    ExprTree *lot_list = file_class_ad->Lookup( attr );
    if (!lot_list) {
        return NEST_UNKNOWN_ERROR;
    }

    string str_lot_list, str_new_lot_list, str_find_lot;

    unparser.Unparse( str_lot_list, lot_list );

    str_find_lot = "\"";
    str_find_lot = str_find_lot + str_lot_handle + "\"";
    
    if ( str_lot_list.find(str_find_lot) == string::npos ) {
        /*
         * Lot is not present in the list and needs to be attached 
         * to the file
         */
        if (NestAdd_entry_to_list( str_new_lot_list, 
                                   str_lot_list,
                                   str_lot_handle ) != NEST_SUCCESS ) {
            return NEST_UNKNOWN_ERROR;
        }
        
        string modify_classad = "";
        modify_classad = modify_classad +
                         " [ " + attr + " = " + str_new_lot_list + " ] ";
        modify_classad = " [ Updates = " + modify_classad + " ] ";
        ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );
        
        if (!this->file_catalog->ModifyClassAd( key, mod_classad )) {
            return NEST_CLASSAD_ERROR;
        }
    }
    
    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::detach_lot_from_file( const char *logical_file,
                                       NestLotHandle *handle ) {
    if(!logical_file || !handle) {
        return NEST_BAD_PARAMETERS;
    }

    string key = STORAGE_OBJECT_KEY_PREFIX;
    key = key + logical_file;
        
    char str_lot_handle[NEST_NAMELEN];
    if(NestHandle_to_string( handle, str_lot_handle, NEST_NAMELEN ) < 0 ) {
        return NEST_LOT_INVALID;
    }
    
    ClassAd *file_class_ad = this->file_catalog->GetClassAd(key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string attr = "AttachedLots";

    ExprTree *lot_list = file_class_ad->Lookup( attr );
    if (!lot_list) {
        return NEST_UNKNOWN_ERROR;
    }

    string str_lot_list, str_new_lot_list, str_find_lot;

    unparser.Unparse( str_lot_list, lot_list );

    str_find_lot = "\"";
    str_find_lot = str_find_lot + str_lot_handle + "\"";
    
    if ( str_lot_list.find(str_find_lot) != string::npos ) {
        /*
         * Lot is present in the list and needs to be detached 
         * from the file
         */
        if (NestDelete_entry_from_list( str_new_lot_list, 
                                        str_lot_list,
                                        str_lot_handle ) != NEST_SUCCESS ) {
            return NEST_UNKNOWN_ERROR;
        }
        
        string modify_classad = "";
        modify_classad = modify_classad +
                         " [ " + attr + " = " + str_new_lot_list + " ] ";
        modify_classad = " [ Updates = " + modify_classad + " ] ";
        ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );
        
        if (!this->file_catalog->ModifyClassAd( key, mod_classad )) {
            return NEST_CLASSAD_ERROR;
        }
    }
    
    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::initialize(ClassAdCollection *resources,
                            const char *rootdir) {

    if (!resources || !rootdir) {
        return NEST_BAD_PARAMETERS;
    }

    this->file_catalog = resources;
    this->root_dir = rootdir;

    int last_byte = this->root_dir.length() - 1;
    if (this->root_dir[last_byte] != '/') {
        this->root_dir = this->root_dir + "/";
    }

    this->query.Bind(resources);
    this->file_catalog_view = NEST_FILE_CATALOG_VIEW;

    NestReplyStatus reply = this->add_entry("/",
                                            rootdir,
                                            "/",
                                            NEST_EVERYBODY,
                                            NEST_NO_GROUP,
                                            NEST_DIR);

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Adding entry for system root failed! %s \n",
                    NestErrorString(reply) );
        return reply;
    }

/*
    char    system_home[NEST_MAXPATHLEN];

    reply = this->get_logical_name(system_home,
                                   "/",
                                   NestSystemUser);
    NEST_REPLY_CHECK(reply);

    reply = this->add_entry(system_home,
                            "/",
                            "/",
                            NEST_EVERYBODY,
                            NEST_NO_GROUP,
                            NEST_DIR);

    NEST_REPLY_CHECK(reply);
*/

    ViewName    parent_view = "root";

    string view_constraint(""),
           view_rank(""),
           view_partition_expr("");

    view_constraint = view_constraint +
                      " other.Type == \"" + STORAGE_OBJECT_KEY_PREFIX + "\" ";

    if (!this->file_catalog->CreateSubView( this->file_catalog_view,
                                            parent_view,
                                              view_constraint,
                                            view_rank,
                                            view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "FileCatalogManager Couldn't create view for itself!\n");
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::recover(ClassAdCollection *resources,
                         const char *rootdir) {

    this->file_catalog = resources;
    this->root_dir = rootdir;
    int last_byte = this->root_dir.length() - 1;
    if (this->root_dir[last_byte] != '/') {
        this->root_dir = this->root_dir + "/";
    }

    this->query.Bind(resources);
    this->file_catalog_view = NEST_FILE_CATALOG_VIEW;

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::synchronize() {

    return NEST_SUCCESS;
}

NestReplyStatus
NestFileCatalog::shutdown() {

    return NEST_SUCCESS;
}








