/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <stdarg.h>

#include "nest_common.h"
// #include "nest_filecache_model_simple.h"
#include "nest_filecache_model_infinite.h"

void
Usage( char *prog, int line ) {
    fprintf( stderr, "Usage: %s:%d -i input[,input2,...] "
                     "cachesize [cachesize ...]\n", 
                     prog, line );
    exit( 1 );
}

vector<nestFilecacheModel *> caches;
vector<nestFilecacheModel *>::iterator cache_itr;
vector<FILE *> files;
vector<FILE *>::iterator file_itr;

struct timeval timer;
long long insertTime = 0;
long long infiniteTime = 0;
long long scanTime = 0;

int pagesize = 4096;
int debug = 0;
int verbose = 0;
int lineNum = 1;

void
printDatapoints( FILE *output, vector<nestFilecacheModel *> *caches ) {
    for( cache_itr = caches->begin(); cache_itr != caches->end(); cache_itr++ ){
        nestFilecacheModel *model = *cache_itr;
        fprintf( output, "%f %f %ld %ld %ld\n", 
                model->getsize(), 
                model->hit_rate() * 100,
                model->gethits(),
                model->getmisses(),
                model->getevictions()
                );
    }
}

float
getRate( int lineNum ) {
    struct timeval now;
    NestGettimeofday( &now );
    long micros = NestTimediff( &timer, &now, __FUNCTION__ );
    float secs  = (float)micros / 1000000;
    float rate  = (float)lineNum / secs;
    return rate;
}

void
readInputFile( FILE *input ) {
        // read the input file
    char action[8];
    char file  [256];
    int offset, length;
    char fileblock[8192];

    while( true ) {
        struct timeval fscanf_s;
        struct timeval fscanf_e;

        NestGettimeofday( &fscanf_s );
        if ( 4 != fscanf( input, "%s %s %d %d",action,file,&offset,&length) ) {
            if ( verbose ) fprintf( stderr, "# done reading at line #%d\n",
                                            lineNum );
            break;
        }
        NestGettimeofday( &fscanf_e );
        scanTime += NestTimediff( &fscanf_s, &fscanf_e, __FUNCTION__ );

        if ( debug ) fprintf( stderr, "read line %d\n", lineNum );
        if ( verbose && lineNum % 500 == 0 ) 
            fprintf( stderr, "# read line #%7d %8.4f lines/sec\n",
                                                 lineNum, getRate( lineNum ) );
        lineNum++;

        int pageNum   = int( offset / pagesize );
        int pageStart = pageNum * pagesize;
        struct timeval insert_s;
        struct timeval insert_e;
                    
        while( offset + length > pageStart ) {

            NestGettimeofday( &fscanf_s );
            snprintf( fileblock, 8192, "%s.%d", file, pageStart );
            NestGettimeofday( &fscanf_e );
            scanTime += NestTimediff( &fscanf_s, &fscanf_e, __FUNCTION__ );

            //now infinite is just a normal looking cache
            //NestGettimeofday( &insert_s );
            //if ( infinite ) infinite->insert( NULL, fileblock, pagesize );
            //NestGettimeofday( &insert_e );
            //infiniteTime += NestTimediff( &insert_s, &insert_e, __FUNCTION__);

            NestGettimeofday( &insert_s );
            for(cache_itr = caches.begin();cache_itr!=caches.end();cache_itr++){
                bool hit = false;
                nestFilecacheModel *model = *cache_itr;
                model->insert( &hit, fileblock, pagesize );
                if ( debug ) {
                    fprintf( stderr, "# insert %s into cache %.2f (%ld): %d\n", 
                        fileblock,
                        model->getsize(),
                        model->getentries(),
                        hit );
                }
            }
            NestGettimeofday( &insert_e );
            insertTime += NestTimediff( &insert_s, &insert_e, __FUNCTION__);

            pageStart  += pagesize;
        }
    }
}

int
main( int argc, char *argv[] ) {

    int c;
    char *prog = argv[0];
    //char *graphfile = "working_set.eps";
    FILE *input = stdin;
    FILE *output = stdout; 

    int hashsize = 102877;
    char *inputfile;

    nestFilecacheModelInfinite *infinite = NULL;
    bool doinfinite = false;

    caches.clear();
    files.clear();

        // read args
    while( ( c = getopt( argc, argv, "Il:g:vhi:o:p:dH:" )) != EOF ) {
        switch( c ) {
        case 'I':
            doinfinite = true;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'd':
            debug = 1;
            break;
        case 'p':
            pagesize = atoi( optarg );
            fprintf( stderr, "Set pagesize to %d.\n", pagesize );
            break;
        case 'i':
            inputfile = strtok( optarg, "," );
            do {
                input = fopen( inputfile, "r" );
                if ( ! input ) {
                    fprintf( stderr, "open %s: %s.\n", inputfile, 
                            strerror( errno) );
                    exit( 1 );
                } else {
                    fprintf( stderr, "Opening input %s.\n", inputfile );
                    files.push_back( input );
                }
            } while( NULL != ( inputfile = strtok( NULL, "," ) ) );
            break;
        case 'g':
        //    graphfile = strdup( optarg );
        //    break;
        case 'o':
        case 'H':
            fprintf( stderr, "# -o and -g and -H options are deprecated.\n" );
            exit( 1 );
            /*
            output = fopen( optarg, "w" );
            if ( ! output ) {
                perror( "fopen" );
                exit( 1 );
            }
            */
            break;
        case 'h':
            Usage( prog, __LINE__ );
            break;
        default:
            Usage( prog, __LINE__ );
            break;
        }
    }
    if ( optind == argc || ( ! output ) ) {
        Usage( prog, __LINE__ );
    }
    if ( files.size() <= 0 ) {
        Usage( prog, __LINE__ );
    }

        // set up our caches
    if ( doinfinite ) {
        infinite = new nestFilecacheModelInfinite(hashsize);
        caches.push_back( infinite );
    }
    while( optind < argc ) {
        float megs = atof( argv[optind++] );
        if ( megs <= 0 ) {
            fprintf( stderr, "# not computing cache for nonpositive size %f.\n",
                    megs );
            continue;
        }
        caches.push_back(new nestFilecacheModel( megs, hashsize ));
    }
        

    NestGettimeofday( &timer );
    for(file_itr = files.begin();file_itr!=files.end();file_itr++){
        readInputFile( *file_itr);
        printDatapoints( stdout, &caches );
    }

    long long totaltime;
    struct timeval end;
    NestGettimeofday( &end );
    totaltime = NestTimediff( &timer, &end, __FUNCTION__  );

    if ( verbose ) fprintf( stderr, 
            "# Totaltime %lld, scanTime %lld, insertTime %lld, infinite %lld\n",
            totaltime / MILLION, scanTime / MILLION, insertTime / MILLION,
            infiniteTime / MILLION );

    return 0;
}

        // print out the output
    /*
    float ideal = infinite->hit_rate();
    fprintf( output,
            "#! /s/std/bin/gnuplot\n"
            "\n"
            "set terminal postscript eps solid 20\n"
            "set output \"%s.eps\"\n"
            "#set logscale x 2\n"
            "#set xtics border mirror norotate 2,4\n"
            "set xlabel \"Cache size (MBs)\"\n"
            "set ylabel \"Hit rate (%%)\" \n"
            "set yrange [ 0.00000 : 100.000 ] noreverse nowriteback\n"
            "plot \"-\" notit with linespoints, "
            "f(x) = y, y = %f, f(x) title \"Infinite cache\"\n",
            graphfile,
            ideal * 100
            );
    printDatapoints( output, &caches );
    fprintf( output, 
            "e\n"
            "set terminal gif\n"
            "set output \"%s.gif\"\n"
            "plot \"-\" notit with linespoints, "
            "f(x) = y, y = %f, f(x) title \"Infinite cache\"\n",
            graphfile,
            ideal * 100
           );
    printDatapoints( output, &caches );
    */
    
