/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_identity_manager.h"
#include "nest_protocol.h"


NestIdentityManager::NestIdentityManager(const char *userlist,
                                         const char *denylist){
    this->Identities = NULL;

    if (!userlist) {
        this->userlist = NEST_EVERYBODY;
    } else {
        this->userlist = userlist;
    }

    if (!denylist) {
        this->denylist = "";
    } else {
        this->denylist = denylist;
    }

    /*
        default case used by this->initialize is that
        nobody was present in either the userlist or denylist, previously!
    */

    this->prev_userlist = "";
    this->prev_denylist = "";

    return;
}

NestIdentityManager::~NestIdentityManager(){
    return;
}

NestReplyStatus
NestIdentityManager::get_total_nest_users( int *total_nest_users ) {
    int temp_total_nest_users;
    string key = TOTAL_USERS_KEY;

    ClassAd *total_users_ad = this->Identities->GetClassAd( key );
    if(total_users_ad){
        bool result = total_users_ad->EvaluateAttrNumber( "TotalNestUsers",
                                                     temp_total_nest_users );
        if(!result){
            *total_nest_users = -1;
            return NEST_CLASSAD_ERROR;
        } else {
            *total_nest_users = temp_total_nest_users;
            return NEST_SUCCESS;
        }

    } else {
        *total_nest_users = -1;
        return NEST_CLASSAD_ERROR;
    }

}

NestReplyStatus
NestIdentityManager::get_total_nest_groups( int *total_nest_groups ) {
    int temp_total_nest_groups;
    string key = TOTAL_GROUPS_KEY;

    ClassAd *total_groups_ad = this->Identities->GetClassAd( key );
    if(total_groups_ad){
        bool result = total_groups_ad->EvaluateAttrNumber( "TotalNestGroups",
                                                     temp_total_nest_groups );
        if(!result){
            *total_nest_groups = -1;
            return NEST_CLASSAD_ERROR;
        } else {
            *total_nest_groups = temp_total_nest_groups;
            return NEST_SUCCESS;
        }

    } else {
        *total_nest_groups = -1;
        return NEST_CLASSAD_ERROR;
    }

}

bool
NestIdentityManager::is_present_in_list(const char *userpool,
                                        const char *user) {

    if (!userpool ||
        !user) {
        return false;
    }

    int     len = NestStrlen(user),
            userpool_len = NestStrlen(userpool);

    char    tmp_user[NEST_NAMELEN],
            *tmp_ptr;

    if (len > userpool_len) {
        return false;
    }

    if (!strcmp(user, userpool)) {
        return true;
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, ",%s,", user);
    if (strstr(userpool, tmp_user)) {
        return true;
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, "%s,", user);
    if ((tmp_ptr = strstr(userpool, tmp_user))) {
        if (tmp_ptr == userpool) {
            return true;
        }
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, ",%s", user);
    if ((tmp_ptr = strstr(userpool, tmp_user))) {
        if ((tmp_ptr + len + 1) == (userpool + userpool_len)) {
            return true;
        }
    }

    return false;
}

bool
NestIdentityManager::is_user_access_allowed (const char *user) {

    if (!NestAdminUser || !user) {
        return false;
    }

    return (strcmp(NestAdminUser, user) == 0);
}

bool
NestIdentityManager::is_group_access_allowed (const char *group,
                                              const char *user,
                                              bool allow_members) {

    if (!group) {
        return false;
    }

    if (!NestAdminUser && !user) {
        return true;
    }

    if (NestAdminUser && user && (!strcmp( NestAdminUser, user ))) {
        return true;
    }

    if (!user) {
        return false;
    }

    if (!strcmp(group, NEST_NO_GROUP)) {
        return false;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_ad = this->Identities->GetClassAd( key );
    if (!group_ad) {
        return false;
    }

    string owner_user;
    if (!group_ad->EvaluateAttrString( "OwnerUser", owner_user )) {
        return false;
    }

    if(!allow_members)
        return (strcmp(owner_user.c_str(), user) == 0);
    else
        return is_user_in_group(user, group);
}

/*
    Refer to file "doc/Userlist.Denylist.Semantics" to look at
    how the following boolean function was derived:

    Result = ((~A (((C + E)(B + ~D)) + CE )) + (ABE ( ~C~D + CD )))

    where   A is "Present in Current-Denylist",
            B is "Present in Previous-Denylist",
            C is "Present in Current-Userlist",
            D is "Present in Previous-Userlist" and
            E is "Valid-Current-User".

    Looks awfully weird, I agree, but really helps to give a pre-defined
    and meaningful semantics with providing both user and deny lists, and
    the ability to add/delete users on the fly! Trust me!!

*/

static const char* bs(bool v) { return ( (v) ? "true" : "false" ); }
#if 0
bool
NestIdentityManager::is_valid_nest_user(const char *user, bool is_init)
{
    /* Refer the /doc/Userlist.Denylist.Semantics file for the valid user
     * semantics. If the current/prev userlist is null then it is assumed to
     * contain all users from the grid-map-file. If the current/prev denylist
     * is null then no users are in that list. One exception is on the very
     * first synchronize call during initialization, if previous userlist is
     * null, then no users are in that list
     * This interpretation is necessary due to the fact that users can be added
     * on fly.
     */

    bool deny_c = this->is_present_in_list(this->denylist.c_str(), user);
	bool deny_p = this->is_present_in_list(this->prev_denylist.c_str(), user);
	bool ulist_c;
	bool ulist_p;
	bool valid = this->is_user_present( user );

    if(!strcmp(this->userlist.c_str(),""))
        ulist_c = true;
    else
        ulist_c = this->is_present_in_list(this->userlist.c_str(), user);

    if(!is_init && !strcmp(this->userlist.c_str(),""))
        ulist_p = true;
    else
        ulist_p = this->is_present_in_list(this->prev_userlist.c_str(), user);

	nest_debug( NEST_DEBUG_AUTH,
				"%s: init=%s deny=%s deny_p=%s ulist_c=%s ulist_p=%s valid=%s\n",
				user, bs(is_init),
				bs(deny), bs(deny_p), bs(ulist_c), bs(ulist_p), bs(valid) );

# if 1
	// Default policy: refuse
	bool	rval = false;

	// Now using hopefully more clear logic..
	if ( !deny_c ) {
		// The user isn't on the "deny" list..
		if ( ulist_c ) {
			// The user in current in the "user" list
			if ( valid ) {
				// The user is currently valid; let 'em in.
				rval = true;
			}
			else if ( deny_p || !ulist_p ) {
				// Not on the prvious deny or previous user list
				rval = true;
			}
		} else if ( valid ) {
			// The user is currently valid
			if ( deny_p || !ulist_p ) {
				rval = true;
			}
		}
	} else {
		// They're on the deny list
		if ( deny_p && valid ) {
			// On the previous deny and current user list
			if ( !ulist_c && !ulist_p ) {
				// Not on current user list & not on previous user list
				rval = true;
			}
			else if ( ulist_c && ulist_p ) {
				// On current user list & on previous user list
				rval = true;
			}
		}
	}

	return rval;

# else
	// More readable version of the Ven's old logic
    return ( ( !deny_c
			   && (
				   ( ( ulist_c || valid ) && ( deny_p || !ulist_p ) )
				   ||
				   ( ulist_c && valid ) )
			   )
			 ||
			 ( deny
			   &&
			   deny_p
			   &&
			   valid
			   && 
			   ( (!ulist_c && !ulist_p)
				 ||
				 (ulist_c && ulist_p) )
			   )
			 );
# endif

}
#else
bool
NestIdentityManager::is_valid_nest_user(const char *user, bool is_init)
{
    /* Refer the /doc/Userlist.Denylist.Semantics file for the valid user
     * semantics. If the current/prev userlist is null then it is assumed to
     * contain all users from the grid-map-file. If the current/prev denylist
     * is null then no users are in that list. One exception is on the very
     * first synchronize call during initialization, if previous userlist is
     * null, then no users are in that list
     * This interpretation is necessary due to the fact that users can be added
     * on fly.
     */

    bool deny = this->is_present_in_list(this->denylist.c_str(), user);
	bool ulist;
	bool valid = this->is_user_present( user );

    if(!strcmp(this->userlist.c_str(), ""))
        ulist = true;
    else
        ulist = this->is_present_in_list(this->userlist.c_str(), user);

	bool	ok = false;

	if ( deny ) {
		// Do nothing
	} else if ( ulist || valid ) {
		ok = true;
	} else {
		// Do nothing
	}

# if 0
	nest_debug( NEST_DEBUG_AUTH,
				"%s: init=%s deny=%s ulist=%s valid=%s ->%s\n",
				user, bs(is_init), bs(deny), bs(ulist), bs(valid), bs(ok) );
# endif

	return ok;
}
#endif

NestReplyStatus
NestIdentityManager::get_actual_nest_group(char *actual_nest_group) {

    if (!actual_nest_group) {
        return NEST_BAD_PARAMETERS;
    }

    group_ent ge;

    if (Nestgetgrpnam( ActualGroup, &ge, SystemGrpFile) < 0) {
        strcpy(actual_nest_group, "root");
        return NEST_SUCCESS;
    }

    strcpy(actual_nest_group, ge.grp_name);
    return NEST_SUCCESS;
}

bool
NestIdentityManager::is_user_present(const char *user) {

    if (!user) {
        return false;
    }

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (!user_ad) {
        return false;
    }

    bool    result,
            valid;

    result = user_ad->EvaluateAttrBool( "ValidEntry", valid );
    if (!result) {
        return false;
    }

    return valid;
}

bool
NestIdentityManager::is_group_present(const char *group) {

    if (!group) {
        return false;
    }

    if (!strcmp(group, NEST_NO_GROUP)) {
        return true;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    return (this->Identities->GetClassAd(key) != NULL);

}


bool
NestIdentityManager::is_user_in_group(const char *user,
                                      const char *group) {

    string  key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return false;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;

    if (!list_itr.CurrentValue( val )) {
        return false;
    }

    string  member;

    do {
        if (val.IsStringValue( member )) {
            if ( member == user ) {
                return true;
            }
        }
    } while( list_itr.NextValue( val ) );

    return false;
}


bool
NestIdentityManager::is_group_in_user(const char *user,
                                      const char *group) {

    string  key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);
    if (!user_class_ad) {
        return false;
    }

    string  attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;

    if (!list_itr.CurrentValue( val )) {
        return false;
    }

    string  grp_entry;

    do {
        if (val.IsStringValue( grp_entry )) {
            if ( grp_entry == group ) {
                return true;
            }
        }
    } while( list_itr.NextValue( val ) );

    return false;
}


NestReplyStatus
NestIdentityManager::add_user_to_group_members(const char *user,
                                               const char *group) {

    if (!strcmp(group, NEST_NO_GROUP)) {
        return NEST_SUCCESS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr("Members");
    ExprTree *member_list = group_class_ad->Lookup( attr );

    if (!member_list) {
        return NEST_CLASSAD_ERROR;
    }

    string members,
           new_members,
           str_user(user);

    unparser.Unparse( members, member_list );
    //  delete member_list;

    NestReplyStatus reply = NestAdd_entry_to_list( new_members,
                                                   members,
                                                   str_user );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_members + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
NestIdentityManager::add_group_to_user_groups(const char *user,
                                              const char *group) {

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);

    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string groups,
           new_groups,
           str_group(group);
    unparser.Unparse( groups, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestAdd_entry_to_list( new_groups,
                                               groups,
                                               str_group );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_groups + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
NestIdentityManager::delete_user_from_group_members(const char *user,
                                                    const char *group) {

    if (!strcmp(group, NEST_NO_GROUP)) {
        return NEST_SUCCESS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string members,
           new_members,
           str_user(user);
    unparser.Unparse( members, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestDelete_entry_from_list( new_members,
                                                    members,
                                                    str_user );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_members + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
NestIdentityManager::delete_group_from_user_groups(const char *user,
                                               const char *group) {

    string  key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);

    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string groups,
           new_groups,
           str_group(group);
    unparser.Unparse( groups, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestDelete_entry_from_list( new_groups,
                                                    groups,
                                                    str_group );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_groups + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


bool
NestIdentityManager::is_active_group(const char *group) {


    if (!strcmp(group, NEST_NO_GROUP)) {
        return true;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return false;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;
    if (!list_itr.CurrentValue( val )) {
        //  delete group_list;
        return false;
    }

    //  Verify!
    string  member;
    if (val.IsStringValue( member )) {
        //  delete group_list;
        return true;
    }

    //  delete group_list;
    return false;
}

NestReplyStatus
NestIdentityManager::get_best_group( string &str_group,
                                     const char *user ) {

    if (!user) {
        return NEST_BAD_PARAMETERS;
    }

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->Identities->GetClassAd( key );
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    ExprTree *group_list = user_class_ad->Lookup("Groups");
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    group_itr((ExprList *) group_list);

    Value   group_val;

    if (!group_itr.CurrentValue( group_val )) {
        return NEST_GROUP_NOT_FOUND;
    }

    if (!group_val.IsStringValue( str_group )) {
        return NEST_GROUP_NOT_FOUND;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::get_group_owner( const char *group,
                                      char *user) {
    if(!group) {
        return NEST_BAD_PARAMETERS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd( key );
    if(!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    bool result;
    string owner;
    result = group_class_ad->EvaluateAttrString( "OwnerUser", owner );
    
    if(!result) {
        return NEST_CLASSAD_ERROR;
    }
       
    strcpy( user, owner.c_str() );
           
    return NEST_SUCCESS;
}  

NestReplyStatus
NestIdentityManager::set_default_group(const char *user,
                                       const char *group,
                                       bool test_presence) {

    if (!user || !group) {
        return NEST_BAD_PARAMETERS;
    }

    if (test_presence) {
        if (!this->is_group_in_user(user, group)) {
            /*
                Group not found in user's Groups list!
            */
            return NEST_GROUP_NOT_FOUND;
        }
    }

    string key = USER_KEY_PREFIX;
    key = key + user;

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ DefaultGroup = \"" + group + "\" ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::reset_default_group(const char *user,
                                         const char *group) {

    if (!user || !group) {
        return NEST_BAD_PARAMETERS;
    }

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string  default_group;

    if (user_ad->EvaluateAttrString( "DefaultGroup", default_group )) {

        /*
            if user's default_group is not group
            don't bother
        */

        if (strcmp(default_group.c_str(), group)) {
            return NEST_SUCCESS;
        }
    }

    string  str_group;

    NestReplyStatus reply = this->get_best_group( str_group,
                                                  user );

    if (reply == NEST_SUCCESS) {
        return this->set_default_group( user,
                                        str_group.c_str(),
                                        false );
    }

    /*
        oops - the user's last group has been removed from its
        "Groups" list

        add user to NEST_NO_GROUP
    */

    reply = this->add_user_in_group( user,
                                     NEST_NO_GROUP,
                                     NestAdminUser );

    NEST_REPLY_CHECK(reply);

    return this->set_default_group( user,
                                    NEST_NO_GROUP,
                                    true );

}

NestReplyStatus
NestIdentityManager::add_user_in_group(const char *user,
                                       const char *group,
                                       const char *cur_user) {

    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add a non-existing user to a group %s\n",
                    user,
                    group );
        return NEST_USER_NOT_FOUND;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add user to a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    /*
        if ((strcmp(group, NEST_NO_GROUP) != 0) &&
            this->is_user_in_group(user, group)) {
    */

    if (this->is_group_in_user(user, group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add an already existing user: %s to group %s\n",
                    user,
                    group );
        return NEST_USER_EXISTS;
    }


    /*
        Start:  keep your fingers crossed!
                it shouldn't crash now
    */

    NestReplyStatus reply = this->add_user_to_group_members(user, group);
    NEST_REPLY_CHECK(reply);

    reply = this->add_group_to_user_groups(user, group);

    /*
        End:    relax - the prayer is over and you are safe
    */

    /*
        Oh yeah! these two operations needs to be atomic!
        NestTime to use a Collection Transaction!
    */

    /*
        Not sure we should have this

    this->set_default_group(user, group, false);
    */

    return reply;
}

NestReplyStatus
NestIdentityManager::delete_user_in_group(const char *user,
                                          const char *group,
                                          const char *cur_user) {

    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete a non-existing user %s from group %s\n",
                    user,
                    group);
        return NEST_USER_NOT_FOUND;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete user to a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    /*
        if ((strcmp(group, NEST_NO_GROUP) != 0)&&
            !this->is_user_in_group(user, group)) {
    */

    if (!this->is_group_in_user(user, group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete a non-member %s from group %s\n",
                    user,
                    group );
        return NEST_USER_NOT_FOUND;
    }

    /*
        Start:  keep your fingers crossed!
                it shouldn't crash now
    */

    NestReplyStatus reply = this->delete_user_from_group_members(user, group);
    NEST_REPLY_CHECK(reply);

    reply = this->delete_group_from_user_groups(user, group);

    /*
        End:    relax - the prayer is over and you are safe
    */

    /*
        Oh yeah! these two operations needs to be atomic!
        NestTime to use a Collection Transaction!
    */

    this->reset_default_group(user, group);

    return reply;
}

NestReplyStatus
NestIdentityManager::add_user(const char *user,
                              const char *group,
                              const char *cur_user,
                              const char *identity,
                              bool log_existing ) {

    if (!user || !group || !identity) {
        return NEST_BAD_PARAMETERS;
    }

    if( strcmp(identity, "") &&
        Nestinsert_grid_mapfile_entry(GridMapFile, identity, user) < 0){
        return NEST_SYSTEM_FILE_UPDATE_FAILED;
    }

    if (!this->is_group_present(group)) {
        if ( log_existing ) {
            nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Trying to add an user %s to a non-existing group %s\n",
                        user,
                        group );
        }
        return NEST_GROUP_NOT_FOUND;
    }

    string  key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (user_ad) {
        bool    result;
        bool    valid;

        result = user_ad->EvaluateAttrBool( "ValidEntry", valid );
        if (!result) {
            return NEST_CLASSAD_ERROR;
        }

        if (valid) {
            if ( log_existing ) {
                nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                            "Trying to add an already existing user %s\n",
                            user);
            }
            return NEST_USER_EXISTS;

        } else {

            string modify_classad1 = "";
            modify_classad1 = modify_classad1 +
                                    "[   Updates  = " +
                                        "[  ValidEntry = true; ] " +
                                    "]";


            ClassAd *mod_classad1 = parser.ParseClassAd( modify_classad1,
                                                         true );

            if (!this->Identities->ModifyClassAd( key, mod_classad1 )) {
                return NEST_CLASSAD_ERROR;
            }

            string modify_classad2 = "";

            modify_classad2 = modify_classad2 +
                                    "[   Updates  = " +
                                        "[ Groups = { \""+group+"\" } ]" +
                                    "]";


            ClassAd *mod_classad2 = parser.ParseClassAd( modify_classad2,
                                                         true );

            if (!this->Identities->ModifyClassAd( key, mod_classad2 )) {
                return NEST_CLASSAD_ERROR;
            }
        }
    }
    else {

        char    str_size[NEST_NAMELEN];
        NestSnprintf(str_size, NEST_NAMELEN, "%f", 0.0);
        char    *local_user = strdup( NestGetUserFromIdentity(user) ); 
        
        string str_user_classad = "";
        str_user_classad = str_user_classad +
                                "[  Type = \"" + IDENTITY_TYPE + "\"; " +
                                "   PerceivedName = \"" + user + "\"; " +
                                "   ActualName = \"" + local_user + "\"; " +
                                "   ValidEntry = true; " +
                                "   IsUser = true; " +
                                "   Groups = {\"" + group + "\"}; " +
                                "   DefaultGroup = \"" + group + "\"; " +
                                "   Members = {}; " +
                                "   StorageSpaceUsed = " + str_size + "; " +
                                "   DefaultLotHandle = \"\" " +
                                "] " ;
        if(local_user) NestFree(local_user);

        user_ad = parser.ParseClassAd( str_user_classad, true );

        if (!user_ad) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "Parsing UserClassAd for user %s failed!\n",
                        user );
            return NEST_CLASSAD_ERROR;
        }

        if (!this->Identities->AddClassAd( key, user_ad )) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "Adding UserClassAd for user %s failed!\n",
                        user );
            return NEST_CLASSAD_ERROR;
        }
    }

    NestReplyStatus reply = this->add_user_to_group_members(user, group);

    if (reply != NEST_SUCCESS) {
        this->delete_user(user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );
        return reply;
    }

    int total_nest_users;

    reply = this->get_total_nest_users(&total_nest_users);

    if (reply != NEST_SUCCESS) {
        this->delete_user(user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );
        return reply;
    }

    total_nest_users++;
    key = TOTAL_USERS_KEY;

    char str_total_nest_users[NEST_NAMELEN];
    NestSnprintf(str_total_nest_users, NEST_NAMELEN, "%d",
                                            total_nest_users );

    string modify_classad3 = "";
    modify_classad3 = modify_classad3 +
                      "[ TotalNestUsers = " + str_total_nest_users + "]";

    modify_classad3 = " [ Updates = " + modify_classad3 + " ] ";

    ClassAd *mod_classad3 = parser.ParseClassAd( modify_classad3, true );

    if(!mod_classad3){
        return NEST_CLASSAD_ERROR;
    }
    if (!this->Identities->ModifyClassAd( key, mod_classad3 )) {
        this->delete_user(user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );

        return NEST_CLASSAD_ERROR;
    }

    NestSync();

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::update_user_usage(const char *user,
                                       double more_size) {
    string key = USER_KEY_PREFIX;
    key = key + user;
    ClassAd *user_ad = this->Identities->GetClassAd(key);
    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    bool result;
    double size;
    result = user_ad->EvaluateAttrNumber( "StorageSpaceUsed", size );
    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    char    str_size[NEST_NAMELEN];
    NestSnprintf(str_size, NEST_NAMELEN, "%f", size + more_size);

    string str_modify_classad = "";
    str_modify_classad = str_modify_classad +
                         "[  StorageSpaceUsed = " + str_size + "]";

    str_modify_classad = " [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, modify_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::update_group_usage(const char *group,
                                       double more_size) {
    string key = GROUP_KEY_PREFIX;
    key = key + group;
    ClassAd *group_ad = this->Identities->GetClassAd(key);
    if (!group_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    bool result;
    double size;
    result = group_ad->EvaluateAttrNumber( "StorageSpaceUsed", size );
    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    char    str_size[NEST_NAMELEN];
    NestSnprintf(str_size, NEST_NAMELEN, "%f", size + more_size);

    string str_modify_classad = "";
    str_modify_classad = str_modify_classad +
                         "[  StorageSpaceUsed = " + str_size + "]";

    str_modify_classad = " [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, modify_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::delete_user(const char *user,
                                 const char *cur_user,
                                 bool log_existing) {

    if (!user) {
        return NEST_BAD_PARAMETERS;
    }

    if (!strcmp(user, NEST_EVERYBODY)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete NEST_EVERYBODY %s\n",
                    NEST_EVERYBODY );
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        if ( log_existing ) {
            nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Trying to delete an already deleted user %s\n",
                        user );
        }
        return NEST_USER_NOT_FOUND;
    }

    string modify_classad1 = "";
    modify_classad1 = modify_classad1 +
                            "[   Updates  = " +
                                "[  ValidEntry = false ] " +
                            "]";

    ClassAd *mod_classad1 = parser.ParseClassAd( modify_classad1, true );

    string key = USER_KEY_PREFIX;
    key = key + user;

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad1 )) {
        return NEST_CLASSAD_ERROR;
    }

    ExprTree *group_list = user_class_ad->Lookup("Groups");
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }
    ExprListIterator    group_itr((ExprList *) group_list);

    Value   group_val;
    string  group_str;

    if (!group_itr.CurrentValue( group_val )) {
        return NEST_SUCCESS;
    }

    do {
        if (group_val.IsStringValue( group_str )) {
            this->delete_user_from_group_members(user, group_str.c_str());
        }
    } while( group_itr.NextValue( group_val ) );

    int total_nest_users;

    NestReplyStatus reply = this->get_total_nest_users(&total_nest_users);

    if(reply!=NEST_SUCCESS){
        return NEST_CLASSAD_ERROR;
    }

    total_nest_users--;
    key = TOTAL_USERS_KEY;

    char str_total_nest_users[NEST_NAMELEN];
    NestSnprintf(str_total_nest_users, NEST_NAMELEN, "%d",
                                        total_nest_users );

    string modify_classad2 = "";
    modify_classad2 = modify_classad2 +
                        "[ TotalNestUsers = " + str_total_nest_users + "]";

    modify_classad2 = " [ Updates = " + modify_classad2 + " ] ";

    ClassAd *mod_classad2 = parser.ParseClassAd( modify_classad2, true );

    if(!mod_classad2){
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad2 )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::add_group(const char *group,
                               const char *cur_user) {

    if (this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
            "Trying to add an existing group %s\n",
            group );
        return NEST_GROUP_EXISTS;
    }

    string str_key,
           str_group_classad = "";

    str_key = GROUP_KEY_PREFIX;
    str_key = str_key + group;
        
    char    str_size[NEST_NAMELEN];
    NestSnprintf(str_size, NEST_NAMELEN, "%f", 0.0);
    
    str_group_classad =  str_group_classad +
                        "[  Type = \"" + IDENTITY_TYPE + "\"; " +
                        "   PerceivedName = \"" + group + "\"; " +
                        "   ActualName = \"" + group + "\"; " +
                        "   ValidEntry = true; " +
                        "   IsUser = false; " +
                        "   OwnerUser = \"" + cur_user + "\"; " +
                        "   Groups = {}; " +
                        "   DefaultGroup = \"\"; " +
                        "   Members = {}; " +
                        "   StorageSpaceUsed = " + str_size + "; " +
                        "   DefaultLotHandle = \"\" " +
                        "]";

    ClassAd *group_classad = parser.ParseClassAd( str_group_classad, true );

    if (!group_classad) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
            "Parsing GroupClassAd for group %s during add-group failed!\n",
            group );
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, group_classad )) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
               "Adding GroupClassAd for group %s during add-group failed!\n",
               group );
        return NEST_CLASSAD_ERROR;
    }
    
    NestReplyStatus reply = add_user_in_group(cur_user,group,cur_user);
    if(reply != NEST_SUCCESS) {
        delete_group(group,cur_user);
    }
   
    int total_nest_groups;
    reply = get_total_nest_groups(&total_nest_groups);
    if(reply != NEST_SUCCESS) {
        delete_user_in_group(cur_user,group,cur_user);
        delete_group(group,cur_user);
    }
    
    total_nest_groups++;
    string key = TOTAL_GROUPS_KEY;

    char str_total_nest_groups[NEST_NAMELEN];
    NestSnprintf(str_total_nest_groups, NEST_NAMELEN, "%d",
                                            total_nest_groups );

    string modify_classad = "";
    modify_classad = modify_classad +
                      "[ TotalNestGroups = " + str_total_nest_groups + "]";

    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if(!mod_classad){
        delete_user_in_group(cur_user,group,cur_user);
        delete_group(group,cur_user);
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        delete_user_in_group(cur_user,group,cur_user);
        delete_group(group,cur_user);
        return NEST_CLASSAD_ERROR;
    }

    /*
        Move the commented portion at the
        end of this file,
        do appropriate modifications,
        when group reservations are enabled!
    */

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::delete_group(const char *group,
                                  const char *cur_user) {

    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to  delete a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    if (this->is_active_group(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete an active group %s\n",
                    group );
        return NEST_GROUP_ACTIVE;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    if (!this->Identities->RemoveClassAd( key )) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "DeleteClassAd failed during group-delete for group %s!\n",
                    group );
        return NEST_CLASSAD_ERROR;
    }

    int total_nest_groups;
    NestReplyStatus reply = get_total_nest_groups(&total_nest_groups);
    NEST_REPLY_CHECK(reply);
    
    total_nest_groups--;
    key = TOTAL_GROUPS_KEY;

    char str_total_nest_groups[NEST_NAMELEN];
    NestSnprintf(str_total_nest_groups, NEST_NAMELEN, "%d",
                                            total_nest_groups );

    string modify_classad = "";
    modify_classad = modify_classad +
                      "[ TotalNestGroups = " + str_total_nest_groups + "]";

    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }
    
    /*
        Move the commented portion at the
        end of this file, when group reservations are enabled!
    */

    return NEST_SUCCESS;
}


/*
    Replicated at NestFileCatalog::get_uid()
    Propogate all changes to there - sorry!
*/

NestReplyStatus
NestIdentityManager::get_uid(user_id_t *uid, const char *user) {

    if (!uid || !user) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestGet_uid(uid, user, SystemPwdFile) < 0) {
        return NEST_USER_NOT_FOUND;
    }

    return NEST_SUCCESS;
}



NestReplyStatus
NestIdentityManager::get_gid(user_id_t *gid, const char *group) {

    if (!gid) {
        return NEST_BAD_PARAMETERS;
    }

    if (!group) {
        group = NEST_NO_GROUP;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr,
            actual_group;

    attr = "ActualName";
    if (!group_class_ad->EvaluateAttrString(attr, actual_group)) {
        return NEST_CLASSAD_ERROR;
    }

    group_ent   grp_ent;
    if (Nestgetgrpnam(actual_group.c_str(), &grp_ent, SystemGrpFile) < 0) {
        return ErrnoToNest(errno);
    }
    *gid = (user_id_t) grp_ent.grp_gid;

    return NEST_SUCCESS;
}

NestReplyStatus
NestIdentityManager::get_space_used(double *size, const char *user, bool is_user) {
    *size = 0.0;
    string key;
    if(is_user) {
        key = USER_KEY_PREFIX;
    } else {
        key = GROUP_KEY_PREFIX;
    }
    key = key + user;

    ClassAd *user_ad = this->Identities->GetClassAd(key);
    if (!user_ad) {
        return NEST_LOT_NOT_FOUND;
    }

    bool result;
    result = user_ad->EvaluateAttrNumber( "StorageSpaceUsed", *size );
    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
NestIdentityManager::initialize(ClassAdCollection *resources ) {

    this->Identities = resources;

    this->query.Bind(resources);
    this->identities_view = NEST_IDENTITIES_VIEW;

    ViewName    parent_view = "root";

    string      view_constraint(""),
                view_rank(""),
                view_partition_expr("");

    view_constraint = view_constraint +
                      "other.Type == \"" + IDENTITY_TYPE + "\"";

    if (!this->Identities->CreateSubView( this->identities_view,
                                            parent_view,
                                          view_constraint,
                                          view_rank,
                                          view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "NestIdentityManager Couldn't create view for itself!\n");
        return NEST_CLASSAD_ERROR;
    }

    string str_key = TOTAL_USERS_KEY;

    string str_total_users_classad = "";
    str_total_users_classad = str_total_users_classad +
                            "[ Type = \"" + TOTAL_USERS_KEY + "\";" +
                            "  TotalNestUsers = 0" +
                            "]";

    ClassAd *total_users_ad = parser.ParseClassAd( str_total_users_classad,
                                                   true );

    if (!total_users_ad) {
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, total_users_ad )) {
        return NEST_CLASSAD_ERROR;
    }

    str_key = TOTAL_GROUPS_KEY;

    string str_total_groups_classad = "";
    str_total_groups_classad = str_total_groups_classad +
                            "[ Type = \"" + TOTAL_GROUPS_KEY + "\";" +
                            "  TotalNestGroups = 0" +
                            "]";

    ClassAd *total_groups_ad = parser.ParseClassAd( str_total_groups_classad,
                                                   true );

    if (!total_groups_ad) {
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, total_groups_ad )) {
        return NEST_CLASSAD_ERROR;
    }

    char    actual_nest_group[NEST_NAMELEN];
    if (this->get_actual_nest_group(actual_nest_group) != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Cannot find a default nest_group!\n" );
        return NEST_GROUP_NOT_FOUND;
    }

    str_key = GROUP_KEY_PREFIX;
    str_key = str_key + NEST_NO_GROUP;

    string str_group_classad =  "";
    str_group_classad = str_group_classad +
                            "[  Type = \"" + IDENTITY_TYPE + "\"; " +
                            "   PerceivedName = \"" + NEST_NO_GROUP + "\"; " +
                            "   ActualName = \"" + actual_nest_group + "\"; "+
                            "   ValidEntry = true; " +
                            "   IsUser = false; " +
                            "   OwnerUser = \"" + actual_nest_group + "\"; " +
                            "   Groups = {}; " +
                            "   DefaultGroup = \"\"; " +
                            "   Members = {}; " +
                            "   DefaultLotHandle = \"\" " +
                            "] " ;

    ClassAd *group_classad = parser.ParseClassAd( str_group_classad, true );

    if (!group_classad) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Parsing GroupClassAd for group %s during init failed!\n",
                    actual_nest_group );
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, group_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    NestReplyStatus reply = this->add_user( NEST_EVERYBODY,
                                            NEST_NO_GROUP,
                                            NestAdminUser,
                                            "" );

    NEST_REPLY_CHECK(reply);

    str_key = USERLIST_KEY;

    string  str_userlist_classad = "";
    str_userlist_classad = str_userlist_classad +
                                "[  Type = \"" + USERLIST_KEY + "\"; " +
                                "   UserList = \"\"; " +
                                "   DenyList = \"\"  " +
                                "] ";

    ClassAd *userlist_ad = parser.ParseClassAd( str_userlist_classad, true );

    if (!userlist_ad) {
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, userlist_ad )) {
        return NEST_CLASSAD_ERROR;
    }

    return this->synchronize(true);
}

NestReplyStatus
NestIdentityManager::recover(ClassAdCollection *resources ) {

    this->Identities = resources;
    this->query.Bind(resources);
    this->identities_view = NEST_IDENTITIES_VIEW;

    return this->synchronize();
}


NestReplyStatus
NestIdentityManager::shutdown() {
    return NEST_SUCCESS;
}


NestReplyStatus
NestIdentityManager::synchronize(bool is_init) {

    string key;
    bool result;

    key = USERLIST_KEY;

    ClassAd *userlist_ad = this->Identities->GetClassAd( key );

    if (!userlist_ad) {
        return NEST_CLASSAD_ERROR;
    }

    result  = userlist_ad->EvaluateAttrString( "UserList",
                                               this->prev_userlist );
    result &= userlist_ad->EvaluateAttrString( "DenyList",
                                               this->prev_denylist );

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    struct grid_mapfile_entry *user;
    string str_user_classad;
    NestReplyStatus reply;

    int state = Nestset_grid_mapfile(GridMapFile);
    if ( state < 0) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Couldn't read system grid-mapfile at %s!\n",
                    GridMapFile );
        return NEST_SYSTEM_FILE_MISSING;
    }

    string str_key;

    while( (user = Nestget_grid_mapfile_entry(state)) != NULL) {

        //  Multiple identities could be mapped to the same user!!

        str_key = USER_KEY_PREFIX;
        str_key = str_key + user->name;
        str_user_classad =  "";

        char *identity = NestGetModifiedIdentity(user->identity);

        if (this->is_valid_nest_user( identity, is_init )) {
            reply = this->add_user( identity,
                                    NEST_NO_GROUP,
                                    NestAdminUser,
                                    "",
                                    false );
            if (reply == NEST_USER_EXISTS) {
                reply = NEST_SUCCESS;
            }
        }
        else {
            reply = this->delete_user( identity,
                                       NestAdminUser,
                                       false );
            if (reply == NEST_USER_NOT_FOUND) {
                reply = NEST_SUCCESS;
            }
        }

        if(identity) {
            NestFree( identity );
        }
        NEST_REPLY_CHECK(reply);
    }   //  End-of-while
    Nestend_grid_mapfile(state);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ UserList = \"" + this->userlist + "\";" +
                        "   DenyList = \"" + this->denylist + "\"] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

/*  //  Add Group!

    //  Move this commented portion to the
    //  add_group(), when group reservations are enabled!

    string constraint = "";

    constraint = constraint +
                    "other.IsUser == false && " +
                    "other.ValidEntry == false";

    ExprTree *constraint_tree = this->parser.ParseExpression(constraint);
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }

    this->query.Query(this->identities_view, constraint_tree);
    this->query.ToFirst();

    string  key;
    if (!this->query.Current(key)) {
        return NEST_GROUP_LIMIT_REACHED;
    }

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return NEST_GROUP_LIMIT_REACHED;
    }

    string  attr,
            actual_groupname;
    attr = "ActualName";
    user_class_ad->EvaluateAttrString( attr , actual_groupname );

    group_ent ge;
    if (Nestgetgrpnam( actual_groupname, &ge, SystemGrpFile ) < 0) {
        return NEST_GROUP_LIMIT_REACHED;
    }

    if (ret_gid) {
        *ret_gid = ge.grp_gid;
    }

    string modify_classad =   "[   Updates  =               \
                                    [   PerceivedName = \""
                                        + group + "\" ;     \
                                        ValidEntry = true ; \
                                        Members = {}        \
                                    ]                       \
                                ]";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }
*/

/*  //  Delete Group!

    //  Move this commented portion to the
    //  delete_group(), when group reservations are enabled!

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string modify_classad =   "[   Updates  =                   \
                                    [   PerceivedName = \"\" ;  \
                                        ValidEntry = false   ;  \
                                        Members = {}            \
                                    ]                           \
                                ]";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }
*/
