/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_protocol_filesystem.h"


nestProtocolFilesystem::nestProtocolFilesystem( int filedescriptor ) {
    this->use_pread = false;
    this->offset = 0;
    this->fdkey  = NULL;
    this->fd_is_cached = false;
    init( filedescriptor );
}

nestProtocolFilesystem::nestProtocolFilesystem( int filedescriptor, 
        off_t offset ) 
{
    this->use_pread = true;
    this->offset    = offset;
    this->fdkey     = NULL;
    this->fd_is_cached = false;
    init( filedescriptor );
}

void
nestProtocolFilesystem::init( int filedescriptor ) {
    this->filedescriptor = filedescriptor;

    *logical_filename = *physical_filename = 0;
    open_flags = 0;
    open_filesize = 0;
    uid = gid = ((user_id_t) -1);
    return;
}

// used by nob
nestProtocol *
nestProtocolFilesystem::clone( ) {

    int dup_fd = NestDup( this->filedescriptor );
    nestProtocolFilesystem *result = new nestProtocolFilesystem( dup_fd,offset);
    result->set_filenames( this->logical_filename, this->physical_filename );
    result->set_ids( this->uid, this->gid );
    result->set_open_flags( this->open_flags );
    result->set_open_filesize( this->open_filesize );
    result->set_fdkey( this->fdkey );

    return (nestProtocol *) result;
}

nestProtocolFilesystem::~nestProtocolFilesystem( ) {

    if ( ! fd_is_cached ) NestClose( filedescriptor );
    if ( fdkey )          NestFree( fdkey );

    return;
}

// NOTE:
// now that the storage manager cache's fds, we must use preads and pwrites
// the cached fds are dups and thus share offsets
// if we see a performance hit because of preads and pwrites we can
// add logic to use them only for cached fds and doing normal reads and writes
// on the original 

ssize_t
nestProtocolFilesystem::Send( const char *buffer, ssize_t n ) {

    ssize_t w = 0;
    
    if (this->uid == ((user_id_t) -1)) {
        if ( use_pread ) w = NestPwrite( filedescriptor, buffer, n, offset );
        else             w = NestWrite(  filedescriptor, buffer, n );
    } else {
        user_id_t   my_id = NestGetuid();
        NestSeteuid(this->uid);
        if ( use_pread ) w = NestPwrite( filedescriptor, buffer, n, offset );
        else             w = NestWrite ( filedescriptor, buffer, n );
        NestSeteuid(my_id);
    }

    if ( w > 0 ) offset += w;
    return w;
}

ssize_t
nestProtocolFilesystem::Receive( char *buffer, ssize_t n ) {
    ssize_t r = 0;
    if ( use_pread ) r = NestPread( filedescriptor, buffer, n, offset );
    else             r = NestRead( filedescriptor, buffer, n );
    if ( use_pread && r > 0 ) {
        offset += r;
    }
    return r;
}

bool
nestProtocolFilesystem::get_fd_cache_status( ) {
    return fd_is_cached;
}


void
nestProtocolFilesystem::set_fd_cache_status( bool status ) {
    fd_is_cached = status;
    return;
}

char *
nestProtocolFilesystem::get_fdkey( ) {
    return fdkey;
}

int
nestProtocolFilesystem::set_fdkey( char *fdkey ) {
    this->fdkey = NestStrdup( fdkey );
    return 1;
}

int
nestProtocolFilesystem::set_open_flags(int flags) {

    open_flags = flags;

    return 0;
}


int
nestProtocolFilesystem::get_open_flags(int *flags) {

    if (!flags) {
        return -1;
    }

    *flags = open_flags;

    return 0;
}

int
nestProtocolFilesystem::set_filenames(char *logical_name, char *physical_name) {
    /*
       // simple storage doesn't use both
    if (!logical_name || !physical_name) {
        return -1;
    }
    */

    if ( logical_name != NULL ) {
        strncpy(logical_filename, logical_name, NEST_MAXPATHLEN);
    }
    if ( physical_name != NULL ) {
        strncpy(physical_filename, physical_name, NEST_MAXPATHLEN);
    }

    return 0;
}


int
nestProtocolFilesystem::get_filenames(char *logical_name, char *physical_name) {

    if ( logical_name != NULL ) {
        strcpy(logical_name, logical_filename);
    }
    if ( physical_name != NULL ) {
        strcpy(physical_name, physical_filename);
    }

    return 0;
}

int
nestProtocolFilesystem::set_open_filesize(off_t size) {

    open_filesize = size;

    return 0;
}

int
nestProtocolFilesystem::get_open_filesize(off_t *size) {

    if (!size) {
        return -1;
    }

    *size = open_filesize;

    return 0;
}

int
nestProtocolFilesystem::set_ids(user_id_t uid, user_id_t gid) {
    this->uid = uid;
    this->gid = gid;
    return 0;
}

int
nestProtocolFilesystem::get_ids(user_id_t *uid, user_id_t *gid) {
    if (uid) {
        *uid = this->uid;
    }

    if (gid) {
        *gid = this->gid;
    }
    return 0;
}

NestReplyStatus
nestProtocolFilesystem::receive_state( int pipe ) {
    char message[NEST_MESSAGESIZE];
    NestConnection nest_con = (NestConnection) pipe;
    NestReplyStatus status;

    if ((status = ReceiveNestMessage( nest_con, message )) != NEST_SUCCESS  ) {
        return status;
    }

    nest_debug(NEST_DEBUG_CONCURRENCY, "Received state message: %s\n", message);
    long l_offset;
    if ( sscanf( message, "%ld %ld %ld\n", 
                &this->uid, &this->gid, &l_offset ) != 3 ) 
    {
        return NEST_BAD_PARAMETERS;
    }
    this->offset = l_offset;
    return NEST_SUCCESS;
}

// used by POP
NestReplyStatus
nestProtocolFilesystem::send_state( int pipe ) {
    char message[NEST_MESSAGESIZE];
    NestConnection nest_con = (NestConnection) pipe;

    NestSnprintf( message,
                  NEST_MESSAGESIZE,
                  "%ld %ld %ld\n",
                  this->uid,
                  this->gid,
                  this->offset );

    nest_debug( NEST_DEBUG_CONCURRENCY,
                "Sending state message: %s\n",
                message);

    if ( ! SendNestMessage( nest_con, message ) ) {
        assert( 0 );
        return NEST_UNKNOWN_ERROR;
    }
    return NEST_SUCCESS;
}



