/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_client.h"
#include "nest_client_pool.h"

#include "nest_common.h"
#include "nest_protocol_nest.h"
#include "network_limit.h"
#include "globus_gss_assist.h"
#include "nest_globus.h"


// static class variables
// why can't these be in the header file ???
static unsigned addr_sz;

int
nestProtocolNest::listen( int port ) {
    
        // now actually listen on the socket
    addr_sz = sizeof( struct sockaddr_in );
    return NestsetSocket( AF_INET, SOCK_STREAM, 0, port );
}

nestProtocol *
nestProtocolNest::accept( int socket, sockaddr_in *address ) {
    int csd =  NestAccept( socket, (sockaddr *)address, &addr_sz );
    return ( csd <= 0 ) ? NULL : new nestProtocolNest( csd );
}

NestReplyStatus
nestProtocolNest::initiateThirdParty( nestClient *client,
        nestProtocol **connection )
{
    NestConnection socket;
    NestReplyStatus status;

        // open connection
    nest_debug(NEST_DEBUG_CHIRP,"Attempting to connect to chirp server at %s\n",
            client->getRequest()->getRemoteHost() );
    status =NestOpenConnection( socket, client->getRequest()->getRemoteHost());
    if ( status != NEST_SUCCESS ) return status;

        // attempt to move connection into transfer-ready state
    if ( client->getRequest()->getType() == THIRD_PARTY_PUT ) {
        status = RequestWriteFile(
                client->getRequest()->getRemoteFile(),
                client->getRequest()->getStart(),
                client->getRequest()->getLength(), 
                socket );
    } else {
        assert( client->getRequest()->getType() == THIRD_PARTY_GET );
        long length = client->getRequest()->getLength();
        status = RequestReadFile( 
                client->getRequest()->getRemoteFile(),
                client->getRequest()->getStart(), 
                length,
                socket );
        client->getRequest()->setEnd(
                length - client->getRequest()->getStart() );
    }
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_CHIRP, 
                "Connection to %s made but transfer request denied: %s\n",
                client->getRequest()->getRemoteHost(),
                NestErrorString( status ) );
        NestCloseConnection( socket );
        return status;
    }

    *connection = new nestProtocolNest( socket );
    return status;
}

nestProtocolNest::nestProtocolNest( int socket ) {
    this->socket    = socket;
    this->msgoffset = 0;
}

nestProtocol *
nestProtocolNest::clone( ) {
    return new nestProtocolNest( NestDup( this->socket ) );
}

nestProtocolNest::~nestProtocolNest() {
    nest_debug( NEST_DEBUG_CHIRP, "Closing chirp socket: %d.\n", socket );
    NestClose( socket );
}

NestReplyStatus
nestProtocolNest::gssAccept( char **username , char **globus_username ) {
    return auth_globus_accept( socket, username, globus_username );
}

bool
nestProtocolNest::endThirdPartyTransfer( int type ) {
    if ( type == THIRD_PARTY_GET ) {
            // nothing to do
        return true;
    } else {
        assert( type == THIRD_PARTY_PUT );
        nest_debug( NEST_DEBUG_CHIRP, 
                    "nestProtocolNest ending 3pt - waiting for ack\n" );
        return ( ReceiveNestAck( socket ) == NEST_SUCCESS );
    }
}
    
ssize_t
nestProtocolNest::Send( const char *buffer, ssize_t n ) {
    return NestWrite( socket, buffer, n );
}

ssize_t
nestProtocolNest::Receive( char *buffer, ssize_t n ) {
    return NestRead( socket, buffer, n );
}

bool
nestProtocolNest::sendAd( nestAd *ad ) {

    char *adstring  = ad->toString();
    int len         = ad->length();

    NestSnprintf( msg, sizeof(msg), "%d %ld", NEST_SUCCESS, len    );
    nest_debug( NEST_DEBUG_CHIRP, "Sending ad len %ld (%.40s).\n", 
				len, msg );
    if ( ! sendMessage( (void *)msg ) ) return false;

    ssize_t sent = NestWriten( socket, (void *)adstring, len );
    return ( sent == len );

    /*
    while( sent < len ) {
        thissend = ( len - sent > NEST_MESSAGESIZE )
                    ? NEST_MESSAGESIZE
                    : len - sent;

        NestSnprintf( msg, sizeof(msg), "%s", &(adstring[sent]) );
        sent += thissend;
        if ( ! sendMessage( (void *)msg ) ) return false;
    }
    return ( sent == len );

    nest_debug( NEST_DEBUG_CHIRP, "sending ack following send nestad\n" );
    return sendAck( );
    */
}

bool
nestProtocolNest::sendDirectoryList( NestFileInfo *list ) {
    while( list != NULL ) {
        NestFileInfo_to_string( list, msg, sizeof(msg) );
        if ( ! sendMessage( (void *)msg ) ) return false;
        list = list->next;
    }
    return sendAck( );
}

#define PROXY_ITERATOR list<char*>::iterator

bool
nestProtocolNest::sendProxyList( list <char *> *proxys ) {
    char *glist;
    long length = 0;

    if ( proxys != NULL ) {
            // go through twice:  first to determine length, then to send
        for(PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++){
            char *name = (char *)*itr;
            length += NestStrlen( name ) + 1;
        }   
    }

        // send the length
    if ( length > 0 ) length += 2;  // add room for null term
    NestSnprintf( msg, sizeof(msg), "%d %ld\n", NEST_SUCCESS, length );
    nest_debug( NEST_DEBUG_PROXY, "Sending proxy list len %ld (%.40s).\n", 
            length, msg );
    if ( ! sendMessage( (void *)msg ) ) return false;

        // cut out early if NULL
    if ( proxys == NULL || proxys->size() <= 0 ) return true;

        // otherwise write the list
    glist = (char *)NestCalloc( length );
    long offset = 0;
    for( PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++ ) {
        char *name = (char *)*itr;
        NestSnprintf( &glist[offset], length + 2 - offset, "%s ", name );
        offset += NestStrlen( name ) + 1;
        nest_debug( NEST_DEBUG_PROXY, "Sending proxy %s.\n", name );
    }
    NestSnprintf( &glist[offset], length - offset, "\n\0" );

        // send the list
    ssize_t sent = NestWriten( socket, (void *)glist, length );
    NestFree( glist );
    if (sent != length ) {
        if ( errno != EPIPE) nest_perror( "NestWriten" );
        return false;
    }

    return true;
}

bool
nestProtocolNest::sendNestLot( NestLot *r) {

    if (!r) {
        sendErrorMessage( __FILE__, __FUNCTION__,  NEST_UNKNOWN_ERROR );
        return false;
    }

    NestSnprintf(msg,
             sizeof(msg),
             "%d %s %s %s %u %s %10.8f %10.8f %10.8f %10.8f %10.8f %10.8f %d %d %s",
             NEST_LOT_SCHEDULED,
             r->lot_handle,
             r->parent_lot_handle,   
             r->owner,
             r->is_user,
             r->dev,
             r->total_size,   
             r->size,
             r->used,
             r->start_time,
             r->duration,
             r->time_left,
             r->cost,
             r->lot_type,
             r->lot_users);

    return sendMessage( (void *) msg);
}

bool
nestProtocolNest::sendLotHandle( NestLotHandle *handle ) {

    rsv_id_t    tmp_id;

    if (NestGet_id_from_handle(handle, &tmp_id) < 0) {
        sendErrorMessage( __FILE__, __FUNCTION__,  NEST_UNKNOWN_ERROR );
        return false;
    }
    nest_debug(NEST_DEBUG_CHIRP, "sendLotHandle: handle=%p, id=%d \n",
			   handle, tmp_id );

    NestSnprintf(msg, sizeof(msg), "%d %lu",
                                NEST_LOT_SCHEDULED,
                                tmp_id);

    return sendMessage( (void *) msg);
}

bool
nestProtocolNest::sendDiskspace(  double avail, double total  ) {
    NestSnprintf( msg, sizeof(msg), "%d %.0f %.0f", NEST_SUCCESS, avail, 
            total );
    return sendMessage( (void *) msg );
}

bool
nestProtocolNest::sendErrorMessage( const char* filename, const char* func_name,   int error ){
    NestSnprintf( msg, sizeof( msg ), "%d", error );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendTransferInfo(  nestRequest *req ){
    NestSnprintf( msg,
                  NEST_MESSAGESIZE,
                  "%d %s %d %ld",
                  NEST_SUCCESS,
                  req->getNestFile(),
                  req->getType(),
                  req->getEnd() );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendTimestamp(  ){
    char buff[NEST_MESSAGESIZE];
    time_t ticks = time( NULL );
    NestSnprintf( buff, sizeof(buff), "%d %ld", NEST_SUCCESS, (long)ticks );
    return sendMessage( buff );
}

/*
bool
nestProtocolNest::sendQuotaInfo(  struct UserUsage *usage ) {
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %ld %ld\n",
            NEST_SUCCESS, usage->current, usage->quota );
    return sendMessage( msg );
}
*/

bool
nestProtocolNest::sendFilesize( long size){
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %ld", NEST_SUCCESS, size );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendPhysicalName( const char *name){
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s", NEST_SUCCESS, name );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendPwd( const char *pwd ) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", NEST_SUCCESS, pwd );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendAcl( const char *user,
                           int is_user,
                           nest_access_rights_t access){
    char msg[NEST_MESSAGESIZE];

    // Format the return string..
    nest_debug( NEST_DEBUG_CHIRP, "Sending access=0x%02x for user '%s'\n",
                access, user );

    // Send the initial success...
    if ( ! sendAck( ) ) {
        nest_debug( NEST_DEBUG_ALWAYS, 
                    "sendACL: Error sending initial ACK\n" );
        return false;
    }

    // Now the data message
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s %d %d",
                  NEST_PARTIAL_MESSAGE, user, is_user, access );

    // Ship it out
    if ( ! sendMessage( msg ) ) {
        return false;
    }
    return sendAck( );
}

bool
nestProtocolNest::sendAcl( const NestACLList *Acl) {
    char msg[NEST_MESSAGESIZE];

    // Send the initial success...
    if ( ! sendAck( ) ) {
        nest_debug( NEST_DEBUG_ALWAYS, 
                    "sendACL: Error sending initial ACK\n" );
        return false;
    }

    // Send the whole list we have...
    nest_debug( NEST_DEBUG_CHIRP, "Sending ACL@%p\n", Acl );
    while( Acl != NULL ) {
        nest_debug( NEST_DEBUG_CHIRP, "Sending access=0x%02x for user '%s'\n",
                    Acl->acl.rights, Acl->acl.name );
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s %d %d",
                      NEST_PARTIAL_MESSAGE,
                      Acl->acl.name, Acl->acl.is_user, Acl->acl.rights );
        if ( ! sendMessage( (void *)msg ) ) {
            nest_debug( NEST_DEBUG_ALWAYS, "sendACL: Error sending ACL\n" );
            return false;
        }
        Acl = Acl->next;
    }
    return sendAck( );
}

bool
nestProtocolNest::sendVersion( float version ) {
    char msg[NEST_MESSAGESIZE];
    nest_debug( NEST_DEBUG_CHIRP, "Sending version %f\n", version );
    NestSnprintf( msg, sizeof(msg), "%d %f", NEST_SUCCESS, version );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendMessage(  const void *vptr ){
    return SendNestMessage( socket, (char *) vptr ); 
}

bool
nestProtocolNest::sendFinalAck( int transferType ){
    nest_debug( NEST_DEBUG_CHIRP,
            "Entering nestProtocolNest::sendFinalAck for type %d: sd %d\n",
            transferType, socket );

    if (   transferType == CLIENT_RECV_FILE 
        || transferType == PROXY_GET 
        || transferType == PROXY_PUT 
        ) 
    {
        /* NO-OP */ 
        return true;
    } else {
        return sendAck( );
    }
}

bool
nestProtocolNest::sendAck(  ) {
    char ack[NEST_MESSAGESIZE];
    nest_debug( NEST_DEBUG_CHIRP,
            "Sending chirp ack on sd %d\n", socket );
    NestSnprintf( ack, sizeof(ack), "%d", NEST_SUCCESS );
    return sendMessage( ack );
}

nestRequest*
nestProtocolNest::receiveRequest( fd_set *ready ){

    if ( ! NestFd_isset( socket, ready ) ) {
        //nest_debug( NEST_DEBUG_CHIRP, "NestSocket %d not ready\n", socket );
        return NULL;
    }

    NestReplyStatus status;
    int read = 0;

    status = ReceiveNestPartialMessage( socket, &(msg[msgoffset]),
                    NEST_MESSAGESIZE - msgoffset, &read );

    if ( msgoffset != 0 ) {
        nest_debug( NEST_DEBUG_CHIRP, 
                "Attempted to resume partial req (sd %d, off %d). "
                "Got %d additional bytes (%s): %s",
                getSocket(), msgoffset, read, 
                NestErrorString( status ), 
                msg );
    }

    msgoffset = ( msgoffset + read ) % NEST_MESSAGESIZE;
    switch( status ) {
    case NEST_SUCCESS:
        assert( msgoffset == 0 );
        break;
    case NEST_PARTIAL_MESSAGE:
        assert( msgoffset > 0 && msgoffset < NEST_MESSAGESIZE );
        msg[msgoffset] = '\0';
        nest_debug( NEST_DEBUG_CHIRP, 
                "WARN: nestProtocol (sock %d) recvd partial msg (len %d): %s\n",
                getSocket(), msgoffset, msg );
        return NULL;
    default:
        if ( status != NEST_NO_CONNECTION ) {
            nest_debug( NEST_DEBUG_CHIRP,
                "WARN: nestProtocol failed to recv request: %s\n",
                NestErrorString( status ) );
        }
            // clients don't need to do anything explicit to close
            // this is gotten for free.  When they close, their
            // fd appears as ready to read but when we read from it
            // we get an error - nice
        return new nestRequest( CLOSE_CONNECTION_REQUEST,
            NULL, NULL, NULL, 0, 0 );
    }

    nest_debug( NEST_DEBUG_CHIRP, "Received chirp req: %s", msg );
    nestRequest *req = parseRequest( msg );
    if ( req == NULL ) {
        nest_debug( NEST_DEBUG_CHIRP, "Bad request received on sd %d: %s\n", 
                getSocket(), msg );
        req = new nestRequest( BAD_REQUEST, 0, 0, 0, NEST_BAD_PARAMETERS, 0 );
    }
    return req;
}

nestRequest*
nestProtocolNest::parseRequest( char *message ) {

        // leave these static so that NestParseLine can free memory later
        // however this does require that this is always done by one thread
    static char **argv=0;   
    static int argc;

    bool valid_request = true;
    
    NestParseLine(message, &argv, &argc);
    if(!argv) return NULL;
    if(!argv[0]) return NULL;
    //if(!*argv[0]) return NULL;    

    int type;
    type = atoi(argv[0]);

        // set up holder variables for all possibilities
    long start = 0, end = 0;

    char *file = NULL;
    char *arg1 = NULL;
    char *arg2 = NULL;
    char *arg3 = NULL;
    char *arg4 = NULL;
    char *arg5 = NULL;
    char *arg6 = NULL;
    /*
    static char file[NEST_MESSAGESIZE];
    static char arg1[NEST_MESSAGESIZE];
    static char arg2[NEST_MESSAGESIZE];
    static char arg3[NEST_MESSAGESIZE];
    static char arg4[NEST_MESSAGESIZE];
    static char arg5[NEST_MESSAGESIZE];
    static char arg6[NEST_MESSAGESIZE];
    */

    /*
    memset( file, 0, NEST_MESSAGESIZE );
    memset( arg1, 0, NEST_MESSAGESIZE );
    memset( arg2, 0, NEST_MESSAGESIZE );
    memset( arg3, 0, NEST_MESSAGESIZE );
    memset( arg4, 0, NEST_MESSAGESIZE );
    memset( arg5, 0, NEST_MESSAGESIZE );
    memset( arg6, 0, NEST_MESSAGESIZE );
    */
    
    switch( type ) {
    case DIR_LIST_REQUEST:
    case AVAILABLE_SPACE_QUERY:
    case TIME_REQUEST:
    case SERVER_TERMINATE_REQUEST:
    case FORMAT_DISK_REQUEST:
    case PWD_REQUEST:
    case NEST_SYNC:
    case REQUEST_NEST_AD:
    case VERSION_REQUEST:
    case PROXY_QUERY:
    case PROXY_INDEX_PUSH:
    case PROXY_FLUSH:
            // nothing more to read from the message
        break;
    case CLIENT_SEND_FILE:
    case CLIENT_RECV_FILE:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(file,argv[1]);
        file = argv[1];
        start = atol(argv[2]);
        end = atol(argv[3]);
            // TODO  this is really bad, but end is supposed to be the
            // actual end not the size, it should really to fixed to be the
            // size.  for the time being, just hack it here
        if ( end != ENTIRE_FILE ) end += start;
        break;
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        end = atol(argv[1]); 
        break;
    case FILE_DELETE_REQUEST:
    case FILESIZE_REQUEST:
    case RMDIR_REQUEST:
    case CWD_REQUEST:
    case MKDIR_REQUEST:
    case PROXY_REGISTER:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(file,argv[1]);
        file = argv[1];
        break;
    case FILE_RENAME_REQUEST:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1,argv[1]);
        //strcpy(arg2,argv[2]);
        arg1 = argv[1];
        arg2 = argv[2];
        break;
    case CHANGE_USER_REQUEST:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1,argv[1]);
        arg1 = argv[1];
        start = atol(argv[2]);
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
        if(!argv[1] || !argv[2] || !argv[3] || !argv[4] || !argv[5] || !argv[6])
        {
            valid_request = false;
            break;
        }
        //strcpy(file,argv[1]);
        //strcpy(arg1,argv[2]);
        //strcpy(arg2,argv[3]);
        file = argv[1];
        arg1 = argv[2];
        arg2 = argv[3];
        start = atol(argv[4]);
        end = atol(argv[5]);
        //strcpy(arg3,argv[6]);
        arg3 = argv[6];
        break;
    case ADD_NEST_USER:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1,argv[1]);
        //strcpy(arg2,argv[2]);
        arg1 = argv[1];
        arg2 = argv[2];
        if(argv[3])
            //strcpy(arg3,argv[3]);
            arg3 = argv[3];
        break;
    case ADD_USER_TO_GROUP:
    case DELETE_USER_FROM_GROUP:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
            return NULL;
        //strcpy(arg1,argv[1]);
        //strcpy(arg2,argv[2]);
        arg1 = argv[1];
        arg2 = argv[2];
        break;

    case DELETE_NEST_USER:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1,argv[1]);
        arg1 = argv[1];
        break;

    case ADD_NEST_GROUP:
    case DELETE_NEST_GROUP:
    case LIST_GROUP_LOTS:
    case GET_LOT_STAT:
    case TERMINATE_LOT:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(arg2,argv[1]);
        arg2 = argv[1];
        break;

    case LIST_USER_LOTS:
        if(argv[1])
            //strcpy(arg3, argv[1]);
            arg3 = argv[1];
        break;

    case REQUEST_USER_LOT:
    case REQUEST_GROUP_LOT:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1, argv[1]);
        //strcpy(file, argv[2]);
        //strcpy(arg2, argv[3]);
        arg1 = argv[1];
        file = argv[2];
        arg2 = argv[3];
    	if(argv[4])
	        //strcpy(arg3, argv[4]);
            arg3 = argv[4];
        if(argv[5])
            //strcpy(arg4, argv[5]);
            arg4 = argv[5];
        if(argv[6]) 
            //strcpy(arg5, argv[6]);
            arg5 = argv[6];
        if(argv[7])
            //strcpy(arg6, argv[7]);
            arg6 = argv[7];
        break;

    case UPDATE_LOT:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(arg2, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy(file, argv[3]);
        arg2 = argv[1];
        arg1 = argv[2];
        file = argv[3];
        break;

    case CHANGE_LOT_T:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(arg2, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy(file, argv[3]);
        arg2 = argv[1];
        arg1 = argv[2];
        file = argv[3];
        break;

    case LOT_MOVE_FILE:
        if(!argv[1] || !argv[2] || !argv[3] || !argv[4]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy(arg2, argv[3]);
        //strcpy(arg3, argv[4]);
        file = argv[1];
        arg1 = argv[2];
        arg2 = argv[3];
        arg3 = argv[4];
        break;

    case ADD_USER_TO_LOT:
    case DELETE_USER_FROM_LOT:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg2, argv[1]);
        //strcpy(arg3, argv[2]);
        arg2 = argv[1];
        arg3 = argv[2];
        break;

    case ATTACH_LOT_TO_FILE:
    case DETACH_LOT_FROM_FILE:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg2, argv[1]);
        //strcpy(file, argv[2]);
        arg2 = argv[1];
        file = argv[2];
        break;

        // Collection ops
    case NAME_COLLECTION:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        file = argv[1];
        arg1 = argv[2];
        break;
    case UNNAME_COLLECTION:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1, argv[1]);
        arg1 = argv[1];
        break;
    case SET_COLLECTION_MTIME:
    case SET_COLLECTION_VERSION:
    case SET_COLLECTION_COMMENT: {
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(arg1, argv[1]);
        arg1 = argv[1];
		int	len = strlen( argv[2] );
		char	*end = argv[2] + len - 1;
		if ( ( len > 0 ) && ( *argv[2] == '\"' ) && ( *end == '\"') ) {
			//strcpy(arg2, argv[2]+1);
            arg2 = argv[2]+1;
			arg2[len - 2] = '\0';
		} else {
			//strcpy(arg2, argv[2]);
            arg2 = argv[2];
		}
        break;
	}

    case COPY_REQUEST:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        file = argv[1];
        arg1 = argv[2];
        start = 0;
        end = ENTIRE_FILE;
        break;
    case LINK_REQUEST:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        file = argv[1];
        arg1 = argv[2];
        start = atol(argv[3]);
        break;

    case PHYSNAME_REQUEST:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        file = argv[1];
        break;

        // ACL Related
    case GET_ACL:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        file = argv[1];
        arg1 = argv[2];
        break;
    case SET_ACL:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy(arg2, argv[3]);
        file = argv[1];
        arg1 = argv[2];
        arg2 = argv[3];
        break;
    case GET_GROUP_ACL:
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy( arg3, "true" );
        file = argv[1];
        arg1 = argv[2];
        arg3 = "true";
        break;
    case SET_GROUP_ACL:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        //strcpy(arg1, argv[2]);
        //strcpy(arg2, argv[3]);
        //strcpy( arg3, "true" );
        file = argv[1];
        arg1 = argv[2];
        arg2 = argv[3];
        arg3 = "true";
        break;

    case DEBUG_REQUEST:
        if(!argv[1]) {
            valid_request = false;
            break;
        }
        //strcpy(file, argv[1]);
        file = argv[1];
        break;

        // we longer have a bandwidth statistics request
    case BANDWIDTH_STATISTICS_REQUEST:
    default:
        nest_debug( NEST_DEBUG_CHIRP,
                "ERROR:  Unknown req type rec'vd: %d.  Full msg: %s\n",
                type, message );
        return new nestRequest(BAD_REQUEST,
                               0,
                               0,
                               0,
                               NEST_UNKNOWN_REQUEST_TYPE,
                               0);
    }

    nestRequest *req = NULL;
    if ( valid_request ) {
        req = new nestRequest( type, file, arg1, arg2, start, end,
            arg3, arg4, arg5, arg6 );
    }
    return req;
}
