#include "nest_proxy.h"
#include "nest_concurrency_manager.h"   // probably should be a nest_server.h
#include "nest_storage_manager.h"       // that contains all server stuff
#include "nest_config.h"

#define MUST_BE_PROXY(name)    if ( ! serverIsProxy() ) {\
                                    nest_warn( __FILE__, __FUNCTION__, \
                                            "%s requested a proxy_operation, "\
                                            "but I'm not a proxy\n",\
                                            (name) ); \
                                    return BAD_REQUEST;\
                                }


/*
   *
   *  This is where all the code responsible for managing 
   *  an individual proxy node goes.
   *
   *  Code for handling proxy connections at a home server
   *  goes in nest_server.C.
   *
*/

enum { PROXY_CONNECTION, COORD_CONNECTION, HOME_CONNECTION };
char *server_types[] = { "proxy", "coordinator", "home" };

typedef struct Proxy_Values {
    char *home;                 // who is the home server
    char *self;                 // who am I?

    nestClient *coordinator;    // connection to the proxy coordinator
    nestClient *home_con;       // connection to the home node
    NestReplyStatus coord_status;   // status of proxy coordinator

    nestClientPool *pending;    // clients waiting on already in progress xfers
    nestClientPool *waiting;    // clients waiting bec no avail connections

    nestProxyClientCache *connections;   // connections home & to other proxies

    DynamicLard *index;         // how to look up the other proxy servers
    struct nest_hash_table *coordinator_map;    // what the coordinator uses
                                                // to track proxies

    pthread_mutex_t mutex;      // for protecting internal state from
                                // asynchronous activity
    pthread_mutex_t auth_mutex;
};

// some parameters that affect how proxies behave
bool coordinator_also_serves = true;    // does coordinator add self to index
bool cooperative_proxies     = true;    // do proxies use each other

    // need to predeclare because registerCoordinatorDone might need to call
NestReplyStatus startRegisterCoordinator( );

    // collection of connections to the home node and other proxies
Proxy_Values *proxys = NULL;

    // state from the main server 
nestClientPool *client_pool = NULL;
fd_set *readfdset = NULL;
int (*transferFunction) (nestClient *, bool );

    // don't think we need these locks anymore since we are more careful to do
    // all data manipulation in the dispatch thread and only allow the
    // concurrency threads to do very simple chores
void
lockAuthMutex( char *server ) {
    // seeing some weird failures from the gsi auth code.  maybe not threadsafe?
    nest_debug( NEST_DEBUG_AUTH, 
            "PID: %d Locking auth mutex to connect to %s\n", 
            NestGetpid(), server );
    nest_pthread_mutex_lock( &(proxys->auth_mutex) );
}

void
unlockAuthMutex( char *server, NestReplyStatus status ) {
    nest_pthread_mutex_unlock( &(proxys->auth_mutex) );
    nest_debug( NEST_DEBUG_AUTH, 
            "PID: %d Unlocking auth mutex after connecting to %s: %s\n", 
            NestGetpid(), server, NestErrorString(status) );
}

void
lockMutex() {
    nest_pthread_mutex_lock( &(proxys->mutex) );
}

void
unlockMutex() {
    nest_pthread_mutex_unlock( &(proxys->mutex) );

}

bool
checkTwoStrings( void *o, void *t ) {
//  nest_debug( NEST_DEBUG_PROXY, "Comparing %s and %s for equality.\n", o, t );

    if ( o == NULL || t == NULL ) return ( o == NULL && t == NULL );
    int stat = strcmp( (char*)o, (char*)t );
    return ( stat == 0 );
}

    // returns whether a nest node is running as a proxy server or as 
    // a "home" node
bool
serverIsProxy( ) {
    return ( proxys != NULL );
}

bool
serverIsProxyCoordinator( ) {
    return ( proxys->coord_status == NEST_SUCCESS 
             && proxys->coordinator == NULL );
}

bool
proxyIsDisconnected( ) {
    return ( proxys->home_con == NULL );
}

int
getProxyCount( ) {
    CHECK_PID;
    return proxys->connections->size();
}

// this add client socket to fdset but not to dispatch client pool
// (probably because already in there)
void
makeDispatcherListen( nestClient *client ) {
    CHECK_PID;
    if ( client->hasSocket() ) {
        int socket = client->getConnection()->getSocket();
        NestFd_set( socket, readfdset );
    }
}

// this adds client socket to fdset and to dispatch client pool
void
makeDispatcherResponsible( nestClient *client) {
    CHECK_PID;
    makeDispatcherListen( client );
    client_pool->addClient( client );
}

void
stopDispatcherListen( nestClient *client ) {
    CHECK_PID;
    if ( client->getConnection()->hasSocket() ) {
        NestFd_clear( client->getConnection()->getSocket(), readfdset );
    }    
}

struct ProxyListArgs {
    bool first;
    char *coord;
};

    // a proxy registers with a home nest and receives a list of all of the
    // other proxies already registered
void
receiveProxyList( char *proxy, void *arg ) {
    struct ProxyListArgs *args = ( struct ProxyListArgs *) arg;
    nest_debug( NEST_DEBUG_PROXY, "%s is a registered proxy\n", proxy );
    if ( args->first ) {
        args->coord = NestStrdup( proxy );
    }
    args->first = false;
}

void
writeProxyAd( nestProtocol *ad ) {
    CHECK_PID;
    if ( ! serverIsProxy() ) return;
    char buffer[NEST_MESSAGESIZE];
    lockMutex();
    int map_size = -1;
    int map_servers = -1;
    if ( proxys->index != NULL ) {
        map_servers = proxys->index->numServers();
        map_size = proxys->index->getSize();
    }
    snprintf( buffer, NEST_MESSAGESIZE,
            "Home_server = \"%s\"\n"
            "Disconnected = %d\n"
            "Proxy_coordinator = \"%s\"\n"
            "Coordinator_status = \"%s\"\n"
            "Number_of_proxy_connections = %d\n"
            "Pending_requests_current = %d\n"
            "Pending_requests_ave = %.2f\n"
            "Pending_requests_max = %d\n"
            "Hash_map_size    = %4d\n"
            "Hash_num_servers = %4d\n",
            proxys->home,
            proxyIsDisconnected(),
            ( proxys->coordinator == NULL ) ? 
                "self" : proxys->coordinator->getIdentity(),
            NestErrorString( proxys->coord_status ),
            getProxyCount(),
            proxys->pending->size() + proxys->waiting->size(),
            proxys->pending->getAveLen() + proxys->waiting->getAveLen(),
            proxys->pending->getMaxLen() + proxys->waiting->getMaxLen(),
            map_size,
            map_servers
            );
    unlockMutex();
    ad->Sendn( buffer, NestStrlen( buffer ) );
}


char *
serverToString( void *server ) {
    return (char *)server;
}

void *
stringToServer( char *string, void *args ) {
    CHECK_PID;
        // TODO:  This will leak, actually if we always call clear first,
        // it might not...
    assert( args == NULL );
    return (void *)NestStrdup(string);
}


// this function flushes a proxy cache.  probably just be used for 
// repeatable measurements against a cold cache
NestReplyStatus 
FlushProxyCache( nestClient *client ) {
    CHECK_PID;
    MUST_BE_PROXY( client->ipAddress() );
    nest_debug( NEST_DEBUG_CLIENT, 
            "Flushing proxy cache at request of %s (%d) from %s.\n",
            client->getIdentity(), client->getId(), client->ipAddress() );
    NestReplyStatus s=nestStorageManager::formatDisk( 
            client->getIdentity(), NULL );
    nest_debug( NEST_DEBUG_CLIENT, 
            "Flushed proxy cache at request of %s (%d) from %s: %s.\n",
            client->getIdentity(), client->getId(), client->ipAddress(),
            NestErrorString(s) );
    return s;
}

const char *
findResponsibleServer(  const char *filepath, 
                        bool was_forwarded, 
                        const char *requestor ) 
{
    CHECK_PID;

    const char *server = NULL;

    lockMutex();
    if ( proxys->index != NULL ) {
        server = (const char *)proxys->index->getServer( filepath );
    } else {
        server = NULL;
    }
    unlockMutex();

    if ( server == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Unable to find a server in my index.  Will go home\n" );
        server = proxys->home;
    } else if ( ! cooperative_proxies ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, but not running coop mode,"
                " so will fetch from home node\n", server, filepath );
        server = proxys->home;
    } else if ( ! strcmp( server, proxys->self ) ) {
        nest_debug( NEST_DEBUG_PROXY,
                "I'm the responsible proxy for %s, will fetch from home\n",
                filepath );
        server = proxys->home;
    } else if ( was_forwarded ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "%s is responsible proxy for %s, but %s thinks I am, "
                "so I'll fetch anyway to avoid loops.\n",
                (char*)server, filepath, requestor );
        server = proxys->home;
    } else {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, fetching accordingly\n" ,
                (char*)server, filepath );
    }
    return server;
}

    // this is the function that receives a new mapping index from the coord
NestReplyStatus
ReceiveProxyIndex( nestClient *client ) {
    MUST_BE_PROXY( client->ipAddress() );
    CHECK_PID;

    if ( client != proxys->coordinator ) {
        nest_warn( __FILE__, __FUNCTION__, "%s is not coordinator, but is sending proxy index.\n",
                client->ipAddress() );
    }

    nest_debug( NEST_DEBUG_PROXY, "About to receive proxy index from %s\n",
            client->getIdentity() );

    lockMutex();

        // maybe clear is dangerous.... what if instead of clear, we free
        // and make a new one?
    delete proxys->index;
    proxys->index = new DynamicLard( NULL );
    //proxys->index->clear( true );
    NestReplyStatus status = NestReceiveIndex( 
                             client->getConnection()->getSocket(),
                             proxys->index,
                             stringToServer,
                             NULL );
    unlockMutex();

    nest_debug( NEST_DEBUG_PROXY, "Received proxy index from %s: %s\n",
            client->getIdentity(), NestErrorString( status ) );

    return status;

}

// TODO:  This is a hack bec the naming got goofy in the new parrot
// parrot always passes absolute names and we assume relative paths
int
skipStorageDir( const char *path ) {
    char *storagedir = nestStorageManager::getInitialClientDir();
    assert( storagedir != NULL );
    int i = 0;
    while( ! strncmp( storagedir, &path[i], strlen(storagedir) ) ) {
        i += strlen(storagedir);
    }
    return i;
}

// first check to see if we have the whole file locally.
// otherwise query the home connection for it.
NestReplyStatus  
proxyFilelength(  long *filesize,       const char *filename,
                  const char *user,     const char *cwd ) 
{
    CHECK_PID;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    *filesize = 0;

        // we have a complete copy local
    if ( ( nestStorageManager::filelength(  filesize,
                                            filename,
                                            user,
                                            cwd )) 
                == NEST_SUCCESS )
    {
        nest_debug( NEST_DEBUG_PROXY, 
                "Request for filesize of %s handled locally: %ld\n",
                filename, *filesize );
        return NEST_SUCCESS;
    }

        // TODO
        // first to see if we've already cached the size for this file.

        // need to query the home connection for it.
    //if ( proxyIsDisconnected() ) 
    if ( proxys->home_con == NULL ) return NEST_TEMPORARILY_UNAVAILABLE;

        // otherwise, go ahead and use the proxy connection to the home
        // node.  This is safe because this is done synchronously.
        // ignore any errors also communicating to the server.  The
        // dispatcher will notice these on the next select and inform us.
    assert( proxys->home_con != NULL );
    assert( proxys->home_con->hasSocket() );
    NestConnection home = proxys->home_con->getConnection()->getSocket();

       // get the real absolute path, then remove our storage dir
    string path;
    status = nestStorageManager::expand_path( path,
                                              (const char*)user,
                                              cwd,
                                              filename );
    const char *queryfile = path.c_str();
    int position = skipStorageDir( queryfile );
    queryfile = &queryfile[position];
    status = NestFilesize( *filesize, (char*)queryfile, home );   
    nest_debug( NEST_DEBUG_PROXY, 
            "Request for filesize of %s handled by querying home: %ld: %s\n",
            queryfile, *filesize, NestErrorString( status ) );
    return status;
}


// returns the port that proxies communicate over
int
getProxyPort() {
    const char *port_string;
    if ( GetConfigValue(Config_ProxyPort, &port_string) ) {
        return( atoi(port_string) );
    } else {
        nest_die( "To run a proxy, must specify proxy_port in config file\n" );
        return -1;
    }

}


/*
// create a connection to the home site, a coordinator or another proxy
NestReplyStatus
openProxyConnection( char *server, nestClient **connection, int type ) {

        // open socket
    int socket;
    NestReplyStatus status;
    lockAuthMutex( server );
    if ( type == PROXY_CONNECTION || type == COORD_CONNECTION ) {
        bool anonymous = false;
        bool user2user = true;
        status = myNestOpenConnection( 
            socket, server, anonymous, user2user, getProxyPort() );
    } else {
        status = NestOpenConnection( socket, server );
    }
    unlockAuthMutex( server, status );

        // make the connection
    if ( status == NEST_SUCCESS ) {
        if ( type == PROXY_CONNECTION || type == COORD_CONNECTION ) {
            *connection = new nestClient( 
                    (nestProtocol *) new nestProtocolProxy( socket ), NULL ); 
        } else {
            *connection = new nestClient( 
                    (nestProtocol *) new nestProtocolNest( socket ), NULL ); 
        }
        (*connection)->setName( server );
        (*connection)->setIp(   server );
    }
	int id = ( (status == NEST_SUCCESS) ? (*connection)->getId() : -1 );
    nest_debug( NEST_DEBUG_PROXY, 
            "PID: %d Made new %s connection to %s (%d): %s\n", 
            NestGetpid(), server_types[type], server, 
			id, NestErrorString(status) );
    return status;
}

NestReplyStatus
openHomeConnection( char *location ) {
    nestClient *connection;
    NestReplyStatus status = openProxyConnection( proxys->home, &connection, 
            HOME_CONNECTION );
    NEST_REPLY_CHECK(status);
    proxys->home_con = connection;
    return status;
}

// create a connection to a cooperating proxy
// TODO: Can this be done synchronously?  
// If so, all that needs to be done is make addProxyTransfer be asynch
NestReplyStatus
openCoordinatorConnection( char *location ) {
    nestClient *coord;
    nest_debug( NEST_DEBUG_PROXY, "Trying to open connection to proxy "
            "coordinator at %s.\n", location );
    NestReplyStatus status = openProxyConnection( location, &coord, 
            COORD_CONNECTION );
    NEST_REPLY_CHECK( status );

        // TODO:  we should register with coordinator to make sure
        // that coordinator is correct coordinator
    status = registerProxyConnection( coord );
    if ( status == NEST_SUCCESS ) {
            // OK, this is tricky.  The coordinator will push indices
            // so we need to listen to him so he must be added as a client
        proxys->coordinator = coord;
        client_pool->addClient( proxys->coordinator );
        makeDispatcherListen( proxys->coordinator );
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Couldn't register with coordinator %s: %s\n",
            location, NestErrorString( status ) );
    }

    return status;
}
*/

nestClient *
removeWaitingClient( nestClientPool *queue, const char *file ) {
    CHECK_PID;
    CLIENT_ITERATOR itr = queue->begin(); 
    while( itr != queue->end() ) {
        nestClient *queued = *itr;
        if ( strcmp( queued->getProxyRequest()->getNestFile(), file ) == 0 ) {
            nest_debug( NEST_DEBUG_PROXY, 
                    "Found client %s (%s) waiting for %s.\n",
                    queued->getIdentity(), queued->ipAddress(), file );
            itr = queue->erase( itr );
            return queued;
        } else {
            itr++;
        }
    }
    return NULL;
}

long proxyTimeout = 5000000;
void
proxyMaintenance( nestClientPool *queue ) {
    CHECK_PID;
    int count = 0;
    lockMutex();
    CLIENT_ITERATOR itr = queue->begin(); 
    while( itr != queue->end() ) {
        count++;
        nestClient *queued = *itr;
        *itr++;
        nestClient *proxy = queued->getProxyClient();
        const char *file = queued->getProxyRequest()->getNestFile(); 
        long millis = queued->getRequest()->getAge();
        const char *proxyname = ( ( proxy ) 
                ? proxy->getIdentity() : "not yet assigned" );
        if ( millis > proxyTimeout ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Client %s from %s (id %d) has been waiting "
                    "for %s from %s for %.3f seconds.\n",
                    queued->getIdentity(),
                    queued->ipAddress(),
                    queued->getId(),
                    file,
                    proxyname,
                    (float)millis/MILLION );
        }
    }
    unlockMutex();
    if ( count > 0 ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "%d clients still waiting on transfers.\n",
            count );
    }
}

void
proxyMaintenance() {
    CHECK_PID;
    if ( ! serverIsProxy() ) return;
    proxyMaintenance( proxys->pending );
    proxyMaintenance( proxys->waiting );
}


// start a transfer (or send error) for everyone who was waiting for a file
int
checkProxyQueue( nestClientPool *queue, const char *file, NestReplyStatus stat){
    CHECK_PID;
    int started = 0;
    nestClient *queued;
    while( ( queued = removeWaitingClient( queue, file ) ) != NULL ) {
        if ( stat == NEST_SUCCESS ) {
            transferFunction( queued, true );
        } else {
            if ( queued->getStorageConnection() != NULL ) {
                nestStorageManager::close( queued->getStorageConnection() );
            }
            queued->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
            if ( queued->getRequest() != NULL ) {
                queued->endRequest( stat );
            }
            makeDispatcherListen( queued );
        }
        started++;
    }
    return started;
}
    
// currently this is called from within the synch operation so the responibility
// to reply to the client becomes the proxy's.  Once all dirty writes are 
// flushed, send the reply.
NestReplyStatus
flushDirtyWrites( nestClient *client ) {
    CHECK_PID;
    NestReplyStatus status = NEST_SUCCESS;

    nestRequest *dirty_write = client->getDirtyWrite( );
    if ( dirty_write ) {
        client->setFlushing( true );
        nest_debug( NEST_DEBUG_PROXY, "Getting set to flush a dirty write "
                "of %s for %s\n",
                dirty_write->getNestFile(),
                client->getIdentity() );
            // addProxyTransfer will add client to the pending queue
            // assigning it the proxy request will allow us to find this
            // client again once the proxy transfer completes
        client->assignProxyRequest( dirty_write );
        status = addProxyTransfer( client );    
        if ( status == NEST_SUCCESS ) {
            return NEST_SUCCESS;
        }
    } else {
        nest_debug( NEST_DEBUG_PROXY, 
                "Done flushing dirty writes for %s\n",
                client->getIdentity() );
            // the error checking is done in checkProxyTransfers
    }

        // if we make it here, either we had a flush error or there are no
        // more dirty writes
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck( );
    }
    client->endRequest( status );
    makeDispatcherListen( client );
    return status;
}

// convenience call which creates a shadow name for incoming files to 
// provide some degree of atomicity and consistency protection
char *
getShadowName( const char *file ) {
    CHECK_PID;
    char *shadow = (char *)NestCalloc( NestStrlen( file ) + 10 );
    NestSnprintf( shadow, NestStrlen( file ) + 10, "%s.proxy", file );
    return shadow;
}


bool
alreadyPendingProxyTransfer( nestClient *client, const char *file ) {
    CHECK_PID;
    for(    CLIENT_ITERATOR itr=proxys->pending->begin();
            itr!=proxys->pending->end(); 
            *itr++)
    {
        nestClient *queued = *itr;
        if ( strcmp( queued->getProxyRequest()->getNestFile(), file ) == 0 ) {
            return true;
        }
    }
    return false;
}

// for caching reads, create a shadow proxyreq
// this is because proxies caches blocks of data and we need to know
// which block to fetch as well as to remember the original parameters
// we actually also create a third request object which is what is 
// assigned to the proxy connection for the purpose of fetching the block 
// (safely synchronous - no network activity)
NestReplyStatus 
proxyTranslateRequest( nestClient *client ) {
    CHECK_PID;

    nestRequest *req = client->getRequest();
        
        // just accept writes as is, do we need a notion of sparse files?
    if ( req->isWrite() ) {
            // TODO
        // client->addDirty(); // actually do this when transfer completes
        nest_debug( NEST_DEBUG_PROXY, 
            "Client %s is writing %s %ld-%ld\n",
            client->getIdentity(),
            req->getNestFile(), req->getStart(), req->getEnd() );
        return NEST_SUCCESS;
    }

        // for third party puts and gets, don't do proxy stuff
    if ( req->isThirdParty() ) {
        return NEST_SUCCESS;
    }
        
    // for reads there are a couple of possibilities
    // 1) we have the whole file because it has been written here
    // 2) we are already fetching the needed block
    // 3) we already have the needed block
    // 4) we need to fetch the needed block from the home server
    // 5) we need to fetch the needed block from another proxy

    // 1) is easy and we just check for that here and leave without doing
    // anything it that is the case
    // 2-5) more complicated.  Make a proxy req asking for the chunk.
    // 2) if the chunk exists, it will be served normally.
    // 3-5) if the chunk doesn't exist, main server will call addProxyTransfer 

    // 1) we have the whole file because it has been written here
    long filelength;
    if ( ( nestStorageManager::filelength( &filelength,
                                            req->getNestFile(),
                                            client->getIdentity(),
                                            client->getCwd() )) 
                == NEST_SUCCESS )
    {
        nest_debug( NEST_DEBUG_PROXY, 
                "Can satisfy req for %s with local whole file copy\n",
                req->getNestFile() );
        return NEST_SUCCESS;
    }
        
        // proxys don't keep whole input files, they only keep chunks
        // translate this file request into a chunk request

        // find the closest start of a chunk which is less than or equal
        // to this req start
    long chunk = 0;
    while( chunk+PROXY_BLOCK_SIZE <=req->getStart() ) chunk += PROXY_BLOCK_SIZE;
    char *newname = (char*)NestMalloc(NEST_MAXPATHLEN);
    NestSnprintf( newname, NEST_MAXPATHLEN, "%s.%ld", req->getNestFile(),chunk);
    long newstart = req->getStart() - chunk;
    long newend   = req->getEnd() - chunk;

        // create a proxy req which is used to request the chunk from
        // either the home node or another proxy
    nest_debug( NEST_DEBUG_PROXY, 
            "Translating req for %s %ld-%ld to %s %ld-%ld\n",
            req->getNestFile(), req->getStart(), req->getEnd(),
            newname, newstart, newend );
    nestRequest *proxyreq = new nestRequest( 
            req->getType(), newname, NULL, NULL, newstart, newend, NULL );
    req->setChunk( chunk );
    proxyreq->setChunk( chunk );
    NestFree( newname );
    client->assignProxyRequest( proxyreq );
    

        // check whether the client has issued a req that will go beyond
        // the end of the current chunk.  If so, this could cause problems
        // for clients doing whole file gets.
    if (        proxyreq->getEnd() == ENTIRE_FILE 
            ||  proxyreq->getEnd() > chunk + PROXY_BLOCK_SIZE ) 
    {
        nest_warn( __FILE__, __FUNCTION__,
                "%s:%s client read will be truncated. Hope they try again\n",
                __FILE__, __FUNCTION__ );
    }
    return NEST_SUCCESS;
}


// this function can be used both to register with the proxy coordinator and
// the home server
NestReplyStatus
registerProxyConnection( nestClient *connection ) {

        // register as a proxy
    NestReplyStatus status = NestProxyRegister( 
                                    proxys->self,
                                    connection->getConnection()->getSocket() );
        
    nest_debug( NEST_DEBUG_PROXY, "Registered as proxy to %s: %s.\n",
            connection->getIdentity(), NestErrorString( status ) );
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "Unable to register as proxy to %s: %s\n",
                connection->getIdentity(), NestErrorString( status ) );
    }
    return status;
}


    // assign a thread the responsibility of opening a connection to home
    // server and registering there as a proxy
    // this one is easier than other asynchronous work because we can just
    // loop forever in this thread since the home node will never change
    // locations.  Later if it can, we need to change this.
NestReplyStatus
asynchronousRegisterHome( nestClient *home ) {
    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    int socket;
    int attempts = 0;
    int sleep_max = 15;

    while( status != NEST_SUCCESS ) {
        if ( attempts++ > 0 ) {
            int sleep_time = ( attempts < sleep_max ) ? attempts : sleep_max;
            nest_warn( __FILE__, __FUNCTION__,  "Unable to register with home node %s: %s.  "
                        "Will sleep %d secs and try again.\n",
                        home->getName(),
                        NestErrorString( status ), sleep_time ); 
            sleep( sleep_time );
        }

            // first try to just open a socket
        status = NestOpenConnection( socket, (char *)home->getIdentity() );
        if ( status == NEST_SUCCESS ) {
            nestProtocolNest *npn = new nestProtocolNest( socket );
            home->assignConnection( npn );

                // then try to actually send a register command
            status = registerProxyConnection( home ); 
            if ( status != NEST_SUCCESS ) {
                home->assignConnection( NULL );
                delete npn;
            }
        }
    }

    return NEST_SUCCESS;
}

NestReplyStatus
registerHomeDone( nestClient *home, NestReplyStatus status ) {
    CHECK_PID;
    assert( status == NEST_SUCCESS );
    nest_debug( NEST_DEBUG_PROXY, 
            "Successfully registered with home %s.\n", home->getIdentity() );
    proxys->home_con = home;
    proxys->home_con->assignTaskFunction( NULL, NULL );
    makeDispatcherResponsible(   proxys->home_con );

        // special case start up.  Start the search for proxy coord now.
    if ( proxys->coord_status == NEST_NOT_YET_INITIALIZED ) {
        startRegisterCoordinator();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
startRegisterHome( ) {
    CHECK_PID;
    nestClient *home = new nestClient( NULL, NULL );
    home->setName( proxys->home );
    home->setIdentity( proxys->home );
    home->assignTaskFunction(  asynchronousRegisterHome, 
                               registerHomeDone );
    return nestConcurrencyManager::addTask( home );
}

NestReplyStatus
registerCoordinatorDone( nestClient *coordinator, NestReplyStatus status ) {
    CHECK_PID;

        // always clear the task function pointers
    coordinator->assignTaskFunction( NULL, NULL );

    if ( status != NEST_SUCCESS ) {
            // need to try again
        delete coordinator;
        return startRegisterCoordinator();
    }

    nest_debug( NEST_DEBUG_PROXY, 
        "Successfully registered with coordinator %s.\n", 
        coordinator->getIdentity() );
    if ( ! strcmp( coordinator->getIdentity(), proxys->self ) ) {

        nest_debug( NEST_DEBUG_PROXY, "Hey, I'm the proxy coordinator!\n" );

        assert( proxys->coordinator == NULL );
        assert( proxys->coordinator_map == NULL );
        assert( proxys->index != NULL );

        proxys->coordinator_map = nest_hash_table_create( 1037, NULL );

            // this actually doesn't seem to be the problem....
            // maybe clear is dangerous.... what if instead of clear, 
            // we free  and make a new one?
        delete proxys->index;
        proxys->index = new DynamicLard( serverToString );
        proxys->index->setServerEqualityFunction( checkTwoStrings );
            /*
            proxys->index->clear( true );   // do we want to clear?
                                            // or keep using same index?
                                            // I think we have to clear...
            */

            // only coordinator needs to bother with maintaining the
            // message string.
        proxys->index->setServerToString( serverToString );
        proxys->index->setName( proxys->self );
        if ( coordinator_also_serves ) {
            proxys->index->addServer( (void*)NestStrdup(proxys->self) );
        }
            
    } else {

            // someone else is the coordinator
        proxys->coordinator = coordinator;
            
            // what if we only make the dispatcher listen, but not be
            // responsible?
        makeDispatcherResponsible( proxys->coordinator );
    }


    proxys->coord_status = status;
    return status;
}

NestReplyStatus
asynchronousRegisterCoordinator( nestClient *coordinator ) {

    NestReplyStatus status;

    struct ProxyListArgs args;
    args.first  = true; 
    args.coord  = NULL;

    
    /*
        // this is maybe a little goofy, but we stashed a duped
        // connection to the home node in this client
        // WATCH OUT: We assume it is a nestProtocolNest connection
    nestProtocolNest *home = (nestProtocolNest*)coordinator->getConnection();
    coordinator->assignConnection( NULL );
    */

        // actually all we have is the home_server name.  try to open 
        // connection so we can fetch proxy list.
        // it may be goofy, but we stashed the home name in the client.
    NestConnection home_con = -1;
    status = NestOpenConnection( home_con, proxys->home );
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "Unable to open home con, so can't discover coord: %s\n",
                NestErrorString( status ) );
    }
    
    if ( status == NEST_SUCCESS ) {

            // ask home who the coordinator is
        status = NestProxyQuery( home_con, receiveProxyList, &args  );
        if ( status != NEST_SUCCESS ) goto DONE;
        if ( args.coord == NULL ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "received empty proxy list.  this shouldn't happen.\n" );
            goto DONE;
        }
        coordinator->setName( args.coord );
            // need to set identity also to be used to open connection
        coordinator->setIdentity( args.coord );

        nest_debug( NEST_DEBUG_PROXY, "Home says coordinator is %s.\n",
                coordinator->getIdentity() );

            // hey, it's me
        if ( ! strcmp( args.coord, proxys->self ) ) {
            goto DONE;
        }

            // open a connection to whoever it is and register with them
        int socket;
        bool anonymous = false;
        bool user2user = true;
        int port = getProxyPort();

            // TODO: This whole next part of code should be shared
            // with asynchronousRegisterHome
        status = myNestOpenConnection(  socket,
                                        (char*)coordinator->getIdentity(),
                                        anonymous,
                                        user2user,
                                        port );
        if ( status != NEST_SUCCESS ) goto DONE;


        nestProtocolNest *npn = new nestProtocolNest( socket );
        coordinator->assignConnection( npn );
        status = registerProxyConnection( coordinator );
        if ( status != NEST_SUCCESS ) {
            coordinator->assignConnection( NULL );
            delete npn;
        }
    }


    DONE:
    if ( status != NEST_SUCCESS ) {
        int sleep_time = 4;
        nest_debug( NEST_DEBUG_PROXY, 
                "Unable to register with coordinator: %s.  "
                "Will try again in %d secs.\n",
                NestErrorString( status ),
                sleep_time );
        sleep( sleep_time );
    }
    if ( args.coord != NULL )   NestFree( args.coord );

    return status;
}

NestReplyStatus
startRegisterCoordinator( ) {
    CHECK_PID;
    nestClient *coordinator = new nestClient( NULL, NULL );

    nest_debug( NEST_DEBUG_PROXY, 
            "Will try to discover / register with coordinator now.\n" );

        // two options.
        // 1) query whether we're disconnected.  if so, add this to some
        //    pending queue when we get connected again.
        // 2) give this a Dup of the current home_con which may be invalid
        //    if this fails to use that Dup, then it sleeps and trys again
        //    and eventually the home_con will be good and so will the Dup
        // 3) actually 2) is no good, bec the dispatcher is listening on 
        //    the home_con and when we ask it to send traffic via the query
        //    proxys request, dispatcher may think it is a ready client and
        //    try to receiveRequest from it.  So, we should just make a new
        //    connection to home_node whenever we need to discover coord

    /*
        // lets do 2)
    if ( ! proxyIsDisconnected() ) {
        assert( proxys->home_con->getConnection()->getType() == NEST_PROTOCOL );

        int socket = NestDup( proxys->home_con->getConnection()->getSocket() );
        nestProtocolNest *npn = new nestProtocolNest( socket );
        coordinator->assignConnection( npn );
    }
    */

    coordinator->assignTaskFunction( asynchronousRegisterCoordinator,
                                     registerCoordinatorDone );
    return nestConcurrencyManager::addTask( coordinator );
}

bool
connectionFailure( NestReplyStatus status ) {
    if ( status == NEST_NO_CONNECTION )          return true;
    if ( status == NEST_COULD_NOT_AUTHENTICATE ) return true;
    return false;
}

NestReplyStatus
asynchronousProxyFetch( nestClient *proxy ) {

    NestReplyStatus status      = NEST_SUCCESS;
    nestProxyFetch *fetch       = proxy->getProxyFetch();
    nestProtocol    *connection = proxy->getConnection();
    long            length      = fetch->getHomeLength(); 
    const char    *file         = fetch->getHomeFile(); 
    long            offset      = fetch->getHomeOffset(); 
    const char    *server       = fetch->getServer();

        // possible the proxy already has a connection
        // otherwise try to open a connection here
    if ( connection == NULL ) {
        int socket;
        int port       = NEST_SERVER_PORT;
        bool anonymous = false;
        bool user2user = false;
        
        if ( ! strcmp( server, proxys->home ) ) {   // home connection
            port = NEST_SERVER_PORT;
            user2user = false;
        } else {                                    // proxy connection
            user2user = true;
            port      = getProxyPort();
        }
        status = myNestOpenConnection(socket,server,anonymous,user2user,port);
        if ( status == NEST_SUCCESS ) {
            proxy->setName( server );
            proxy->setIp(   server );
            connection = (nestProtocol*) new nestProtocolNest( socket );
            proxy->assignConnection( connection );
        }
        nest_debug( NEST_DEBUG_PROXY, "Opened connection to %s: %s\n",
                    server, NestErrorString( status ) );
    }
    
        // ask for permission to read this block
    if ( status == NEST_SUCCESS ) {
        assert( connection != NULL );
        int position = skipStorageDir( file );
        status = RequestReadFile(&file[position], offset, length, 
                connection->getSocket());
        if ( status != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_PROXY,
                    "Unable to request read of %s (%d) from %s: %s\n",
                    file, offset, server, NestErrorString( status ) );
        }
    }

        // go ahead and read the block
        // TODO: this is where it would be nice to use concurrency manager to
        // actually schedule this transfer and get the accounting as well.
    if ( status == NEST_SUCCESS ) {
        proxy->startTransfer( PROXY_GET, length );
        status = proxy->getTransfer()->BlockingTransfer();
    }

        // if no connection, need to return to dispatcher to rediscover
        // who is responsible for this block before we try again.
    if ( connectionFailure( status ) ) {
        // TODO: probably should not sleep here.  Add another 
        // chore type to just sleep.  Have proxy translator determine.
        // sleep here and hope things improve.  eventually need to die.
        fetch->incrementAttempts();
        long sleep_secs = 2 * fetch->getAttempts(); 
        nest_debug( NEST_DEBUG_PROXY,
                "Unable to connect to %s for %s (%ld). Try again in %d secs.\n",
                server, file, offset, sleep_secs );
        sleep( sleep_secs ); 
        nest_debug( NEST_DEBUG_PROXY,
                "Slept after failure for %s (%ld) from %s. Should try again.\n",
                file, offset, server );
    }

    return status;
}

// if connectionFailure, clean up and try again
// otherwise, go ahead and start any pending clients waiting on this
NestReplyStatus
proxyFetchDone( nestClient *proxy, NestReplyStatus status ) {
    CHECK_PID;
    nestProxyFetch *fetch = proxy->getProxyFetch();
    assert( fetch != NULL );

    nest_debug( NEST_DEBUG_PROXY, 
            "Proxy fetch of %s done: %s\n",
            fetch->getProxyFile(),
            NestErrorString( status ) );

        // clean up the storage stuff regardless of status
    nestProtocol *storage = proxy->getStorageConnection();
    proxy->assignStorageConnection( NULL );
    nestStorageManager::close( storage, proxy );
    delete storage;

        // either rename or delete the temporary incoming name
    if ( status != NEST_SUCCESS ) {
        nestStorageManager::deleteFile( fetch->getClientName(),
                                        fetch->getClientCwd(), 
                                        fetch->getShadowFile() );
    } else {
        NestReplyStatus renameStatus = nestStorageManager::renameFile( 
                        fetch->getClientIdentity(), 
                        fetch->getClientUid(),
                        fetch->getShadowFile(),
                        fetch->getProxyFile(),
                        fetch->getClientCwd(),
                        false );
        nest_debug( NEST_DEBUG_PROXY, "Renamed proxyfile %s to %s: %s.\n",
                        fetch->getShadowFile(),
                        fetch->getProxyFile(),
                        NestErrorString( renameStatus ) ); 
        if ( renameStatus != NEST_SUCCESS ) {
            nest_die( "Failed to rename proxyfile: %s\n",
                    NestErrorString( renameStatus ) );
        }
    }

        // see if we need to try again
    if ( connectionFailure( status ) ) {
            // try again
        proxy->endTransfer( status );
        delete proxy;
        nest_debug( NEST_DEBUG_PROXY,
                "Proxy fetch of %s failed %d times.  Will try again.\n",
                fetch->getProxyFile(),
                fetch->getAttempts() );
        NestReplyStatus status = startProxyFetch( fetch );
        assert( status == NEST_SUCCESS );
    } else {
        // if we make it here, either success or some probably permanent error
            // check to see if any of the pending clients were waiting for this
            // there'd better be at least one!
            // pending must be checked before waiting
            //  this ensures at least minimal forward progress
        int started = 0;
        lockMutex();
        const char *file = fetch->getProxyFile();
        started += checkProxyQueue( proxys->pending, file, status);
        started += checkProxyQueue( proxys->waiting, file, status);
        unlockMutex();
        assert( started != 0 );

            // this proxy is in good shape, let's cache it
        proxy->assignTaskFunction( NULL, NULL );
        proxy->assignProxyFetch( NULL );
        assert( ! strcmp( proxy->getName(), fetch->getServer() ) );
        proxys->connections->insertProxyClient( 
                    (char*)proxy->getName(), proxy ); 
        delete fetch;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
startProxyFetch( nestProxyFetch *fetch ) {
        
        // 1) we need to open a storage connection for this object
        // 2) we need to create a proxy client to do the transfer

    NestReplyStatus status   = NEST_SUCCESS; 
    nestProtocol    *storage = NULL;
    nestClient      *proxy   = NULL;
    const char      *server  = NULL;
    int weight               = 1;
    int link_count           = 0;

        // figure out who the server is.
    server = findResponsibleServer( fetch->getProxyFile(),
                                    fetch->wasForwarded(),
                                    fetch->getClientName() ); 
    fetch->setServer( server );
    if ( ! strcmp( server, proxys->home ) ) {
        // if home is the server, then this proxy is responsible
        // so up the weight to avoid this block from getting removed
        // from the cache to make it so that primary copies are valued
        // above secondary
        nest_debug( NEST_DEBUG_PROXY,
                "Upping weight of %s since its a primary copy\n", 
                fetch->getProxyFile() );
        weight = 2;
    }

        // actually get the storage connection
    status = nestStorageManager::open(  &storage,
                                        fetch->getClientIdentity(),
                                        fetch->getClientUid(),
                                        NULL,
                                        fetch->getShadowFile(),
                                        WRITE_FLAGS,
                                        0,
                                        fetch->getHomeLength(),
                                        fetch->getClientCwd(),
                                        link_count,
                                        weight );
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Could not open an fd for %s: %s.\n",
                fetch->getShadowFile(), NestErrorString( status ) );
        return status;
    }
    assert( status == NEST_SUCCESS );

        // try to find a cached connection to this server
    proxy = proxys->connections->getCachedProxy( server );
    if ( proxy == NULL ) {
        nest_debug( NEST_DEBUG_PROXY,
                "Unable to find cached proxy to %s, creating new one.\n",
                server );
        proxy = new nestClient( NULL, NULL );
    }

    proxy->assignProxyFetch( fetch );
    proxy->assignStorageConnection( storage );
    proxy->assignTaskFunction(  asynchronousProxyFetch, 
                                proxyFetchDone );

    status = nestConcurrencyManager::addTask( proxy );
    return status;
}

    
    // are we here because 
    // 1) the file is missing and we need to fetch it
    // 1a) either from the home node
    // 1b) or from another proxy
    // 2) are we here because the file is already currently being fetched?
    // 3) are we here to flush a dirty write back to the home server?
    // -- in any event, don't do anything here that could block.
    // especially nothing that could block on another proxy bec of deadlock
NestReplyStatus
addProxyTransfer( nestClient *client ) {

    NestReplyStatus  status      = NEST_SUCCESS;
    nestProxyFetch   *fetch      = NULL;
    nestRequest      *proxyreq   = client->getProxyRequest();
    assert( proxyreq != client->getRequest() );

        // don't know how to do flush writes anymore.
    if( proxyreq->isWrite() ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

        // 2) are we here because the file is already currently being fetched?
    if ( alreadyPendingProxyTransfer( client, proxyreq->getNestFile() ) ) {
       nest_debug( NEST_DEBUG_PROXY, "Adding another pending req for %s\n",
                proxyreq->getNestFile() );
        status = NEST_SUCCESS;
    } else {
            // need to go out and get it ourselves 
        fetch = NULL;
        const char *homefile  = client->getRequest()->getNestFile();
        const char *proxyfile = client->getProxyRequest()->getNestFile();
        char *shadowname      = getShadowName( proxyfile );
        long home_off         = client->getRequest()->getChunk();
        long length           = PROXY_BLOCK_SIZE;
        bool forwarded_req    = (client->getConnection()->getType()
                                 ==PROXY_PROTOCOL);
        nest_debug( NEST_DEBUG_PROXY, "Adding proxy fetch for user '%s'\n", 
                client->getIdentity() );
        fetch = new nestProxyFetch( homefile, 
                                    proxyfile, 
                                    (const char *)shadowname, 
                                    home_off,
                                    length, 
                                    client,
                                    forwarded_req );
        NestFree( shadowname );
        status = startProxyFetch( fetch );
    }

        // make sure to accept responsibility for this client so main server
        // won't accept on it anymore.  dispatcher turns back on at taskDone.
        // also make sure to remember that we're waiting on this file.
    if ( status == NEST_SUCCESS ) {
        stopDispatcherListen( client );
        proxys->pending->addClient( client );
    } else {
        if ( fetch != NULL ) delete fetch;
    } 
    return status;

    /*
        // otherwise find an available connection, (or make one)
        // and try to open a storage connection for the proxy file
        // and request permission from the home server to start proxy transfer
    status = openProxyStorageConnection( client, proxyreq, &storage );
    if ( status != NEST_SUCCESS ) {
        return status;
    } 
    client->assignStorageConnection( storage );

        // need to remember here that this client is waiting for this
        // so that only one thread will request it
        // this does need to be done synchronously
        // however, this means that for failed transfers, we need to remove
        // client from the pending list
    lockMutex();
    proxys->pending->addClient( client );
    unlockMutex();

        // this is the code that needs to be put into a thread
    nest_debug( NEST_DEBUG_PROXY, "Assigning asycnInitProxyTransfer func ptr "
            "to client %s (%d) to start transfer of %s.\n",
            client->getIdentity(), client->getId(), proxyreq->getNestFile() );
    client->assignTaskFunction( 
            asynchronouslyInitiateProxyTransfer,
            asynchronousInitCleanup );
    status = nestConcurrencyManager::addTask( client );
    return status;
    */
}

// TODO: This can segfault when it trys to send it to a client that has
// already crashed and been removed....
int
sendProxyIndexToOne( nestClient *client ) {
    CHECK_PID;
    assert( proxys->index != NULL );
    NestReplyStatus status = client->sendProxyIndex( proxys->index );
    nest_debug( NEST_DEBUG_PROXY, "Pushed revised index to %s: %s\n",
            client->ipAddress(), NestErrorString( status ) ); 
    if ( status != NEST_SUCCESS ) {
            // ok to ignore this here?
            // probably.  assume that only reason for failure is that
            // no connection, in which case the client is lost and soon
            // the dispatcher will tell us about it.
        if ( ! connectionFailure( status ) ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Coordinator not dealing with proxy failure: %s\n",
                    NestErrorString( status ) );
        }
    }
    return 1;
}

int
sendProxyIndex( const char *key, void *item, void *arg ) {
    assert( arg == NULL );
    nestClient *client = (nestClient *)item;
    return sendProxyIndexToOne( client );
}

// TODO:  Does coordinator activity need to be asynchronous?
// I think not bec as long as there is only one coordinator, only that one
// coordinator will ever do active initiations to the other proxies as long
// as the other proxies active initiations are done asynchronously
//
// OK.  Admittedly there can be deadlock if two different proxies ever think
// they are each the coordinator...can this happen?
NestReplyStatus
sendProxyIndexToAll( ) {
    nest_debug( NEST_DEBUG_PROXY, "Sending proxy index to all\n" );
    nest_hash_table_iterate( proxys->coordinator_map,
            sendProxyIndex, NULL );
    return NEST_SUCCESS;
}


// this is where we check to see if we've lost any proxy connections that
// are important to us.
NestReplyStatus
removeClientFromProxy( nestClient *client ) {
    CHECK_PID;

    bool isProxy = ( client->getConnection()->getType() == PROXY_PROTOCOL ); 

    if ( ! isProxy )  {
        // check see if this normal client was was waiting on anything
        // however do we want to remove from the pending and waiting queues?
        // TODO remove real clients from waiting and pending lists removeClient
        // actually, if they stay in there, eventually the proxy transfer
        // will complete and we'll discover they down when we try to send
        // them the proxy file
		lockMutex();
		bool uh_oh = (bool)
			(proxys->pending->contains( client ) 
			 + proxys->waiting->contains(client));
		if ( uh_oh ) {
			nest_warn( __FILE__, __FUNCTION__, 
                    "Need to remove client %s (%d) from %s waiting "
					"on %s (%d) from pending list\n",
					client->getIdentity(), client->getId(), client->ipAddress(), 
					client->getRequest()->getNestFile(),
					client->getRequest()->getChunk() );
		}
		unlockMutex();
    }

    if ( client == proxys->home_con ) {
            // lost home connection
        nest_warn( __FILE__, __FUNCTION__, 
                "Oh shit!  Lost our home connection %s at %s. "
                    "Will reconnect\n",
                    client->getIdentity(), client->ipAddress() );
        proxys->home_con = NULL;
        startRegisterHome();

    } 

    if ( client == proxys->coordinator ) {
            // lost coordinator connection
        assert( proxys->coordinator != NULL );
        nest_warn( __FILE__, __FUNCTION__, "Oh shit!  Lost our coordinator %s at %s.\n",
                        client->getIdentity(), client->ipAddress() );
        proxys->coordinator = NULL;
        proxys->coord_status = NEST_NO_CONNECTION; 
        startRegisterCoordinator();
    } 
    
    if ( serverIsProxyCoordinator() && isProxy ) {
            // if we make it here, this proxy is the current coordinator
            // and this client is another proxy
        assert( proxys->coordinator_map != NULL );
        lockMutex();
        assert( proxys->index != NULL );
        nest_debug( NEST_DEBUG_PROXY, "Removing server %s from proxy map.\n",
                client->ipAddress() );
        proxys->index->removeServer( (void *)client->ipAddress(), true );
        nest_hash_table_remove( 
                proxys->coordinator_map, client->ipAddress() );
        sendProxyIndexToAll();
        unlockMutex();
    }

    return NEST_SUCCESS;
}


// proxies register themselves with who they think is the coordinator
NestReplyStatus
proxyRegisterProxyClient( nestClient *client, char *proxy ) {
    CHECK_PID;

    if ( ! serverIsProxyCoordinator() ) {
        nest_warn( __FILE__, __FUNCTION__, "%s (%s) tried to register as proxy, but I'm not coord.\n",
                    client->getIdentity(), client->ipAddress() );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_NOT_YET_INITIALIZED ); 
        return NEST_SUCCESS;
    } else {
            // make sure to send the ack before sending the index
        client->sendAck();
    }

        // if we make it here, this proxy is the current coordinator
        // and this client is another proxy
    //client->setName( client->ipAddress() );
    lockMutex();
    assert( proxys->coordinator_map != NULL );
    if (!nest_hash_table_lookup( proxys->coordinator_map, client->ipAddress())){
        nest_debug( NEST_DEBUG_PROXY, 
                "New proxy (%s) from client %s (%d) registering\n",
                proxy, client->ipAddress(), client->getId() );
        proxys->index->addServer( (void*)(NestStrdup(client->ipAddress())) );
        nest_hash_table_insert( proxys->coordinator_map, 
                client->ipAddress(),
                client );
        sendProxyIndexToAll();
    } else {
        nest_debug( NEST_DEBUG_PROXY, "New proxy %s (%d) already registered\n",
                client->getIdentity(), client->getId() );
        sendProxyIndexToOne( client );
    }
    unlockMutex();
    return NEST_SUCCESS;
}

NestReplyStatus
addNewClientToProxy( nestClient *client ) {

        // nothing to do here now.
    return NEST_SUCCESS;
}

// find out who the coordinator is.  may need to do this if the 
// current coordinator dies as well.
// this currently just keeps looping until it correctly finds one
/*
NestReplyStatus
findCoordinator( nestClient *home ) {
    struct ProxyListArgs args;
    args.self  = proxys->self;
    NestReplyStatus status = NEST_UNKNOWN_ERROR;

    int attempts = 0;
    do {
        if ( attempts > 0 ) {
            long sleep_time = 500000000 * attempts;
            nest_warn( __FILE__, __FUNCTION__, "Unable to connect to %s: %s.  "
                       "Will sleep for %.3f and try again.\n", 
                       args.coord, ((float)sleep_time / BILLION ));
            NestNanosleep( sleep_time );
        }
        args.first = true;
        args.coord = NULL;

        if ( proxys->coordinator != NULL ) {
            nest_die( 
                "Trying to find new coordinator but old one still exists.\n");
        }

        status = NestProxyQuery( home->getConnection()->getSocket(),
                receiveProxyList, &args  );
        if ( args.coord == NULL ) {
            nest_die( "Couldn't discover proxy coordinator.\n" );
        }
        if ( status != NEST_SUCCESS ) break;
        
        if ( ! strcmp( args.coord, proxys->self ) ) {
            nest_debug( NEST_DEBUG_PROXY, "Hey!  I'm the proxy coordinator\n" );
            lockMutex();
            proxys->coordinator = NULL;
            if ( proxys->coordinator_map ) {
                nest_hash_table_delete( proxys->coordinator_map );
            }
            proxys->coordinator_map = nest_hash_table_create( 1037, NULL );
            proxys->index->clear( true );

            if ( coordinator_also_serves ) {
                proxys->index->addServer( (void*)NestStrdup(proxys->self) );
            }
            unlockMutex();
        } else {
            status = openCoordinatorConnection( args.coord );
        }
        
        if ( args.coord ) NestFree( args.coord );
    } while ( status != NEST_SUCCESS && attempts < 10 );

    return status;
}
*/

char *
selfIpAddress() {

    struct utsname utsname_buf;
    int ret = uname( &utsname_buf );
    if ( ret < 0 ) nest_die( "Couldn't discover self ip address.\n" );
//    printf( "Name is %s.\n", utsname_buf.nodename );
    struct hostent *host_info = gethostbyname( utsname_buf.nodename );
    char *ip = inet_ntoa( *(struct in_addr *)*host_info->h_addr_list );
//    printf( "ip is %s.\n", ip );

    return ip;
}


void
proxyHelp() {
    nest_printf("--home home, to make a proxy server for a home node\n" );
    nest_printf("--standalone, to make proxies not cooperate\n" );
}


    // set up proxy structures and connect to the home server
int
initProxy( int( *transferFP )(nestClient*,bool),
            nestClientPool *clients,
            fd_set *read_fds )
{
    CHECK_PID;
    
    const char *home = NULL;

        // save the state passed from the main server
    transferFunction  = transferFP;
    client_pool = clients;
    readfdset = read_fds;

	// Read the proxy host from the config file...
    const char *value;
    if ( GetConfigValue( Config_ProxyHost, &value ) && *value != 0) {
        home = value;
    }
    nest_debug( NEST_DEBUG_PROXY, "Config for proxy host says home is %s\n",
            home );

    if ( GetConfigValue( Config_ProxysStandAlone, NULL ) ) {
        cooperative_proxies = false;
    }

    if ( home == NULL ) return 0;
        
        // set up the structure
    assert( proxys == NULL );
    proxys = (Proxy_Values *) NestCalloc( sizeof( *proxys ) );
    proxys->self = NestStrdup( selfIpAddress() );
    proxys->home = NestStrdup( home );
    proxys->home_con = NULL;
    proxys->pending = new nestClientPool( false );
    proxys->waiting = new nestClientPool( false );
    proxys->connections = new nestProxyClientCache( );
    proxys->coordinator = NULL;
    proxys->coordinator_map = NULL;
    proxys->coord_status = NEST_NOT_YET_INITIALIZED;
    proxys->index = NULL;
    proxys->index = new DynamicLard( NULL );
    assert( proxys->index != NULL );
    proxys->index->setServerEqualityFunction( checkTwoStrings );
    nest_pthread_mutex_init( &(proxys->mutex), NULL );
    nest_pthread_mutex_init( &(proxys->auth_mutex), NULL );
    nest_debug( NEST_DEBUG_PROXY, "Initing proxy data for proxy of %s.\n",
            proxys->home );
        

        // make an asynchronous task to register with the server
        // we _could_ do it synchronously without deadlock but this
        // would unnecessarily starve clients during any disconnection
    startRegisterHome();
    return 0;

        /*
        // open connection to the home node and register with it
    status = openHomeConnection( proxys->home );
    if ( status == NEST_SUCCESS ) {
        status = registerProxyConnection( proxys->home_con );
    }
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not set up as proxy to %s: %s.\n",
                home, NestErrorString( status ) );
    }

        // open connection to the coordinator 
    status = findCoordinator( proxys->home_con );
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not find the proxy coordinator: %s\n",
                NestErrorString( status ) );
    }
    nest_debug( NEST_DEBUG_PROXY, "Set up as proxy of %s: %s\n",
            home, NestErrorString( status ) );
        */
}


/////////////////////// OLD CODE /////////////////////////////////////

    /*

// find an available connection, (or make one)
// TODO: This is where to put the logic about whether to fetch directly from
// home or from other proxy servers
nestClient *
findAvailableProxyConnection( const char *filepath ) {

    nestClient *cursor = NULL;
    int type = PROXY_CONNECTION;

        // get the name of responsible server
        // then figure out if whether we use that server or route home 
    lockMutex();
    char *server = (char *)proxys->index->getServer( filepath );
    unlockMutex();

    if ( ! cooperative_proxies ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, but not running coop mode,"
                " so will fetch from home node\n", server, filepath );
        server = proxys->home;
        type   = HOME_CONNECTION;
    } else if ( ! strcmp( server, proxys->self ) ) {
        nest_debug( NEST_DEBUG_PROXY,
                "I'm the responsible proxy for %s, will fetch from home\n",
                filepath );
        server = proxys->home;
        type   = HOME_CONNECTION;
    } else {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, fetching accordingly\n" ,
                (char*)server, filepath );
    }

        // see if we have any connections cached to this server
        // if not, make one and cache it for possible future use
    lockMutex();
    cursor = proxys->connections->getCachedProxy( server );
    unlockMutex();
    if ( cursor == NULL ) {
        NestReplyStatus status = openProxyConnection( server, &cursor, type );
        if ( status != NEST_SUCCESS ) {
                // TODO: This can fail and kill the thread but the rest
                // of the ghost is still OK.  But this transfer, and others
                // which may depend on it, will hang indefinitely.
            nest_die( "TODO: need to deal with proxy failure\n" );
        }
        nest_debug( NEST_DEBUG_PROXY, 
                "Created new connection to %s (%d) to fetch %s: %s\n", 
                server, cursor->getId(), filepath, NestErrorString( status ));
        cursor->setName( server );
    }
    nest_debug( NEST_DEBUG_PROXY, 
            "Will use proxy client %s (%d) to fetch %s.\n",
            cursor->getIdentity(), cursor->getId(), filepath );
    return cursor;
}

// TODO:  needs to be async
// first set parameters according to whether we're reading or writing
// then go ahead and open the storage connection
// the logic may seem backward.  For a proxy write, we open a read 
// connection to local storage (and vice versa)
NestReplyStatus
openProxyStorageConnection( 
        nestClient *client,
        nestRequest *proxyreq, 
        nestProtocol **storage)
{
    char *proxyfile = NULL;
    int openflags;
    NestReplyStatus status = NEST_UNKNOWN_ERROR;
    long length = 0;

    if ( proxyreq->isWrite() ) {
        proxyfile = (char *)proxyreq->getNestFile();
        openflags = READ_FLAGS;
        if ( ( status = nestStorageManager::filelength( &length,
                                                    client->getIdentity(),
                                                    client->getCwd() )) 
                != NEST_SUCCESS )
        {
            nest_die( "Tried to flush a dirty write which no longer exists\n");
            return status;
        }
    } else {
            // make a temp name for the transfer
            // this is done so when others may ask for the file, it will not
            // be found and they can be put in the pending list
            // ask for the entire proxy chunk
        proxyfile = getShadowName( proxyreq->getNestFile() );
        openflags = WRITE_FLAGS;
        length    = PROXY_BLOCK_SIZE;
    }

        // go ahead and open the storage connection
    status = nestStorageManager::open(  storage,
                                        client->getIdentity(),
                                        NULL,
                                        proxyfile,
                                        openflags,
                                        0,
                                        length,
                                        client->getCwd() );
    if ( ! proxyreq->isWrite() ) {
        NestFree( proxyfile );
    }

    return status;
}


NestReplyStatus
startProxyWrite(    NestConnection connection,
                    nestRequest *ghostreq, 
                    nestRequest **proxyreq ) 
{
    long length = ghostreq->getLength();
    if ( length == ENTIRE_FILE ) {
        nest_die( "Need to add some code at %s:%s:%d\n",
                __FILE__, __FUNCTION__, __LINE__ );
    }
    NestReplyStatus status = RequestWriteFile( 
                (char *)ghostreq->getNestFile(), 
                ghostreq->getStart(), 
                ghostreq->getLength(),
                connection );
    nest_debug( NEST_DEBUG_PROXY, "Requested permission at home (%s) to "
            "flush dirty write of %s (%ld-%ld): %s\n",
            proxys->home,
            ghostreq->getNestFile(),
            ghostreq->getStart(),
            ghostreq->getLength(),
            NestErrorString( status ) );
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_PROXY, "RequestWriteFile %s failed: %s.\n",
                ghostreq->getNestFile(), NestErrorString( status ) );
        return status;
    }

    *proxyreq = new nestRequest(   PROXY_PUT, 
                                    //CLIENT_SEND_FILE, // sends acks
                                    (char *)ghostreq->getNestFile(),
                                    NULL, NULL, 0, length, NULL );
    return NEST_SUCCESS;
}

// req is the original unmodified client request
// proxyreq is the translated request since proxies store blocks not whole files
// proxyreq is what we create to fetch the block from the home node
NestReplyStatus
startProxyRead(     NestConnection connection,
                    nestRequest *req,   
                    nestRequest *ghostreq, 
                    nestRequest **proxyreq)
{

         // try to move connection into transfer-ready state
        // TODO this is maybe not great since we do it synchronously . . . 
        // also, if we fail to allocate space we have to tear it down

        // need to remember the real file name and the real offset
    long length    = PROXY_BLOCK_SIZE;
    NestReplyStatus status = RequestReadFile( 
            (char*)req->getNestFile(), 
            ghostreq->getChunk(), 
            length,
            connection );
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_PROXY, "RequestReadFile %s failed: %s.\n",
                req->getNestFile(), NestErrorString( status ) );
        return status;
    }
        // what if length has changed since our prediction, we'll need to
        // do something about that here
    if ( PROXY_BLOCK_SIZE != length ) {
        nest_warn( __FILE__, __FUNCTION__,  
        "Predicted length of proxy transfer != real length: %ld != %ld\n",
        (long)PROXY_BLOCK_SIZE, length );
    // close the storage connection, then try to reopen with the correct len
            // the connection must be destroyed because it can't be used
            // again since the home server will be sending bytes down it
            // remove it from the list
        nest_warn( __FILE__, __FUNCTION__, "should remove proxy connection, but not doing so.\n" );
        //assert( removeProxy( cursor ) == NEST_SUCCESS );
    }

        // create a server task and start it up
    *proxyreq = new nestRequest(   PROXY_GET, 
                                    //CLIENT_SEND_FILE, // sends acks
                                    (char *)ghostreq->getNestFile(),
                                    NULL, NULL, 0, length, NULL );
    return status;
}


// TODO:  needs to be async
// this needs to be pushed asynchronously...
// danger is that proxies can get deadlocked waiting on each other.
// which means that even opening proxy connections must be done asynch.
NestReplyStatus
startProxyOp( nestClient *client, nestClient *proxy ) {
    NestReplyStatus status = NEST_UNKNOWN_ERROR;

    nestRequest *req = client->getRequest();
    nestRequest *proxyreq = client->getProxyRequest();
    nestRequest *proxy_req = NULL;
    NestConnection con = proxy->getConnection()->getSocket();

    if ( proxyreq->isWrite() ) {    // proxy writes are flushing to home
        status = startProxyWrite( con, proxyreq, &proxy_req );
    } else {
        status = startProxyRead ( con, req, proxyreq, &proxy_req );
    }

    if ( status == NEST_SUCCESS ) {
            // assign the req to a proxy connection and start it
        nest_debug( NEST_DEBUG_PROXY, "Adding proxy transfer of %s.\n", 
                proxyreq->getNestFile() );
        proxy->assignRequest( proxy_req );
        proxy->setCwd( client->getCwd() );
    } else {
        if ( status == NEST_NO_CONNECTION ) {
            nest_die( "Need code here: %s:%s\n", __FILE__, __FUNCTION__ );
        }
    }

    return status;
}

// this is the function pointer that is passed to the concurrency architecture
NestReplyStatus
asynchronouslyInitiateProxyTransfer( nestClient *client ) {

    NestReplyStatus status;
    nestClient *proxy          = NULL;
    nestRequest *proxyreq      = client->getProxyRequest();

    proxy = findAvailableProxyConnection( proxyreq->getNestFile() );
    proxy->assignStorageConnection( client->getStorageConnection() );

        // notice that startProxyOp assigns a req to the proxy
        // therefore in wasProxyTask, we need to end this req
    client->assignProxyClient( proxy );
    status = startProxyOp( client, proxy );
    nest_debug( NEST_DEBUG_PROXY, "%s for %s: %s\n", 
            __FUNCTION__, client->getIdentity(), NestErrorString( status ) );
    return status;
}

// this is called at taskDone in dispatcher and allows proxy to continue the
// work of any clients that were waiting for the asynchronous initiation
NestReplyStatus
asynchronousInitCleanup( nestClient *client, NestReplyStatus status ) {

    taskFunction *chore = client->getTaskFunction();
    if ( chore == NULL ) {
        return false;
    }

	lockMutex();

    nestClient *proxy = client->getProxyClient();
    assert( proxy != NULL );
    client->assignTaskFunction( NULL, NULL );
    client->assignProxyClient( NULL );

    if (&asynchronouslyInitiateProxyTransfer == chore ) {
        if ( status != NEST_SUCCESS ) {
                // clean everything up that has been opened so far
                // need to remove from pending list
            char *file = NestStrdup( client->getProxyRequest()->getNestFile() );
            checkProxyQueue( proxys->pending, file, status );
            checkProxyQueue( proxys->pending, file, status );
            NestFree( file );
            //proxy->endRequest( status ); proxy req is just a copy of clients
            proxys->connections->insertProxyClient( 
                    (char*)proxy->getIdentity(), proxy ); 
        } else {
            status = nestConcurrencyManager::addTask( proxy ); 
        }
        nest_debug( NEST_DEBUG_PROXY, 
            "Asynch init done: %s, "
            "Starting a proxy operation for client %s (%d) "
            "using proxy client %s (%d)\n",
            NestErrorString( status ),
            client->getIdentity(),
            client->getId(),
            proxy->getIdentity(),
            proxy->getId() );
    } else {
        assert( 0 );
    }

	unlockMutex();
    return status;

 //       // this old code tried to deal with case when a req for a file
 //       // didn't fit in available space in the proxy.
 //   } else {
 //       nest_debug( NEST_DEBUG_PROXY, 
 //               "Unable to open storage connection for %s: %s\n",
 //               proxyreq->getNestFile(), NestErrorString( status ) );
 //           // the connection must be destroyed because it can't be used
 //           // again since the home server will be sending bytes down it
 //           // remove it from the list
 //       nest_warn( __FILE__, __FUNCTION__, "should remove proxy connection, but not doing so.\n" );
 //       //assert( removeProxy( cursor ) == NEST_SUCCESS );
//
 //           // right here check to see why we failed
//            // if we failed bec not enough space due to locked files then
//            // we can either return EINTR to the client or we can queue it
//        if ( status == NEST_INSUFFICIENT_SPACE && ! proxyreq->isWrite() ) {
//
//                // is reason too many locked files or this file is too big?
//            long long int avail, total;
//            nestStorageManager::bytesAvailable( IS_USER, &avail, &total );
//            if ( (long long int) length <= total ) {
//                status = NEST_INTR;
//                proxys->waiting->addClient( client );
//                nest_debug( NEST_DEBUG_PROXY, 
//                    "Not enough free space in cache for %s, will queue req.\n",
//                    proxyreq->getNestFile() );
//                status = NEST_SUCCESS;
//            }
//        }
//    }
}

// this is called also from dispatcher taskDone
// this is called whenever a transfer completes
// if it was a proxy operation, start anyone waiting for it
// and start any one waiting for this proxy transfer
NestReplyStatus
checkProxyTransfers( nestClient *client, NestReplyStatus status ) {

        // make sure we're in proxy mode
    if ( ! serverIsProxy() ) {
        return NEST_SUCCESS;
    }

        // set up some state
    const char *file = client->getLastRequest()->getNestFile();
//    const char *cwd  = client->getCwd();
    int type         = client->getLastRequest()->getType();

    nest_debug( NEST_DEBUG_PROXY, "Checking proxy transfers after complete "
            "transfer of %s for (or by) %s (%d): %s\n",
            file,
            client->getIdentity(),
            client->getId(),
            NestErrorString( status ) );

        // check whether this was a proxy proxy get
    if ( type == PROXY_GET ) {
            
            // should never get here anymore
        assert( 0 );

            // rename the file to its permanent position
        char *proxyfile = getShadowName( file );
        if ( status == NEST_SUCCESS ) {
            NestReplyStatus renameStatus = 
                    nestStorageManager::renameFile( 
                            "proxy", cwd, proxyfile, file );
            nest_debug( NEST_DEBUG_PROXY, "Renamed proxyfile %s to %s: %s.\n",
                    proxyfile, file, NestErrorString( renameStatus ) ); 
            if ( renameStatus != NEST_SUCCESS ) {
                nest_die( "Failed to rename proxyfile: %s\n",
                        NestErrorString( renameStatus ) );
            }
        } else {
                nestStorageManager::deleteFile( "proxy", cwd, proxyfile );
        }
        NestFree( proxyfile );

            // check to see if any of the pending clients were waiting for this
            // there'd better be at least one!
            // pending must be checked before waiting
            //  this ensures at least minimal forward progress
        int started = 0;
        lockMutex();
        started += checkProxyQueue( proxys->pending, file, status );
        started += checkProxyQueue( proxys->waiting, file, status );
        unlockMutex();
        assert( started != 0 );

    } else if ( type == PROXY_PUT ) {
        lockMutex();
        nestClient *real_client = removeWaitingClient( proxys->pending, file );
        unlockMutex();
        assert( real_client != NULL );
        nest_debug( NEST_DEBUG_PROXY, 
                "Found %s waiting for a proxy flush, will keep flushing.\n",
                real_client->getIdentity() );

            // after writing a file, need to receive an ack that server got it
            // TODO: make async?
        if ( status == NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_PROXY, 
                    "Receiving an ack after flushing a write.\n" );
            status = ReceiveNestAck( client->getConnection()->getSocket() );
        }
        if ( status == NEST_SUCCESS ) {
            status = flushDirtyWrites( real_client );
        } else {
            real_client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
        
    } else if ( nestRequest::isWrite( type ) ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "Remembering dirty write of %s for %s\n",
                client->getLastRequest()->getNestFile(),
                client->getIdentity() );
        client->addDirtyWrite( client->getLastRequest() );
    }

        // return this proxy client back into active duty.
    if ( type == PROXY_GET || type == PROXY_PUT ) {
        lockMutex();
        proxys->connections->insertProxyClient( 
                (char*)client->getIdentity(), client );
        unlockMutex();
    }

        // starvation
        // if this next thing is only called following completed proxy
        // transfers instead of after any transfer, then we'll have less 
        // starvation but many more evictions and slower performance

        // see if any waiting proxys are now able to go, this will often
        // be the case if previously they were held due to file locks
        // now that taskDone, there will be one more unlocked file
	lockMutex();
	int waiters = proxys->waiting->size();
	unlockMutex();
    if ( waiters > 0 ) {
        lockMutex();
        nestClient *client = proxys->waiting->getFront();
        unlockMutex();
        status = addProxyTransfer( client );
        nest_debug( NEST_DEBUG_PROXY, 
                    "Started queued proxy request %s for %s: %s.\n",
                    client->getRequest()->getNestFile(),
                    client->getIdentity(),
                    NestErrorString( status ) ); 
        if ( status != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_PROXY, 
                    "Failed to start queued proxy request %s: %s.\n",
                    file, NestErrorString( status ) ); 
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
            client->endRequest( status );
        }
    }

    return status;
}

    // each proxy maintains possibly multiple connections to the home node 
    // when a connection needs to be killed (perhaps bec it is in the middle
    // of partially completed transfer), do so here
NestReplyStatus
removeProxyConnection( nestClient *cursor ) {
    NestReplyStatus openStatus = NEST_SUCCESS; 
    
        // remember whether this was our registered one, then remove
    bool registered = cursor->registered;
    proxys->connections->removeClient( cursor );
    nest_debug( NEST_DEBUG_PROXY, "PID: %d Removing proxy: %d (sd %d).\n", 
            NestGetpid(),
            proxys->connections->size,
            cursor->connection->getConnection()->getSocket() );

        // clear the fd so server won't select on it anymore
        // it got set in taskDone
    NestFd_clear( cursor->connection->getConnection()->getSocket(), readfdset );
    delete( cursor->connection );
    delete( cursor );
            
        // now reregister if necessary
    if ( registered ) {
        nest_debug( NEST_DEBUG_PROXY, 
            "Opening new proxy con to replace old registered one.\n");
        nestClient *new_connection;
        openStatus = openHomeConnection( &new_connection );
        if ( openStatus != NEST_SUCCESS ) {
            nest_die( "Lost connection to home %s.\n", proxys->home );
        } else {
            registerProxyConnection( new_connection, false );
        }
    }
    return openStatus;
}

    // connect to a monitor process 
    // the monitor tracks load at the various proxys and is responsible
    // for the ratio of proxy to compute nodes
    // the monitor also pushes updated mappings to the compute nodes
    //
    // this function is deprecated in favor of self-organizing, 
    // self-coordination by the proxies
NestReplyStatus
connectMonitor( char * monitor ) {
    // this function is deprecated in favor of self-organizing, 
    // self-coordination by the proxies
    NestConnection con;
    NestReplyStatus status =NestOpenAnonymously(con,monitor,PROXY_MONITOR_PORT);
    if ( status == NEST_SUCCESS ) {
        nestClient *mon = new nestClient( new nestProtocolNest(con), NULL );
        client_pool->addClient( mon );
        NestFd_set( con, readfdset );
    } else {
        nest_warn( __FILE__, __FUNCTION__, "Unable to open connection to proxy monitor at %s.\n",
                monitor );
        status = NEST_SUCCESS;
    }
    
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not connect to monitor at %s: %s.\n",
                monitor, NestErrorString( status ) );
    }
    return status;
}
    */
