/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_common.h"
#include "nest_proxy.h"

#include "nest_concurrency_manager.h"
#include "nest_collection_manager.h"
#include "nest_storage_manager.h"
#include "nest_client_pool.h"
#include "nest_protocol_manager.h"
#include "nest_advertiser.h"
#include "nest_protocol_advertise.h"
#include "nest_config.h"
#include "nest_hash_table.h"
#include "nest_filecache_model.h"

#include <sys/utsname.h>

    // time between any regular maintenance (in millisecs)
#define MAINTENANCE_INTERVAL 100000

    // handlers for the various services provided to clients
nestRequest* ReceiveRequest( nestClient *client, fd_set *ready );
int  HandleRequests( nestClientPool *requests, fd_set *readfds );
int  HandleRequest( nestClient *client );
NestReplyStatus NewClient( nestClient *newclient );
bool RemoveClient ( nestClient *client, fd_set *reads );
int  SynchronousRequestHandler( nestClient *client );
NestReplyStatus SendNestAd( nestClient *client );
NestReplyStatus openStorageConnections( nestProtocol **storage,
                                        nestProtocol **otherStorage,
                                        nestClient   *client);
NestReplyStatus SendDirectoryList( nestClient *client );
NestReplyStatus SendPwd( nestClient *client );
NestReplyStatus MakeLink( nestClient *client );
NestReplyStatus CollectionOp( nestClient *client, int type );
NestReplyStatus DirOp( nestClient *client, int type );
NestReplyStatus SendVersion( nestClient *client );
int             SendBandwidthStatistics( nestClient *client );
int             TransferFile( nestClient *client, bool already_trans = false );
NestReplyStatus CopyFile ( nestClient *client );
NestReplyStatus DeleteFile ( nestClient *client );
NestReplyStatus RenameFile ( nestClient *client );
NestReplyStatus SendFileinfo( nestClient *client );
NestReplyStatus SetFileAttr( nestClient *client );
NestReplyStatus SendFilesize( nestClient *client );
int             NestSync( nestClient *client );
NestReplyStatus QueryProxys ( nestClient *client );
NestReplyStatus RegisterProxyClient ( nestClient *client );
NestReplyStatus RemoveProxyClient ( nestClient *client );
NestReplyStatus FormatDisk ( nestClient *client );
NestReplyStatus AuthenticateUser( nestClient *client );
NestReplyStatus AddUser( nestClient *client );
NestReplyStatus AddGroup( nestClient *client );
NestReplyStatus DeleteUser( nestClient *client );
NestReplyStatus DeleteGroup( nestClient *client );
NestReplyStatus AddUserToGroup( nestClient *client );
NestReplyStatus DeleteUserFromGroup( nestClient *client );
NestReplyStatus ReserveUserSpace( nestClient *client );
NestReplyStatus ReserveGroupSpace( nestClient *client );
NestReplyStatus ListUserLots( nestClient *client );
NestReplyStatus ListGroupLots( nestClient *client );
NestReplyStatus GetLotStat( nestClient *client );
NestReplyStatus UpdateLot( nestClient *client );
NestReplyStatus ChangeLotType( nestClient *client );
NestReplyStatus MoveFileLot( nestClient *client );
NestReplyStatus AddUserToLot( nestClient *client );
NestReplyStatus DeleteUserFromLot( nestClient *client );
NestReplyStatus AttachLotToFile( nestClient *client );
NestReplyStatus DetachLotFromFile( nestClient *client );
NestReplyStatus SendPhysName( nestClient *client );
NestReplyStatus SendUserAcl( nestClient *client );
NestReplyStatus SetUserAcl( nestClient *client );
NestReplyStatus TerminateLot( nestClient *client );
NestReplyStatus DebugRequest( nestClient *client );
NestReplyStatus TerminateServer( nestClient *client );

void taskDone( nestClient *client, NestReplyStatus status );


NestReplyStatus GetFilehandle( nestClient *client );
NestReplyStatus SendAvailableSpace( nestClient *client );

    // other misc helper functions
NestReplyStatus RegularMaintenance( );
struct timeval* setMaintenanceTimeout( struct timeval *timeout );
int  getMaxfd( );
void installSignalHandlers( );
void sigChild( int signal, siginfo_t *info, void *context );
int  main_init( );
NestReplyStatus WriteAd( );
void debug_init( bool Daemonize );
void daemon_init( const char *pname, int facility );

    // timer to do regular maintenance
struct timeval last_maintenance;
int maintenance_interval = MAINTENANCE_INTERVAL;

    // client connection states
enum {  CLOSE_CONNECTION,       // client is no longer in the building
        ACTIVE_CONNECTION,      // client is being handled by concurrency mgr
        PASSIVE_CONNECTION      // client should be monitored by server
};

const char *config_file = NULL;

    // for a main nest to know which proxys there are
    // this hash table is an stl list for each client registering proxies
struct nest_hash_table *registered_proxys = NULL;
char *home_server = NULL;   // this is flag set to run as proxy server

    // acceptable authentication models
int allowedAuthentication = NEST_AUTH_GSS;

    // collection of currently connected clients
nestClientPool *clients = NULL;

    // collection of server initiated tasks
nestClientPool *serverTasks = NULL;

    // the pid of the main server thread
    // used to make sure that only main server thread performs protected duties
extern int admin_pid;

    // used to make a nest appear remote
unsigned long latency = 0;

    // needs to be global now for taskDone
fd_set writefds, readfds, r_readyfds, w_readyfds;

int
main( int argc, char *argv[] ) {

        // set up config file must occur before initing other stuff
    ReadConfigFile(argc, argv);
    main_init( );

        // enter the main server loop after doing first RegularMaintenance
        // first handle any messages from our dispatched worker processes
        // then check for any new incoming client requests
    RegularMaintenance();
    nestClientPool *requests = new nestClientPool();
    nest_printf( "Initializing done.  Ready to service requests.\n" );
    while( true ) {

            // merge rpc readlist with our own, then enter select loop
        r_readyfds = svc_fdset;
        NestMergeFdsets( &r_readyfds, &readfds, getMaxfd() + 1 );
        nest_debug( NEST_DEBUG_FDSETS, "NestRead set ( %x ): ", &readfds );
        NestFd_print( NEST_DEBUG_FDSETS, &readfds, getMaxfd() + 1 );
        w_readyfds = writefds;
        nest_debug( NEST_DEBUG_FDSETS,
                "maxfd is %d, entering select\n", (getMaxfd()+1) );
        int readies = NestSelect(getMaxfd()+1, &r_readyfds,&w_readyfds, NULL,
                            NestNextTimeout( maintenance_interval, 
                                &last_maintenance ) );
        if ( readies <= 0 ) {
            if ( readies == 0 ) {
                    // TODO need slow timeouts for regular maintenance
                    // and fast timeouts for scheduling
                RegularMaintenance();
            } else nest_warn( __FILE__, __FUNCTION__, 
                    "select failed: %s\n", strerror( errno ) );
            continue;
        }
            // appear remote if specified
            // half of latency should be used because clients will be expected
            // to arrive during some previous sleep?
        if ( latency != 0 ) {
            NestNanosleep( latency * 1000 * 1000 );
        }
        nest_debug( NEST_DEBUG_FDSETS, "Woke from select with %d readies.\n",
                readies );

            // then accept any new clients - listen on each protocol socket
        for( int protocol_socket  = nestProtocolManager::firstProtocolSocket();
             protocol_socket != nestProtocolManager::lastProtocolSocket();
             protocol_socket  = nestProtocolManager::nextProtocolSocket() )
        {
                // ignore any sockets that were unable to listen correctly
            if ( protocol_socket <= 0 ) continue;

                // some protocols may have multiple clients to give . . .
                // some protocols may have none and just use the socket as
                // a way of waking up the main thread (i.e. gftp)
            nestClient *newclient;
            int accepts = 0;
            while ( ( newclient =
                    nestProtocolManager::accept(
                        protocol_socket, &r_readyfds, accepts ))
                != NULL )
            {

                accepts++;

                // clear the socket since ftp and gftp share one
                // I don't really understand this anymore . . .
                //NestFd_clear( protocol_socket, &r_readyfds );

                    // add new connection state to appropriate structures
                NewClient( newclient );

            }
        }

            // then queue all requests from existing clients
        nestClient *client;
        nestRequest *req;
        CLIENT_ITERATOR itr = clients->begin();
        int size;
        if ( ( size = requests->clear() ) != 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Requests list was not previously cleared. Size %d.\n",
                    size );
        }
        while( itr != clients->end( ) ) {
            client = *itr++;
            nest_debug( NEST_DEBUG_CLIENT, "Checking client %d (%s)\n",
                    client->getId(), client->getIdentity() );

                // pass the r_readyfds to each client, if they are ready
                // they can create a request object
            req = ReceiveRequest( client, &r_readyfds );
            if ( req != NULL ) {
                requests->addClient( client );
                nest_debug( NEST_DEBUG_CLIENT, "Client %s (%d) request: %s\n",
                        client->getIdentity(),
                        client->getId(),
                        req->toString() );
            }
        } 

            // now iterate through new requests and handle them here
        HandleRequests( requests, &readfds );

            // work on any concurrent transfers last
            // this is important to be last to allow anticipatory scheduling
        nestConcurrencyManager::makeProgress( );

    } // end of main service loop

        // we should never get here
    nest_die( "Big trouble in Little China.\n" );
    return ERROR_EXIT;
}

NestReplyStatus
NewClient( nestClient *newclient ) {
    int size = clients->addClient( newclient );
    nest_debug( NEST_DEBUG_CLIENT,
            "added a new client, client count is %d\n", size );
    if ( newclient->hasSocket() ) {
        nestProtocol *connection = newclient->getConnection();
            // may be an optimization to add to ready_fds
        NestFd_set( connection->getSocket(),&readfds);
            // I don't think this is an optimization
        //NestFd_set( connection->getSocket(), &r_readyfds );
    }
    return NEST_SUCCESS;
}

void
debug_init( bool Daemonize ) {

    const char *log_file = NULL;

    if ( Daemonize ) {
        nest_debug_init( NULL, true );
    } else {
        nestProtocol *logfile = NULL;
        if( GetConfigValue( Config_LogFile, &log_file ) ) {
            assert( log_file != NULL );
            nest_printf( "Logging to %s.\n", log_file );
            logfile = nestStorageManager::openMetaFile( (char *)log_file, 
                                                                APPEND_FLAGS );
            assert( logfile != NULL );
        }
        if ( GetConfigValue( Config_DebugStderr, NULL ) ) {
            if ( logfile ) {
                nest_printf( "Logging to stderr and ignoring conflicting "
                        "directive to log to %s.\n", log_file );
            }
            logfile = NULL;
        }
        nest_debug_init( logfile, false );
    }
}

#define MAXFD 64

void
daemon_init( const char *pname, int facility ) {
    pid_t pid = 0;

    if ((  pid = fork() ) < 0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // parent
        exit( 0 );
    }

        // first child continues
    setsid();

    NestSignal( SIGHUP, SIG_IGN ); // ignore sig_pipes

        // forking twice is not working
    if ( ( pid = fork() ) <  0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // first child terminates
        exit( 0 );
    }

    umask( 0 );     // clear our file mode creation mask . . . ?

        // should we attempt to close all open descriptors . . . ?
    fprintf( stderr, "NeST has been daemonized.\n" );
    for( int i = 3; i < MAXFD; i++ ) {
        close( i );
    }
        // should we open /dev/null for reading and writing and dup it 
        // to stdin and stdout
    int devnull = NestOpen( "/dev/null", O_RDWR ); 
    NestDup2( devnull, STDIN_FILENO );
    NestDup2( devnull, STDOUT_FILENO );
        
    openlog( pname, LOG_PID | LOG_CONS, facility );
}

void 
taskDone( nestClient *client, NestReplyStatus status ) {

        // special case for transfers that put data but don't prespecify 
        // a transfer length.  we can only allocate lot space for these
        // transfers incrementally.
    if ( status == NEST_LOT_NOT_ENOUGH ) {
        status = nestStorageManager::increaseAllocation( client );
        if ( status == NEST_SUCCESS ) {
            status = nestConcurrencyManager::addTask( client );
            return;
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    }


        // sometimes a proxy may add other special non-transfer requests
        // to the concurrency manager.  Allow them to clear up after them.
    if ( client->getTaskFunction() != NULL ) {
        taskCleanupFunction *cleanup = client->getTaskCleanupFunction();
        assert( cleanup != NULL );
        //NestReplyStatus cleanup_status = 
        cleanup( client, status );
        return;
    }

        // this persistent thing is for a weird case with http clients and
        // the pot concurrency.  It was possible for the taskDone to get
        // called after the client was removed.
        //
        // also however nfs clients are not persistent . . . 
        //
        // I think there still may be a race condition if the client object
        // is deleted while we're operating in taskDone
        // I think it is important to not call client->endTransfer until
        // ___after__ we've returned control to the server
        // because once client->endTransfer is called it is possible that
        // the client could be deleted.  So we can't reference it again.
        // 
    if ( client->hasSocket() && client->persistent() ){
        NestFd_set( client->getConnection()->getSocket(), &readfds );
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_FDSETS, 
                "Task_done adding %d to readfds %x\n", 
                client->getConnection()->getSocket(), &readfds );
    }



        // close the storage connections
    if ( client->getStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getStorageConnection();
        nestStorageManager::close( storage_con, client );
        delete( storage_con );
        client->assignStorageConnection( NULL ); 
    }

    if ( client->getOtherStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getOtherStorageConnection();
        nestStorageManager::close( storage_con ,client );
        delete( storage_con );
        client->assignOtherStorageConnection( NULL ); 
    }

        // TODO:  Be nice to maybe cache these.
    if ( client->getThirdPartyConnection() != NULL ) {
        nestProtocol *third = client->getThirdPartyConnection();
        delete( third );
        client->assignThirdPartyConnection( NULL );
    }
        // this seems weird to me (jb), why do we handle this one particular
        // error specially?
    if ( status == NEST_INSUFFICIENT_SPACE ) { 
        client->sendErrorMessage( __FILE__, __FUNCTION__, status);
    }
        // client endTransfer must be last thing to happen
        // otherwise, client might suddenly disappear
    client->endTransfer( status );

        // if we're running as proxy, start any who were waiting for this
        // and clean this one up
        // no longer needed, bec proxy taskDone is called directly 
        // above using the cleanup function pointer
    /*
    if ( serverIsProxy() ) {
        checkProxyTransfers( client, status ); 
    }
    */

}

int
HandleRequests( nestClientPool *requests, fd_set *readfds ) {
    nestClient *client;
    CLIENT_ITERATOR itr = requests->begin();
    while ( itr != requests->end() ) {
        client = *itr;
        itr = requests->erase( itr );

        switch( HandleRequest( client ) ) {
        case CLOSE_CONNECTION:
            RemoveClient( client, readfds );
            break;
        case PASSIVE_CONNECTION:
        case ACTIVE_CONNECTION:
                // nothing to do, we used to clear the client fd from the
                // read_fd set we pass to select here for an active
                // connection.  But now this is done in the concurrency
                // manager.
            break;
        default:
            assert( 0 );
            break;
        }
    } // end of client service loop
    return 0;
}

bool
RemoveClient( nestClient *client, fd_set *readfds ) {
    nest_debug( NEST_DEBUG_CLIENT, "Removing client %d\n", client->getId() );

        // if this client has a socket, stop listening to it 
    if ( client->hasSocket() ) {
        NestFd_clear( client->getConnection()->getSocket(), readfds);
    }

        // if this client registered as a proxy, remove them from proxy list
    if ( client->isProxy( ) ) {
        RemoveProxyClient( client );
    }

    if ( serverIsProxy() ) {
        removeClientFromProxy( client );
    }

    clients->removeClient( client );
    delete client;
    return true;
}

void
setNewGroup() {
        // set ourselves up as a new group
    // doesn't work   arghh!
    //NestSetpgrp();
    /*
    if ( setpgid( 0, 37 ) ) {
        perror( "setpgid" );
    }
    if ( setsid() < 0 ) {
        perror( "setsid" );
    }
    */
}

int  
main_init( ) {

    bool Daemonize = false;
    const char *value = NULL;

    setNewGroup();

        // set up our fd sets
    FD_ZERO( &writefds );
    FD_ZERO( &readfds  );

        // turn off all debugging unless instructed otherwise
    set_nest_debug_level( NEST_DEBUG_NEVER );

        // instantiate our client pool and our serverTasks
        // these need to be done before args, bec if we're a proxy,
        // we're gonna add the proxy monitor as a client
    clients     = new nestClientPool();
    serverTasks = new nestClientPool();


        // read config values
    if (GetConfigValue( Config_Debug, &value)) {
        if ( add_nest_debug_level(value) != NEST_SUCCESS ) {
            nest_die( "Unknown debug value specified: %s\n", value );
        }
    }

    if ( GetConfigValue( Config_PrintHelp, &value)) {
        PrintHelp(value);
        exit( 0 );
    }

    if ( GetConfigValue( Config_DumpConfigFile, NULL)) {
        DumpSampleConfigFile();
        exit( 0 );
    }

    if ( GetConfigValue( Config_Latency, &value ) ) {
        latency = atol( value );
        nest_printf( "Latency set at %ld ms.\n", latency );
    }

    if ( GetConfigValue(Config_MaintenanceInterval, &value ) ) {
        maintenance_interval = atoi(value);
    }

    if ( GetConfigValue( Config_ProcessUser, &value ) ) {
        if ( NestChangeUser( value ) < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to switch to user %s\n", value );
        }
    }

    if ( GetConfigValue( Config_Daemonize, NULL ) ) {
        Daemonize = true;
        daemon_init( "NeST", LOG_USER );
        //debug_init( true );
    }
    
    if ( GetConfigValue( Config_DisableAuthentication, NULL ) ) {
        allowedAuthentication = NEST_AUTH_NONE;
        nest_warn( __FILE__, __FUNCTION__, 
                "Running with authentication disabled is strongly "
                "discouraged!!\n" );
    }


    initCommonMutex();

        // set admin pid value, must be done after daemonizing
    admin_pid = NestGetpid();
    nest_debug( NEST_DEBUG_ADMIN, "Admin pid is %d\n", admin_pid );

        // debug_init must be called after storage and before concurrency
        //   unless Daemonize is true
        // protocolManager must proceed storageManager
    nestStorageManager        ::init( );
    /*if ( !Daemonize )*/ debug_init( false );
    nestProtocolManager     ::init( );
    nestAdvertiser          ::init( );
    nestConcurrencyManager  ::init( &readfds, 
                                    &writefds, 
                                    &r_readyfds, 
                                    &w_readyfds,
                                    taskDone );
    initProxy( TransferFile, clients, &readfds ); 

        // install signal handlers
    installSignalHandlers( );

        // add each available socket to our read set
        // the protocol manager has listened on one socket for each protocol
    for( int protocol_socket = nestProtocolManager::firstProtocolSocket();
         protocol_socket != nestProtocolManager::lastProtocolSocket();
         protocol_socket = nestProtocolManager::nextProtocolSocket() )
    {
        if ( protocol_socket > 0 ) {
            NestFd_set( protocol_socket, &readfds );
        }
    }

        // ask the ad manager to write the initial ad
    WriteAd();


    return 0;
}

// the max of the clients, the concurrencies and the protocols
// now that we add rpc fds to our select loop, I'm not sure how to set
// the maxfd . . . 
int
getMaxfd( ) {
    int maxfd = 
        max( nestConcurrencyManager::getMaxfd( ), clients->getMaxfd());
    maxfd = max( maxfd, nestProtocolManager::maxSocket( ) );
    return maxfd;
}


NestReplyStatus
advertise( ) {
    NestReplyStatus status = NEST_SUCCESS;


        // update ad
    WriteAd();

    /*
    char *host = "coral.cs.wisc.edu";
    int  port  = 9618;
    long len;
    nestProtocol *adfile = nestStorageManager::openMetaFile(AD_NAME,READ_FLAGS);
    len = nestStorageManager::metaFilelength( AD_NAME );
    nestProtocolAdvertise *collector = NULL;
    
    if ( collector == NULL ) {
        collector = new nestProtocolAdvertise( );
        if ( ! collector->connect( host, port ) ) {
            delete( collector );
            collector = NULL;
            nest_warn( __FILE__, __FUNCTION__, "Unable to advertise to %s:%d\n", host, port );
            delete( adfile );
            return NEST_NO_CONNECTION;
        }
    }

        // create a fake client and assign it this transfer
    nestClient *client = new nestClient( collector, NULL );
    nestRequest *req   = new nestRequest( CLIENT_RECV_FILE, AD_NAME, NULL, NULL,
            0, len, NULL );
    client->assignRequest( req );
    client->assignStorageConnection( adfile );

    serverTasks->addClient( client );
    status = nestConcurrencyManager::addTask( client );
    nest_debug( NEST_DEBUG_SERVER,
            "Added advertise task to concurrencyMgr: %s\n", 
            NestErrorString( status ) );
    delete( adfile );
    return status;
    */
            
    const char *condor_advertise;
    char command[NEST_MESSAGESIZE];
    int retval;

        // condor_advertise
    if ( GetConfigValue( Config_CondorAdvertise, &condor_advertise ) ) {
        snprintf( command, NEST_MESSAGESIZE, "%s UPDATE_STORAGE_AD %s",
            condor_advertise, nestStorageManager::metaFilepath( AD_NAME ) );
        retval = system( command );
        nest_debug( NEST_DEBUG_SERVER, "Exec'd system %s: %d\n",
                command, retval );
        if ( retval < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to exec command \"%s\"\n", command );
        }
    }

    return status;
}


NestReplyStatus 
RegularMaintenance( ) {
    NestReplyStatus status = NEST_SUCCESS;

    nest_debug( NEST_DEBUG_SERVER, "Performing regular maintenance now\n" );
    NestGettimeofday( &last_maintenance );

        // only do concurrency::makeProgress for now
        // TODO need slow timeouts for regular maintenance
        // and fast timeouts for scheduling
        // turn off anticipatory scheduling for now

        // for a while with the proxies, this was needed bec the pot pipe
        // was somehow maybe not waking up the server
        // but I think that was a race condition otherwise solved.
    //nestConcurrencyManager::makeProgress();

    advertise();

    nestStorageManager::maintenance();

    proxyMaintenance();

    return status;
}

/*
void
sigInt( int signal ) {
    nest_printf( "Exiting due to SIGINT.\n" );
    _exit( 0 );
}
*/

// ignore sigpipes
// pass sigchilds to the concurrency manager
void
installSignalHandlers( ) {

    NestSignal( SIGPIPE, SIG_IGN ); // ignore sig_pipes
 //   NestSignal( SIGINT,  sigInt );

    struct sigaction sa;
    sa.sa_sigaction = sigChild;
    sigemptyset( &sa.sa_mask );
    sa.sa_flags = SA_SIGINFO;

    if ( sigaction( SIGCHLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
    if ( sigaction( SIGCLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
}

void
sigChild( int signal, siginfo_t *info, void *context ) {
    nestConcurrencyManager::sigChild( signal, info, context );    
}

nestRequest *
ReceiveRequest( nestClient *client, fd_set *ready ) {
    struct timeval start, end;
    nestRequest *req;

    NestGettimeofday( &start );
    req = client->receiveRequest( ready );
    NestGettimeofday( &end );
    if ( NestTimediff( &start, &end, __FUNCTION__ ) > 1000000 ) {
        nest_warn( __FILE__, __FUNCTION__, "Receive request took %.3f secs.\n", 
                (float)NestTimediff( &start, &end, __FUNCTION__ ) / 1000000 );
    }
    return req;
}

int
HandleRequest( nestClient *client ) {
    switch( client->getRequest()->getType() ) {
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
    case THIRD_PARTY_PUT:
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
        return TransferFile( client );
    case COPY_REQUEST:
        return CopyFile(client);
    case CLOSE_CONNECTION_REQUEST:
        client->endRequest( NEST_SUCCESS );
        return CLOSE_CONNECTION;
    case NEST_SYNC:
        return NestSync( client );
    default:
        return SynchronousRequestHandler( client );
    }

}

// these quick simple requests are just handled immediately by the main server
int
SynchronousRequestHandler( nestClient *client ) {

    int connection_state = PASSIVE_CONNECTION;
    nest_debug( NEST_DEBUG_SERVER, "SynchronousRequestHandler for req %d\n",
            client->getRequest()->getType() );
    NestReplyStatus status = NEST_SUCCESS;

        // respond appropriately to each request type
    switch( client->getRequest()->getType() ) {
    case PROXY_REGISTER:
        status = RegisterProxyClient( client );
        break;
    case PROXY_QUERY:
        status = QueryProxys( client );
        break;
    case PROXY_FLUSH:
        status = FlushProxyCache( client );
        break;
    case PROXY_INDEX_PUSH:
        status = ReceiveProxyIndex( client );
        break;
    case FORMAT_DISK_REQUEST:
        status = FormatDisk( client );
        break;
    case BAD_REQUEST:
        status = client->sendErrorMessage( __FILE__, __FUNCTION__,  
                client->getRequest()->getErrorCode() );
        break;
    case VERSION_REQUEST:
        status = SendVersion( client );
        break;
    case REQUEST_NEST_AD:
        status = SendNestAd( client );
        break;
    case GET_FILEHANDLE:
        status = GetFilehandle( client );
        break;
    case AVAILABLE_SPACE_QUERY:
        status = SendAvailableSpace( client );
        break;
    case TIME_REQUEST:
        status = client->sendTimestamp( );
        break;
    case CONNECTION_REQUEST:
    case CHANGE_USER_REQUEST:
        status = AuthenticateUser( client );
        break;
    case ADD_NEST_USER:
        status = AddUser( client );
        break;
    case ADD_NEST_GROUP:
        status = AddGroup( client );
        break;
    case DELETE_NEST_USER:
        status = DeleteUser( client );
        break;
    case DELETE_NEST_GROUP:
        status = DeleteGroup( client );
        break;
    case ADD_USER_TO_GROUP:
        status = AddUserToGroup( client );
        break;
    case DELETE_USER_FROM_GROUP:
        status = DeleteUserFromGroup( client );
        break;
    case REQUEST_USER_LOT:
        status = ReserveUserSpace( client );
        break;
    case REQUEST_GROUP_LOT:
        status = ReserveGroupSpace( client );
        break;
    case LIST_USER_LOTS:
        status = ListUserLots( client );
        break;
    case LIST_GROUP_LOTS:
        status = ListGroupLots( client );
        break;
    case GET_LOT_STAT:
        status = GetLotStat( client );
        break;
    case UPDATE_LOT:
        status = UpdateLot( client );
        break;
    case CHANGE_LOT_T:
        status = ChangeLotType( client );
        break;
    case LOT_MOVE_FILE:
        status = MoveFileLot( client );
        break;
    case ADD_USER_TO_LOT:
        status = AddUserToLot( client );
        break;
    case DELETE_USER_FROM_LOT:
        status = DeleteUserFromLot( client );
        break;
    case ATTACH_LOT_TO_FILE:
        status = AttachLotToFile( client );
        break;
    case DETACH_LOT_FROM_FILE:
        status = DetachLotFromFile( client );
        break;
    case TERMINATE_LOT:
        status = TerminateLot( client );
        break;
    case FILE_DELETE_REQUEST:
        status = DeleteFile( client );
        break;
    case FILE_RENAME_REQUEST:
        status = RenameFile( client );
        break;
    case BANDWIDTH_STATISTICS_REQUEST:
        status = SendBandwidthStatistics( client );
        break;
    case DIR_LIST_REQUEST:
        status = SendDirectoryList( client );
        break;
    case FILEINFO_REQUEST:
        status = SendFileinfo( client );
        break;
    case SETATTR_REQUEST:
        status = SetFileAttr( client );
        break;
    case FILESIZE_REQUEST:
        status = SendFilesize( client );
        break;
    /** This has been replaced by the copy function that uses concurrency
        And so it appears along with the other transfer functions
    case COPY_REQUEST:
        status = CopyFile( client );
        break;
    */
    case LINK_REQUEST:
        status = MakeLink( client );
        break;
    case PWD_REQUEST:
        status = SendPwd( client );
        break;
    case MKDIR_REQUEST:
    case CWD_REQUEST:
    case RMDIR_REQUEST:
        status = DirOp( client, client->getRequest()->getType() );
        break;
    case PHYSNAME_REQUEST:
        nest_debug( NEST_DEBUG_SERVER, "PhysName request\n" );
        status = SendPhysName( client );
        break;
    case GET_ACL:
    case GET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "GetUserAcl request\n" );
        status = SendUserAcl( client );
        break;
    case SET_ACL:
    case SET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "SetUserAcl request\n" );
        status = SetUserAcl( client );
        break;
    case NAME_COLLECTION:
    case UNNAME_COLLECTION:
    case QUERY_COLLECTION:
	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
		status = CollectionOp( client, client->getRequest()->getType() );
        WriteAd();
        RegularMaintenance();
        break;
    case SERVER_TERMINATE_REQUEST:
        if ( TerminateServer( client ) == NEST_SUCCESS ) {
            exit( 0 );
        }
        break;
    case DEBUG_REQUEST:
        status = DebugRequest( client );
        break;
    default:
        assert( 0 );
        return CLOSE_CONNECTION;
    }

    client->endRequest( status );

    if ( connection_state == PASSIVE_CONNECTION ) {
            // this means the server is responsible for listening to the client
        if ( ! client->persistent() ) {
                // client is a one time visitor (e.g. nfs or http)
            connection_state = CLOSE_CONNECTION;
        }
    }
    return connection_state;
}

NestReplyStatus
TerminateServer( nestClient *client ) {

    NestReplyStatus status = NEST_NONPRIVELEDGED_USER;
    const char *tmp_admin, *admin;

        // check whether user is the registered admin user
    if ( ! GetConfigValue( Config_AdminUser, &tmp_admin ) ) {
        status = NEST_NONPRIVELEDGED_USER;
    } else {
        admin = NestGetModifiedIdentity(tmp_admin);
        if ( !strcmp( client->getIdentity(), admin ) ) {
            status = NEST_SUCCESS;
        } else {
            status = NEST_NONPRIVELEDGED_USER;
        }
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
DebugRequest( nestClient *client ) {
    NestReplyStatus status
        = add_nest_debug_level( client->getRequest()->getDebugValue() );
    if ( status == NEST_SUCCESS ) {
        nestConcurrencyManager::debugLevel( get_nest_debug_level() );
        nest_debug( NEST_DEBUG_DEBUG, "nest_debug set to %s\n",
            client->getRequest()->getDebugValue() );
        client->sendAck();
    } else {
        nest_debug( NEST_DEBUG_DEBUG, "could not set nest_debug set %s\n",
            client->getRequest()->getDebugValue() );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
WriteAd( ) {
    NestReplyStatus status;

    nestProtocol *ad = nestStorageManager::openMetaFile
        ( AD_NAME, WRITE_FLAGS | O_TRUNC );

    if ( ad == NULL ) {
        nest_die( "Unable to open metafile %s\n", AD_NAME );
        status = NEST_NONPRIVELEDGED_SERVER;
    } else {
//        status = ad->Sendn( "[\n", 2 ); this is new classads
        status = nestAdvertiser::writeAd( ad );

            // now write some server stuff
        char buffer[NEST_MESSAGESIZE];
        snprintf( buffer, NEST_MESSAGESIZE, 
                "Current_clients   = %d\n"
                "Maintenance_interval = %.2f\n",
                ( ( clients != NULL ) ? clients->size() : 0 ),
                (float)maintenance_interval / THOUSAND );

        if ( serverIsProxy() ) {
            writeProxyAd( ad );
        }
            //this is new classads
//        snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "]\n" ); 
        snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "\n" ); 
        
        ad->Sendn( buffer, NestStrlen( buffer ) );
        delete( ad );
    }

    return status;
}

NestReplyStatus 
GetFilehandle( nestClient *client ) {
    NestFileInfo info;

    NestReplyStatus status = nestStorageManager::fileinfo( 
            &info, 
            client->getRequest()->getNestFile(),
            client->getIdentity(),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendFilehandle( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SendFileinfo( nestClient *client ) {

    NestFileInfo info;

    NestReplyStatus status = nestStorageManager::fileinfo( 
            &info, 
            client->getRequest()->getNestFile(),
            client->getIdentity(),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendFileinfo( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SetFileAttr( nestClient *client ) {
    NestReplyStatus     retval;
    char                *phys_name;

    retval = nestStorageManager::get_physical_filename(
        &phys_name,
        client->getRequest()->getNestFile(),    
        client->getIdentity(),
        client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
   
    NestFileInfo info; 
    retval = nestStorageManager::fileinfo( 
            &info, 
            client->getRequest()->getNestFile(),
            client->getIdentity(),
            client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
      
    client->setFileattr(&info, 
                        phys_name,
                        client->getRequest()->getFileattr() );
    NestFree( phys_name );

    return retval;
}

        
NestReplyStatus
SendFilesize( nestClient *client ) {

    long filesize;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    if ( serverIsProxy() ) {
        status = proxyFilelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getIdentity(),
            client->getCwd() );
    } else {
        status = nestStorageManager::filelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getIdentity(),
            client->getCwd() );

    }

    if ( status == NEST_SUCCESS ) {
        client->sendFilesize( filesize );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SendPhysName( nestClient *client ) {

    NestReplyStatus        retval;
    char                *phys_name;

    retval = nestStorageManager::get_physical_filename(
        &phys_name,
        client->getRequest()->getNestFile(),    
        client->getIdentity(),
        client->getCwd() );

    if ( retval == NEST_SUCCESS ) {
        client->sendPhysName( phys_name );
        NestFree( phys_name );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }

    return retval;
}

NestReplyStatus
SendUserAcl( nestClient *client ) {
    NestReplyStatus     retval;
    const char          *target_user;       // Target user / group name
    bool                is_group;

    // Target user specified?
    target_user = client->getRequest()->getTargetUser();
    is_group = client->getRequest()->getAclIsGroup();
    nest_debug( NEST_DEBUG_SERVER,
                "SendUserAcl: User=%p='%s' Group=%c, File='%s'\n",
                target_user,
                target_user,
                is_group ? 't' : 'f',
                client->getRequest()->getNestFile() );

    // Check for all users?
    if (  ( target_user == NULL ) || ( *target_user == '\0' )  ) {
        NestACLList    *Acl;

        retval = nestStorageManager::get_all_acl(
                    &Acl,
                    client->getRequest()->getNestFile(),
                    client->getIdentity(),
                    client->getCwd() );
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[ACL]: retval=%d, Acl @@ %p\n", retval, Acl );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( Acl );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    } else {
        // Just a single (specified) user -or- group
        nest_access_rights_t    access;
        if ( ! is_group ) {
            retval = nestStorageManager::get_user_acl(
                &access,
                target_user,
                client->getRequest()->getNestFile(),
                client->getIdentity(),
                client->getCwd() );
        } else {
            retval = nestStorageManager::get_group_acl(
                &access,
                target_user,
                client->getRequest()->getNestFile(),
                client->getIdentity(),
                client->getCwd() );
        }
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[User='%s']: Access=0x%02x\n", 
                    target_user, access );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( target_user, is_group ? 0 : 1, access );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    }

    return retval;
}

NestReplyStatus
SetUserAcl( nestClient *client ) {
    NestReplyStatus         status;
    const char              *target_user;
    nest_access_rights_t    rights;
    bool                    is_group;

    // Extract the file name
    string  filename = client->getRequest()->getNestFile();

    // Get request parameters
    target_user = client->getRequest()->getTargetUser();
    is_group = client->getRequest()->getAclIsGroup();
    rights = (nest_access_rights_t)
              atoi( client->getRequest()->getTargetAcl() );

    nest_debug( NEST_DEBUG_SERVER,
                "SetUserAcl: User='%s', IsGroup=%c, File='%s', ACL=0x%02x\n",
                target_user, is_group ? 't' : 'f',
                filename.c_str(), rights );

    // Make the check
    if ( ! is_group ) {
        status = nestStorageManager::set_user_acl( rights,
                                                   target_user,
                                                   filename.c_str(),
                                                   client->getIdentity(),
                                                   client->getCwd() );
    } else {
        status = nestStorageManager::set_group_acl( rights,
                                                    target_user,
                                                    filename.c_str(),
                                                    client->getIdentity(),
                                                    client->getCwd() );
    }

    nest_debug( NEST_DEBUG_SERVER, "SetUserAcl: status=%d\n", status );

    // Done
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
}

NestReplyStatus
SendVersion( nestClient *client ) {
    client->sendVersion( NEST_VERSION );
    return NEST_SUCCESS;
}


NestReplyStatus
SendPwd( nestClient *client ) {
    string pwd;

    NestReplyStatus reply = nestStorageManager::expand_path( 
            pwd, client->getIdentity(),
            client->getCwd(), "" );

    if (reply != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__, reply);
    } else {
        client->sendPwd( pwd.c_str() );
    }

    return reply;
}

NestReplyStatus
CollectionOp( nestClient *client, int type ) {

    const char *user = client->getIdentity();
    const char *path = client->getRequest()->getNestFile();
    string name = client->getRequest()->getCollectionName();
	bool			sendack = false;
	NestCollectionManager	*colman = nestStorageManager::collection_manager;

    NestReplyStatus status = NEST_SUCCESS;
    nest_obj_type_t filetype;

    switch( type ) {

    case NAME_COLLECTION:
        status = nestStorageManager::get_type( &filetype,
                                               path,
                                               user,
                                               client->getCwd() );

        if (status == NEST_SUCCESS ) {
            string pathname;
            nestStorageManager::expand_path( pathname,
                                             user,
                                             client->getCwd(),
                                             path );

            status = colman->createAssociation( name, pathname );
        }
		sendack = true;
        break;

    case UNNAME_COLLECTION:
        status = colman->removeAssociation( name );
		sendack = true;
        break;

	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT: {
		sendack = true;
		NestCollection	*col = colman->getCollection( name );
		if ( ! col ) {
			return NEST_LOCAL_FILE_NOT_FOUND;
		}

		switch( type ) {
		case SET_COLLECTION_MTIME: {
			time_t	mtime = client->getRequest()->getCollectionMtime( );
			colman->updateMtime( col, mtime );
			break;
		}
		case SET_COLLECTION_VERSION: {
			const char *version = client->getRequest()->getCollectionVersion( );
			col->setVersion( version );
			break;
		}
		case SET_COLLECTION_COMMENT: {
			const char *comment = client->getRequest()->getCollectionComment( );
			col->setComment( comment );
			break;
		}
		case GET_COLLECTION_MTIME:
		case GET_COLLECTION_VERSION:
		case GET_COLLECTION_COMMENT:
			break;
		}
		delete col;
		break;
	}

    case QUERY_COLLECTION:
    default:
        assert( 0 );
    }

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else if ( sendack ) {
		client->sendAck( );
	} else {
		assert( 0 );		// QUERY_COLLECTION not yet in chirp
    }

    return status;
}

NestReplyStatus
DirOp( nestClient *client, int type ) {

    char *new_cwd;
    const char *dir  = client->getRequest()->getNestFile();
    const char *user = client->getIdentity();
    const char *cwd  = client->getCwd();

    NestReplyStatus status;

    switch( type ) {
    case RMDIR_REQUEST:
        status = nestStorageManager::rmDir( user, cwd, dir );
        break;
    case CWD_REQUEST:
        status = nestStorageManager::chDir( &new_cwd, user, cwd, dir );
        if (status == NEST_SUCCESS) {
            status = client->setCwd( new_cwd );
            NestFree( new_cwd );
        }
        break;
    case MKDIR_REQUEST:
        status = nestStorageManager::mkDir( user, cwd, dir, client->getUid() );
        break;
    default:
        assert( 0 );
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck( );
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
AuthenticateUser( nestClient *client ) {

    char *username = NULL;
    char *identity = NULL;
    char *modified_identity = NULL;
    NestReplyStatus stat = NEST_SUCCESS;

        // check whether requested authentication method is allowed
    if ( ! (client->getRequest()->getAuthType() & allowedAuthentication) ) {
        nest_debug( 
                NEST_DEBUG_AUTH,
                "Client %s requested unallowed authentication type: %d\n",
                client->getRequest()->getUser(),
                client->getRequest()->getAuthType() );
        stat = NEST_INVALID_AUTHENTICATION;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  stat ); 
        return stat;
    } else {
        nest_debug( NEST_DEBUG_AUTH, "Client %s (%s) requesting auth type %d\n",
                client->getRequest()->getUser(),
                client->ipAddress(),
                client->getRequest()->getAuthType() );
    }

        // check requested authentication method
    int type = client->getRequest()->getAuthType();
    switch( type ) {
    case NEST_AUTH_GSS:
        stat = client->getConnection()->gssAccept( &username , &identity );
        if ( stat != NEST_SUCCESS ) {
            nest_debug(NEST_DEBUG_AUTH, 
                    "Client %s (%s) couldn't authenticate: %s\n",
                        client->getRequest()->getUser(),
                        client->ipAddress(),
                        NestErrorString( stat ) );
            client->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
        }
        break;
    case NEST_AUTH_NONE:
        username = NestStrdup( (char*)client->getRequest()->getUser() );
        identity = NestStrdup( username );
        break;
        /*
    case NEST_AUTH_IP:
            // TODO:  put in IP based authentication ala Kangaroo
        */
    default:
        nest_warn( __FILE__, __FUNCTION__, "Unknown auth type requested\n" );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  
                NEST_INVALID_AUTHENTICATION );
        stat = NEST_INVALID_AUTHENTICATION;
    }

        // if we make it here, the authentication was successful
        // now go ahead and set the user name accordingly
    if ( stat == NEST_SUCCESS ) {
        assert( username != NULL );
        assert( identity != NULL || (allowedAuthentication & NEST_AUTH_NONE) );
        
        modified_identity = NestGetModifiedIdentity( identity );

        user_id_t uid;
        if ( nestStorageManager::is_valid_user( modified_identity, &uid ) ) {
            client->setName( username );
            client->setIdentity( modified_identity );
            client->setUid( uid );
            nest_debug( NEST_DEBUG_SERVER | NEST_DEBUG_CLIENT,
                "Client identity set to %s (name %s) (uid %d)\n", 
                client->getIdentity(), client->getName(), client->getUid() );
            client->sendAck( );
        } else {
            nest_warn( __FILE__, __FUNCTION__, 
                    "WARN: unable to change to user: %s\n", identity );
            client->sendErrorMessage(__FILE__, __FUNCTION__, NEST_UNKNOWN_USER);
        }
        if ( username != NULL ) NestFree( username );
        if ( identity != NULL ) NestFree( identity );
        if ( modified_identity != NULL ) NestFree( modified_identity );


            // proxy servers may need to do something special with
            // this client
        if ( serverIsProxy() ) {
            addNewClientToProxy( client ); 
        }

    }

    return stat;
}

NestReplyStatus
AddUser( nestClient *client ) {

    NestReplyStatus status;
    const char      *identity = client->getRequest()->getIdentity();

    if(identity)
        status = nestStorageManager::add_user( client->getRequest()->getUser(),
                                               client->getRequest()->getGroup(),
                                               client->getIdentity(),
                                               identity);
    else
        status = nestStorageManager::add_user( client->getRequest()->getUser(),
                                               client->getRequest()->getGroup(),
                                               client->getIdentity(),
                                               "");
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteUser( nestClient *client ) {
    NestReplyStatus status;
    status = nestStorageManager::delete_user( client->getRequest()->getUser(),
                                              client->getIdentity() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddGroup( nestClient *client ) {
    NestReplyStatus status;
    status = nestStorageManager::add_group( client->getRequest()->getGroup(),
                                            client->getIdentity() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteGroup( nestClient *client ) {
    NestReplyStatus status;
    status = nestStorageManager::delete_group(client->getRequest()->getGroup(),
                                              client->getIdentity());
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddUserToGroup( nestClient *client ) {
    NestReplyStatus status;
    status = nestStorageManager::add_user_to_group(
                                            client->getRequest()->getUser(),
                                            client->getRequest()->getGroup(),
                                            client->getIdentity() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteUserFromGroup( nestClient *client ) {
    NestReplyStatus status;
    status = nestStorageManager::delete_user_from_group(
                                            client->getRequest()->getUser(),
                                            client->getRequest()->getGroup(),
                                            client->getIdentity() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
ReserveUserSpace( nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;
    double duration = client->getRequest()->getLotDuration();

    if (duration < 0) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }
    int lot_type = client->getRequest()->getLotType();

    NestLotHandle handle, parent_handle;

    const char* user = client->getRequest()->getLotUser();
    const char* parent_user = client->getRequest()->getParentLotUser();
    const char* parent_group = client->getRequest()->getParentLotGroup();
    rsv_id_t rsv_id = client->getRequest()->getParentLot();
    NestAdd_id_to_handle(&parent_handle,rsv_id);
    if(!parent_user)
        parent_user = strdup("");
    if(!parent_group)
        parent_group = strdup("");

    if(user)
        status = nestStorageManager::request_user_lot(
                        &handle,
                        client->getIdentity(),
                        client->getRequest()->getLotSize(),
                        duration,
                        user,
                        parent_user,
                        parent_group,
                        &parent_handle,
                        lot_type
                 );
    else{
        status = nestStorageManager::request_user_lot(
                        &handle,
                        client->getIdentity(),
                        client->getRequest()->getLotSize(),
                        duration,
                        client->getIdentity(),
                        parent_user,
                        parent_group,
                        &parent_handle,
                        lot_type
                 );
    }

	{
		rsv_id_t	tmpid;
		NestGet_id_from_handle( &handle, &tmpid );
		nest_debug( NEST_DEBUG_SERVER, "ReserveUserSpace: Handle=%p, id=%d\n",
					&handle, tmpid );
	}
                        
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
ReserveGroupSpace( nestClient *client ) {
    NestReplyStatus status;
    const char* group = client->getRequest()->getLotUser();
    int lot_type = client->getRequest()->getLotType();    
    double duration = client->getRequest()->getLotDuration();

    if (duration < 0) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }
    NestLotHandle handle, parent_handle;

    const char* parent_user = client->getRequest()->getParentLotUser();
    if(!parent_user) {
        parent_user = strdup("");
    }
    const char* parent_group = client->getRequest()->getParentLotGroup();
    if(!parent_group) {
        parent_group = strdup("");
    }
    rsv_id_t rsv_id = client->getRequest()->getParentLot();
    NestAdd_id_to_handle(&parent_handle,rsv_id);
    
    status = nestStorageManager::request_group_lot(
                &handle,
                client->getIdentity(),
                client->getRequest()->getLotSize(),
                duration,
                group,
                parent_user,
                parent_group,
                &parent_handle,            
                lot_type
             );
            
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
ListUserLots( nestClient *client ) {

    NestLotList     *result  = NULL;
    const char      *user = client->getRequest()->getLotUser();
    NestReplyStatus status;

    if(user)
        status = nestStorageManager::list_user_lots(&result,
                                                    client->getIdentity(),
                                                    user);
    else
        status = nestStorageManager::list_user_lots(&result,
                                                    client->getIdentity(),
                                                    client->getIdentity());                                                        

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotList *tmp = result->next;
            NestFree(result);
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotList *tmp = result->next;
            client->sendNestLot( &result->stat );
            NestFree(result);
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
ListGroupLots( nestClient *client ) {

    NestLotList     *result = NULL;
    NestReplyStatus
    status = nestStorageManager::list_group_lots(
                                         &result,
                                        client->getIdentity(),
                                        client->getRequest()->getGroup());

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotList *tmp = result->next;
            NestFree(result);
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotList *tmp = result->next;
            client->sendNestLot( &result->stat );
            NestFree(result);
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
GetLotStat( nestClient *client ) {

    NestReplyStatus ret_val;
    NestLot   rstat;
    NestLotHandle handle;

    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    ret_val = nestStorageManager::get_lot_status(&rstat,
                                                 &handle,
                                                 client->getIdentity() );

    if (ret_val != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  ret_val );
    } else {
        client->sendAck();
        client->sendNestLot( &rstat );
    }
    return ret_val;
}

NestReplyStatus
UpdateLot( nestClient *client ) {

    NestReplyStatus status;
    double duration = client->getRequest()->getLotDuration();

    NestLotHandle   handle;

    if (client->getRequest()->getLotHandle(&handle) < 0) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = nestStorageManager::update_lot(
            &handle,
            client->getIdentity(),
            client->getRequest()->getLotSize(),
            duration);

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
ChangeLotType( nestClient *client) {
    NestReplyStatus status;
    double duration = client->getRequest()->getLotDuration();
    
    NestLotHandle   handle;
    
    if (duration < 0 || client->getRequest()->getLotHandle(&handle) < 0) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = nestStorageManager::change_lot_type(
                &handle,
                client->getIdentity(),
                client->getRequest()->getNewLotType(),
                duration);
    
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
MoveFileLot( nestClient *client ) {

    NestReplyStatus status;
    double movebytes = client->getRequest()->getLotMoveBytes();
    NestLotHandle source_lot, dest_lot;

    // Extract the file name..

    // Check the Bytes to move
    if (movebytes < 0 || 
        client->getRequest()->getSourceLotHandle(&source_lot) < 0 ||
        client->getRequest()->getDestLotHandle(&dest_lot) < 0) 
    {
        status = NEST_BAD_PARAMETERS; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = nestStorageManager::modify_lot_usage(
        client->getIdentity(),
        client->getRequest()->getNestFile(),
        client->getCwd(),
        &source_lot,
        movebytes,
        &dest_lot );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
TerminateLot( nestClient *client ) {

    NestReplyStatus status;
    NestLotHandle handle;

    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    status = nestStorageManager::terminate_lot(&handle,
                                               client->getIdentity());
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddUserToLot( nestClient *client ) {
    NestReplyStatus status; 
    NestLotHandle handle;
        
    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    const char* user = client->getRequest()->getLotUser();
    
    status = nestStorageManager::add_user_to_lot(&handle,
                                                 client->getIdentity(),
                                                 user);
    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;    
}

NestReplyStatus
DeleteUserFromLot( nestClient *client ) {
    NestReplyStatus status; 
    NestLotHandle handle;
        
    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    const char* user = client->getRequest()->getLotUser();
    
    status = nestStorageManager::delete_user_from_lot(&handle,
                                                      client->getIdentity(),
                                                      user);
    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;    
}

NestReplyStatus
AttachLotToFile( nestClient *client ) {
    NestReplyStatus status;
    NestLotHandle handle;

    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    const char* file = client->getRequest()->getNestFile();

    status = nestStorageManager::attach_lot_to_file( &handle,
                                                     file,
                                                     client->getIdentity(),
                                                     client->getCwd() );
    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}                                 
                                                    
NestReplyStatus
DetachLotFromFile( nestClient *client ) {
    NestReplyStatus status;
    NestLotHandle handle;

    if (client->getRequest()->getLotHandle(&handle) < 0) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  NEST_BAD_PARAMETERS );
    }

    const char* file = client->getRequest()->getNestFile();

    status = nestStorageManager::detach_lot_from_file( &handle,
                                                     file,
                                                     client->getIdentity(),
                                                     client->getCwd() );
    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}            

NestReplyStatus
MakeLink( nestClient *client ) {

    string  oldfile = "",
            newfile = "";

    NestReplyStatus status = nestStorageManager::Link(
            client->getIdentity(),
            client->getRequest()->getNestFile(),
            client->getRequest()->getLinkName(),
            client->getRequest()->getValue(),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
CopyFile( nestClient *client ) {
/** This copy function in storage manager is no longer used
    
    NestReplyStatus status =
        nestStorageManager::copyFile( client->getIdentity(),
                                      client->getRequest()->getNestFile(),
                                      client->getRequest()->getLinkName(),
                                      client->getCwd() );

    if( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
*/

    NestReplyStatus status;

    nestProtocol *storage, *otherStorage;

    status = openStorageConnections( &storage, &otherStorage, client );
    if(status == NEST_SUCCESS){
        client->assignStorageConnection( storage );
        client->assignOtherStorageConnection( otherStorage );
        status = nestConcurrencyManager::addTask( client );
    }
        
    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "copy transfer added to concurrency manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "copy transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }

}

NestReplyStatus
DeleteFile( nestClient *client ) {

    const char *file = client->getRequest()->getNestFile();

    NestReplyStatus status = 
        nestStorageManager::deleteFile( client->getIdentity(), 
                                        file,
                                        client->getCwd() );

    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
}

NestReplyStatus
RenameFile( nestClient *client ) {
    
    const char *oldfile = client->getRequest()->getOldNestFile(),
               *newfile = client->getRequest()->getNewNestFile();

    NestReplyStatus status =
        nestStorageManager::renameFile( client->getIdentity(),
                                        client->getUid(),
                                        oldfile,
                                        newfile,
                                        client->getCwd()
                                      );
                                
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
    
}


NestReplyStatus
openStorageConnections( nestProtocol **storage,
                        nestProtocol **otherStorage,
                        nestClient   *client) {

    NestReplyStatus     reply;
    nestProtocol *storage_con, *other_storage_con;

    nestRequest *req = client->getProxyRequest();

    off_t offset = (off_t) req->getStart();
	int importance = 1;
	int link_count = 0;
    
    long filesize;
    reply = nestStorageManager::filelength(
                                    &filesize,
                                    req->getNestFile(),
                                    client->getIdentity(),
                                    client->getCwd() );
    NEST_REPLY_CHECK(reply);
    req->setEnd(filesize);

    reply = nestStorageManager::open( &storage_con,
                                      client->getIdentity(),
                                      client->getUid(),
                                      NULL,
                                      req->getLinkName(),
                                      WRITE_FLAGS,
                                      offset,
                                      filesize,
                                      client->getCwd(),
									  link_count,
									  importance );
    *storage = (nestProtocol *)storage_con;
    NEST_REPLY_CHECK(reply);

    reply = nestStorageManager::open( &other_storage_con,
                                      client->getIdentity(),
                                      client->getUid(),
                                      NULL,
                                      req->getNestFile(),
                                      READ_FLAGS,
                                      offset,
                                      filesize,
                                      client->getCwd(),
									  link_count,
									  importance );
    *otherStorage = (nestProtocol *)other_storage_con;
    NEST_REPLY_CHECK(reply);

    nest_debug(NEST_DEBUG_TRANSFER, "Created copy transfer: %s->%s\n",
                                    req->getNestFile(),req->getLinkName());

    return reply;
}

NestReplyStatus
openStorageConnection( nestProtocol **storage, nestClient *client ) {

    NestReplyStatus     reply = NEST_SUCCESS;
    off_t    offset;
    ssize_t filesize;

    nestRequest *req = client->getProxyRequest();
	int importance = 1;
	int link_count = 0;

        // TODO: check with consistency manager
    switch( req->getType() ) {
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
		if ( serverIsProxy() ) {
				// don't delete explicitly put files
			importance = INFINITE_IMPORTANCE;
			nest_debug( NEST_DEBUG_PROXY, "Setting importance of %s to %d.\n",
					req->getNestFile(),
					importance );
		}
        offset = (off_t) req->getStart();
        filesize = (ssize_t) req->getEnd();
        reply = nestStorageManager::open( storage,
                                          client->getIdentity(),
                                          client->getUid(),
                                          NULL,
                                          req->getNestFile(),
                                          WRITE_FLAGS,
                                          offset,
                                          filesize,
                                          client->getCwd(),
										  link_count,
										  importance );
        break;
    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        offset = (off_t) req->getStart();
        filesize = -1;
        reply = nestStorageManager::open( storage,
                                          client->getIdentity(),
                                          client->getUid(),
                                          NULL,
                                          req->getNestFile(),
                                          READ_FLAGS,
                                          offset,
                                          filesize,
                                          client->getCwd(),
										  link_count,
										  importance );
        break;
    default:
        *storage = NULL;
        break;
    }

    return reply;
}

    // this function is called by both the home server and proxy servers
    // so a pointer is passed to the proxyInit function
    //
    // when we run in proxy mode, we translate client requests into
    // requests for chunks.  however, we also will use transfer file
    // to send the proxy chunk once it arrives from the server
    // so we only want to translate in the first case, but not the second
int
TransferFile( nestClient *client, bool already_translated ) {

    nestRequest *req = NULL;
    if ( serverIsProxy() ) {
        if ( ! already_translated ) {
            proxyTranslateRequest( client );
        }
        req = client->getProxyRequest();
    } else {
        req = client->getRequest();
    }

    NestReplyStatus status = NEST_SUCCESS;
    long end          = req->getEnd();
    const char *file  = req->getNestFile();
    int  type         = req->getType();

    nest_debug( NEST_DEBUG_TRANSFER, "TransferFile %s for %s (%s) (len %d), "
            "type %d\n",
            file, client->getIdentity(), client->ipAddress(), end, type );

    switch( type ) {
    case CLIENT_SEND_FILE:
            // try to reserve the necessary space
            // status = MakeReservation( client );
        break;
    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        long filelength;
            // get the filelength, if file not found, report this and return
        if ( ( status = nestStorageManager::filelength( &filelength,
                                                    file,
                                                    client->getIdentity(),
                                                    client->getCwd() )) 
                != NEST_SUCCESS )
        {
            break;
        }

            // check if user is requesting to read the entire file or to read
            // past eof, adjust that here
            // don't forget to adjust in both proxy and normal reqs
        if ( end == ENTIRE_FILE ) {
            end = filelength;
            req->setEnd( filelength );
            client->getRequest()->setEnd( filelength );
        } else if ( end > filelength ) {
            nest_debug( NEST_DEBUG_PROXY, 
                        "Req end %ld > filelength %ld. Fixing\n",
                        end, filelength );
            end = filelength;
            //long diff = req->getEnd() - filelength;
                // these next two lines are what proxys should do...
            //req->setEnd( req->getEnd() - diff );
            //client->getRequest()->setEnd(client->getRequest()->getEnd() - diff);
            req->setEnd( filelength + req->getChunk() );
            client->getRequest()->setEnd( filelength 
                    + client->getRequest()->getChunk() );
        }
            // special case, user trying to read past eof
        if ( req->getStart() >= filelength + req->getChunk() ) {
                // make it a 0 byte transfer so they will go away
            req->setStart( req->getEnd() );
            client->getRequest()->setStart( client->getRequest()->getEnd() ); 
        }
        break;
    case THIRD_PARTY_GET:
            // nothing to do, except when we find out the size, maybe
            // we need to try to make a reservation . . .
        break;
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
            // nothing to do here
        break;
    default:
        status = NEST_UNKNOWN_REQUEST_TYPE;
        break;
    }

        // deal with proxy logic here to retrieve files we don't have
        // probably should never see this NEST_FLOCK error
    if (serverIsProxy() && (status==NEST_FILE_NOT_FOUND || status==NEST_FLOCK)){
        if ( status == NEST_FLOCK ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Possibly mishandling an unexpected error at %s,%s: %s",
                    __FILE__, __FUNCTION__, NestErrorString( status ) );
        }
        assert( type == CLIENT_RECV_FILE ); 
        nest_debug( NEST_DEBUG_PROXY, "Will proxy request %s for %s.\n",
                file, client->getIdentity() );
        status = addProxyTransfer( client );
        if ( status == NEST_SUCCESS ) {
            return ACTIVE_CONNECTION;
        } else {
            nest_debug( NEST_DEBUG_PROXY, "Failed to proxy request %s: %s.\n",
                    file, NestErrorString( status ) ); 
                // fall through to sendErrorMessage
        }
    }

        // make sure parameters are OK
        // but don't overwrite a previously set error value
    if ( status == NEST_SUCCESS ) {
        if (req->getStart()<0 || (end != ENTIRE_FILE && end < req->getStart())){
            nest_warn(__FILE__,__FUNCTION__, "bad params, %ld end, %ld start\n",
                    end, req->getStart());
            status = NEST_BAD_PARAMETERS;
        }
    }
    
        // if no errors yet, add transfer request into concurrency manager
    if ( status == NEST_SUCCESS ) {
            // concurrencyManager now responsible for client->endRequest
        nestProtocol *storage;
        status = openStorageConnection( &storage, client );

        /* This is for transfers that don't send filesize info*/
        if(req->getEnd() == ENTIRE_FILE){
            status = nestStorageManager::increaseAllocation( client );
            if(status != NEST_SUCCESS) {
                nestStorageManager::cleanupGftpTransfer(client);
            }
        }

        if ( status == NEST_SUCCESS ) {
            client->assignStorageConnection( storage );
            status = nestConcurrencyManager::addTask( client );
        }
    }

    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "transfer added to concurrency manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }
}

NestReplyStatus
SendAvailableSpace( nestClient *client ) {
    long long int avail, total;
    NestReplyStatus status = nestStorageManager::bytesAvailable(IS_USER,
            &avail, &total );
    if ( status == NEST_SUCCESS ) {
        client->sendDiskspace( avail, total );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

int
NestSync( nestClient *client ) {
    int connection = PASSIVE_CONNECTION;
    struct timeval before, after;
    nest_debug( NEST_DEBUG_SERVER, "client %s requested a sync\n",
            client->getIdentity() );
    NestGettimeofday( &before );
    NestReplyStatus status = nestStorageManager::sync( client->getIdentity() );
    NestGettimeofday( &after );

    if( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_SERVER, "sync successful, took %ld millis\n",
                NestTimediff( &before, &after, __FUNCTION__ ) );
        if ( serverIsProxy() ) {
            status = flushDirtyWrites( client ); 
                // client is now responsibility of proxy server 
            connection = ACTIVE_CONNECTION; 
        } else {
            client->sendAck();
        }
    } else {
        nest_warn( __FILE__, __FUNCTION__, "sync failure: %s\n", NestErrorString( status ) );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    if ( connection == PASSIVE_CONNECTION ) {
            // otherwise, the proxy code must eventually end the request
        client->endRequest( status );
    }
    return connection;
}


#define PROXY_ITERATOR list<char*>::iterator

// a client (or a proxy monitor) can query a home nest to know which proxy
// nodes have registered themselves for a particular client
NestReplyStatus
QueryProxys( nestClient *client ) {
    list< char *> *proxys;

    if ( registered_proxys == NULL ) {
        proxys = NULL;
    } else {
        proxys = (list<char*>*)
            nest_hash_table_lookup( registered_proxys, client->getIdentity());
    }

    client->getConnection()->sendProxyList( proxys );

    return NEST_SUCCESS;
    
}

NestReplyStatus
RemoveProxyClient( nestClient *client ) {

    list< char *> *proxys;
    const char *proxy = client->proxyAddress();

    nest_debug( NEST_DEBUG_PROXY, "Removing proxy %s for client %s.\n",
                proxy, client->getIdentity() );

        // find the list of proxys registered for this client
    assert( registered_proxys != NULL );
    void *entry = nest_hash_table_lookup( registered_proxys, client->getIdentity());
    proxys = (list<char *> *)entry;
    assert( proxys != NULL );

        // remove this particular proxy
    bool found = false;
    PROXY_ITERATOR itr = proxys->begin();
    while ( itr != proxys->end() ) {
        char *g = (char *)*itr;
        if ( ! strcmp( g, proxy ) ) {
            found = true;
            itr = proxys->erase( itr );
            NestFree( g );
            break;
        }
        itr++;
    }
    if ( ! found ) {
        nest_warn( __FILE__, __FUNCTION__, "Tried to remove proxy %s for client %s, not found.\n",
                proxy, client->getIdentity() );
        //assert( found );
    }

    return NEST_SUCCESS;
}

    
// there are two reasons why clients register with a nest:
// 1) nest is home node and client is a proxy server
// 2) nest is proxy coordinator node and client is a proxy server
NestReplyStatus
RegisterProxyClient( nestClient *client ) {

    list< char *> *proxys;      // the list of proxys known for a client
    const char *proxy;          // the proxy location being added now

    proxy = client->getRequest()->getProxyLocation();
    client->setProxy( proxy );
    nest_debug(NEST_DEBUG_PROXY,"Client %s (%d) from %s registering proxy %s\n",
                client->getIdentity(), client->getId(), client->ipAddress(), proxy);

        // init the main structure if there already isn't one
    if ( registered_proxys == NULL ) {
        registered_proxys = nest_hash_table_create( 1037, NULL );
    }

        // see if this client has previously registered any proxys
        // if not, create a list for them
    void *entry = nest_hash_table_lookup( registered_proxys, client->getIdentity());
    if ( entry == NULL ) {
        proxys = new list< char *>;
        proxys->clear();
        nest_hash_table_insert( registered_proxys, 
                                client->getIdentity(), 
                                (void *)proxys );
    } else {
        proxys = (list<char *> *)entry;
    }

        // see whether we need to add this proxy to the list 
    bool found = false;
    for( PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++ ) {
        if ( ! strcmp( *itr, proxy ) ) {
            found = true;
            break;
        }
    }
        // important to do a push_back and not a push_front so that
        // new dynamicly larding pfs clients put proxys into same buckets
        // as old pfs dynamically larding clients
    if ( ! found ) proxys->push_back( NestStrdup( proxy ) );

        // if the proxy is coordinator needs to know about this
    if ( serverIsProxy() ) {
            // proxy will deal with sending ack or error reply
        return proxyRegisterProxyClient( client, (char*)proxy );
    }

    //return QueryProxys( client );
    client->sendAck();
    return NEST_SUCCESS;
}

NestReplyStatus
FormatDisk( nestClient *client ) {
        // delete all entries and report available space
    NestReplyStatus status = 
        nestStorageManager::formatDisk( client->getIdentity() , NULL );
    SendAvailableSpace( client );
    return status;
}

NestReplyStatus
SendNestAd( nestClient *client ) {

    NestReplyStatus status = NEST_SUCCESS;
        // should we update it first maybe?
    WriteAd();
    
    nestProtocol *adfile = nestStorageManager::openMetaFile(AD_NAME,READ_FLAGS);

    if ( adfile == NULL ) {
        status = NEST_REMOTE_FILE_NOT_FOUND; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        int adlen = nestStorageManager::metaFilelength( AD_NAME );
        nestAd *ad = nestAdvertiser::readAd( adfile, adlen );
        if ( ad == NULL ) {
            status = NEST_REMOTE_FILE_NOT_FOUND; 
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        } else {
            client->sendAd( ad );
            delete( ad );
        }
        delete( adfile );
    }
    return status; 
}

NestReplyStatus
SendDirectoryList( nestClient *client ) {

    NestFileInfo *list =
        nestStorageManager::directoryContents( client->getIdentity(),
                                               client->getCwd() );

    client->sendDirectoryList( list );
    while( list != NULL ) {
        NestFileInfo *tmp = list;
        list = list->next;
        NestFree( tmp );
    }
    return NEST_SUCCESS;
}

NestReplyStatus
SendBandwidthStatistics( nestClient *client ) {
    client->sendBandwidthStatistics( );
    return NEST_SUCCESS;
}
