/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_protocol_manager.h"
#include "nest_speak.h"
#include "nest_speak.i.h"

// prototypes for private functions
int  ClientLogMessage( char * msg );
void printBandwidthStatistics( struct timeval *request, struct timeval *start,
        struct timeval *end, long size, char *req_type );

NestReplyStatus DirOp( const char *dir, int req_type, NestConnection con );

NestReplyStatus ThirdPartyTransfer(  char *nestfile, char *remotehost,
        char *remotefile, NestConnection con, int transfer_type,
        char *remoteprotocol );

bool badArgs( long fileoffset, long length );

NestReplyStatus NestWriteBytes( NestConnection con, char *remotefile,
        long offset,  char *buffer, int size, bool ack_wanted,
        bool completeTransfer );

NestReplyStatus NestReadBytes( NestConnection con, char *remotefile,
        long offset, char *buf, int size, int &bytes_read,
        bool completeTransfer );

// bool SendQuotaRequest( char *user, long new_q, NestConnection con );

// ------------------ functions start here -------------------

NestReplyStatus
myNestOpenConnection( NestConnection &con, const char *nestlocation, bool anon,
        bool user_user_auth, int port ) 
{
    NestReplyStatus status;
    char user[NEST_MESSAGESIZE];

    hostent *hostp = NestGethostbyname( nestlocation );
    if ( hostp == NULL ) return NEST_NO_CONNECTION;

    sockaddr_in sin;
    memset( &sin, 0, sizeof( sockaddr_in ) );
    sin.sin_family = AF_INET;
    sin.sin_port   = htons( (unsigned short) port );
    memcpy( &sin.sin_addr, hostp->h_addr, hostp->h_length );

    con = NestSocket( AF_INET, SOCK_STREAM, 0 );
    if ( con < 0 ) {
        status = NEST_NO_CONNECTION;
        goto DONE;
    }

    status = NestConnect( con, (sockaddr *)&sin, sizeof( sockaddr_in ) );
    if ( status != NEST_SUCCESS ) {
        NestClose( con );
        con = -1;
    } else {
        if ( !anon ) { // that's all we do anonymously
            char *server = ( user_user_auth ) ? NULL : hostp->h_name;
            status = NestChangeUserID( NestGetUserName( user ), con, server );
        }
    }

    DONE:
    NestFree( hostp );
    return status;
}

NestReplyStatus
NestOpenAnonymously( NestConnection &con, const char *nestlocation, 
        int port ) 
{
    NestReplyStatus status = myNestOpenConnection( 
            con, nestlocation, true, false, port);
    if ( status != NEST_SUCCESS && status != NEST_NO_CONNECTION ) {
        NestClose( con );
        nest_debug( NEST_DEBUG_AUTH, 
                "Unable to authenticate to host-based server: %s.\n"
                "\tWill try user to user.\n",
                NestErrorString( status ) );
        status = myNestOpenConnection( con, nestlocation, true, true, port );
    }
    return status;
}

NestReplyStatus
NestOpenConnection( NestConnection &con, const char *nestlocation, 
        int port ) 
{
    NestReplyStatus status = myNestOpenConnection( 
            con, nestlocation, false, false, port);
    if ( status != NEST_SUCCESS ) {
        if(con!=-1) NestClose( con );
        nest_debug( NEST_DEBUG_AUTH, 
        //nest_warn( 
                __FILE__,
                __FUNCTION__,
                "Unable to authenticate to host-based server: %s.\n"
                "\tWill try user to user.\n",
                NestErrorString( status ) );
        status = myNestOpenConnection( con, nestlocation, false, true, port );
    }
    return status;
}

NestReplyStatus
NestCloseConnection( NestConnection con ) {
    if ( NestClose( con ) < 0 ) {
        return NEST_NO_CONNECTION;
    } else {
        return NEST_SUCCESS;
    }
}

NestReplyStatus
NestSendFile ( char *localfile, char *remotefile, NestConnection con ) {
    return NestSendFile( localfile, remotefile, con, NULL, NULL );
}

NestReplyStatus
NestSendFile( char *localfile, char *remotefile, NestConnection con,
        NestProgressFunction *progress, void *args )
{
    int fd;
    NestReplyStatus serverReply;
    long filelength;
    long offset = 0;
    long remaining = 0;
    static char buffer[FILEBUFFER];
    struct timeval requesttime, transferstart, transferend;

        // sanity check on the connection
    if ( con < 0 ) return NEST_NO_CONNECTION;

    fd = NestOpen( localfile, O_RDONLY );
    if(fd<0) return ErrnoToNest(errno);

    filelength = lseek(fd,0,SEEK_END);
    lseek(fd,0,SEEK_SET);

    if(filelength<0) {
        NestClose(fd);
        return NEST_NOT_A_FILE;
    }
        // ask for permission from the server
        // this will reserve space for the file as well
    NestGettimeofday( &requesttime );
    serverReply = RequestWriteFile( remotefile, offset, filelength, con );
    if ( serverReply != NEST_SUCCESS ) {
        return serverReply;
    }

        // if we make it here, server gave us permission, so send the file
        // just do a succession of NestWriteBytes
    NestGettimeofday( &transferstart );

    offset    = 0;
    remaining = filelength;
    while( remaining > 0 ) {
        /* maybe this lseek can be removed ? */
        if ( offset != lseek( fd, offset, SEEK_SET ) ) {
            nest_perror( "lseek" );
            serverReply = ErrnoToNest( errno );
            break;
        }
            // read into buffer
        int readsize = ( remaining < FILEBUFFER ) ? remaining : FILEBUFFER;
        if ( NestReadn( fd, buffer, readsize ) < 0 ) {
            nest_perror( "NestReadn" );
            serverReply = ErrnoToNest( errno );
            break;
        }

            // call the users progress function
        if ( progress != NULL ) {
            progress( filelength - remaining + readsize, filelength, args );
        }
        
            // call NestWriteBytes
        bool completeTransfer = false;
        serverReply = NestWriteBytes( con, remotefile, offset, buffer,
                readsize, true, completeTransfer );
        if ( serverReply != NEST_SUCCESS ) break;

            // call the users progress function again
        if ( progress != NULL ) {
            progress( filelength - remaining + readsize, filelength, args );
        }
        
            // adjust values appropriately and repeat if necessary
        remaining -= readsize;
        offset    += readsize; 
    }
    NestClose( fd );

        // make sure server received file OK
    if ( serverReply == NEST_SUCCESS ) {
        assert( remaining == 0 );
        assert( offset == filelength );
            // try to receive a final ack
        serverReply = ReceiveNestAck( con );
        if ( serverReply == NEST_SUCCESS ) {
            NestGettimeofday( &transferend );
            printBandwidthStatistics( &requesttime, &transferstart, 
                    &transferend, filelength, "S" );
        }
    }

    return serverReply;
}

NestReplyStatus 
NestWriteBytes( NestConnection con, char *remotefile, long offset, 
                char *buffer, int size, bool ack_wanted ) 
{
    bool completeTransfer = true;
    return NestWriteBytes( con, remotefile, offset, buffer, size, ack_wanted,
            completeTransfer );
}

NestReplyStatus
NestWriteBytes( NestConnection con, char *remotefile, long offset,  
                char *buffer, int size, bool ack_wanted,
                bool completeTransfer )
{
    NestReplyStatus serverReply = NEST_SUCCESS;

    if ( ! ack_wanted ) {
        nest_die( "You really should ask for an ack.\n" );
    }

        // if this is a complete transfer, need to get permission first
    if ( completeTransfer ) {
        serverReply = RequestWriteFile( remotefile, offset, size, con );
        if ( serverReply != NEST_SUCCESS ) return serverReply;
    }

        // if we make it here, server gave us permission so go ahead and write
    if ( NestWriten( con, buffer, size ) != size ) {
        return NEST_NO_CONNECTION;
    }

        // if this is a complete transfer, check for ack here
    if ( completeTransfer ) {
        serverReply = ReceiveNestAck( con );
    }

    return serverReply; 
}

NestReplyStatus 
NestThirdPartyGet( char *nestfile, char *remotehost, 
        char *remotefile, char *remoteprotocol, NestConnection con ) 
{
    return ThirdPartyTransfer( nestfile, remotehost, remotefile, con, 
            THIRD_PARTY_GET, remoteprotocol );
}

NestReplyStatus
NestThirdPartyPut( char *nestfile, char *remotehost, 
        char *remotefile, char *remoteprotocol, NestConnection con ) 
{
    return ThirdPartyTransfer( nestfile, remotehost, remotefile, con, 
            THIRD_PARTY_PUT, remoteprotocol );
}

NestReplyStatus 
ThirdPartyTransfer(  char *nestfile, char *remotehost, 
        char *remotefile, NestConnection con, int transfer_type,
        char *remoteprotocol ) 
{
    char msg[NEST_MESSAGESIZE];
    long offset = 0;
    long length = ENTIRE_FILE;

    NestSnprintf( msg, sizeof(msg), "%d %s %s %s %ld %ld %s",
                transfer_type, nestfile, remotehost, remotefile, offset,
                length, remoteprotocol );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con ); 
}

NestReplyStatus
NestReceiveFile( char *remotefile, char *localfile, NestConnection con ) {    
    return NestReceiveFile( remotefile, localfile, con, NULL, NULL );
}

NestReplyStatus
NestReceiveFile( char *remotefile, char *localfile, NestConnection con,
        NestProgressFunction *progress, void *args ) {
    NestReplyStatus serverReply;
    struct timeval requesttime, transferstart, transferend;
    long filelength = ENTIRE_FILE;
    long offset = 0;
    long remaining;
    int  nbytes;
    static char buffer[FILEBUFFER];
    int  local_fp;

        // sanity check on connection
    if ( con < 0 ) return NEST_NO_CONNECTION;

        // get permission
    NestGettimeofday( &requesttime );
    serverReply = RequestReadFile( remotefile, offset, filelength, con );
    if ( serverReply != NEST_SUCCESS ) {
        return serverReply;
    }    
    assert( filelength >= 0 );

        // if we make it here, we got permission
        // first open our local file then
        // just string together a procession of NestReadBytes
    NestGettimeofday( &transferstart );
    local_fp = NestOpen( localfile, WRITE_FLAGS | O_TRUNC );
    if ( local_fp < 0 ) {
        nest_perror( "open" );
        return NEST_LOCAL_FILE_NOT_FOUND;
    }

    remaining = filelength;
    while( remaining > 0 ) {
        int readsize = ( remaining < FILEBUFFER ) ? remaining : FILEBUFFER;

            // read bytes from server
        bool completeTransfer = false;
        serverReply = NestReadBytes( con, remotefile, offset, buffer, 
                readsize, nbytes, completeTransfer );
        if ( serverReply != NEST_SUCCESS ) break;
        assert( readsize == nbytes );

            // call the user's progress function
        if ( progress != NULL ) {
            progress( filelength - remaining + readsize, filelength, args ); 
        }

            // write to file 
        NestWriten( local_fp, buffer, nbytes );

            // call the user's progress function again
        if ( progress != NULL ) {
            progress( filelength - remaining + readsize, filelength, args );
        }

            // adjust variables appropriately
        offset     += nbytes;
        remaining -= nbytes;
    }
    NestClose( local_fp );

    // no ack needed for a receive

    if ( serverReply == NEST_SUCCESS ) {
            // TODO - this assertion is sometimes hit when another file of the
            // same name but a different length is being overwritten 
        if ( NestFilelength( localfile ) != filelength ) {
            if ( strcmp( localfile, "/dev/null" ) ) {
                    // don't print this is user was just testing speed
                nest_warn( __FILE__, __FUNCTION__, 
                     "Unexpected filelen: transferred %ld != filesize %ld\n", 
                        filelength, NestFilelength( localfile ) );
            }
        }
        NestGettimeofday( &transferend );
        printBandwidthStatistics( &requesttime, &transferstart, 
                &transferend, filelength, "R" );
    }

    return serverReply;
}

NestReplyStatus 
NestReadBytes( NestConnection con, char *remotefile, long offset, char *buf, 
                int size, int &bytes_read ) 
{
    bool completeTransfer = true;
    return NestReadBytes( con, remotefile, offset, buf, size, bytes_read, 
            completeTransfer );
}

NestReplyStatus 
NestReadBytes( NestConnection con, char *remotefile, long offset, char *buf,
                int size, int &bytes_read, bool completeTransfer ) 
{
    NestReplyStatus serverReply = NEST_SUCCESS;
    long bytes_avail;

        // ask for permission if this is a completeTransfer
    if ( completeTransfer ) {
        bytes_read = size;
        bytes_avail = bytes_read;
        serverReply = RequestReadFile( remotefile, offset, bytes_avail, con );
        if ( serverReply != NEST_SUCCESS ) {
            return serverReply;
        }    

    bytes_read = bytes_avail;

        assert( bytes_read <= size );
            // if request is beyond eof, then append a NULL
        if ( bytes_read < size ) {
            buf[bytes_read] = '\0';
        }
        size = bytes_read;
    }

        // if we make it here, server gave us permission so go ahead and read
    if ( ( bytes_read = NestReadn( con, buf, size ) ) != size ) {
        serverReply = NEST_NO_CONNECTION;
    }

    return serverReply;
}

NestReplyStatus
NestRemoveFile( char *remotefile, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg,
                  sizeof(msg),
                  "%d %s",
                  FILE_DELETE_REQUEST,
                  remotefile );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestRenameFile( char *oldfile, char *newfile, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg,
                  sizeof(msg),
                  "%d %s %s",
                  FILE_RENAME_REQUEST,
                  oldfile,  
                  newfile );
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus 
NestNameCollection( const char *path, const char *name, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %s", NAME_COLLECTION, path, name );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestUnnameCollection( const char *name, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", UNNAME_COLLECTION, name );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus 
NestDateCollection( const char *name, time_t date, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %d", SET_COLLECTION_MTIME, name, date );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus 
NestVersionCollection( const char *name, const char *version, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %s", SET_COLLECTION_VERSION, name, version );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus 
NestCommentCollection( const char *name, const char *comment, NestConnection con ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %s", SET_COLLECTION_COMMENT, name, comment );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestDebug( char *debug_level, NestConnection con ) {

    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof( msg ), "%d %s", DEBUG_REQUEST, debug_level );
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
receiveProxyList( NestConnection con,
        NestProxyListFunction *func, 
        void *args ) 
{

    char msg[NEST_MESSAGESIZE];
    long listsize;
    NestReplyStatus serverReply;
    if ( ( serverReply = ReceiveNestMessage( con, msg ) ) != NEST_SUCCESS ) {
        return serverReply;
    }

        // first scan out the reply status, if ok, scan out the list size
    if ( sscanf( msg, "%d", &serverReply ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }
    if ( serverReply != NEST_SUCCESS ) {
        return serverReply;
    }
    if ( sscanf( msg, "%*d %ld", &listsize ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }

        // cut out early if no list
    if ( listsize == 0 ) {
        return NEST_SUCCESS;
    }

        // malloc size for the list, then read it in
    char *list = (char *)NestCalloc( listsize );
    if ( ! list ) return NEST_OUT_OF_RESOURCES;

    if ( NestReadn( con, (void *)list, listsize ) != listsize ) {
        return NEST_NO_CONNECTION;
    }

        // foreach token, call the user function
    char *proxy = strtok( list, " \t\n" );
    while( proxy ) {
        func( proxy, args );
        proxy = strtok( NULL, " \t\n" );
    }
    
    NestFree( list );
    return NEST_SUCCESS;

}

NestReplyStatus 
NestProxyRegister( char *nestlocation, NestConnection con )
{
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof( msg ), "%d %s", PROXY_REGISTER, nestlocation );
    nest_debug( NEST_DEBUG_PROXY,"Sending '%s' msg to register as proxy.\n", msg);
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con ); 
    //return receiveProxyList( con, func, args );
}

NestReplyStatus 
NestProxyQuery( NestConnection con, NestProxyListFunction *func, void *args  ) {
    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof( msg ), "%d", PROXY_QUERY );
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return receiveProxyList( con, func, args );
}

NestReplyStatus
NestFilesize( long &filesize, char *remotefile, NestConnection con ) {

    if ( con < 0 ) return NEST_NO_CONNECTION;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", FILESIZE_REQUEST, remotefile );
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

        // read the server's reply
    NestReplyStatus serverReply;
    if ( ( serverReply = ReceiveNestMessage( con, msg ) ) != NEST_SUCCESS ) {
        return serverReply;
    } 
    
        // first scan out the reply status, if ok, scan out the size
    if ( sscanf( msg, "%d", &serverReply ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }
    if ( serverReply != NEST_SUCCESS ) {
        return serverReply;
    }
    if ( sscanf( msg, "%*d %ld", &filesize ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }

    return serverReply;

}

int
NestReplyStatusToErrno( NestReplyStatus error ) {
    return NestToErrno( error );
}

NestReplyStatus
NestSync( NestConnection con ) {
    if ( con < 0 || ! SendSimpleRequest( con, NEST_SYNC ) ) {
        return NEST_NO_CONNECTION;
    }    
    return ReceiveNestAck( con ); 
}

NestReplyStatus
NestFormat( float &bytesavailable, float &totalbytes, NestConnection con ) {

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d", FORMAT_DISK_REQUEST );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

        // read server's reply
    NestReplyStatus serverReply;
    if ( ( serverReply = ReceiveNestMessage( con, msg ) ) != NEST_SUCCESS ) {
        return serverReply;
    }

    if (sscanf(msg, "%d %f %f", &serverReply, &bytesavailable, &totalbytes)!=3){
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }
    return serverReply;
}

NestReplyStatus
NestRemoteShutdown( NestConnection con ) {
    if ( con < 0 || ! SendSimpleRequest( con, SERVER_TERMINATE_REQUEST) ) {
        return NEST_NO_CONNECTION;
    }
    return ReceiveNestAck( con );
}

NestReplyStatus
NestAvailableSpace(float &bytesavailable, float &totalbytes,NestConnection con){

    if ( con < 0 || ! SendSimpleRequest( con, AVAILABLE_SPACE_QUERY ) ) {
        return NEST_NO_CONNECTION;
    }

    char msg[NEST_MESSAGESIZE];
    int serverReply;
    if ( ( serverReply = ReceiveNestMessage (con, msg ) != NEST_SUCCESS ) ) {
        return serverReply;
    }

    if (sscanf(msg, "%d %f %f", &serverReply, &bytesavailable, &totalbytes)!=3){
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }

    return serverReply;
}

NestReplyStatus
NestRequestAd( NestConnection con, char **ad, unsigned *adlen ) {

    if ( con < 0 || ! SendSimpleRequest( con, REQUEST_NEST_AD ) ) {
        return NEST_NO_CONNECTION;
    }

    NestReplyStatus serverReply;
    char msg[NEST_MESSAGESIZE];
    long adsize;
    if ( ( serverReply = ReceiveNestMessage( con, msg ) ) != NEST_SUCCESS ) {
        return serverReply;
    }

        // first scan out the reply status, if ok, scan out the list size
    if ( sscanf( msg, "%d", &serverReply ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }
    if ( serverReply != NEST_SUCCESS ) {
        return serverReply;
    }
    if ( sscanf( msg, "%*d %ld", &adsize ) != 1 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }

        // cut out early if no list
    if ( adsize == 0 ) {
        return NEST_SUCCESS;
    }

        // malloc size for the list, then read it in
    long safesize = adsize + 1;     // add one for null term
    *ad = (char *)NestCalloc( safesize );
    if ( ! *ad ) return NEST_OUT_OF_RESOURCES;

    if ( NestReadn( con, (void *)*ad, adsize ) != adsize ) {
        return NEST_NO_CONNECTION;
    }
    *adlen = safesize;
    return NEST_SUCCESS;

    /*
#define BUILDING_FOR_PFS 1

#if( BUILDING_FOR_PFS )
    return NEST_NOT_YET_IMPLEMENTED;
#else
    ClassAdParser *cap;
    cap = new ClassAdParser();
    *ad = cap->ParseClassAd( adc );
    return NEST_SUCCESS;
#endif
    */

}

NestReplyStatus
NestAddUser( NestConnection con, char *user, char *group, 
             char *identity) {
    char msg[NEST_MESSAGESIZE];

    if(identity)
        NestSnprintf( msg, sizeof(msg), "%d %s %s %s", ADD_NEST_USER,
                      user, group, identity );
    else
        NestSnprintf( msg, sizeof(msg), "%d %s %s", ADD_NEST_USER,
                      user, group);
        
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestDelUser( NestConnection con, char *user) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", DELETE_NEST_USER,
                    user);

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestAddGroup( NestConnection con, char *group) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", ADD_NEST_GROUP,
                    group);
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestDelGroup( NestConnection con, char *group) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", DELETE_NEST_GROUP,
                    group);

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestAddUserToGroup( NestConnection con, char *user, char *group) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %s", ADD_USER_TO_GROUP,
                    user, group);

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestDelUserFromGroup( NestConnection con, char *user, char *group) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %s", DELETE_USER_FROM_GROUP,
                    user, group);

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
GetLotHandle( NestLotHandle *handle,
              char *msg) {

    NestReplyStatus reply;
    rsv_id_t    tmp_rsv_id;

    if (!handle || !msg) {
        return NEST_BAD_PARAMETERS;
    }

    sscanf( msg, "%d %lu\n",
                 &reply,
                 &tmp_rsv_id );

    if (NestAdd_id_to_handle( handle, tmp_rsv_id ) < 0) {
        return NEST_UNKNOWN_ERROR;
    }

    return reply;
}


NestReplyStatus
NestRequestUserLot( NestLotHandle *handle,
                    NestConnection con,
                    double size_in_MB,
                    double minutes,
                    char *user,
                    char *parent_user,
                    char *parent_group,
                    NestLotHandle *parent_handle,
                    int lot_type) {

    char msg[NEST_MESSAGESIZE];
    NestReplyStatus reply;
    rsv_id_t rsv_id=0;

    if (!handle || parent_user && parent_group) {
        return NEST_BAD_PARAMETERS;
    }
    if(parent_handle)
        NestGet_id_from_handle(parent_handle, &rsv_id);
    if(!user)
        user = strdup("0");
    if(!parent_user)
        parent_user = strdup("0");
    if(!parent_group)
        parent_group = strdup("0");

    NestSnprintf( msg, sizeof(msg), "%d %f %f %d %s %s %s %lu",
                                REQUEST_USER_LOT,
                                size_in_MB,
                                minutes,
                                lot_type,
                                user,
                                parent_user,
                                parent_group,
                                rsv_id);
      
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    reply = ReceiveNestAck( con );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    reply = ReceiveNestMessage( con, msg );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    reply = GetLotHandle( handle, msg );
    return NEST_SUCCESS;
}

NestReplyStatus
NestRequestGroupLot( NestLotHandle *handle,
                     NestConnection con,
                     double size_in_MB,
                     double minutes,
                     char *group,
                     char *parent_user,
                     char *parent_group,
                     NestLotHandle *parent_handle,
                     int lot_type) {

    char msg[NEST_MESSAGESIZE];
    NestReplyStatus reply;
    rsv_id_t rsv_id=0;

    if (!handle || parent_user && parent_group) {
        return NEST_BAD_PARAMETERS;
    }

    if(parent_handle)
        NestGet_id_from_handle(parent_handle, &rsv_id);
    if(!group) {
        return NEST_BAD_PARAMETERS;
    }
    if(!parent_user)
        parent_user = strdup("0");
    if(!parent_group)
        parent_group = strdup("0");

    NestSnprintf( msg, sizeof(msg), "%d %f %f %d %s %s %s %lu",
                                REQUEST_GROUP_LOT,
                                size_in_MB,
                                minutes,
                                lot_type,
                                group,
                                parent_user,
                                parent_group,
                                rsv_id);

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    reply = ReceiveNestAck( con );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    reply = ReceiveNestMessage( con, msg );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    reply = GetLotHandle( handle, msg );
    return NEST_SUCCESS;
}

NestReplyStatus
NestGetLot ( NestLot *result, char *message) {
    if (!result || !message) {
        return NEST_BAD_PARAMETERS;
    }

    NestReplyStatus reply;
    sscanf( message, "%d %s %s %s %u %s %lf %lf %lf %lf %lf %lf %d %d %[^\n]\n",
            &reply,
            result->lot_handle,
            result->parent_lot_handle,
            result->owner,
            &result->is_user,
            result->dev,
            &result->total_size,
            &result->size,
            &result->used,
            &result->start_time,
            &result->duration,
            &result->time_left,
            &result->cost,
            &result->lot_type,
            result->lot_users);
    return reply;
}

NestReplyStatus
NestListUserLots( NestLotList **list, NestConnection con, 
                  const char *user) {
    char msg[NEST_MESSAGESIZE];
    NestReplyStatus reply;

    if(user)
        NestSnprintf( msg, sizeof(msg), "%d %s",
                                LIST_USER_LOTS,
                                user);
    else
        NestSnprintf( msg, sizeof(msg), "%d",
                                LIST_USER_LOTS);
        

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    reply = ReceiveNestAck( con );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    *list = NULL;

    do {
        NestLotList *tmp;
        reply = ReceiveNestMessage( con, msg );
        if (reply != NEST_SUCCESS) {
            return reply;
        }
        sscanf(msg, "%d", &reply);

        if (reply != NEST_SUCCESS) {
            tmp = (NestLotList *) NestMalloc (sizeof(NestLotList));
            reply = NestGetLot( &tmp->stat, msg );
            tmp->next = *list;
            *list = tmp;
        }
    } while (reply != NEST_SUCCESS);

    return NEST_SUCCESS;
}

NestReplyStatus
NestListGroupLots(NestLotList **list, NestConnection con, char *group) {
    char msg[NEST_MESSAGESIZE];
    NestReplyStatus reply;

    NestSnprintf( msg, sizeof(msg), "%d %s",
                                LIST_GROUP_LOTS,
                                group);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    reply = ReceiveNestAck( con );

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    *list = NULL;

    do {
        NestLotList *tmp;
        reply = ReceiveNestMessage( con, msg );
        if (reply != NEST_SUCCESS) {
            return reply;
        }

        sscanf(msg, "%d", &reply);

        if (reply != NEST_SUCCESS) {
            tmp = (NestLotList *) NestMalloc (sizeof(NestLotList));
            reply = NestGetLot( &tmp->stat, msg );
            tmp->next = *list;
            *list = tmp;
        }

    } while (reply != NEST_SUCCESS);

    return NEST_SUCCESS;
}

NestReplyStatus
NestGetLotStat( NestLot *result, 
            NestConnection con, 
            NestLotHandle *lot_handle) {

    char     msg[NEST_MESSAGESIZE];
    char    str_handle[NEST_MESSAGESIZE];

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf( msg, sizeof(msg), "%d %s",
                                GET_LOT_STAT,
                                str_handle);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    NestReplyStatus reply;
    reply = ReceiveNestAck( con );
    if (reply != NEST_SUCCESS) {
        return reply;
    }

    reply = ReceiveNestMessage( con, msg );
    reply = NestGetLot( result, msg );
    return NEST_SUCCESS;
}

NestReplyStatus
NestUpdateLot( NestConnection con,
           NestLotHandle *lot_handle,
           double more_size_in_MB,
           double more_minutes) {

    char     msg[NEST_MESSAGESIZE];
    char    str_handle[NEST_MESSAGESIZE];

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf( msg, sizeof(msg), "%d %s %f %f",
                                UPDATE_LOT,
                                str_handle,
                                more_size_in_MB,
                                more_minutes);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestChangeLotType(NestConnection con,
                  NestLotHandle *lot_handle,
                  int lot_type,
                  double more_minutes) {

    char     msg[NEST_MESSAGESIZE];
    char     str_handle[NEST_MESSAGESIZE];

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf( msg, sizeof(msg), "%d %s %d %f",
                                    CHANGE_LOT_T,
                                    str_handle,
                                    lot_type,
                                    more_minutes);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }            

    return ReceiveNestAck( con );
}

NestReplyStatus
NestLotMoveFile( NestConnection con,
                 const char *file,
                 NestLotHandle *source,
                 NestLotHandle *dest,
                 double move_size_in_bytes ) {

    char    msg[NEST_MESSAGESIZE];
    char    source_str[NEST_MESSAGESIZE];
    char    dest_str[NEST_MESSAGESIZE];

    // Convert the handles into passable strings
    if (NestHandle_to_string(source, source_str, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }
    if (NestHandle_to_string(dest, dest_str, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf( msg, sizeof(msg), "%d %s %s %s %f",
                  LOT_MOVE_FILE,
                  file,
                  source_str,
                  dest_str,
                  move_size_in_bytes );

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestTerminateLot( NestConnection con, 
              NestLotHandle *lot_handle) {
    char     msg[NEST_MESSAGESIZE];
    char    str_handle[NEST_MESSAGESIZE];

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf( msg, sizeof(msg), "%d %s",
                                TERMINATE_LOT,
                                str_handle);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestAddUserToLot( NestConnection con,
                  NestLotHandle *lot_handle,
                  const char *user) {
    char msg[NEST_MESSAGESIZE];
    char str_handle[NEST_NAMELEN];

    if(!lot_handle || !user) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }
    
    NestSnprintf( msg, sizeof(msg), "%d %s %s",
                                    ADD_USER_TO_LOT,
                                    str_handle,
                                    user);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION; 
    }

    return ReceiveNestAck( con );
    
}

NestReplyStatus
NestDeleteUserFromLot( NestConnection con,
                       NestLotHandle *lot_handle,
                       const char *user) {
    char msg[NEST_MESSAGESIZE];
    char str_handle[NEST_NAMELEN];

    if(!lot_handle || !user) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }
    
    NestSnprintf( msg, sizeof(msg), "%d %s %s",
                                    DELETE_USER_FROM_LOT,
                                    str_handle,
                                    user);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION; 
    }

    return ReceiveNestAck( con );
    
}

NestReplyStatus
NestAttachLotToFile( NestConnection con,
                     NestLotHandle *lot_handle,
                     const char *file) {
    char msg[NEST_MESSAGESIZE];
    char str_handle[NEST_NAMELEN];

    if(!lot_handle || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }
    
    NestSnprintf( msg, sizeof(msg), "%d %s %s",
                                    ATTACH_LOT_TO_FILE,
                                    str_handle,
                                    file);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION; 
    }

    return ReceiveNestAck( con );
    
}

NestReplyStatus
NestDetachLotFromFile( NestConnection con,
                     NestLotHandle *lot_handle,
                     const char *file) {
    char msg[NEST_MESSAGESIZE];
    char str_handle[NEST_NAMELEN];

    if(!lot_handle || !file) {
        return NEST_BAD_PARAMETERS;
    }

    if (NestHandle_to_string(lot_handle, str_handle, NEST_MESSAGESIZE) < 0) {
        return NEST_BAD_PARAMETERS;
    }
    
    NestSnprintf( msg, sizeof(msg), "%d %s %s",
                                    DETACH_LOT_FROM_FILE,
                                    str_handle,
                                    file);

    if (! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION; 
    }

    return ReceiveNestAck( con );
    
}

NestReplyStatus
GlobusAuthentication( char *user, NestConnection con, char *server ) {
    char msg[NEST_MESSAGESIZE];

        // first send a message asking to authenticate via globus
        // then actually do mutual authentication
    NestSnprintf( msg, sizeof(msg), "%d %s %d",
            CHANGE_USER_REQUEST, user, NEST_AUTH_GSS );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }
    if ( auth_globus_assert( con, server ) == 0 ) {
        return NEST_COULD_NOT_AUTHENTICATE;
    }
    
    return ReceiveNestAck( con );
}

NestReplyStatus
SimpleAuthentication( char *user, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s %d",
            CHANGE_USER_REQUEST, user, NEST_AUTH_NONE );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }
    return ReceiveNestAck( con );
}

NestReplyStatus
NestChangeUserID( char *user, NestConnection con, char *server ) {
        // first try simple authentication, then globus
    NestReplyStatus status;
    status = SimpleAuthentication( user, con ); 
    if ( status == NEST_SUCCESS ) {
        return NEST_SUCCESS;
    } else {
        status = GlobusAuthentication( user, con, server );
    }
        // first try server auth, then try user-user auth
    /*
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "Could not authenticate to %s: %s\n",
                server,
                NestErrorString( status ) );
            // now try user-user auth
        status = GlobusAuthentication( user, con, NULL );
    }
    */
    return status;
}


NestReplyStatus
NestGetTimestamp( long &timestamp, NestConnection con ) {

    if ( con < 0 || ! SendSimpleRequest( con, TIME_REQUEST ) ) {
        return NEST_NO_CONNECTION;
    }

    char msg[NEST_MESSAGESIZE];
    int serverReply;
    if ( ( serverReply = ReceiveNestMessage ( con, msg ) != NEST_SUCCESS ) ) {
        return serverReply;
    }

    if ( sscanf( msg, "%d %ld", &serverReply, &timestamp ) != 2 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }

    return serverReply;

}

NestReplyStatus
NestGetPhysicalName( char **buf, const char *file, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    NestReplyStatus serverReply;

    NestSnprintf( msg, sizeof(msg), "%d %s", PHYSNAME_REQUEST, file );
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    // Read the reply from the sender...
    if ( (serverReply = ReceiveNestAck ( con, msg )) != NEST_SUCCESS ) {
        return serverReply;
    }

    *buf = (char *) NestMalloc( NEST_MESSAGESIZE );
    if ( sscanf( msg, "%d %s", &serverReply, *(buf) ) != 2 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        NestFree( *buf );
        return NEST_NO_CONNECTION;
    }
    if ( serverReply != NEST_SUCCESS ) {
        NestFree( *buf );
    }
    return serverReply;
}

NestReplyStatus
NestGetAcl( NestConnection con,
            const char *file,
            const char *user,
            NestAclFunction *Callback,
            void *CbData,
            bool is_group) {
    char                msg[NEST_MESSAGESIZE];
    NestACL                Acl;
    NestReplyStatus        serverReply;
    int                    rights;
    int                    status;

    // User field must be valid for "group"; check it here
    if (  ( is_group ) && ( NULL == user )  ) {
        return NEST_BAD_PARAMETERS;
    }

    // Format & send the message
    NestSnprintf( msg,
                  sizeof(msg),
                  "%d %s %s",
                  is_group ? GET_GROUP_ACL : GET_ACL,
                  file,
                  user ? user : "");
    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    // Get the initial status back..
    if (  ( status = ReceiveNestAck( con )  ) != NEST_SUCCESS ) {
        return status;
    }

    // Wait for & process all replies (ACK terminates the list)
    while( ReceiveNestAck( con, msg ) != NEST_SUCCESS ) {
        if ( sscanf( msg, "%d %s %d %d",
                     &serverReply, Acl.name, &Acl.is_user, &rights ) != 4 ) {
            return ErrnoToNest( errno );
        }
        Acl.rights = rights;
        if ( Callback != NULL ) {
            Callback( &Acl, CbData );
        }
    }
    return NEST_SUCCESS;
}

NestReplyStatus
NestSetAcl( NestConnection       con,
            const char           *file,
            const char           *user,
            nest_access_rights_t access,
            bool                 is_group) {
    char            msg[NEST_MESSAGESIZE];

    NestSnprintf( msg,
                  sizeof(msg),
                  "%d %s %s %d",
                  is_group ? SET_GROUP_ACL : SET_ACL,
                  file,
                  user,
                  access );

    if ( ! SendNestMessage( con, msg ) ) {
        return NEST_NO_CONNECTION;
    }

    return ReceiveNestAck( con );
}

NestReplyStatus
NestGetPwd( char **buf, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    NestReplyStatus serverReply;

    if ( con < 0 || ! SendSimpleRequest( con, PWD_REQUEST ) ) {
        return NEST_NO_CONNECTION;
    }

    if ( ( serverReply = ReceiveNestMessage ( con, msg ) != NEST_SUCCESS ) ) {
        return serverReply;
    }

    *buf = (char *) NestMalloc( NEST_MESSAGESIZE );
    if ( sscanf( msg, "%d %s", &serverReply, *(buf) ) != 2 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        NestFree( *buf );
        return NEST_NO_CONNECTION;
    }
    if ( serverReply != NEST_SUCCESS ) {
        NestFree( *buf );
    }
    return serverReply;
}

NestReplyStatus
NestGetVersion( float *version, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    NestReplyStatus serverReply;

    if ( con < 0 || ! SendSimpleRequest( con, VERSION_REQUEST ) ) {
        return NEST_NO_CONNECTION;
    }

    if ( ( serverReply = ReceiveNestMessage ( con, msg ) != NEST_SUCCESS ) ) {
        return serverReply;
    }

    if ( sscanf( msg, "%d %f", &serverReply, version ) != 2 ) {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    }
    return serverReply;
}

NestReplyStatus
NestLink( const char *oldPath, const char *newPath, int isSymbolic,
           NestConnection con )
{
    char msg[NEST_MESSAGESIZE];
    NestSnprintf(msg, sizeof(msg), "%d %s %s %d",
				 LINK_REQUEST, oldPath, newPath, isSymbolic);
    if(!SendNestMessage(con,msg)) return NEST_NO_CONNECTION;
    return ReceiveNestAck(con,msg);
}

NestReplyStatus
NestCopy( const char *oldPath, const char *newPath, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf(msg, sizeof(msg), "%d %s %s",
				 COPY_REQUEST, oldPath, newPath);
    if(!SendNestMessage(con,msg)) return NEST_NO_CONNECTION;
    return ReceiveNestAck(con,msg);
}



void NestFormatAccessString( nest_access_rights_t access,
                             char *access_string ) {

    // Format the access string..
    *access_string++ = (access & NEST_READ_ACCESS) ? 'r' : '-';
    *access_string++ = (access & NEST_WRITE_ACCESS) ? 'w' : '-';
    *access_string++ = (access & NEST_LIST_ACCESS) ? 'l' : '-';
    *access_string++ = (access & NEST_DELETE_ACCESS) ? 'd' : '-';
    *access_string++ = (access & NEST_INSERT_ACCESS) ? 'i' : '-';
    *access_string++ = (access & NEST_MODIFYAC_ACCESS) ? 'a' : '-';
    *access_string = '\0';
}

NestReplyStatus NestParseAccessString( nest_access_rights_t *access,
                                       const char *string ) {

    char    c;

    // Check passed in pointers..
    if (  ( NULL == string )  || ( NULL == access )  ) {
        return NEST_BAD_PARAMETERS;
    }

    *access = 0;

    if (!strcasecmp(string,"none")) {
        return NEST_SUCCESS;
    }

    // Walk through the string, one char at a time
    while (  ( c = toupper (*string) ) != '\0' )
    {
        string++;
        if ( 'R' == c ) *access |= NEST_READ_ACCESS;
        else if ( 'W' == c ) *access |= NEST_WRITE_ACCESS;
        else if ( 'L' == c ) *access |= NEST_LIST_ACCESS;
        else if ( 'D' == c ) *access |= NEST_DELETE_ACCESS;
        else if ( 'I' == c ) *access |= NEST_INSERT_ACCESS;
        else if ( 'A' == c ) *access |= NEST_MODIFYAC_ACCESS;
        else return NEST_BAD_PARAMETERS;
    }

    // Done
    return NEST_SUCCESS;
}


/////////////////////////// private functions start here /////////////////////

NestReplyStatus
RequestReadFile( const char *remotefile, long offset, long &length, int con ) {

    char    msg[NEST_MESSAGESIZE];
    char verify[NEST_MESSAGESIZE];

    long end;

    NestReplyStatus serverReply;
    int  requestType;

    if ( badArgs( offset, length ) ) return NEST_BAD_PARAMETERS;

        // send request
    memset( msg, 0, NEST_MESSAGESIZE );
    NestSnprintf( msg, sizeof( msg ), "%d %s %ld %ld",
        CLIENT_RECV_FILE, remotefile, offset, length );
    if ( ! SendNestMessage( con, msg )) return NEST_NO_CONNECTION;

        // get server's answer
    if ( ( serverReply = ReceiveNestAck( con, msg ) ) != NEST_SUCCESS ) {
        return serverReply;
    }

        // parse server's answer
    if ( sscanf( msg, "%d %s %d %ld",
            &serverReply, verify, &requestType, &end ) != 4 )
    {
        fprintf( stderr, "%s:%d Bad formatted serverReply: %s.\n", 
                __FILE__, __LINE__, msg );
        return NEST_NO_CONNECTION;
    } else {
        assert( requestType == CLIENT_RECV_FILE );
        assert( strcmp( remotefile, verify ) == 0 );
    }

        // special case, user requested an offset beyond eof
    if ( offset > end ) {
        end = offset;
    }

    length = end - offset;
    return serverReply;
}

NestReplyStatus
RequestWriteFile( const char *filename, long offset, long length, int con ) {

    if ( badArgs( offset, length ) ) return NEST_BAD_PARAMETERS;

    char msg[NEST_MESSAGESIZE];
    memset( msg, 0, NEST_MESSAGESIZE );
    NestSnprintf( msg, sizeof( msg ), "%d %s %ld %ld",
                CLIENT_SEND_FILE, filename, offset, length );

        // send request
    if ( ! SendNestMessage( con, msg ) ) return NEST_NO_CONNECTION;

        // read the server's reply to see whether we were given permission
    return ReceiveNestAck( con, msg ); 
}

NestReplyStatus
NestChangeDir( const char *dir, NestConnection con ) {
    return DirOp( dir, CWD_REQUEST, con );
}

NestReplyStatus
NestMkDir( const char *dir, NestConnection con ) {
    return DirOp( dir, MKDIR_REQUEST, con );
}

NestReplyStatus
NestRmDir( const char *dir, NestConnection con ) {
    return DirOp( dir, RMDIR_REQUEST, con );
}

static void addToList( NestFileInfo *info, void *arg )
{
	list<string> *dlist = (list<string>*) arg;
	string s(info->name);
	if(!strcmp(info->name,".") || !strcmp(info->name,"..")) {
		/* do nothing */
	} else {
		dlist->push_front(s);
	}
}

NestReplyStatus
NestRmRecursive( const char *n, NestConnection con ) {
	NestReplyStatus r;
	int failures=0;
	long size;

	r = NestFilesize(size,(char*)n,con);
	if(r==NEST_SUCCESS) {
		if(size>=0) {
			r = NestRemoveFile((char*)n,con);
			if(r!=NEST_SUCCESS) failures++;
		} else {
			r = NestChangeDir((char*)n,con);
			if(r==NEST_SUCCESS) {
				list<string> dlist;
				r = NestListFiles(con,addToList,&dlist);
				if(r==NEST_SUCCESS) {
					while(!dlist.empty()) {
						string sname = dlist.front();
						dlist.pop_front();
						char *name = (char*) sname.c_str();
						r = NestRmRecursive(name,con);
						if(r!=NEST_SUCCESS) failures++;
					}
				} else {
					failures++;
				}
				r = NestChangeDir("..",con);
				if(r!=NEST_SUCCESS) {
					fprintf(stderr,"lost hold of parent directory\n");
					abort();
				}
				r = NestRmDir((char*)n,con);
				if(r!=NEST_SUCCESS) failures++;
			} else {
				failures++;
			}
		}
	} else {
		if(r==NEST_FILE_NOT_FOUND) {
			/* no problem! */
		} else {
			failures++;
		}
	}

	if(failures>0) {
		return NEST_NOT_EMPTY;
	} else {
		return NEST_SUCCESS;
	}
}

NestReplyStatus 
DirOp( const char *dir, int req_type, NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    memset( msg, 0, NEST_MESSAGESIZE );
    NestSnprintf( msg, sizeof( msg ), "%d %s",
            req_type, dir );

        // send request
    if ( ! SendNestMessage( con, msg ) ) return NEST_NO_CONNECTION;

        // read the server's reply 
    return ReceiveNestAck( con, msg ); 
}

bool
badArgs( long fileoffset, long length ) {
    return ( fileoffset < 0  
             || ( length != ENTIRE_FILE && length < 0 ) );
}

void
printBandwidthStatistics( struct timeval *request, struct timeval *start,
        struct timeval *end, long size, char *req_type )
{
    long interval;
    long latency;
    char msg[NEST_MESSAGESIZE];

    latency  = NestTimediff( request, start, __FUNCTION__ );
    interval = NestTimediff( start, end, __FUNCTION__ );
    float bw = NestBandwidth( size,  interval );
    interval = NestTimediff( request, end, __FUNCTION__ );
    float bw2 = NestBandwidth( size, interval );
    NestSnprintf( msg, sizeof(msg),
        "Cl %d bytes %ld bw %.4f lat %ld bw2 %.4f req %s\n",
        getpid(), size, bw, latency, bw2, req_type );
        // turn off client log file
    //ClientLogMessage( msg );
    //fprintf( stderr, "%.4f\t%ld\t%s\n", bw, latency, req_type );
}

int
ClientLogMessage( char * msg ) {
    int  fd;
    char filename[NEST_MESSAGESIZE];
    int  retval;
    bool writefile = true;

    if ( ! writefile ) return -1;

    NestSnprintf( filename, sizeof(filename), "Client%d.log", getpid() );
    fd = NestOpen( filename, O_WRONLY | O_CREAT | O_APPEND );
    retval = NestLogMessage( fd, msg );
    NestClose( fd );

    return retval;
}

bool
SendNestMessage( NestConnection con, const void *vptr ) {
    char msgbuf[NEST_MESSAGESIZE];

	// Copy the message to a temp buffer
	memset( msgbuf, 0, sizeof( msgbuf ) );
	strncpy( msgbuf, ( const char *) vptr, NEST_MESSAGESIZE - 2 );

	// Debugging
    nest_debug( NEST_DEBUG_CHIRP, "Sending message '%s' on %d for %d\n",
				msgbuf, con, getpid() );

	// Get the length of the original message
	int msg_length = NestStrlen( (const char *) vptr );
    if ( msg_length > NEST_MESSAGESIZE ) {
        nest_die( "Gonna have to do more work in SendNestMessage\n" );
    }

	// Add the '\n' to it
	strcat( msgbuf, "\n" );

	// And, send it out
    ssize_t sent = NestWriten( con, msgbuf, NEST_MESSAGESIZE );
    if (sent != NEST_MESSAGESIZE) {
        if ( errno != EPIPE) nest_perror( "NestWriten" );
        return false;
    }

	// Done
    return true;
}

NestReplyStatus
ReceiveNestPartialMessage( NestConnection con, void *vptr, int size,
                            int *read){
    if ( ( *read = NestRead( con, vptr, size ) ) <= 0 ) {
        return NEST_NO_CONNECTION;
    } else {
        return ( *read == size ) ? NEST_SUCCESS : NEST_PARTIAL_MESSAGE;
    }
}

// receive a message over the connection
NestReplyStatus
ReceiveNestMessage( NestConnection con, void *vptr ) {
    NestReplyStatus status =
        ( NestReadn( con, vptr, NEST_MESSAGESIZE ) == NEST_MESSAGESIZE )
        ? NEST_SUCCESS
        : NEST_NO_CONNECTION;
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_CHIRP, "Unable to receive message\n" );
    } else {
        nest_debug( NEST_DEBUG_CHIRP, "Received message %s\n", (char*)vptr );
    }
    return status;
}

// request a directory list
NestReplyStatus
NestListFiles(NestConnection con, NestListDirFunction *foreachfile, void *args){
    char msg[NEST_MESSAGESIZE];
    NestFileInfo    finfo;

    if ( con < 0 || ! SendSimpleRequest( con, DIR_LIST_REQUEST ) ) {
        return NEST_NO_CONNECTION;
    }

    while( ReceiveNestAck( con, msg ) != NEST_SUCCESS ) {
        if ( NestString_to_NestFileInfo( &finfo, msg ) != 3 ) {
            return ErrnoToNest( errno );
        }
        if ( foreachfile != NULL ) {
            foreachfile( &finfo, args );
        }
    }
    return NEST_SUCCESS;
}


// these types of requests are only implemented currently in NEST_PROTOCOL
// TODO this should return NestReplyStatus
bool 
SendSimpleRequest( NestConnection con, int request ) {
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d", request );
    return SendNestMessage( con, msg );
}

bool
PrintFromSocket( NestConnection con ) {
    return PrintFromSocket( con, stdout );
}

// see the comment in sendSimpleRequest, this function is its counterpart
bool
PrintFromSocket( NestConnection con, FILE *out ) {
    char recvline[NEST_MESSAGESIZE];
    while( ReceiveNestAck( con, recvline ) != NEST_SUCCESS ) {
        if ( fputs( recvline, out ) == EOF ) {
            nest_perror( "fputs error\n" );
        }
    }
    return true;
}

NestReplyStatus
ReceiveNestAck( NestConnection con ) {
    char msg[NEST_MESSAGESIZE];
    return ReceiveNestAck( con, msg );
}

NestReplyStatus
ReceiveNestAck( NestConnection con, void *vptr ) {
    NestReplyStatus status;
    nest_debug( NEST_DEBUG_CHIRP, "Receiving chirp ack from sd %d.\n", con );
    //fprintf( stderr, "Waiting for ack\n" );
    if ( ( status = ReceiveNestMessage( con, vptr ) ) != NEST_SUCCESS ) {
        return status;
    }
    if ( sscanf( (char*)vptr, "%d", &status ) != 1 ) {
        return NEST_NO_CONNECTION;
    }
    return status;
}

// first send the request to make the other proxy ready to receive
// then send the actual index
NestReplyStatus 
NestSendIndex( NestConnection con, DynamicLard *index ) {

    if ( ! SendSimpleRequest( con, PROXY_INDEX_PUSH ) ) {
        return NEST_NO_CONNECTION;
    }

    int stat = index->send( con );
    if ( stat < 0 ) {
        return NEST_NO_CONNECTION;
    } else {
        return NEST_SUCCESS;
    }
}

NestReplyStatus 
NestReceiveIndex( NestConnection con, DynamicLard *index,
                                  StringToServerFunc *func,
                                  void *args )
{
    int stat = index->receive( con, func, args );
    if ( stat < 0 ) {
        return NEST_NO_CONNECTION;
    } else {
        return NEST_SUCCESS;
    }
}
