/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_storage_manager.h"
#include "nest_protocol_manager.h"
#include "nest_protocol_filesystem.h"
#include "nest_config.h"

#include "nest_storage_fd_cache.h"

// the flat directory space where the server stores its files

static char*            rootDir;

double                  nestStorageManager::max_lot_duration =
                                            INFINITE_LOT_DURATION;
double                  nestStorageManager::default_lot_duration =
                                            DEFAULT_LOT_DURATION;
double                  nestStorageManager::storage_space = -1;
mode_t                  nestStorageManager::file_create_mode = FILE_MODE;

char                    *nestStorageManager::userlist = "";
char                    *nestStorageManager::denylist = "";

NestIdentityManager     *nestStorageManager::identity_manager = NULL;
NestLotManager          *nestStorageManager::lot_manager = NULL;
NestFileCatalog         *nestStorageManager::file_catalog = NULL;
NestCollectionManager *nestStorageManager::collection_manager = NULL;

ClassAdCollection       *nestStorageManager::Resources = NULL;
ClassAdParser           nestStorageManager::parser;
ClassAdUnParser         nestStorageManager::unparser;
LocalCollectionQuery    nestStorageManager::query;

    // used by NFS
struct nest_hash_table *     nestStorageManager::fh_hash;

    // cache fds for open performance
    // however, right now it doesn't work correctly bec it can give out
    // the same fd multiple times to a user who has multiple connections.
    // since these fds share a seek pointer, they can get screwy.
    // change this:
    // old version on open:
    //  duped the fd and gave out a copy (pointers were shared)
    // new version:
    //  open give away the cached copy, retain it at close
nestFdCache *fd_cache = NULL;
int fd_cache_size = 0;


void
nestStorageManager::writeAd( nestProtocol *dest ) {
    char buffer[NEST_MESSAGESIZE];

    long long int capacity, total;
    if ( bytesAvailable(IS_USER, &capacity, &total ) 
            == NEST_SUCCESS )
    {
        NestSnprintf( buffer, NEST_MESSAGESIZE, "Disk = %lld\nFree = %lld\n", 
                total, capacity);
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }

        // now lots?
    if(nestStorageManager::is_lots_enabled())
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = true\n");
    else
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = false\n");
    dest->Sendn( buffer, NestStrlen( buffer ) );
}

int
nestStorageManager::init( ) {

    bool recovery = true;

    const char *StorageDir = NULL;
    rootDir = NULL;

        // needed for NFS
    fh_hash = nest_hash_table_create( 1037, NULL );
    assert( fh_hash != NULL );

    /*
        Must have all these inits that initializes various global
        variables before any of the ->initialize() s are called!
    */

    const char *value;

    if ( GetConfigValue(Config_FdCacheSize, &value) && *value != 0) {
        fd_cache = new nestFdCache( (unsigned)atoi( value ) );
    }

    if ( GetConfigValue(Config_StorageDir, &value) && *value != 0) {
        StorageDir = value;
    } else {
        nest_die( "Must have StorageDir defined.\n" );
    }

    if ( GetConfigValue(Config_DisableLots, NULL)) {
        NestEnableLots = false;
    }

    if ( GetConfigValue(Config_MaxLotDuration, &value) && *value != 0) {
        sscanf(value, "%lf", &nestStorageManager::max_lot_duration);
    }

    if ( GetConfigValue(Config_DefaultLotDuration, &value) && *value != 0) {
        sscanf(value, "%lf", &nestStorageManager::default_lot_duration);
    }

    if ( GetConfigValue(Config_StorageSpace, &value) && *value != 0) {
        sscanf(value, "%lf", &nestStorageManager::storage_space);
    }

    if ( GetConfigValue(Config_UserList, &value) && *value != 0) {
        char *identity_list = NestGetModifiedIdentity(value);
        nestStorageManager::userlist = NestStrdup(identity_list);
    }
    if ( GetConfigValue(Config_DenyList, &value) && *value != 0) {
        char *identity_list = NestGetModifiedIdentity(value);
        nestStorageManager::denylist = NestStrdup(identity_list);
    }

    if ( GetConfigValue(Config_AdminUser, &value) && *value != 0) {
        NestAdminUser = NestGetModifiedIdentity(value);
    }

    if ( GetConfigValue(Config_SystemUser, &value) && *value != 0) {
        NestSystemUser = value;
    }

    if ( GetConfigValue(Config_AnonymousUser, &value) && *value != 0) {
        NestAnonymousUser = value;
    }

    if ( GetConfigValue(Config_PasswordFile, &value) && *value != 0) {
        SystemPwdFile = value;
    }

    if ( GetConfigValue(Config_GroupFile, &value) && *value != 0) {
        SystemGrpFile = value;
    }

    if ( GetConfigValue(Config_GridMapFile, &value) && *value != 0) {
        GridMapFile = value;
    }

    if ( GetConfigValue(Config_MetaDir, &value) && *value != 0) {
        MetaDir = value;
    }

    if ( GetConfigValue(Config_ActualGroup, &value) && *value != 0) {
        ActualGroup = value;
    }

    if ( GetConfigValue(Config_FileCreateMode, &value) && *value != 0) {
        nestStorageManager::file_create_mode = (unsigned)strtol( value, NULL, 8 );
    }

    AdminUid = NestGetuid();
    if (strcmp(NestAdminUser, "")) {
        user_id_t   uid;
        if (NestGet_uid( &uid, NestAdminUser, SystemPwdFile ) >= 0) {
            AdminUid = uid;
        }
    }

    ActualGid = NestGetgid();
    if (strcmp(ActualGroup, "")) {
        group_ent   ge;
        if (Nestgetgrpnam( ActualGroup, &ge, SystemGrpFile ) >= 0) {
            ActualGid = ge.grp_gid;
        }
    }

    // make and move into the storage (root) dir
    nest_debug( NEST_DEBUG_STORAGE, "StorageDir is %s\n", StorageDir );
    if ( NestMkdir( StorageDir, DIR_MODE ) < 0 ) {
        if ( errno != EEXIST ) {
            nest_die( "mkdir %s: %s\n", StorageDir, strerror( errno ) );
        }
    }
    NestChmod( StorageDir, 0777 );
    NestChown( StorageDir, AdminUid, ActualGid );
    NestChdir( StorageDir );

        // remember what the rootDir is
    char buffer[NEST_MAXPATHLEN];
    if ( ( NestGetcwd( buffer, NEST_MAXPATHLEN ) ) != NEST_SUCCESS ) {
        nest_die( "getcwd failed" );
    }
    rootDir = NestStrdup( buffer );
    nest_debug( NEST_DEBUG_ALWAYS, "Storage Manager's cwd is %s\n", rootDir );

        // create our meta data dir
    nest_debug( NEST_DEBUG_STORAGE, "MetaDir is %s\n", MetaDir );
    if ( NestMkdir( MetaDir, DIR_MODE ) < 0 ) {
        if ( errno != EEXIST ) {
            nest_perror( "mkdir" );
            nest_die( "mkdir failed" );
        }
    }
    NestChmod( MetaDir, DIR_MODE );
    NestChown( MetaDir, AdminUid, ActualGid );

    //  lots won't work unless we are run as root
    /*  This is not true anymore! 
    if ( NestGetuid() != 0 ) {
        NestEnableLots = false;
    }
    */

    //Resources = new ClassAdCollectionServer();
    ClassAdCollection *r = new ClassAdCollection();
    Resources = r; 
    if ( Resources == NULL ) {
        nest_die("Couldn't instantiate a ClassAdCollection!\n");
    }
    query.Bind(Resources);

    string  log_file = "";
    log_file = log_file + MetaDir + "/" + CLASSAD_LOGFILE;

    nest_stat   tmp_stat_buf;

    if (NestStat( log_file.c_str(), &tmp_stat_buf ) < 0) {
        recovery = false;
    }

    // see whether required system-files are present!

    if (NestStat( SystemPwdFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the password file at %s\n",
                 SystemPwdFile);
    }

    if (NestStat( SystemGrpFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the group file at %s\n",
                 SystemGrpFile);
    }

    if (NestStat( GridMapFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the grid-mapfile at %s\n",
                 GridMapFile);
    }

    if (!Resources->InitializeFromLog(log_file, "", "" )) {
		nest_debug( NEST_DEBUG_ALWAYS, "ClassAd error: %s\n",
					CondorErrMsg.c_str() );
        nest_die("Couldn't recover ClassAdCollection from log!\n");
    }
    NestChmod( log_file.c_str(), FILE_MODE );
    NestChown( log_file.c_str(), AdminUid, ActualGid );

    NestReplyStatus reply;
    identity_manager = new NestIdentityManager(userlist,
                                               denylist);
    if (!identity_manager) {
        nest_die("Couldn't instantiate NestIdentityManager!\n");
    }

    if (recovery) {
        reply = identity_manager->recover(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_die( "IdentityManager 'recover' failed! - %s\n",
                        NestErrorString(reply) );
        }
    } else {
        reply = identity_manager->initialize(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "IdentityManager 'initialize' failed! - %s\n",
                        NestErrorString(reply) );
            nest_die( "IdentityManager 'initialize' failed!\n" );
        }
    }

    lot_manager = new NestLotManager(rootDir,
                                     storage_space,
                                     max_lot_duration);
    if (!lot_manager) {
        nest_die("Couldn't instantiate NestLotManager!\n");
    }

    /*
        dont bother about lot_manager::init or recovery is its disabled!
    */
    if (NestEnableLots) {
        lot_manager->init();
        if (recovery) {
            if ( (reply = lot_manager->recover(Resources)) != NEST_SUCCESS ) {
                nest_debug( NEST_DEBUG_ALWAYS,
                            "LotManager 'recover' failed - %s\n",
                            NestErrorString(reply) );
                nest_warn( __FILE__, __FUNCTION__,
                        "No LotManager => disabling lots\n" );
                if (reply == NEST_NO_QUOTA_ENFORCEMENT) {
                    nest_warn( __FILE__, __FUNCTION__,
                            "The quota system has been disabled.\n"
                              "Enable quotas using \"$/sbin/quotaon -a\" "
                              "to enable lots!\n" );
                }
                nest_debug( NEST_DEBUG_ALWAYS,
                            "Error returned is %s\n", NestErrorString(reply));
                NestEnableLots = false;
            }
        } else {
            if ( (reply = lot_manager->initialize(Resources))
                        != NEST_SUCCESS ) {
                nest_debug( NEST_DEBUG_ALWAYS,
                            "LotManager 'initialize' failed - %s\n",
                            NestErrorString(reply) );
                nest_warn( __FILE__, __FUNCTION__,
                        "No LotManager => disabling lots\n" );
                if (reply == NEST_NO_QUOTA_ENFORCEMENT) {
                    nest_warn( __FILE__, __FUNCTION__,
                            "The quota system has been disabled.\n"
                            "Enable quotas using \"$/sbin/quotaon -a\" "
                            "to enable lots!\n" );
                }
                NestEnableLots = false;
            }
        }
    }

    file_catalog = new NestFileCatalog();
    if (!file_catalog) {
        nest_die("Couldn't instantiate NestFileCatalog!\n");
    }

    if (recovery) {
        reply = file_catalog->recover(Resources, rootDir);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "FileCatalog 'recover' failed - %s\n",
                        NestErrorString(reply) );
            nest_die( "FileCatalog 'recover' failed!\n" );
        }
    } else {
        reply = file_catalog->initialize(Resources, rootDir);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "FileCatalog 'initialize' failed - %s\n",
                        NestErrorString(reply) );
            nest_die( "FileCatalog 'initialize' failed!\n" );
        }
    }

	// Create & initialize the collection manager.
    collection_manager = new NestCollectionManager( );
    if (!collection_manager) {
        nest_die("Couldn't instantiate NestCollectionManager!\n");
    }
    if (recovery) {
        reply = collection_manager->recover(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_die( "CollectionManager 'recover' failed! - %s\n",
                        NestErrorString(reply) );
        }
    } else {
        reply = collection_manager->initialize(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "CollectionManager 'initialize' failed! - %s\n",
                        NestErrorString(reply) );
            nest_die( "CollectionManager 'initialize' failed!\n" );
        }
    }

    reply = initialize();
    if ( reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "StorageManager 'initialize' failed - %s\n",
                     NestErrorString(reply));
        nest_die( "StorageManager 'initialize' failed!\n" );
    }

    return 0;
}

NestReplyStatus
nestStorageManager::initialize() {

    int     sync_status = synchronize( );
    if ( sync_status < 0 ) {
        nest_debug( NEST_DEBUG_ALWAYS, "Synchronization error\n" );
        return sync_status;
    }

    //  Got to take care of aborted trasfers in progress

    string  constraint = "other.TransferInProgress == true ";
    ExprTree *constraint_tree = parser.ParseExpression( constraint );
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }
    ViewName    file_view = NEST_FILE_CATALOG_VIEW;

    query.Query( file_view, constraint_tree );
    query.ToFirst();

    string key;
    if (!query.Current(key)) {
        return NEST_SUCCESS;
    }

    ClassAd             *file_ad;
    nestProtocol        *storage_con;
    string              str_user,
                        str_group,
                        logical_filename;
    char                filename[NEST_MAXPATHLEN];
    bool                result;
    int                 count = 0;

    do {
        file_ad = Resources->GetClassAd( key );

        if (!file_ad) {
            continue;
        }

        result  = file_ad->EvaluateAttrString( "OwnerUser", str_user );
        result &= file_ad->EvaluateAttrString( "OwnerGroup", str_group );
        result &= file_ad->EvaluateAttrString( "LogicalName",
                                               logical_filename );

        if (!result) {
            continue;
        }

        file_catalog->get_reverse_logical_name( filename,
                                                logical_filename.c_str() );

        nest_debug( NEST_DEBUG_STORAGE,
                    "Transfer to file %s owned by %s was in progress during "
                    "the previous crash\n"
                    "Fixing it ...\n",
                    filename,
                    str_user.c_str() );

        NestReplyStatus status;
        status = open( &storage_con,
                       str_user.c_str(),
                       AdminUid,
                       str_group.c_str(),
                       filename,
                       O_WRONLY,
                       0,
                       -1,
                       "~/",
                       0 );
        if ( NEST_SUCCESS == status ) {
            close ( storage_con );

            delete storage_con;
        }
        storage_con = NULL;

        count ++;

    } while (query.Next(key));

    nest_debug ( NEST_DEBUG_STORAGE,
                 "%d transfers were in progress during the previous crash!\n",
                 count );
    return NEST_SUCCESS;
}

nestProtocol *
nestStorageManager::openMetaFile( char *file,
                                  int flags ) {

        // move into the meta dir
    if ( chMetaDir() < 0 ) return NULL;

        // open the file
    int fd = NestOpen( file, flags );
    NestChmod( file, FILE_MODE );
    NestChown( file, AdminUid, ActualGid );

        // move back to the rootDir
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }

        // wrap the descriptor in a protocol object
    return ( fd < 0 ) ? NULL : new nestProtocolFilesystem( fd );
}


NestReplyStatus
nestStorageManager::renameMetaFile( char *oldpath,
                                    char *newpath) {

    if (!oldpath || !newpath) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return ErrnoToNest( errno );
    }

    NestRename( oldpath, newpath );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::statMetaFile( char *file,
                                  nest_stat *st) {

    if (!file) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return NEST_UNKNOWN_ERROR;
    }

    NestLstat( file, st );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::unlinkMetaFile( char *file ) {

    if (!file) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return NEST_UNKNOWN_ERROR;
    }

    NestUnlink( (const char *)file );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}


long
nestStorageManager::metaFilelength( char *file ) {
        // move into meta dir
        // get the length, move back to rootDir, return length
    if ( chMetaDir() < 0 ) return -1;
    long len = NestFilelength( file );
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }
    return len;
}

char *
nestStorageManager::metaFilepath( char *file ) {
    static char path[NEST_MESSAGESIZE];
    memset( path, 0, NEST_MESSAGESIZE );

        // make sure file exists
    if ( metaFilelength( file ) < 0 ) return NULL;

        // move into meta dir and read cwd
    if ( chMetaDir() < 0 ) return NULL;
    NestReplyStatus status = NestGetcwd( path, NEST_MESSAGESIZE );
    if ( status != NEST_SUCCESS ) {
        chRootDir();
        return NULL;
    }

        // move back to root dir
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }

        // add a slash and the filename
    NestSnprintf( &(path[NestStrlen(path)]), 
            NEST_MESSAGESIZE - NestStrlen(path),
            "/%s", file );
    return path;
}

char *
nestStorageManager::getInitialClientDir() {
    return "~/";
}

/*
    For file management, access control is done here

    For user management, access control is done here

    For group management, access control is done inside the identity_manager.

    For Lot management, access control is done part here and part
                        inside the lot_manager
*/


NestReplyStatus
nestStorageManager::open( nestProtocol **connection,
                          const char *user,
                          user_id_t  uid,
                          const char *group,
                          const char *file,
                          int flags,
                          off_t offset,
                          ssize_t filesize,
                          const char *cwd,
                          int link_count,
                          int importance ) 
{
        // other variables
    NestReplyStatus reply;
	nestProtocolFilesystem *result;
    char    logical_filename[NEST_MAXPATHLEN];
    int     fd;
    off_t   filesize_bytes,
            more_bytes;
    bool    new_file_created = false;
    string  filename;
    char    fdkey[NEST_MAXPATHLEN * 2];
    bool    system_file = false;
    
    nest_debug( NEST_DEBUG_STORAGE,
                "opening %s - user %s - flags %d\n",
                file,
                user,
                flags );

    reply = expand_path(filename,
                        user,
                        cwd,
                        file);
    NEST_REPLY_CHECK(reply);

    /*
        ignore return value - as while creating new files
        this would return NEST_FILE_NOT_FOUND
    */
    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    system_file = file_catalog->is_system_file(logical_filename);
    
        // see if we have this fd already cached
        // make sure to expand_path and get_logical_name first
        // if the fd is cached and we skip down, these must be set
    int snp_ret = NestSnprintf( fdkey, NEST_MAXPATHLEN * 2, "%s.%s.%d",
            user, filename.c_str(), flags );
    if ( snp_ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__,
                "Unable to cache fd for %s.  Path is too long.\n",
                filename.c_str() );
    } else {
        if ( fd_cache && ( fd = fd_cache->lookup( fdkey ) ) > 0 ) {
            nest_debug( NEST_DEBUG_STORAGE, "Using cached fd %d for %s\n",
                fd, fdkey );
            goto USE_CACHE_FD;
        }
    }

    /*
        it should be "O_RDONLY ==" and not "O_RDONLY &"-
        stupid O_RDONLY is defined as '0'
        took me 3 hours of debugging to fix this ://
    */

    if (flags == O_RDONLY) {
        if (file_catalog->is_file_present (logical_filename)) {
            if (!file_catalog->is_access_allowed(logical_filename,
                                                 user,
                                                 NEST_READ_ACCESS)) {
                return NEST_NONPRIVELEDGED_USER;
            }
        } else {
            if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                          user,
                                                          NEST_LIST_ACCESS)) {
                return NEST_FILE_NOT_FOUND;
            }
            return NEST_NONPRIVELEDGED_USER;
        }
    }

    if (flags & O_CREAT) {

        if (!file_catalog->is_access_allowed_in_parent(logical_filename,
                                                       user,
                                                       NEST_INSERT_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }

        if(!system_file) {
            reply = file_catalog->add_entry(logical_filename,
                                            user,
                                            NEST_FILE);

            if (reply != NEST_SUCCESS &&
                reply != NEST_FILE_EXISTS) {
                return reply;
            }

            new_file_created =  (reply == NEST_SUCCESS);
        } else 
            new_file_created = true;
    }

    if (flags & O_WRONLY) {
        if (file_catalog->is_file_present (logical_filename)) {
            if (!file_catalog->is_access_allowed(logical_filename,
                                                 user,
                                                 NEST_WRITE_ACCESS)) {
                return NEST_NONPRIVELEDGED_USER;
            }
        }
        else {
            if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                          user,
                                                          NEST_WRITE_ACCESS)) {
                return NEST_FILE_NOT_FOUND;
            }
            return NEST_NONPRIVELEDGED_USER;
        }
    }

    if (flags & O_RDWR)  {
        if (file_catalog->is_file_present (logical_filename)) {
            if (!file_catalog->is_access_allowed(logical_filename,
                                                 user,
                                                 NEST_READ_ACCESS |
                                                 NEST_WRITE_ACCESS)) {
                return NEST_NONPRIVELEDGED_USER;
            }
        }
        else {
            if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                          user,
                                                          NEST_READ_ACCESS |
                                                          NEST_WRITE_ACCESS)) {
                return NEST_FILE_NOT_FOUND;
            }
            return NEST_NONPRIVELEDGED_USER;
        }
    }

    char    physical_filename[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_filename);

    if ( reply != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while translating "
                    "logical name to physical name!\n" );
        return reply;
    }

    nest_obj_type_t type;
    reply = file_catalog->get_type(&type, logical_filename);

    if (reply != NEST_SUCCESS) {
        goto delete_and_out;
    }

    if ((type == NEST_LINK) && !system_file) {
        /*
            sym-link ==> need to act accordingly
        */

        if (link_count++ > NEST_LINK_MAXLOOP) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Too many levels of symbolic links!\n" );
            return NEST_TOO_MANY_LINKS;
        }

        return nestStorageManager::open(connection,
                                        user,
                                        uid,
                                        group,
                                        physical_filename,
                                        flags,
                                        offset,
                                        filesize,
                                        cwd,
                                        link_count,
                                        importance);
    }

    if ((type != NEST_FILE) &&
        (type != NEST_LINK)) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to open a non-file object!\n" );
        reply = NEST_BAD_PARAMETERS;
        goto delete_and_out;
    }

    fd = NestOpen( physical_filename, flags, nestStorageManager::file_create_mode );
    if ( fd < 0 ) {
        reply = ErrnoToNest(errno);
        goto delete_and_out;
    }

USE_CACHE_FD:
    *connection = new nestProtocolFilesystem( fd, (off_t)offset );
	result = (nestProtocolFilesystem*)*connection;

    if (! result ) {
        nest_debug( NEST_DEBUG_STORAGE,
            "Couldn't instantiate a nestProtocolFilesystem in open!\n" );
        reply = NEST_OUT_OF_RESOURCES;
        goto close_and_out;
    }

    //  adding the open_flags
    result->set_open_flags(flags);

        // remember the hash key so we can cache it at close and reuse it
    result->set_fdkey( fdkey );

    if (flags == O_RDONLY) {
        return NEST_SUCCESS;
    }

    if (new_file_created) {
        NestChmod( physical_filename, nestStorageManager::file_create_mode );
        NestChown( physical_filename, uid, ActualGid );
    }

    //  adding the logical_filename & physical_filename to the *result
    result->set_filenames(logical_filename, physical_filename);

    //  adding the open_filesize
    reply = file_catalog->get_size( &filesize_bytes,
                                    logical_filename );
    if (reply != NEST_SUCCESS) {
        goto free_close_and_out;
    }

    more_bytes = filesize >= 0 ? (filesize + offset - filesize_bytes)
                              : -1;

    long long int avail_bytes;
    if (NestEnableLots && !system_file) {
        /*          
            We have to do this for file transfers that overwrite a file 
            but do not transfer the size of the file 
        */
        if(filesize == ENTIRE_FILE) {
            reply = resetAllocation(file, user, cwd, offset);
            if (reply != NEST_SUCCESS) {
                goto free_close_and_out;
            }
        }

        avail_bytes = (long long int)
                      (lot_manager->available_lot_space(user,logical_filename) *
                       1024 * 1024);

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail user_lot_space(%s) = %lld & filesize = %ld\n",
                    user,
                    avail_bytes,
                    more_bytes );
	
        if ((avail_bytes == 0) && (more_bytes > 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available Lot space for user %s is 0 bytes!\n",
                        user );
            reply = NEST_LOT_NOT_ENOUGH;
            goto free_close_and_out;
        }

        if ((more_bytes > 0) && (avail_bytes < more_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the user's Lots!\n");
            reply = NEST_LOT_NOT_ENOUGH;
            goto free_close_and_out;
        }
	
	/* Zero byte files are not attached to any lot and hence do not need a lot. 
	 */
        if (more_bytes > 0) {
            double  more_bytes_MB = more_bytes;
            more_bytes_MB /= (1024 * 1024);

            reply = lot_manager->update_lot_usage(logical_filename,
                                          more_bytes_MB);
            if(reply != NEST_SUCCESS) {
                goto free_close_and_out;
            }

            filesize_bytes += more_bytes;
            file_catalog->set_size( logical_filename,
                                    filesize_bytes );
        }
    } else {
        long long int total_bytes;
        if ( (reply = bytesAvailable( IS_USER, &avail_bytes, &total_bytes )) !=
                NEST_SUCCESS ) 
        {
            goto free_close_and_out; 
        }

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail device_space = %lld & filesize = %ld\n",
                    avail_bytes,
                    more_bytes );

        if ((avail_bytes == 0) && (more_bytes != 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available disk space is 0 bytes!\n" );
            reply = NEST_DISK_FULL;
            goto free_close_and_out;
        }

        if ((more_bytes > 0) && (avail_bytes < more_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the device!\n");
            reply = NEST_DISK_FULL;
            goto free_close_and_out;
        }
    }

    nest_debug( NEST_DEBUG_STORAGE,
                "Setting open_filesize %ld in nestStorageManager::open()\n",
                filesize_bytes );
    result->set_open_filesize(filesize_bytes);

    if (!NestEnableLots) {
        return NEST_SUCCESS;
    }


    //  adding user and group IDs
    nest_debug( NEST_DEBUG_STORAGE,
                "Setting ids in nestStorageManager::open() UID %d GID %d\n",
                uid,
                ActualGid );
    result->set_ids( uid, ActualGid );

    /*
        Its a PUT -
        Lots are enabled -
        (*result) is all set -

        Turn on the "TransferInProgress" flag
    */

    if(!system_file) {
        reply = file_catalog->set_transfer_flag( logical_filename,
                                                 true );
        if (reply != NEST_SUCCESS) {
            goto free_close_and_out;
        }
    }

    return NEST_SUCCESS;

free_close_and_out:
    delete (*connection);
    *connection = NULL;

close_and_out:
    NestClose(fd);

delete_and_out:
    if (new_file_created) {
        if(!system_file)
            file_catalog->delete_entry( logical_filename );
        NestUnlink(physical_filename);
    }

    return reply;
}

char *
nestStorageManager::filehandleToPath( NestFilehandle *uniquifier ) {
    char *path = (char *) nest_hash_table_lookup( fh_hash, (char*)uniquifier ); 
    nest_debug( NEST_DEBUG_NFS, "found hash of %d to %s\n", *uniquifier, path );
    return path;
}

NestReplyStatus
nestStorageManager::pathToFilehandle(   const char *user,
                                        const char *group,
                                        const char *file,
                                        const char *cwd,
                                        NestFilehandle    *uniquifier,
                                        int        link_count ) 
{

        // this is no longer used
    /*
    nest_debug( NEST_DEBUG_STORAGE,
                "opening %s - user %s \n",
                file,
                user );

    NestReplyStatus reply;
    char    logical_filename[NEST_MAXPATHLEN];
    string  filename;

    *uniquifier = -1;

    reply = expand_path(filename,
                        user,
                        cwd,
                        file);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);

    NEST_REPLY_CHECK(reply);

    if (file_catalog->is_file_present (logical_filename)) {
        if (!file_catalog->is_access_allowed(logical_filename,
                                             user,
                                             NEST_READ_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                      user,
                                                      NEST_LIST_ACCESS)) {
            return NEST_FILE_NOT_FOUND;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    char    physical_filename[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_filename);

    if ( reply != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while translating "
                    "logical name to physical name!\n" );
        return reply;
    }

    nest_obj_type_t type;
    reply = file_catalog->get_type(&type, logical_filename);

    if (reply != NEST_SUCCESS) {
        return reply;
    }

    if ((type == NEST_LINK) &&
        !file_catalog->is_system_file(logical_filename)) 
    {
            // sym-link ==> need to act accordingly (recurse)

        if (link_count++ > NEST_LINK_MAXLOOP) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Too many levels of symbolic links!\n" );
            return NEST_TOO_MANY_LINKS;
        }

        return nestStorageManager::pathToFilehandle(user,
                                               group,
                                               physical_filename,
                                               cwd,
                                               uniquifier, 
                                               link_count);
    }

    if ((type != NEST_FILE) &&
        (type != NEST_LINK) &&
        (type != NEST_DIR) ) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to open a non-file object!\n" );
        reply = NEST_BAD_PARAMETERS;
        return reply;
    }

    struct stat stat_buf;
    if ( NestLstat(physical_filename, &stat_buf ) < 0) {
        reply = ErrnoToNest(errno);
        return reply;
    } else {
        *uniquifier = stat_buf.st_ino;
    }

        // if we make it here, we succeeded in the lookup, cache the result
    cache_uniquifier( logical_filename, uniquifier ); 

    return reply;
    */
    assert( 0 );
    return NEST_NOT_YET_IMPLEMENTED;
}

void
nestStorageManager::cache_uniquifier(   string &filename,
                                        NestFilehandle *uniquifier,
                                        nest_obj_type_t type ) 
{

        // if the file is a dir, add a slash onto the end
        // this won't work for symbolic links . . .
    if ( type == NEST_DIR ) {
        nest_debug( NEST_DEBUG_NFS, "adding slash to end of lookup dir\n" );
        filename += '/';
    }
    if (nest_hash_table_lookup( fh_hash, (char*)uniquifier) == NULL) {
        nest_hash_table_insert( fh_hash, (char*)uniquifier, 
                NestStrdup(filename.c_str()) );
        nest_debug( NEST_DEBUG_NFS, "adding hash of %d to %s (ht size %d)\n", 
            *uniquifier, filename.c_str(), nest_hash_table_size( fh_hash ) );
    } else {
        nest_debug( NEST_DEBUG_NFS, "not inserting a duplicate hash for %d\n", 
            *uniquifier, filename.c_str() );
    }
}

// contract
// storageConnection cannot be destroyed until _after_ calling this function
NestReplyStatus
nestStorageManager::close( nestProtocol *connection,
                           nestClient *client ) 
{
    nestRequest     *req;
    ssize_t         filesize_bytes = 0; 
	nestProtocolStorage *storageConnection = NULL;

    if (!connection) {
        return NEST_BAD_PARAMETERS;
    }
    
    if(!client) {
        return NEST_BAD_PARAMETERS;
    }

	if ( connection->getType() != STORAGE_PROTOCOL ) {
		nest_die( "Weirdness in nestStorageManager\n" );
	} else {
		storageConnection = (nestProtocolStorage *)connection;
	}

    req = client->getProxyRequest();
    if(!req) {
         return NEST_BAD_PARAMETERS;
    }
    filesize_bytes = (ssize_t) req->getEnd();
    
    int open_flags,
        fd = storageConnection->getSocket();

    char *fdkey = storageConnection->get_fdkey();
    if ( fd_cache && fdkey ) {
        if( fd_cache->stow( fd, fdkey ) ) {
            storageConnection->set_fd_cache_status( true );   
        } else {
            nest_warn( __FILE__, __FUNCTION__,
                    "unable to cache %s -> %d\n", fdkey, fd );
        }
    }

    storageConnection->get_open_flags( &open_flags );

    if  ((open_flags & O_WRONLY) ||
         (open_flags & O_RDWR)) {

        char    logical_filename[NEST_MAXPATHLEN],
                physical_filename[NEST_MAXPATHLEN];

        storageConnection->get_filenames(logical_filename,
                                         physical_filename);

        if (file_catalog->is_system_file( logical_filename )) {
            return NEST_SUCCESS;
            /*
            if (file_catalog->get_logical_name( logical_filename,
                                                physical_filename )
                != NEST_SUCCESS) {
                    //the 'put' is to a non-nest system file
                return NEST_SUCCESS;
            }
            */
        }

        nest_stat   stat_buf;
        if (NestFstat(fd, &stat_buf) < 0) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Next to impossible!\n" );
            stat_buf.st_size = 0;
        }

        off_t   open_filesize_bytes,
                close_filesize_bytes  = stat_buf.st_size;

        storageConnection->get_open_filesize(&open_filesize_bytes);

        file_catalog->set_size( logical_filename,
                                close_filesize_bytes );

        if (NestEnableLots) {
            char    user[NEST_NAMELEN],
                    file[NEST_MAXPATHLEN];
            NestReplyStatus error_return;
            double  lot_space = 0,
                    more_space_MB;

            if ( file_catalog->get_owner_user( user,
                                               logical_filename )
                 == NEST_SUCCESS ) {

                lot_space = lot_manager->available_lot_space(user,
                                                             logical_filename);
            }

            if(filesize_bytes != ENTIRE_FILE) {
                /* 
                    If the file size is known before hand,
                    at the end of transfer, we check the filesize
                    again to make sure nothing has gone wrong.
                    If more data than the file size is sent,
                    the lot is updated if space is available
                    or the file is removed.
                 */

                more_space_MB  = (close_filesize_bytes - open_filesize_bytes);
                more_space_MB /= (1024 * 1024);     //  in MB

                if ( lot_space >= more_space_MB) {

                    nest_debug( NEST_DEBUG_STORAGE,
                                "Lot_space %f >= filesize %f\n",
                                lot_space,
                                more_space_MB );

                    error_return = lot_manager->update_lot_usage(
                            logical_filename,
                            more_space_MB);

                    if ( error_return == NEST_SUCCESS) {

                        /*
                            return value purposefully ignored
                                -   doesn't really matter or
                                    really can't do anything at all

                                -   in all other cases the file will
                                get deleted, so dont bother!
                        */

                        file_catalog->set_transfer_flag( logical_filename,
                                                         false );
    
                        /*
                        if (NestClose( storageConnection->getSocket() ) < 0) {
                            return ErrnoToNest(errno);
                        }
                        */

                        return NEST_SUCCESS;
                    }
                    else {
                        nest_debug( NEST_DEBUG_STORAGE,
                                    "File-Lot attachment failed!"
                                    " - deleting the file %s!\n",
                                    logical_filename );
                    }
                }
                else {
                    nest_debug( NEST_DEBUG_STORAGE,
                                "Not enough space available in user's Lots"
                                " - deleting file %s!\n",
                                logical_filename );
                    error_return = NEST_LOT_NOT_ENOUGH;
                }
            } 
            else {
                if(client && req) {
                    /*
                        If the file size is not known before hand,
                        we cannot use the open_filesize_bytes value as
                        it will be zero. Instead we compare close_filesize_bytes
                        with the allocated lot space and allocate more if needed
                        or delete the file as in previous case.
                    */
                   
                    /* 
                        hopefully the allocated lot space is larger than
                        actual filesize - cos lot space is allocated in MBs when
                        dynamically allocated.
                    */
                
                    double          allocated_space_MB = 0;

                    error_return = fileusage(&allocated_space_MB, 
                                             req->getNestFile(),
                                             client->getIdentity(),
                                             client->getCwd() );  

                    if(error_return == NEST_SUCCESS) {
                        more_space_MB = allocated_space_MB -
                                            (double) close_filesize_bytes 
                                            / ( 1024*1024 );
                        /*
                            -if more_space_MB is positive - we have over allocated.
                             Reduce the lot allcation to file
                            -if more_space_MB is negative - we have under allocated.
                             Increase the lot allocation to file if lot space is
                             available, else return error.
                        */
                        if(more_space_MB > 0) {
                            lot_manager->update_lot_usage( logical_filename,
                                                           -more_space_MB );
                            file_catalog->set_transfer_flag( logical_filename,
                                                             false );
                            return NEST_SUCCESS;
                        }
                        else {
                            if ( lot_space >= more_space_MB) {
                                nest_debug( NEST_DEBUG_STORAGE,
                                            "Lot_space %f >= filesize %f\n",
                                            lot_space,
                                            more_space_MB );

                                error_return = lot_manager->update_lot_usage(
                                            logical_filename,
                                            -more_space_MB);

                                if ( error_return == NEST_SUCCESS) {

                                    file_catalog->set_transfer_flag( 
                                            logical_filename,
                                            false );

                                    return NEST_SUCCESS;
                                }
                                else {
                                    nest_debug( NEST_DEBUG_STORAGE,
                                                "File-Lot attachment failed!"
                                                " - deleting the file %s!\n",
                                                logical_filename );
                                    goto delete_file;
                                }

                            }
                            else {
                                nest_debug( NEST_DEBUG_STORAGE,
                                        "Not enough space available in user's "
                                        "Lots - deleting file %s!\n",
                                        logical_filename );
                                error_return = NEST_LOT_NOT_ENOUGH;
                                goto delete_file;
                            }
                        }
                    }
                }
                else {
                    error_return = NEST_BAD_PARAMETERS;
                }

                nest_debug( NEST_DEBUG_STORAGE,
                            "Error on closing the file - deleting file %s.\n ",
                            logical_filename );
            }


delete_file:

            file_catalog->get_reverse_logical_name(file,
                                                   logical_filename);

            /*
                as the file is going to be deleted
                and the open_filesize would have been already accounted
                to relevant Lots,
                set the filesize to open_filesize and call deleteFile()
            */

            file_catalog->set_size( logical_filename,
                                    open_filesize_bytes );
            deleteFile(user,
                       file,
                       "");
            file_catalog->set_transfer_flag( logical_filename,
                                             false );
            //NestClose( storageConnection->getSocket());
            return error_return;
        
        }
    }

    /*
    if ( NestClose( storageConnection->getSocket() ) < 0) {
        return ErrnoToNest(errno);
    }
    */

    return NEST_SUCCESS;
}

/*
    Only NEST_LINKS - deletes for ~system links dont come here
*/
NestReplyStatus
nestStorageManager::deleteLink( const char *logical_link ) {
    return file_catalog->delete_entry(logical_link);
}

NestReplyStatus
nestStorageManager::renameFile(const char *user,
                               user_id_t  uid,
                               const char *oldpath,
                               const char *newpath,
                               const char *cwd,
                               bool update_lot ) {

    nest_debug( NEST_DEBUG_STORAGE, 
                "rename called - %s, %s -> %s\n",
                user, oldpath, newpath );
    
    if ( !user || !oldpath || !newpath ) {
        return NEST_BAD_PARAMETERS;
    }

    string oldfilename,newfilename;
    NestReplyStatus reply = expand_path( oldfilename,
                                         user,
                                         cwd,
                                         oldpath );
    NEST_REPLY_CHECK(reply);
    reply = expand_path( newfilename,
                         user,
                         cwd,
                         newpath );
    NEST_REPLY_CHECK(reply);

    char    logical_oldfilename[NEST_MAXPATHLEN],
            logical_newfilename[NEST_MAXPATHLEN];

    reply = file_catalog->get_logical_name( logical_oldfilename,
                                            oldfilename.c_str(),
                                            user);
    NEST_REPLY_CHECK(reply); 
    /*
        ignoring return value explicitly as the file entry would
        not be created yet
     */
    reply = file_catalog->get_logical_name( logical_newfilename,
                                            newfilename.c_str(),
                                            user);

    bool old_is_system_file = file_catalog->is_system_file(logical_oldfilename),
         new_is_system_file = file_catalog->is_system_file(logical_newfilename);

    if(old_is_system_file && !new_is_system_file || 
       !old_is_system_file && new_is_system_file ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (file_catalog->is_file_present (logical_oldfilename)) {
        if (!file_catalog->is_access_allowed(logical_oldfilename,
                                             user,
                                             NEST_READ_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (file_catalog->is_access_allowed_in_parent(logical_oldfilename,
                                                      user,
                                                      NEST_LIST_ACCESS)) {
            return NEST_FILE_NOT_FOUND;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->is_file_present (logical_newfilename)) {
        if (!file_catalog->is_access_allowed(logical_oldfilename,
                                             user,
                                             NEST_WRITE_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (file_catalog->is_access_allowed_in_parent(logical_oldfilename,
                                                      user,
                                                      NEST_LIST_ACCESS)) {
            return NEST_FILE_EXISTS;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    nest_obj_type_t type;
    reply = file_catalog->get_type(&type, logical_oldfilename);
    NEST_REPLY_CHECK(reply);

    if (type != NEST_FILE) {
        return NEST_NOT_YET_IMPLEMENTED;
    }
  
    off_t filesize_bytes;
    reply = file_catalog->get_size(&filesize_bytes,
                                   logical_oldfilename); 
    NEST_REPLY_CHECK(reply);
    double  filesize;
    filesize = (double)filesize_bytes / (1024 * 1024);  //  in MB
    
    /* 
        The following operations must be atomic. And I don't know
        the correct way to do it. For now, I am ignoring errors.
        The operations:
            1) Update file catalog entries
            2) Rename file
            3) Update lot usage
     */

    // Add entry in file catalog
    reply = file_catalog->add_entry(logical_newfilename,
                                   user,
                                   NEST_FILE);
    NEST_REPLY_CHECK(reply);
    reply = file_catalog->set_size(logical_newfilename,
                                   filesize_bytes);
    if(reply != NEST_SUCCESS) {
        goto delete_file_entry_and_return;
    }

    // Do actual rename
    char    physical_oldfilename[NEST_MAXPATHLEN],
            physical_newfilename[NEST_MAXPATHLEN];
    file_catalog->get_physical_name( physical_oldfilename,
                                     logical_oldfilename);
    file_catalog->get_physical_name( physical_newfilename,
                                     logical_newfilename);

    if( rename(physical_oldfilename, physical_newfilename) < 0 ) {
        goto delete_file_entry_and_return;
    }
    
    NestChmod( physical_newfilename, nestStorageManager::file_create_mode );
    NestChown( physical_newfilename, uid, ActualGid );

    // update our fd_cache
    if ( fd_cache ) {
        fd_cache->invalidateAll( oldfilename.c_str() );
    }

    // Update lot usage
    if (NestEnableLots && update_lot && !old_is_system_file) {
        reply = lot_manager->update_lot_usage(logical_oldfilename,
                                              -filesize);
        if(reply != NEST_SUCCESS) {
            goto undo_rename_and_return;
        }
        reply = lot_manager->update_lot_usage(logical_newfilename,
                                              filesize);
        if(reply != NEST_SUCCESS) {
            goto redo_lot_usage_and_return;
        }
    }
   
    // Delete entry in file catalog 
    reply = file_catalog->delete_entry(logical_oldfilename);
    if(reply != NEST_SUCCESS) {
        goto undo_lot_usage_and_return;
    }

    /* 
        FIX ME: Correct this part when moving across NeST and 
        ~system file systems are implemented

    long long int avail_bytes;
    if (NestEnableLots) {

        avail_bytes = (long long int)
                      (lot_manager->available_user_lot_space(user) *
                       1024 * 1024);

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail user_lot_space(%s) = %lld & filesize = %ld\n",
                    user,
                    avail_bytes,
                    filesize_bytes );
	
        if ((avail_bytes == 0) && (filesize_bytes > 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available Lot space for user %s is 0 bytes!\n",
                        user );
            reply = NEST_LOT_NOT_ENOUGH;
            goto delete_file_entry_and_out;
        }

        if ((filesize_bytes > 0) && (avail_bytes < filesize_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the user's Lots!\n");
            reply = NEST_LOT_NOT_ENOUGH;
            goto delete_file_entry_and_out;
        }
	
	// Zero byte files are not attached to any lot and hence do not need a lot. 
	 
        if (filesize_bytes > 0) {
            double  more_bytes_MB = filesize_bytes;
            more_bytes_MB /= (1024 * 1024);

            reply = lot_manager->update_lot_usage(logical_newfilename,
                                                  more_bytes_MB);
            if(reply != NEST_SUCCESS) {
                goto delete_file_entry_and_out;
            }

            reply = file_catalog->set_size( logical_newfilename,
                                            filesize_bytes );
            if(reply != NEST_SUCCESS)
                goto delete_file_entry_and_out;

        }
    } else {
        long long int total_bytes;
        if ( (reply = bytesAvailable( IS_USER, &avail_bytes, &total_bytes )) !=
                NEST_SUCCESS ) 
        {
            goto delete_file_entry_and_out; 
        }

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail device_space = %lld & filesize = %ld\n",
                    avail_bytes,
                    filesize_bytes );

        if ((avail_bytes == 0) && (filesize_bytes != 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available disk space is 0 bytes!\n" );
            reply = NEST_DISK_FULL;
            goto delete_file_entry_and_out;
        }

        if ((filesize_bytes > 0) && (avail_bytes < filesize_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the device!\n");
            reply = NEST_DISK_FULL;
            goto delete_file_entry_and_out;
        }
        
        reply = file_catalog->set_size( logical_newfilename,
                                        filesize_bytes );
        if(reply != NEST_SUCCESS)
            goto delete_file_entry_and_out;
        
    }

    */

    return NEST_SUCCESS;

undo_lot_usage_and_return:
    lot_manager->update_lot_usage(logical_newfilename,
                                  -filesize);

redo_lot_usage_and_return:
    lot_manager->update_lot_usage(logical_oldfilename,
                                  filesize); 
    
undo_rename_and_return:
    rename(physical_newfilename, physical_oldfilename);


delete_file_entry_and_return:
    file_catalog->delete_entry( logical_newfilename );

    return reply;
}

NestReplyStatus
nestStorageManager::deleteFile( const char *user,
                                const char *file,
                                const char *cwd,
                                bool update_lot) {

    if (!user ||
        !file) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    // update our fd_cache
    if ( fd_cache ) {
        fd_cache->invalidateAll( filename.c_str() );
    }


    char    logical_filename[NEST_MAXPATHLEN];
    bool    delete_access_present = false;

    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user);

    delete_access_present = file_catalog->is_access_allowed_in_parent(
                                                   logical_filename,
                                                   user,
                                                   NEST_DELETE_ACCESS);

    if (reply != NEST_SUCCESS) {
        if (delete_access_present) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!delete_access_present) {
        return NEST_NONPRIVELEDGED_USER;
    }

    nest_obj_type_t type;
    char    physical_filename[NEST_MAXPATHLEN],
            owner[NEST_MAXPATHLEN];

    reply = file_catalog->get_type(&type,
                                   logical_filename);

    NEST_REPLY_CHECK(reply);

    if (type == NEST_DIR) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Cannot delete dir with the SM::deleteFile() !\n" );
        return NEST_BAD_PARAMETERS;
    }

    if (file_catalog->is_system_file(logical_filename)) {

        reply = file_catalog->get_physical_name( physical_filename,
                                                 logical_filename );

        NEST_REPLY_CHECK(reply);

        /*
            system-links can never point to nest-files
            in this version
        */

        if (type == NEST_LINK) {
            if (NestUnlink(physical_filename) < 0) {
                return ErrnoToNest(errno);
            }
            return NEST_SUCCESS;
        }

        /*
            Its a system-file - OK
            See, if its a physical_file of some logical nest-file
        */

        reply = file_catalog->get_logical_name( logical_filename,
                                                physical_filename );

        if (reply != NEST_SUCCESS) {

            /*
                if somebody is deleting a non-NeST file
                dont care about the Lot status!

                use TruckUnlink anyway!
            */

            if (NestTruncUnlink(physical_filename) < 0) {
                return ErrnoToNest(errno);
            }

            return NEST_SUCCESS;
        }
    }

    /*
        only system-files, nest-files or nest-links
        might come here!
    */

    if (type == NEST_LINK) {
        return deleteLink( logical_filename );
    }

    /*
        only NEST_FILES come here
    */

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_filename);
    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while translating %s to physical name!\n",
                    logical_filename );
        nest_debug( NEST_DEBUG_STORAGE,
                    "Probably, accessing a non-NeST file!\n" );
        return NEST_FILE_NOT_FOUND;
    }

    reply = file_catalog->get_owner_user( owner,
                                          logical_filename );

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while getting owner name for %s \n",
                    logical_filename );
        return NEST_FILE_NOT_FOUND;
    }

    off_t   filesize_bytes;
    reply = file_catalog->get_size( &filesize_bytes,
                                    logical_filename );
    NEST_REPLY_CHECK(reply);

    double  filesize = filesize_bytes;
    filesize /= (1024 * 1024);  //  in MB

    /*
        I really dont know the best way to do this

        Basically,
            the lot_usage should be (conditionally) updated,
            the file_catalog entry should be deleted and
            the physical file should be Unlinked

        I don't know how to make these changes atomic

        I'm ignoring errors
    */

    if (NestEnableLots && update_lot) {
        reply = lot_manager->update_lot_usage(logical_filename,
                                              -filesize);
        NEST_REPLY_CHECK(reply);
    }

    int retval = NestTruncUnlink(physical_filename);
    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "NestTruncUnlink() failed - %s!\n",
                    strerror(errno) );
    }

    reply = file_catalog->delete_entry(logical_filename);
    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "FileCatalog->delete_entry(%s) failed: %s"
                    " - IGNORING ERROR\n",
                    logical_filename,
                    NestErrorString( reply ) );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::get_type( nest_obj_type_t *result,
                              const char *file,
                              const char *user,
                              const char *cwd ) {

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char logical_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user);

    if (reply != NEST_SUCCESS) {
        if (file_catalog->is_access_allowed_in_parent( logical_filename,
                                                       user,
                                                       NEST_LIST_ACCESS )) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->is_access_allowed( logical_filename,
                                          user,
                                          NEST_READ_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_type(result,
                                   logical_filename);

    if (reply != NEST_FILE_NOT_FOUND) {
        return reply;
    }

    /*
        Hmm .. could be a NEST_DIR and the logical_filename
        is missing a '/' at the end - oh boy!
    */

    int len = NestStrlen(logical_filename);
    if (logical_filename[len - 1] == '/') {
        return NEST_FILE_NOT_FOUND;
    }

    logical_filename[len] = '/';
    logical_filename[len+1] = 0;

    return file_catalog->get_type( result,
                                   logical_filename );
}

NestReplyStatus
nestStorageManager::get_physical_filename(char **physical,
                                          const char *file,
                                          const char *user,
                                          const char *cwd,
                                          int link_count) {

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_name[NEST_MAXPATHLEN];
    char    physical_name[NEST_MAXPATHLEN];

    reply = file_catalog->get_logical_name( logical_name,
                                            filename.c_str(),
                                            user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->is_access_allowed_in_parent( logical_name,
                                                       user,
                                                       NEST_LIST_ACCESS )) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->is_access_allowed(logical_name,
                                         user,
                                         NEST_READ_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_physical_name( physical_name,
                                             logical_name );
    NEST_REPLY_CHECK(reply);

    if (!file_catalog->is_system_file( logical_name )) {
        nest_obj_type_t     filetype;

        reply = file_catalog->get_type( &filetype,
                                        logical_name );
        NEST_REPLY_CHECK(reply);

        if (filetype == NEST_LINK) {

            if (link_count++ > NEST_LINK_MAXLOOP) {
                nest_debug( NEST_DEBUG_STORAGE,
                            "Too many levels of symbolic links!\n" );
                return NEST_TOO_MANY_LINKS;
            }

            return get_physical_filename( physical,
                                          physical_name,
                                          user,
                                          cwd,
                                          link_count );
        }
    }

    *physical = NestStrdup( physical_name );
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::Link(const char *user,
                         const char *oldpath,
                         const char *newpath,
                         int val,
                         const char *cwd) {

    if (!user || !oldpath || !newpath) {
        return NEST_BAD_PARAMETERS;
    }

    if (!val) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    string  oldpath_file,
            newpath_file;

    /*
        ignoring return value intentionally
    */

    NestReplyStatus reply = expand_path( oldpath_file,
                                         user,
                                         cwd,
                                         oldpath );
    /*
        ignoring return value intentionally
    */

    reply = expand_path( newpath_file,
                         user,
                         cwd,
                         newpath );

    char    old_logical[NEST_MAXPATHLEN],
            new_logical[NEST_MAXPATHLEN];

    /*
        ignoring return value intentionally
    */
    reply = file_catalog->get_logical_name( old_logical,
                                            oldpath_file.c_str(),
                                            user );

    /*
        ignoring return value intentionally
    */
    reply = file_catalog->get_logical_name( new_logical,
                                            newpath_file.c_str(),
                                            user );

    nest_obj_type_t     type;

    reply = file_catalog->get_type( &type,
                                    old_logical );
    
    NEST_REPLY_CHECK(reply);
    
    if (type == NEST_DIR) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (!file_catalog->is_access_allowed_in_parent( new_logical,
                                                    user,
                                                    NEST_INSERT_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    /*
    if (!file_catalog->is_access_allowed( old_logical,
                                          user,
                                          NEST_READ_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }
    */


    if (file_catalog->is_file_present( new_logical )) {
        return NEST_FILE_EXISTS;
    }

    char    old_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_reverse_logical_name(old_filename,
                                                   old_logical);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->add_entry(new_logical,
                                    old_filename,
                                    user,
                                    NEST_LINK);

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Couldn't add-entry into FileCatalog for NEST_LINK! %s\n",
                    NestErrorString(reply));
        return reply;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::sync(const char *user) {
    NestSync();
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::synchronize() {

    int state = Nestset_grid_mapfile( GridMapFile );

    if ( state < 0) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Couldn't read grid-mapfile at %s!\n",
                    GridMapFile );
        return NEST_SYSTEM_FILE_MISSING;
    }

    grid_mapfile_entry  *user;
    NestReplyStatus     reply;

    while( (user = Nestget_grid_mapfile_entry(state)) != NULL) {

        char *identity = NestGetModifiedIdentity( user->identity );
        
        if (identity_manager->is_valid_nest_user( identity )) {
            reply = add_home_dir( identity,
                                  NEST_NO_GROUP );
        }
        else {
            reply = delete_home_dir( identity );
        }

        NEST_REPLY_CHECK(reply);
    }
    Nestend_grid_mapfile(state);

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::maintenance() {
    nest_debug( NEST_DEBUG_STORAGE,
                "Executing Storage manager maintenance - "
                "Truncating ClassAd log.\n" );
    Resources->TruncateLog();
    identity_manager->synchronize();
    synchronize();
    file_catalog->synchronize();
    if (NestEnableLots) {
        //lot_manager->synchronize();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::shutdown() {
    identity_manager->shutdown();
    file_catalog->shutdown();
    if (NestEnableLots) {
        lot_manager->shutdown();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::get_acl(nest_access_rights_t *result,
                            const char *for_name,
                            int is_user,
                            const char *file,
                            const char *user,
                            const char *cwd) {

    if (!result || !for_name || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user );

    if (!file_catalog->is_access_allowed_in_parent( logical_filename,
                                                    user,
                                                    NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_acl(logical_filename,
                                  for_name,
                                  is_user,
                                  result);

    return reply;
}

NestReplyStatus
nestStorageManager::get_all_acl(NestACLList **result,
                                const char *file,
                                const char *user,
                                const char *cwd) {

    if (!result || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user );

    if (!file_catalog->is_access_allowed_in_parent( logical_filename,
                                                    user,
                                                    NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    *result = NULL;

    reply = file_catalog->get_all_acl(result,
                                      logical_filename,
                                      IS_USER);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_all_acl(result,
                                      logical_filename,
                                      IS_GROUP);

    return reply;
}

NestReplyStatus
nestStorageManager::set_acl(nest_access_rights_t arg,
                            const char *to_name,
                            int is_user,
                            const char *file,
                            const char *user,
                            const char *cwd) {

	char    owner[NEST_NAMELEN];


    if (!to_name || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                      user,
                                                      NEST_MODIFYAC_ACCESS)) {
            return reply;
        }

        return NEST_NONPRIVELEDGED_USER;
    }

	reply = file_catalog->get_owner_user(owner,
										 logical_filename);
	NEST_REPLY_CHECK(reply);

    if (!file_catalog->is_access_allowed( logical_filename,
                                          user,
                                          NEST_MODIFYAC_ACCESS)
		&& strcmp(owner,user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->set_acl(logical_filename,
                                  to_name,
                                  is_user,
                                  arg);

    return reply;
}

NestReplyStatus
nestStorageManager::get_user_acl(nest_access_rights_t *result,
                                 const char *for_user,
                                 const char *file,
                                 const char *user,
                                 const char *cwd) {

    if (!identity_manager->is_user_present(for_user)) {
        return NEST_USER_NOT_FOUND;
    }

    return get_acl( result,
                    for_user,
                    IS_USER,
                    file,
                    user,
                    cwd );
}

NestReplyStatus
nestStorageManager::set_user_acl(nest_access_rights_t arg,
                                 const char *to_user,
                                 const char *file,
                                 const char *user,
                                 const char *cwd) {

    if (!identity_manager->is_user_present(to_user)) {
        return NEST_USER_NOT_FOUND;
    }

    return set_acl( arg,
                    to_user,
                    IS_USER,
                    file,
                    user,
                    cwd );
}

NestReplyStatus
nestStorageManager::get_group_acl(nest_access_rights_t *result,
                                  const char *for_group,
                                  const char *file,
                                  const char *user,
                                  const char *cwd) {

    if (!identity_manager->is_group_present(for_group)) {
        return NEST_GROUP_NOT_FOUND;
    }

    return get_acl( result,
                    for_group,
                    IS_GROUP,
                    file,
                    user,
                    cwd );
}

NestReplyStatus
nestStorageManager::set_group_acl(nest_access_rights_t arg,
                                  const char *to_group,
                                  const char *file,
                                  const char *user,
                                  const char *cwd) {

    if (!identity_manager->is_group_present(to_group)) {
        return NEST_GROUP_NOT_FOUND;
    }

    return set_acl( arg,
                    to_group,
                    IS_GROUP,
                    file,
                    user,
                    cwd );
}

NestReplyStatus
nestStorageManager::frame_full_path( string &result,
                                     const char *user,
                                     const char *file ) {

    result = "";

    if (!file || !file[0]) {
        //result = result + "~" + user + "/";
        result = result + "~" + "/";
        return NEST_SUCCESS;
    }

    if (!strncmp( file, "~/", 2 )) {
        //result = result + "~" + user + (file + 1);
        result = result + "~" + (file + 1);
    } else {
        result = file;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::expand_path ( string &result,
                                  const char *user,
                                  const char *cwd,
                                  const char *file ) {
   
    NestReplyStatus reply;
    char    final_path[NEST_MAXPATHLEN];
    nest_debug( NEST_DEBUG_STORAGE, "expand_path: cwd='%s'\n", cwd );

    if (file[0] && (file[0] == '~')) {
        reply = file_catalog->get_dest_file( final_path,
                                             file, user );
        NEST_REPLY_CHECK(reply);

        return frame_full_path( result, user, final_path );
    }

    /*
        the filename is not absolute anymore

        cwd needn't be collapsed as it would never have .s or ..s
    */

    reply = frame_full_path( result, user, cwd);

    NEST_REPLY_CHECK(reply);

    result = result + file;
    nest_debug( NEST_DEBUG_STORAGE, "expand_path: result='%s'\n", result.c_str() );

    /*
        this needs to be collapsed!!
    */
    reply = file_catalog->get_dest_file( final_path,
                                         result.c_str(),
										 user );

    NEST_REPLY_CHECK(reply);
    nest_debug( NEST_DEBUG_STORAGE, "expand_path: final='%s'\n", final_path );

    result = final_path;
    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::get_dest_dir( char *dest,
                                  const char *cwd,
                                  const char *dir,
                                  const char *user ) {

    if (!dest || !cwd || !dir) {
        return NEST_BAD_PARAMETERS;
    }

    if (!strcmp(dir, "~")) {
        strcpy( dest, "~/" );
        return NEST_SUCCESS;
    }

    char    tmp_dest[NEST_MAXPATHLEN];

    if (dir[0] == '~') {
        if (NestSnprintf( tmp_dest, NEST_MAXPATHLEN, "%s", dir ) < 0) {
            return NEST_FILENAME_TOO_LONG;
        }
    }
    else {
        if (cwd[NestStrlen(cwd)-1] != '/') {
            if (NestSnprintf( tmp_dest,
                              NEST_MAXPATHLEN,
                              "%s/%s",
                              cwd,
                              dir ) < 0) {
                return NEST_FILENAME_TOO_LONG;
            }
        }
        else {
            if (NestSnprintf( tmp_dest,
                              NEST_MAXPATHLEN,
                              "%s%s",
                              cwd,
                              dir ) < 0) {
                return NEST_FILENAME_TOO_LONG;
            }
        }
    }

    /*
        takes care of ../ ./ and //
    */

    NestReplyStatus reply = file_catalog->get_dest_file( dest,
														 tmp_dest,
														 user );
    NEST_REPLY_CHECK(reply);

    int last_byte = NestStrlen(dest) - 1;

    if ((last_byte < 0) || (last_byte >= NEST_MAXPATHLEN)) {
        return NEST_BAD_PARAMETERS;
    }

    if (dest[last_byte] != '/') {
        dest[last_byte + 1] = '/';
        dest[last_byte + 2] = 0;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::chDir(char **new_cwd,
                          const char *user,
                          const char *cwd,
                          const char *dir){

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN];

    nest_debug( NEST_DEBUG_STORAGE,
                "chdir called - %s,%s\n",
                user,
                //cwd,
                dir );

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
                                          user );

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->is_access_allowed_in_parent(logical_dest,
                                                      user,
                                                      NEST_LIST_ACCESS)) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->is_access_allowed(logical_dest,
                                         user,
                                         NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    *new_cwd = NestStrdup( dest );

    nest_debug( NEST_DEBUG_STORAGE,
                "chDir to %s\n",
                *new_cwd );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::mkDir( const char *user,
                           const char *cwd,
                           const char *dir,
                           user_id_t     uid ) {

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN];

    nest_debug( NEST_DEBUG_STORAGE,
                "mkdir called - %s,%s\n",
                user,
                //cwd,
                dir );

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
                                          user );
    NEST_REPLY_CHECK(reply);

    nest_debug( NEST_DEBUG_STORAGE,
                "mkdir %s [%d] - user %s\n",
                dest,
                NEST_MAXPATHLEN,
                user );

    string  tmp_full_path,
            tmp_userhome;

    reply = frame_full_path(tmp_userhome,
                            user,
                            "");
    NEST_REPLY_CHECK(reply);

    reply = frame_full_path(tmp_full_path,
                            user,
                            dest);
    NEST_REPLY_CHECK(reply);

    if (tmp_userhome == tmp_full_path) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Disallowing userhome_dir %s creation!\n",
                    tmp_userhome.c_str() );
        return NEST_BAD_PARAMETERS;
    }

    /*
        ignoring return value intentionally!
        similar to the one in ::open()
    */

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    bool is_system_file = file_catalog->is_system_file( logical_dest );
        
    if (!file_catalog->is_access_allowed_in_parent( logical_dest,
                                                    user,
                                                    NEST_INSERT_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if( !is_system_file ) {
        reply = file_catalog->add_entry( logical_dest,
                                         user,
                                         NEST_DIR);
        NEST_REPLY_CHECK(reply);
    }

    char    physical_filename[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_dest);
    NEST_REPLY_CHECK(reply);

    if ( NestMkdir( physical_filename, DIR_MODE ) < 0 ) {
        if (errno != EEXIST) {
            nest_warn( __FILE__, __FUNCTION__, "mkdir of %s failed: %s\n",
                       physical_filename,
                       strerror( errno ) );

            if( !is_system_file ) {
                file_catalog->delete_entry( logical_dest );
                return ErrnoToNest( errno );
            }
        }
    }
    NestChmod( physical_filename, DIR_MODE );
    NestChown( physical_filename, uid, ActualGid );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::deleteDirContents( const char *user,
                                       const char *logical_dir) {

    string constraint = "";
    constraint = constraint +
                    "other.DirName == \"" + logical_dir + "\"";

    ViewName    file_view = NEST_FILE_CATALOG_VIEW;
    ExprTree    *constraint_tree = parser.ParseExpression(constraint);
    query.Query(file_view, constraint_tree);
    query.ToFirst();

    string query_key;

    if (!query.Current(query_key)) {
        return NEST_SUCCESS;
    }

    ClassAd         *file_ad;

    do {
        file_ad = Resources->GetClassAd( query_key );

        if (!file_ad) {
            continue;
        }

        string  logical_filename;
        if (!file_ad->EvaluateAttrString( "LogicalName",
                                          logical_filename )) {
            continue;
        }
        string  type;
        if (!file_ad->EvaluateAttrString( "FileType",
                                          type )) {
            continue;
        }

        if (type == "NEST_DIR") {
            deleteDirContents( user,
                               logical_filename.c_str());
        }
        else {
            deleteFile(user,
                       logical_filename.c_str(),
                       "");
        }

    } while (query.Next(query_key));

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::rmDir( const char *user,
                           const char *cwd,
                           const char *dir,
                           bool recursive ) {

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN],
            logical_cwd[NEST_MAXPATHLEN];

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
										  user );

    if (!strcmp(dest, "~/")) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to delete user's homedir!\n" );
        return NEST_BAD_PARAMETERS;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    if (file_catalog->is_system_file( logical_dest )) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (!file_catalog->is_access_allowed_in_parent(logical_dest,
                                                   user,
                                                   NEST_DELETE_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_cwd,
                                            cwd,
                                            user );

    NEST_REPLY_CHECK(reply);

    /*
    if (!strcmp(logical_dest, logical_cwd)) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to delete the current working dir!\n" );
        return NEST_BAD_PARAMETERS;
    }
    */

    if (recursive) {
        reply = deleteDirContents( user,
                                   logical_dest);

        NEST_REPLY_CHECK(reply);
    }

    char    physical_filename[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_dest);
    NEST_REPLY_CHECK(reply);

    if ( NestRmdir( physical_filename ) < 0 ) {
        reply = ErrnoToNest( errno );
    }

    if (file_catalog->delete_entry(logical_dest) != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while deleting NEST_DIR entry from FileCatalog!\n");
    }

    return reply;
}

char *
nestStorageManager::getRootDir( ) {
    return rootDir;
}

NestFileInfo *
nestStorageManager::directoryContents( const char *user,
                                       const char *dir) {

    char    dest_dir[NEST_MAXPATHLEN],
            logical_dir[NEST_MAXPATHLEN];
    NestReplyStatus reply;

    reply = get_dest_dir( dest_dir,
                          dest_dir,
                          dir,
                          user );
    if (reply != NEST_SUCCESS) {
        return NULL;
    }

    reply = file_catalog->get_logical_name(logical_dir,
                                           dest_dir,
                                           user);
    
    if (reply != NEST_SUCCESS) {
        return NULL;
    }
 
    if (!file_catalog->is_access_allowed(logical_dir,
                                         user,
                                         NEST_LIST_ACCESS)) {
        return NULL;
    }

    NestFileInfo *list = NULL;

    reply = file_catalog->dir_list( &list,
                                    logical_dir );

    if (reply != NEST_SUCCESS) {
        return NULL;
    }

    return list;
}

NestReplyStatus
nestStorageManager::add_user(const char *username,
                             const char *groupname,
                             const char *cur_user,
                             const char *identity){

    if (!identity_manager->is_user_access_allowed( cur_user )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NestReplyStatus reply = identity_manager->add_user(username,
                                                       groupname,
                                                       cur_user,
                                                       identity);
    NEST_REPLY_CHECK(reply);
	
    return add_home_dir( username,
                         groupname );
}

NestReplyStatus
nestStorageManager::add_user(const char *username,
                             const char *cur_user) {

    return add_user( username, NEST_NO_GROUP, cur_user, "" );
}

NestReplyStatus
nestStorageManager::delete_user(const char *username,
                                const char *cur_user){

    if (!identity_manager->is_user_access_allowed( cur_user )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NestReplyStatus reply = delete_home_dir( username );

    NEST_REPLY_CHECK(reply);

    return identity_manager->delete_user(username, cur_user);
}

NestReplyStatus
nestStorageManager::increaseAllocation(nestClient *client) {
    nestRequest         *req;
    double              avl,
                        allocation_MB, allocation;
    NestReplyStatus     reply;
    
    /* the allocated space is increased in multiples of 1 MB */
   
    if(!client) {
        return NEST_BAD_PARAMETERS;
    }

    if( !(req = client->getProxyRequest() ) ) {
        return NEST_BAD_PARAMETERS;
    }

    if (!NestEnableLots || lot_manager->kernel_quota_enabled()) {
        /*  
         *  allocate a very large chunk 1GB - no need to care of lots :-)
         *  Yet to proceed with the transfer the allocation parameter
         *  has to be set and can be set to arbitrarily large values to 
         *  minimize the number of times increaseAllocation is called.
         */  
        allocation = 1024 * 1024 * 1024;
        client->setAllocation(allocation);
        nest_debug(NEST_DEBUG_LOTS,
                   "Allocating %f lot space for gftp-type transfers.\n",
                   allocation);  
        return NEST_SUCCESS;
    }

    string filename;
    char   logical_filename[NEST_MAXPATHLEN];

    reply = expand_path(filename,
                        client->getIdentity(),
                        client->getCwd(),
                        req->getNestFile() );
    NEST_REPLY_CHECK(reply);
    /*
        ignore return value - as while creating new files
        this would return NEST_FILE_NOT_FOUND
    */
    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           client->getIdentity());
    
    avl = lot_manager->available_lot_space(client->getIdentity(), 
                                           logical_filename);

    if(avl <= 0.0)
        return NEST_INSUFFICIENT_SPACE;
    if(avl < 1.0)
        allocation_MB = avl;
    else
        allocation_MB = 1;

    allocation = allocation_MB * 1024 * 1024;
    nest_debug(NEST_DEBUG_LOTS,
               "Allocating %f lot space for gftp-type transfers.\n",
               allocation);  
    client->setAllocation(allocation);
       
    reply = lot_manager->update_lot_usage(logical_filename,
                                          allocation_MB);
    return reply;
}

NestReplyStatus
nestStorageManager::resetAllocation(const char *file,
                                    const char *user,
                                    const char *cwd,
                                    long offset){
    NestReplyStatus reply;
    double          size;
    
    reply = fileusage(&size, file, user, cwd);
    if(reply == NEST_FILE_NOT_FOUND){
        return NEST_SUCCESS;
    }

    string filename;
    char   logical_filename[NEST_MAXPATHLEN];

    reply = expand_path(filename, user, cwd, file);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    NEST_REPLY_CHECK(reply);

    if(offset <  size) {
        reply = lot_manager->update_lot_usage(logical_filename,
                                              offset - size);
    } 
    
    return reply;
}

NestReplyStatus
nestStorageManager::cleanupGftpTransfer(
                            nestClient *client) {
    const char      *file, *user, *cwd;
    NestReplyStatus reply;
    char            logical_filename[NEST_MAXPATHLEN],
                    physical_filename[NEST_MAXPATHLEN];
    string          filename;

    file = client->getRequest()->getNestFile();
    user = client->getIdentity();
    cwd  = client->getCwd();

    reply = expand_path( filename, user, cwd, file ); 
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_filename);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->delete_entry( logical_filename );
    NEST_REPLY_CHECK(reply);

    if (NestUnlink(physical_filename) < 0) {
        return ErrnoToNest(errno);
    }
    return NEST_SUCCESS;
    
}
        

NestReplyStatus
nestStorageManager::add_group(const char *groupname,
                              const char *cur_user){
    return identity_manager->add_group(groupname, cur_user);
}

NestReplyStatus
nestStorageManager::delete_group(const char *groupname,
                                 const char *cur_user){
    return identity_manager->delete_group(groupname, cur_user);
}

NestReplyStatus
nestStorageManager::add_user_to_group(const char *username,
                                      const char *groupname,
                                      const char *cur_user){
    return identity_manager->add_user_in_group(username,
                                               groupname,
                                               cur_user);
}

NestReplyStatus
nestStorageManager::delete_user_from_group(const char *username,
                                           const char *groupname,
                                           const char *cur_user){
    return identity_manager->delete_user_in_group(username,
                                                  groupname,
                                                  cur_user);
}

NestReplyStatus
nestStorageManager::request_user_lot(NestLotHandle *handle,
                                     const char *user,
                                     double size,
                                     double duration,
                                     const char* lot_user,
                                     const char* parent_user,
                                     const char* parent_group,
                                     NestLotHandle *parent_handle,   
                                     int lot_type,
                                     int cost) {

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    return lot_manager->request_user_lot(handle,
                                         user,
                                         size,
                                         ((int)duration) ?
                                         duration :
                                         default_lot_duration,
                                         lot_user,
                                         parent_user,
                                         parent_group,
                                         parent_handle,
                                         lot_type,
                                         cost);
}


NestReplyStatus
nestStorageManager::request_group_lot(NestLotHandle *handle,
                                      const char *user,
                                      double size,
                                      double duration,
                                      const char* lot_group,
                                      const char* parent_user,
                                      const char* parent_group,
                                      NestLotHandle *parent_handle,
                                      int lot_type,
                                      int cost) {
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    return lot_manager->request_group_lot(handle,
                                          user, 
                                          size,
                                          ((int)duration) ?
                                          duration :
                                          default_lot_duration,
                                          lot_group,
                                          parent_user,
                                          parent_group,
                                          parent_handle,
                                          lot_type,
                                          cost);
}

NestReplyStatus
nestStorageManager::list_user_lots(NestLotList **result,
                                   const char *user,
                                   const char *lot_user ) {
    NestReplyStatus reply;
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if(!identity_manager->is_valid_nest_user(lot_user))
        return NEST_USER_NOT_FOUND;

    reply = lot_manager->list_user_lots(result, user, lot_user);
    return reply;
}

NestReplyStatus
nestStorageManager::list_group_lots(NestLotList **result,
                                    const char *req_user,
                                    const char *group ) {

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if(!identity_manager->is_group_present(group))
        return NEST_GROUP_NOT_FOUND;

    return lot_manager->list_group_lots(result, req_user, group);
}

NestReplyStatus
nestStorageManager::update_lot(NestLotHandle   *handle,
                               const char *user,
                               double more_size,
                               double more_duration,
                               int more_cost) {
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    return lot_manager->update_lot(handle,
                                   user,
                                   more_size,
                                   more_duration,
                                   more_cost);
}

NestReplyStatus
nestStorageManager::change_lot_type(NestLotHandle   *handle,
                                    const char *user,
                                    int lot_type,
                                    double duration) {
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }
    
    return lot_manager->change_lot_type(handle,
                                        user,
                                        lot_type,
                                        (int)duration ?
                                        duration :
                                        default_lot_duration);
}                        

NestReplyStatus
nestStorageManager::modify_lot_usage(const char *user,
                                     const char *file,
                                     const char *cwd,
                                     NestLotHandle  *from_lot,
                                     double from_share,
                                     NestLotHandle  *to_lot) {

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( logical_filename,
                                            filename.c_str(),
                                            user );

    if (reply == NEST_SUCCESS) {
        if (!file_catalog->is_access_allowed(logical_filename,
                                            user,
                                            NEST_WRITE_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    }
    else {
        if (file_catalog->is_access_allowed_in_parent(logical_filename,
                                                      user,
                                                      NEST_WRITE_ACCESS)) {
            return NEST_FILE_NOT_FOUND;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    return lot_manager->modify_lot_usage(user,
                                         logical_filename,
                                         from_lot,
                                         from_share,
                                         to_lot);
}


NestReplyStatus
nestStorageManager::get_lot_status(NestLot *result,
                                   NestLotHandle *handle,
                                   const char *user) {
    if (!NestEnableLots) 
        return NEST_LOT_DISABLED;
    
    return lot_manager->get_lot_status(result, handle, user);
    
}

NestReplyStatus
nestStorageManager::set_user_default_lot(NestLotHandle *handle,
                                         const char *user) {

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    return (lot_manager->set_user_default_lot( handle,
                                               user ));

}

NestReplyStatus
nestStorageManager::terminate_lot(NestLotHandle *handle,
                                  const char *user) 
{
    if (!NestEnableLots) 
        return NEST_LOT_DISABLED;

    return lot_manager->terminate_lot(handle,
                                      user);
}

NestReplyStatus
nestStorageManager::add_user_to_lot(NestLotHandle *handle,
                                    const char *cur_user,
                                    const char *user)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }
    
    return lot_manager->add_user_to_lot(handle,
                                        cur_user,
                                        user);
}

NestReplyStatus
nestStorageManager::delete_user_from_lot(NestLotHandle *handle,
                                         const char *cur_user,
                                         const char *user)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }
    
    return lot_manager->delete_user_from_lot(handle,
                                             cur_user,
                                             user);    
}

NestReplyStatus
nestStorageManager::attach_lot_to_file(NestLotHandle *handle,
                                       const char *file,
                                       const char *user,
                                       const char *cwd)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    NestReplyStatus reply;
    string          filename;
    char            logical_filename[NEST_MAXPATHLEN];
    bool            new_file_created = false;

    reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);
    
    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);

    if( file_catalog->is_system_file(logical_filename) ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    /*
     * Check if the user has the rights to create the file
     * if it is not already present.
     */
    if (!file_catalog->is_access_allowed_in_parent(logical_filename,
                                                   user,
                                                   NEST_INSERT_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }
    reply = file_catalog->add_entry(logical_filename,
                                    user,
                                    NEST_FILE);
    if (reply != NEST_SUCCESS && 
        reply != NEST_FILE_EXISTS) {
        return reply;
    }
    new_file_created =  (reply == NEST_SUCCESS);

    /* Check if the user has the right to attach a lot to the file.
     * If the file entry was just created, we know that the right is
     * present.
     * Since NEST_MODIFYAC_ACCESS gives the access permissions to 
     * modify acls, this right is checked instead of adding a new
     * acl, say, NEST_ATTACH_LOT_ACCESS
     */
    if(!new_file_created) {
        if(!file_catalog->is_access_allowed(logical_filename,
                                            user,
                                            NEST_MODIFYAC_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    }
    
    /*
     * Check if the user has the rights to the lot 
     * so that it can be attached to the file.
     */
    NestLot lot;
    reply = lot_manager->get_lot_status(&lot, handle, user);
    if(reply != NEST_SUCCESS) {
        if(new_file_created) {
            file_catalog->delete_entry(logical_filename);
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    /*
     * Attach the lot to the file.
     */
    reply = file_catalog->attach_lot_to_file( logical_filename, handle );
    if(reply != NEST_SUCCESS) {
        if(new_file_created) {
            file_catalog->delete_entry(logical_filename);
        }
        return reply;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::detach_lot_from_file(NestLotHandle *handle,
                                         const char *file,
                                         const char *user,
                                         const char *cwd)
{
    if(!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    NestReplyStatus reply;
    string          filename;
    char            logical_filename[NEST_MAXPATHLEN];

    reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    NEST_REPLY_CHECK(reply);
   
    if( file_catalog->is_system_file(logical_filename) ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }
    
    /*
     * Check if the user has the right to detach a lot from the file.
     */
    if(!file_catalog->is_access_allowed(logical_filename,
                                        user,
                                        NEST_MODIFYAC_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }
    
    /*
     * Check if the user has the rights to the lot 
     * so that it can be attached to the file.
     */
    NestLot lot;
    reply = lot_manager->get_lot_status(&lot, handle, user);
    if(reply != NEST_SUCCESS && reply != NEST_LOT_NOT_FOUND)
        return reply;

    /*
     * Detach the lot from the file.
     */
    return file_catalog->detach_lot_from_file( logical_filename, handle );
}

/*
    Returns the MAX(free_space) among all the devices available.
*/
NestReplyStatus
nestStorageManager::bytesAvailable(int is_user, long long int *free, 
        long long int *total ) 
{
    double avail_in_MB;
    double total_in_MB;

    if (NestEnableLots) {
        avail_in_MB = lot_manager->free_space(is_user);
        total_in_MB = lot_manager->total_space(is_user);
        *free  = (long long int) avail_in_MB * 1024 * 1024;
        *total = (long long int) total_in_MB * 1024 * 1024;
        return NEST_SUCCESS;
    } else {
        nest_fsstat fs_buf;

        if (NestStatFS(rootDir, &fs_buf) >= 0) {
            *free  = (long long int) fs_buf.f_bfree  * fs_buf.f_bsize;
            *total = (long long int) fs_buf.f_blocks * fs_buf.f_bsize; 
            return NEST_SUCCESS;
        } else {
            nest_debug( NEST_DEBUG_STORAGE,
                    "statfs failed in bytesAvailable()!\n");
            nest_perror( "statfs" );
            return ErrnoToNest( errno );
        }
    }
}


NestReplyStatus
nestStorageManager::filelength(   long *filesize, 
                                  const char *file,
                                  const char *user,
                                  const char *cwd )
{
    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_filename[NEST_MAXPATHLEN],
            physical_filename[NEST_MAXPATHLEN];

    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    NEST_REPLY_CHECK(reply);

    if( file_catalog->is_system_file(logical_filename) ) {

        if (!file_catalog->is_access_allowed_in_parent(logical_filename,
                                                       user,
                                                       NEST_LIST_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
        NEST_REPLY_CHECK(reply);

        reply = file_catalog->get_physical_name(physical_filename,
                                                logical_filename);
        NEST_REPLY_CHECK(reply);

        nest_stat   stat_buf;
        if (NestStat(physical_filename, &stat_buf) < 0) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Next to impossible!\n" );
            *filesize = 0;
            return NEST_UNKNOWN_ERROR; 
        }
   
        *filesize = stat_buf.st_size; 
        return NEST_SUCCESS;
    }
    else {
        NestFileInfo info;
        reply = fileinfo( &info, file, user, cwd );
        if( reply == NEST_SUCCESS ) {
            *filesize = info.size;
        }
        return reply;
    }
}

NestReplyStatus
nestStorageManager::fileusage( double *size,
                               const char *file,
                               const char *user,
                               const char *cwd )
{
    NestFileInfo    info;
    NestReplyStatus status;
    *size = 0;

    status = fileinfo( &info,
                       file,
                       user,
                       cwd );
    NEST_REPLY_CHECK(status);

    ClassAd         *file_ad=info.classad;
    ExprTree *lot_list = file_ad->Lookup( "LotHandle" );
    if (!lot_list) {
        NestFree(info.classad);
        NestFree(&info);
        return NEST_CLASSAD_ERROR;
    }

    string str_lot_list = "";
    unparser.Unparse( str_lot_list, lot_list );
    *size = NestStrLotList_to_double( str_lot_list.c_str() );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::fileinfo(   NestFileInfo *info, 
                                const char *file,
                                const char *user,
                                const char *cwd,
                                int link_count ) 
{


    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    nest_debug( NEST_DEBUG_STORAGE,
                "file info %s %s %s - %s\n",
                file,
                user,
                cwd,
                filename.c_str() );

    if (!info) {
        return NEST_BAD_PARAMETERS;
    }

    char    logical_filename[NEST_MAXPATHLEN],
            physical_filename[NEST_MAXPATHLEN];


    reply = file_catalog->get_logical_name(logical_filename,
                                           filename.c_str(),
                                           user);
    NEST_REPLY_CHECK(reply);

    if (!file_catalog->is_access_allowed_in_parent(logical_filename,
                                                   user,
                                                   NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_physical_name(physical_filename,
                                            logical_filename);
    NEST_REPLY_CHECK(reply);

    nest_obj_type_t     filetype;

    reply = file_catalog->get_type( &filetype,
                                    logical_filename );
    NEST_REPLY_CHECK(reply);
    info->type = filetype;

    if ((filetype == NEST_LINK) &&
        !file_catalog->is_system_file(logical_filename)) 
    {

        if (link_count++ > NEST_LINK_MAXLOOP) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Too many levels of symbolic links!\n" );
            return NEST_TOO_MANY_LINKS;
        }

            // recursion
        return fileinfo( info,
                           physical_filename,
                           user,
                           cwd,
                           link_count );
    }

        // set classad and uniquifier
    if(!file_catalog->is_system_file(logical_filename)) {
        reply = file_catalog->get_file_classad(logical_filename, &(info->classad) );
        NEST_REPLY_CHECK(reply);
    }
    reply = file_catalog->get_file_uniquifier( 
            physical_filename, &(info->uniquifier) );
    NEST_REPLY_CHECK(reply);

    info->size = NestFilelength( physical_filename );

        // if we make it here, we succeeded in the lookup, cache the result
    cache_uniquifier( filename, &(info->uniquifier), info->type ); 
    
    return NEST_SUCCESS;
}

bool
nestStorageManager::is_valid_user( const char *user, user_id_t *uid ) {
    // first find the uid for this user
    // then make sure the identity manager knows about them
    
    /*
        We need an UID for anonymous_user's puts also - not just for regular
        puts!

        Ideally, we need UIDs for anonymous users too, or simply
        whatever name anonymous_user points too should be a real username
        with a valid UID attached to it and default quotas set-up for
        anonymous users!

        For the time being, let all files created by anonymous users
        belong to AdminUser, if there is anonymous_username is invalid
    */
    
        // Problem.  A client can fail to connect if
        // identity_manager->get_uid fails.
        // So we need to make sure that dynamically created users
        // will have a uid assigned to them by identity_manager->get_uid.
    if ( NestAnonymousUser && (!strcmp( NestAnonymousUser, user)) ) {
        *uid = AdminUid;
        return true;
    } else {
		char	*tmp = NestGetUserFromIdentity( user );
		if ( ! tmp ) {
			return false;
		}
        char *local_user = NestStrdup( tmp );
        NestReplyStatus reply = identity_manager->get_uid( uid, local_user );
        if(local_user) NestFree(local_user);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_AUTH | NEST_DEBUG_STORAGE,
                    "Unable to find a uid for %s.\n", user );
            return false;
        }
        return identity_manager->is_user_present(user);
    }

    //return identity_manager->is_user_present(user);
}

NestReplyStatus
nestStorageManager::formatDisk( const char *user,
                                const char *group ) {

    return NEST_NOT_YET_IMPLEMENTED;

/*
    char    logical_home[NEST_MAXPATHLEN];
    NestReplyStatus reply = file_catalog->get_logical_name(logical_home,
                                                           "/",
                                                           user);

    NEST_REPLY_CHECK(reply);

    //  TODO: fix this

    return rmDir ( user,
                   logical_home,
                   "..",
                   true );
*/

}

int
nestStorageManager::chRootDir( ) {
    int ret = chdir( rootDir );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "Couldn't chdir to rootdir %s: %s\n", rootDir,
                strerror( errno ) );
    }
    return ret;
}

int
nestStorageManager::chMetaDir( ) {
    if ( chRootDir() < 0 || chdir( MetaDir ) < 0 ) {
        nest_warn( __FILE__, __FUNCTION__,"chdir to MetaDir %s: %s\n", MetaDir, strerror(errno));
        return -1;
    }
    return 0;
}

NestReplyStatus
nestStorageManager::get_home_dir( string &result,
                                  const char *username ) {

    /*
        File Catalog manager - depends heavily on this
        format for the homedir!

        If you change this - make sure NestFileCatalog
        is aware of this and is happy
    */

    result = "";
    result = result + rootDir + "/" + username + ":";

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::add_home_dir(const char *username,
                                 const char *groupname) {

    char    logical_home[NEST_MAXPATHLEN];
    /*
        ignoring return value intentionally
    */
    NestReplyStatus reply = file_catalog->get_logical_name(logical_home,
                                                           "/",
                                                           username);
    string  physical_home;
    reply = get_home_dir( physical_home, username );
    NEST_REPLY_CHECK(reply);

    if ( NestMkdir( physical_home.c_str(), DIR_MODE ) < 0 ) {
        if (errno != EEXIST) {
            nest_warn( __FILE__, __FUNCTION__, "mkdir of %s failed: %s\n",
                       physical_home.c_str(),
                       strerror( errno ) );
            return ErrnoToNest( errno );
        }
    }

    NestChmod( physical_home.c_str(), DIR_MODE );

    user_id_t   uid = (user_id_t) -1;
    is_valid_user( username, &uid );

    NestChown( physical_home.c_str(), uid, ActualGid );

    reply = file_catalog->add_entry( logical_home,
                                     physical_home.c_str(),
                                     "/",
                                     username,
                                     groupname,
                                     NEST_DIR);

    if (reply == NEST_FILE_EXISTS) {
        reply = NEST_SUCCESS;
    }

    return reply;
}

NestReplyStatus
nestStorageManager::delete_home_dir(const char *username) {

    char    logical_home[NEST_MAXPATHLEN];
    /*
        ignoring return value intentionally
    */

    NestReplyStatus reply = file_catalog->get_logical_name(logical_home,
                                                           "/",
                                                           username);

    reply = file_catalog->delete_entry( logical_home );

    if (reply != NEST_SUCCESS &&
        reply != NEST_FILE_NOT_FOUND) {
        return reply;
    }

    string  physical_home;
    get_home_dir( physical_home, username );

    if ( NestRmdir( physical_home.c_str() ) < 0 ) {
        if (errno != ENOENT) {
            nest_warn( __FILE__, __FUNCTION__, "rmdir of %s failed: %s\n",
                       physical_home.c_str(),
                       strerror( errno ) );
            return ErrnoToNest( errno );
        }
    }

    return NEST_SUCCESS;
}
    
NestReplyStatus  
nestStorageManager::rememberLength( const char *user, const char *cwd, 
        const char *file, long *length ) 
{
        // this is for proxys to predict the length of previously fetched files.
        // not necessary on a "real" dedicated nest
    return NEST_NOT_YET_IMPLEMENTED;
}

char *
nestStorageManager::type( ) {
    return "NeST";
}

bool
nestStorageManager::is_lots_enabled( ) {
    return NestEnableLots;
}
