/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_worker_pool.h"
#include "nest_concurrency_manager.h"
#include "nest_filetransfer.h"

// NOTE that nest_worker and nest_worker_pool object speak nest_protocol
// between them, regardless of what protocol is spoken by the client

#define WORKER_ITERATOR list< nestWorker *>::iterator

nestWorker::nestWorker( pid_t pid, int stream, int id ) {
    this->pid    = pid;
    this->id     = id;
    this->idle   = WORKER_IDLE;
    this->stream  = stream;
}

pid_t
nestWorker::getPid() {
    return pid;
} 

bool
nestWorker::isIdle() {
    return( idle == WORKER_IDLE );
}

void
nestWorker::setIdle( int state ) {
    idle = state;
}

bool
nestWorker::sendMessage( const char * msg ) {
    nest_debug( NEST_DEBUG_POP, "Gonna send %s to worker %d\n", msg, id );
    return SendNestMessage( stream, msg );
}

bool
nestWorker::canReceiveMessage( fd_set *readyfds ) {
    if ( NestFd_isset( stream, readyfds ) ) {
        NestFd_clear( stream, readyfds ); // so we don't loop
        return true;
    } else {
        return false;
    }
}

int 
nestWorker::getID( ) {
    return id;
}

int
nestWorker::getStream() {
    return stream;
}

NestReplyStatus
nestWorker::assignTask( nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;

    this->client         = client;

    if ( client->getConnection()->getType() == GFTP_PROTOCOL
            || client->getConnection()->getType() == FTP_PROTOCOL ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    char msg[NEST_MESSAGESIZE];

        // TODO:  Instead of sending weird stuff here, we should have a
        // function that sends a client and a function that receives one
        // we'll need more client info to pass authentication for 3pt's
    int type = client->getRequest()->getType();
    switch( type ) {
    case PROXY_PUT:
    case PROXY_GET:
        nest_debug( NEST_DEBUG_PROXY, "Assigning a proxy task to worker\n" );
            // fall through
    case CLIENT_SEND_FILE:
    case CLIENT_RECV_FILE:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %d %s %ld %ld\n",
            type, client->getConnection()->getType(),
            client->getStorageConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getNestFile(),
            client->getRequest()->getStart(),
            client->getRequest()->getLength() );
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %d %s %s %s %ld\n",
                type, client->getConnection()->getType(),
                client->getStorageConnection()->getType(),
                client->getId(), client->getConnection()->getSocket(),
                client->getRequest()->getNestFile(),
                client->getRequest()->getRemoteFile(),
                client->getRequest()->getRemoteHost(),
                client->getRequest()->getLength() );
        break;
    case TEST_SERVER_TO_CLIENT:
    case TEST_CLIENT_TO_SERVER:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %ld\n",
            type, client->getConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getLength() );
        break;
    case COPY_REQUEST:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %d %d %s %s %ld %ld\n",
            type, client->getConnection()->getType(),
            client->getStorageConnection()->getType(),
            client->getOtherStorageConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getNestFile(),
            client->getRequest()->getLinkName( ),
            client->getRequest()->getStart(),
            client->getRequest()->getLength() ); 
            break;
    default:
        nest_die( "unknown transfer type: %d\n", type );
    }

        // send the message and then send the necessary protocol connections
    nest_debug( NEST_DEBUG_POP, "Sending msg %s to w%d\n", msg, getID() );
    sendMessage( msg );
    if ( type != TEST_SERVER_TO_CLIENT && type != TEST_CLIENT_TO_SERVER ) {
        status = nestProtocolManager::SendConnection(
            stream, client->getStorageConnection() );
        NEST_REPLY_CHECK(status);
        if(type == COPY_REQUEST){
            status = nestProtocolManager::SendConnection(
            stream, client->getOtherStorageConnection() );
            NEST_REPLY_CHECK(status);
        }
            
    }
    return nestProtocolManager::SendConnection(stream, client->getConnection());
}

void
nestWorker::endTask( bool success, long bytes_transferred ) {
        // TODO: one time this client pointer was bad in gdb...?
    client->getTransfer()->setBytesTransferred( bytes_transferred );
    //client->endTransfer( success );   // taskDone will do this
    this->client = NULL;
    setIdle( true );
}

nestClient *
nestWorker::getClient( ) {
    return client;
}


nestWorkerPool::nestWorkerPool( int quantity, fd_set *watchfds,
                int (*worker_loop)( int ) )
{
    maxfd = 0;
    worker_id = 0;
    this->worker_loop = worker_loop;
    createWorkers( quantity, watchfds );
}

int
nestWorkerPool::createWorkers( int quantity, fd_set *watchfds ) {

        int fd[2];
        pid_t pid;
        char  idstring[NEST_MESSAGESIZE];

        memset( idstring, 0, NEST_MESSAGESIZE );

        for( int i = 0; i < quantity; worker_id++, i++ ) {

                // set up stream pipe for the child
            if ( NestSpipe( fd ) < 0 ) {
                nest_perror( "s_pipe" );
                assert( 0 );
                return ERROR_EXIT;
            }

                // fork for each child
            if( ( pid = fork() ) < 0 ) {
                   nest_perror( "fork" );
                assert( 0 );
                return ERROR_EXIT;
            } else if ( pid > 0) {                  /* parent */
                NestClose( fd[1] );
                addWorker( pid, fd[0], worker_id, watchfds );
            } else {                                /* child */
                    // NestClose the unneeded pipe and setup the other pipes
                    // arrange the other pipes as stdin and stdout for children
                NestClose( fd[0] );
                if ( fd[1] != STDIN_FILENO ) {
                    NestDup2( fd[1], STDIN_FILENO );
                }

                if ( fd[1] != STDOUT_FILENO ) {
                    NestDup2( fd[1], STDOUT_FILENO );
                }

                NestSnprintf( idstring, sizeof(idstring), "%d", worker_id );
                worker_loop( worker_id );

                    // child should never get here
                nest_warn( __FILE__, __FUNCTION__, "Weirdness in nestWorkers %d\n", __LINE__ );
            }
            
        }
    return 0;
}

int
nestWorkerPool::addWorker( pid_t pid, int stream, int id, fd_set *watchfds ) {

    nestWorker *newWorker = new nestWorker( pid, stream, id );

    workers.push_back( newWorker );
    maxfd = ( maxfd > stream ) ? maxfd : stream;

    if ( watchfds ) NestFd_set( stream, watchfds ); 

    return workers.size();
}

int
nestWorkerPool::getMaxfds( ) {
    return maxfd;
}

ssize_t
nestWorkerPool::size( ) {
        return workers.size();
}

int
nestWorkerPool::removeWorker( pid_t pid, fd_set *watchfds, fd_set *readyfds ) {
    
    nestWorker *worker = getWorker( pid );

    if (worker == NULL) {
        return ERROR_EXIT;
    }

    if ( worker->getClient() != NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "Removing a worker %d with active client connection.\n",pid);
    }

    if ( watchfds ) NestFd_clear( worker->getStream(), watchfds );
    if ( readyfds ) NestFd_clear( worker->getStream(), readyfds );
    NestClose( worker->getStream() );
        // be careful about removing the workers because we may be in a signal
        // handler and when we return execution, we might have screwed up our
        // iterator, just set the guy busy so we won't use him anymore
    //workers.remove( worker );
    //delete( worker );
    nest_debug( NEST_DEBUG_POP, "Removing dead worker %d\n", worker->getID() );
    worker->setIdle( WORKER_DEAD );
        // TODO when a worker dies while performing a transfer, we should 
        // check client connection.  If still good, server needs to listen 
        // to client again?  Otherwise, client gets stuck. 
    return 0;
}

void
nestWorkerPool::debugLevel( long level ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    int busy = 0, failures = 0;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %ld\n", DEBUG_REQUEST, level );

    while( p != workers.end() ) {
        worker = *p;
        if ( ! worker->isIdle() ) {
            busy++;
        } else {
            worker->sendMessage( msg );
                // explicitly wait for ack here
            int unused;
            struct IO_Measurement unused_measure;
            if (NEST_SUCCESS != ReceiveWorkerMessage( 
                        worker->getStream(), unused, unused, &unused_measure ) )
            {
                failures++;
            }
        }
        p++; 
    }
    if ( busy ) nest_warn( __FILE__, __FUNCTION__, "Unable to set debug in %d busy workers\n",
            busy );
    if ( failures ) nest_warn( __FILE__, __FUNCTION__, "Failed to set debug in %d workers\n",
            failures );
    return;
}


nestWorker*
nestWorkerPool::getWorkerFromID( int id ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    
    while( p != workers.end() ) {
        worker = *p;
        if ( worker->getID() == id ) return worker;
        p++;
    }
    return NULL;
}


nestWorker*
nestWorkerPool::getWorker( pid_t pid ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    
    while( p != workers.end() ) {
        worker = *p;
        if ( worker->getPid() == pid ) return worker;
        p++;
    }
    return NULL;
}

// getting an idle worker automatically turns idle off, so make sure to use
// the worker
// eventually, we should add timestamps to when idle was turned off and reclaim
// any slackers
// TODO: make this constant time ??
nestWorker* 
nestWorkerPool::getIdleWorker( ) {

        WORKER_ITERATOR p = workers.begin();
        nestWorker *worker;

        while( p != workers.end() ) {
            worker = *p;
            if ( worker->isIdle() ) {
                worker->setIdle( false );    
                return worker;
            }
            p++;
        }
        return NULL;
}

/* 
returns a ptr to the worker which sent the message which is written into
the msg parameter, returns NULL if there are no more worker messages
*/
nestWorker *
nestWorkerPool::canReceiveMessage( fd_set *readyfds ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;

    while( p != workers.end() ) {
        worker = *p;
        if ( worker->canReceiveMessage( readyfds ) ) {
            return worker;
        }
        p++;
    }
    return NULL;
}

// used by worker processes
bool
nestWorkerPool::SendWorkerMessage( NestConnection con, int t_id, int client_id,
        int client_socket, int status, nestFiletransfer *transfer )
{
    char msg[NEST_MESSAGESIZE];
    long bytes_transferred = 0;
    int first_in_time = 0, first_in_size = 0;
    int first_out_time = 0, first_out_size = 0;
    long total_out_time = 0, total_in_time = 0;

    if ( transfer != NULL ) {
        struct IO_Measurement *measures = transfer->getMeasurements();
        assert( measures != NULL );
        bytes_transferred = transfer->getBytesTransferred();
        first_in_time = measures->first_in_time;
        first_out_time = measures->first_out_time;
        total_in_time = measures->total_in_time;
        total_out_time = measures->total_out_time;
        first_in_size = measures->first_in_size;
        first_out_size = measures->first_out_size;
    }

    
    NestSnprintf( msg, NEST_MESSAGESIZE, 
            "%d %d %d %d %d %d %ld %d %d %ld %ld %d %d\n",
            NEST_SUCCESS, 
            t_id, 
            REQUEST_DONE, 
            status, 
            client_id,
            client_socket, 
            bytes_transferred,
            first_in_time,
            first_out_time,
            total_in_time,
            total_out_time,
            first_in_size,
            first_out_size );
    nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_POP, 
            "W%d: Sending %s to boss.\n", t_id, msg );
    return SendNestMessage( con, msg );
}

NestReplyStatus
nestWorkerPool::ReceiveWorkerMessage( NestConnection con, 
        int &client_id, int &client_socket, struct IO_Measurement *measures )
{
    char msg[NEST_MESSAGESIZE];
    int worker_id; 
    NestReplyStatus worker_status, success, status;

        // receive message
    status = ReceiveNestMessage( con, msg );
    if ( status != NEST_SUCCESS ) return status;

        // parse message
    if ( sscanf( msg, "%d %d %d %d %d %d %ld %d %d %ld %ld %d %d",
            &status, 
            &worker_id, 
            &worker_status, 
            &success,
            &client_id, 
            &client_socket, 
            &(measures->total_bytes),
            &(measures->first_in_time), 
            &(measures->first_out_time),
            &(measures->total_in_time),
            &(measures->total_out_time), 
            &(measures->first_in_size),
            &(measures->first_out_size) ) != 13 ) 
    {
        nest_warn( __FILE__, __FUNCTION__, "recvWorkerMsg: bad format msg: %s\n", msg );
        return NEST_NO_CONNECTION;
    } else {
        nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_POP,
                "recv'd worker msg: %s\n", msg );
    }
 
        // check return values
    if ( status != NEST_SUCCESS ) return status; 
    if ( worker_status != REQUEST_DONE ) return NEST_UNKNOWN_ERROR;

    return success; 
}           
