/******************************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "nest_speak.h"

/* For platforms that don't have this bit, ignore it. */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

/*-----------------------------------------------------------------------------
 *
 * Private data structures
 *
 *---------------------------------------------------------------------------*/

struct parameters
{
    bool  verbose;
    char  *server;
    char  *chirp_port;
    char  *http_port;
    char  *ftp_port;
    char  *gftp_port;
    bool  test_versions;
    bool  test_dirs;
    bool  test_file;
    bool  test_xfers;
    bool  test_lots;
    bool  test_robustness;
    bool  test_user_mgmt;
    bool  test_misc;
    bool  test_stable_funcs;
    bool  test_all_funcs;
    bool  help;
    bool  loop;
};

/*-----------------------------------------------------------------------------
 *
 * Private global variables
 *
 *---------------------------------------------------------------------------*/

static int number_of_errors;

/*-----------------------------------------------------------------------------
 *
 * Private functions
 *
 *---------------------------------------------------------------------------*/

static void parse_command_line(int argc, char **argv,
                               struct parameters *parameters);
static void usage(char *program);
static void run_tests(struct parameters *parameters,NestConnection connection);

static void test_versions(struct parameters *parameters,
                          NestConnection connection);

static bool make_test_file(char *filename);
static bool compare_files(char *filename1, char *filename2);
static double get_avail_nest_space(NestConnection connection, int line);
static bool check_nest_sanity(struct parameters *parameters, int line,
                              char* protocol = "chirp");

static void test_directories(struct parameters *parameters,
                             NestConnection connection);
static bool get_directory(char **name, struct parameters *parameters,
                          NestConnection connection, int line);
static bool make_directory(char *name, struct parameters *parameters,
                           NestConnection connection, int line);
static bool change_directory(char *name, struct parameters *parameters,
                             NestConnection connection, int line);
static bool remove_directory(char *name, struct parameters *parameters,
                             NestConnection connection, int line);
static bool check_directories(char *name1, char *name2,
                              struct parameters *parameters, int line);

static void test_transfers(struct parameters *parameters,
                           NestConnection connection);
static void test_send_file(struct parameters *parameters,
                           NestConnection connection,
                           int line,
                           bool should_fail = false);
static void test_send_large_file(struct parameters *parameters,
                                 NestConnection connection,
                                 int line,
                                 bool should_fail = false);
static bool put_file(char *localfile, char *remotefile,
                     NestConnection connection, int line,
                     bool should_fail = false);
static bool get_file(char *remotefile, char *localfile, NestReplyStatus *status,
                     NestConnection connection, int line);
static bool remove_file(char *filename, NestConnection connection, int line);

static void test_lots(struct parameters *parameters, NestConnection connection);
static bool check_lots_enabled(NestConnection connection);
       bool get_lot(NestConnection connection, NestLotHandle *handle,
                    double size_in_MB, double minutes, int line);
       bool return_lot(NestConnection connection, NestLotHandle *handle,
                       int line);
static bool return_all_lots(NestConnection connection, int line);
static bool list_user_lots(NestConnection connection, int line);

static void test_misc(struct parameters *parameters,
                      NestConnection connection, bool all_cases = true);
static void test_get_physical_name(struct parameters *parameters,
                                   NestConnection connection);
static void test_links(struct parameters *paramters,
                       NestConnection connection, bool all_cases = true);
static bool sync_nest(NestConnection connection, int line);
static bool get_nest_server_time(NestConnection connection, int line);
static bool get_physical_name(NestConnection connection, char **physicalname,
                              char *logicalname, int line);
static bool create_link(char *oldPath, char *newPath,
                        NestConnection connection, int line);

static void test_user_mgmt(struct parameters *parameters,
                           NestConnection connection);
static void clean_test_users(NestConnection connection);
static bool check_admin_user(NestConnection connection);
static bool add_group(NestConnection connection, char *group, int line,
                      bool should_fail = false);
static bool del_group(NestConnection connection, char *group, int line,
                      bool should_fail = false);
static bool add_user_to_group(NestConnection connection, char *user,
                              char *group, int line, bool should_fail = false);
static bool del_user_from_group(NestConnection connection, char *user,
                              char *group, int line, bool should_fail = false);
static bool add_user(NestConnection connection, char *user,
                     char *group, int line, bool should_fail = false);
static bool del_user(NestConnection connection, char *user, int line,
                     bool should_fail = false);

static void test_robustness(struct parameters *parameters,
                 NestConnection connection, bool loop_endlessly = false);
static bool test_protocol_robustness(struct parameters *paramters,
                                     NestConnection connection,
                                     char *protocol = "chirp");
static void fill_buffer(char *buffer, int len);

static void test_file_operations(struct parameters *parameters,
                                 NestConnection connection,
                                 bool all_cases = true);

int
main(int argc, char *argv[])
{
    NestConnection     connection;
    NestReplyStatus    status;
    struct parameters  parameters;

    parse_command_line(argc, argv, &parameters);

    number_of_errors = 0;

    // Connect to the Nest server
    printf("Connecting to %s...\n\n", parameters.server);
    status = NestOpenConnection(connection, parameters.server);
    if(status != NEST_SUCCESS) {
        printf("FAILED: Couldn't connect to %s: %s in line %d\n",
                parameters.server, NestErrorString(status), __LINE__);
        number_of_errors++;
    } else {
        printf("OK: Connected to %s in line %d.\n",
                    parameters.server, __LINE__);

        run_tests(&parameters, connection);

        status = NestCloseConnection(connection);
        if(status != NEST_SUCCESS) {
            printf("FAILED: Couldn't close connection to %s: %s in line %d\n",
                   parameters.server, NestErrorString(status), __LINE__);
            number_of_errors++;
        } else {
            printf("OK: Connection closed in line %d.\n", __LINE__);
        }
    }

    printf("\n");
    if (number_of_errors == 0) {
        printf("OK: No errors.\n");
    } else if (number_of_errors == 1) {
        printf("FAILED: 1 error.\n");
    } else {
        printf("FAILED %d errors.\n", number_of_errors);
    }

    return number_of_errors;
}


/******************************************************************************
 *
 * Function: parse_command_line
 * Purpose:  Read the command line arguments and fill in
 *           the parameters structure to reflect them.
 *
 *****************************************************************************/
static void
parse_command_line(
     int                argc,
     char               **argv,
     struct parameters  *parameters)
{
    int   argument_index;
    bool  dump_usage;

    dump_usage = false;

    parameters->verbose             = false;
    parameters->server              = NULL;
    parameters->chirp_port          = strdup("7878");
    parameters->http_port           = strdup("80");
    parameters->ftp_port            = strdup("21");
    parameters->gftp_port           = strdup("2811");
    parameters->test_versions       = false;
    parameters->test_dirs           = false;
    parameters->test_file           = false;
    parameters->test_xfers          = false;
    parameters->test_lots           = false;
    parameters->test_user_mgmt      = false;
    parameters->test_robustness     = false;
    parameters->test_stable_funcs   = false;
    parameters->test_misc           = false;
    parameters->test_all_funcs      = false;
    parameters->help                = false;
    parameters->loop                = false;

    argument_index = 1;

    while (argument_index < argc) {
        if (!strcmp(argv[argument_index], "-v")
            || !strcmp(argv[argument_index], "-verbose")) {
            parameters->verbose = !parameters->verbose;
        } else if (!strcmp(argv[argument_index], "-versions")
            || !strcmp(argv[argument_index], "-V")) {
            parameters->test_versions = !parameters->test_versions;
        } else if (!strcmp(argv[argument_index], "-dirs")
            || !strcmp(argv[argument_index], "-d")) {
            parameters->test_dirs = !parameters->test_dirs;
        } else if (!strcmp(argv[argument_index], "-file")
            || !strcmp(argv[argument_index], "-F")) {
            parameters->test_file = !parameters->test_file;
        } else if (!strcmp(argv[argument_index], "-xfers")
            || !strcmp(argv[argument_index], "-x")) {
            parameters->test_xfers = !parameters->test_xfers;
        } else if (!strcmp(argv[argument_index], "-lots")
            || !strcmp(argv[argument_index], "-l")) {
            parameters->test_lots = !parameters->test_lots;
        } else if (!strcmp(argv[argument_index], "-user")
            || !strcmp(argv[argument_index], "-u")) {
            parameters->test_user_mgmt = !parameters->test_user_mgmt;
        } else if (!strcmp(argv[argument_index], "-min")
            || !strcmp(argv[argument_index], "-m")) {
            parameters->test_stable_funcs = !parameters->test_stable_funcs;
        } else if (!strcmp(argv[argument_index], "-robustness")
            || !strcmp(argv[argument_index], "-r")) {
            parameters->test_robustness = !parameters->test_robustness;
        } else if (!strcmp(argv[argument_index], "-misc")
            || !strcmp(argv[argument_index], "-M")) {
            parameters->test_misc = !parameters->test_misc;
        } else if (!strcmp(argv[argument_index], "-all")
            || !strcmp(argv[argument_index], "-a")) {
            parameters->test_all_funcs = !parameters->test_all_funcs;
        } else if (!strcmp(argv[argument_index], "-loop")
            || !strcmp(argv[argument_index], "-L")) {
            parameters->loop = !parameters->loop;
        } else if (!strcmp(argv[argument_index], "-H")
            || !strcmp(argv[argument_index], "-help")
            || !strcmp(argv[argument_index], "-?")) {
            parameters->help = !parameters->help;
        } else if (!strcmp(argv[argument_index], "-s")
                   || !strcmp(argv[argument_index], "-server")) {
            argument_index++;
            if (argument_index >= argc) {
                fprintf(stderr,
                        "The %s argument requires the name of the server.\n",
                        argv[argument_index-1]);
            } else {
                parameters->server = strdup(argv[argument_index]);
            }
        } else if (!strcmp(argv[argument_index], "-c")
                   || !strcmp(argv[argument_index], "-chirp-port")){
            argument_index++;
            if (argument_index >= argc) {
                fprintf(stderr,
                        "The %s argument requires the port number.\n",
                        argv[argument_index-1]);
            } else {
                parameters->chirp_port = strdup(argv[argument_index]);
            }
        } else if (!strcmp(argv[argument_index], "-h")
                   || !strcmp(argv[argument_index], "-http-port")){
            argument_index++;
            if (argument_index >= argc) {
                fprintf(stderr,
                        "The %s argument requires the port number.\n",
                        argv[argument_index-1]);
            } else {
                parameters->http_port = strdup(argv[argument_index]);
            }
        } else if (!strcmp(argv[argument_index], "-f")
                   || !strcmp(argv[argument_index], "-ftp-port")){
            argument_index++;
            if (argument_index >= argc) {
                fprintf(stderr,
                        "The %s argument requires the port number.\n",
                        argv[argument_index-1]);
            } else {
                parameters->ftp_port = strdup(argv[argument_index]);
            }
        } else if (!strcmp(argv[argument_index], "-g")
                   || !strcmp(argv[argument_index], "-gftp-port")){
            argument_index++;
            if (argument_index >= argc) {
                fprintf(stderr,
                        "The %s argument requires the port number.\n",
                        argv[argument_index-1]);
            } else {
                parameters->gftp_port = strdup(argv[argument_index]);
            }
        } else {
            fprintf(stderr, "Unknown argument: %s\n",
                    argv[argument_index]);
            dump_usage = true;

        }
        argument_index++;
    }

    if (!parameters->help && !dump_usage && parameters->server == NULL) {
        char server[4096];
        printf( "Enter server to test: " );
        if ( fgets( server, 4096, stdin )  != NULL ) {
            server[ strlen(server) - 1 ] = '\0';    // chomp
            parameters->server = NestStrdup( server );
        } else {
            perror( "fgets" );
            exit( -1 );
        }
    }

    if( !parameters->test_versions && !parameters->test_dirs
        && !parameters->test_xfers && !parameters->test_lots
        && !parameters->test_stable_funcs && !parameters->test_misc
        && !parameters->test_all_funcs && !parameters->test_robustness
        && !parameters->test_user_mgmt  && !parameters->test_file) {
        parameters->test_all_funcs = !parameters->test_all_funcs;
    }

    if (dump_usage || parameters->help) {
        usage(argv[0]);
        exit(1);
    }

    return;
}


/******************************************************************************
 *
 * Function: usage
 * Purpose:  Prints the help message
 *
 *****************************************************************************/
static void
usage(
    char *program)
{
        fprintf(stderr, "Usage: %s [OPTIONS] "
                        "<-s | -server servername>\n", program);
        fprintf(stderr, "Tests the NeST functions. "
                        "Options -a and -m take precedence "
                        "over other options for tests.\n");
        fprintf(stderr, "By default all the functions are tested.\n");
        fprintf(stderr, "       [-v | -verbose]                 "
                        "Print oodles of info.\n");
        fprintf(stderr, "       [-H | -help | -?]               "
                        "Print this message.\n");
        fprintf(stderr, "       <-s | -server servername>       "
                        "Connect to nest_server servername\n");
        fprintf(stderr, "       <-c | -chirp_port portnumber>   "
                        "Connect using chirp_port portnumber\n");
        fprintf(stderr, "       <-h | -http-port portnumber>    "
                        "Connect using http_port portnumber\n");
        fprintf(stderr, "       <-f | -ftp-port portnumber>     "
                        "Connect using ftp_port portnumber\n");
        fprintf(stderr, "       <-g | -gftp-port portnumber>    "
                        "Connect using gftp_port portnumber\n");
        fprintf(stderr, "       [-V | -versions]                "
                        "Print the NeST server and client versions.\n");
        fprintf(stderr, "       [-L | -loop]                    "
                        "Run the tests in a endless loop!.\n");
        fprintf(stderr, "                                       "
                        "Used along with -r option to continuously "
                        "blast packets to the server till it fails.\n");
        fprintf(stderr, "       [-d | -dirs]                    "
                        "Test the directory handling related APIs.\n");
        fprintf(stderr, "       [-F | -file]                    "
                        "Test the basic file system operations.\n");
        fprintf(stderr, "       [-x | -xfers]                   "
                        "Test the Get and Put functions "
                        "with lots disabled at server.\n");
        fprintf(stderr, "       [-l | -lots]                    "
                        "Test the Lot management related functions.\n");
        fprintf(stderr, "       [-u | -user]                    "
                        "Test the user management related functions.\n");
        fprintf(stderr, "       [-r | -robustness]              "
                        "Test for robustness.\n");
        fprintf(stderr, "       [-M | -misc]                    "
                        "Test miscellaneous nest API function calls.\n");
        fprintf(stderr, "       [-m | -min]                     "
                        "Test only the minimum set of functions "
                        "that are known to be bug free!.\n");
        fprintf(stderr, "       [-a | -all]                     "
                        "Test all the functions.\n");
}


/******************************************************************************
 *
 * Function: run_tests
 * Purpose:  Run all of the various tests that we do.
 *
 *****************************************************************************/
static void
run_tests(
    struct parameters *parameters,
    NestConnection     connection)
{

    bool lots_enabled = check_lots_enabled(connection);

    if(parameters->test_all_funcs){
        test_versions(parameters, connection);
        test_directories(parameters, connection);
        test_file_operations(parameters, connection, true);
        if(lots_enabled)
            test_lots(parameters, connection);
        else
            test_transfers(parameters, connection);
        test_misc(parameters, connection);
        test_user_mgmt(parameters, connection);
        test_robustness(parameters, connection);
    } else if(parameters->test_stable_funcs){
        test_versions(parameters, connection);
        test_directories(parameters, connection);
        test_file_operations(parameters, connection, false);
        if(lots_enabled)
            test_lots(parameters, connection);
        else
            test_transfers(parameters, connection);
        test_misc(parameters, connection, false);
        test_user_mgmt(parameters, connection);
    } else {
        if(parameters->test_lots){
            if(lots_enabled)
                test_lots(parameters, connection);
        }
        if(parameters->test_xfers)
            test_transfers(parameters, connection);
        if(parameters->test_dirs)
            test_directories(parameters, connection);
        if(parameters->test_file)
            test_file_operations(parameters, connection);
        if(parameters->test_versions)
            test_versions(parameters, connection);
        if(parameters->test_user_mgmt)
            test_user_mgmt(parameters, connection);
        if(parameters->test_robustness)
            if(parameters->loop)
                test_robustness(parameters, connection, true);
            else
                test_robustness(parameters, connection);
        if(parameters->test_misc)
            test_misc(parameters, connection);
        
    }

    return;
}


/******************************************************************************
 *
 * Function: make_test_file
 * Purpose:  This makes a file and fills it with 100 lines of the
 *           alphabet. It should be generalized.
 *
 *****************************************************************************/
static bool
make_test_file(char *filename)
{
    FILE  *test_file;
    bool  succeeded;

    unlink(filename);
    test_file = fopen(filename, "w");
    if (test_file == NULL) {
        printf("FAILED: Couldn't make test file \"%s\" in line %d\n",
               filename, __LINE__);
        succeeded = false;
        number_of_errors++;
    } else {
        for (int i = 0; i < 100; i++) {
            fprintf(test_file, "abcdefghijklmnopqrstuvwxyz\n");
        }
        fclose(test_file);
        succeeded = true;
    }
    return succeeded;
}



/******************************************************************************
 *
 * Function: compare_files
 * Purpose: This function compares two files and returns true if
 *             they have the same contents else it returns false
 *
 *****************************************************************************/
static bool
compare_files(
    char *filename1,
    char *filename2 )
{
    FILE *file1;
    FILE *file2;
    char buf1[MAXPATHLEN];
    char buf2[MAXPATHLEN];
    int read_count_1;
    int read_count_2;

    file1 = fopen(filename1, "r");
    file2 = fopen(filename2, "r");

    while( !feof(file1) ) {
        read_count_1 = fread(buf1, sizeof(char), MAXPATHLEN, file1);
        read_count_2 = fread(buf2, sizeof(char), MAXPATHLEN, file2);
        if( memcmp(buf1, buf2,
                   read_count_1>read_count_2 ? read_count_1 : read_count_2 ) ){
            return false;
        }
    }
    return true;
}



/******************************************************************************
 *
 * Function: get_avail_nest_space
 * Purpose : Returns the available Nest Space in MBs
 *
 *****************************************************************************/
static double
get_avail_nest_space(
    NestConnection connection,
    int            line)
{
    NestReplyStatus status;
    float avail_bytes, total_bytes;

    status = NestAvailableSpace( avail_bytes, total_bytes, connection );
    if ( status != NEST_SUCCESS ) {
        printf("FAILED: AvailableNestSpace (%s) in line %d.\n",
                    NestErrorString(status), line );
        number_of_errors++;
    } else {
        printf("OK: Available Nest Space: %f MB.\n",
                    (double) ( avail_bytes / (1024 * 1024) ) );
    }

    return (double) ( avail_bytes / (1024 * 1024 ) );
}


/******************************************************************************
 *
 * Function: check_nest_sanity
 * Purpose : Sends some requests to check whether the server is up and
 *           servicing requests. Currently just sends a request for the
 *           server time stamp using the chirp protocol. Should be modified
 *           to send requests using other protocols too.
 *
 *****************************************************************************/
static
bool check_nest_sanity(
    struct parameters *parameters,
    int               line,
    char              *protocol)
{
    NestReplyStatus status;
    NestConnection  connection;
    int             portnumber;
    int             succeeded = true;

    if( !strcmp(protocol, "chirp") )
        portnumber = atoi(parameters->chirp_port);
    else if( !strcmp(protocol, "http") )
        portnumber = atoi(parameters->http_port);
    else if( !strcmp(protocol, "ftp") )
        portnumber = atoi(parameters->ftp_port);
    else if( !strcmp(protocol, "gftp") )
        portnumber = atoi(parameters->gftp_port);

    status = NestOpenConnection(connection, parameters->server, portnumber);
    if ( status != NEST_SUCCESS ) {
        printf("FAILED: NestOpenConnection (%s) in line %d.\n",
                    NestErrorString(status), __LINE__ );
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Opened a new connection to NeST using %s port.\n",
                protocol);
    }

    if( get_nest_server_time(connection, __LINE__) )
        succeeded = true;
    else
        succeeded = false;

    status = NestCloseConnection(connection);
    if ( status != NEST_SUCCESS ) {
        printf("FAILED: NestCloseConnection (%s) in line %d.\n",
                    NestErrorString(status), __LINE__ );
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Closed %s connection to NeST.\n",
                protocol);
    }

    if(succeeded)
        printf("OK: NeST sanity check succeeded.\n");
    else{
        printf("FAILED: NeST sanity check falied.\n");
        number_of_errors++;
    }

    return succeeded;

}



/******************************************************************************
 *
 * Function: test_versions
 * Purpose:  Gets the version of the server.
 *
 *****************************************************************************/
static void
test_versions(
    struct parameters *parameters,
    NestConnection     connection)
{
    float            nest_version;
    NestReplyStatus  status;

    /* ----- Check versions ----- */

    printf("OK: Nest Client Protocol Version: %f\n", (float) NEST_VERSION);

    status = NestGetVersion(&nest_version, connection);
    if (status == NEST_SUCCESS) {
        printf("OK: Nest Server Protocol Version: %f\n", nest_version);
    } else {
        printf("FAILED: NestGetVersion (%s) in line %d\n",
               NestErrorString(status), __LINE__);
        number_of_errors++;
    }
    return;
}


/*-----------------------------------------------------------------------------
 *
 * Functions to twiddle with directories
 *
 *---------------------------------------------------------------------------*/

/******************************************************************************
 *
 * Function: test_directories
 * Purpose:  Does a bunch of operations on directories.
 *
 *****************************************************************************/
static void
test_directories(
    struct parameters *parameters,
    NestConnection     connection)
{
    char *initial_directory, *return_directory;
    char *test1_directory, *test1_return_directory;
    char *test2_directory, *test2_return_directory;
    char *test2_1_directory, *test2_1_return_directory;

    if (get_directory(&initial_directory, parameters, connection, __LINE__)) {
        // Make directory hierarchy
        make_directory("test1", parameters, connection, __LINE__);
        make_directory("test2", parameters, connection, __LINE__);
        make_directory("test2/test2.1", parameters, connection, __LINE__);

        // Make sure I can change into test1 and back
        change_directory("test1", parameters, connection, __LINE__);
        get_directory(&test1_directory, parameters, connection, __LINE__);
        change_directory("..", parameters, connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);

        // Make sure I can change into test2, test2.1, and back
        change_directory("test2", parameters, connection, __LINE__);
        get_directory(&test2_directory, parameters, connection, __LINE__);
        change_directory("test2.1", parameters, connection, __LINE__);
        get_directory(&test2_1_directory, parameters, connection, __LINE__);
        change_directory("..", parameters, connection, __LINE__);
        get_directory(&test2_return_directory, parameters,
                      connection, __LINE__);
        check_directories(test2_directory, test2_return_directory,
                          parameters, __LINE__);

        // Make sure I can change into ~
        change_directory("~", parameters, connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);

        // Make sure I can change into ~/
        // We make sure we're somewhere else first.
        change_directory("test1", parameters, connection, __LINE__);
        change_directory("~/", parameters, connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);

        // Make sure I can change into test2.1, then immediately into test1
        change_directory("test2/test2.1", parameters, connection, __LINE__);
        get_directory(&test2_1_return_directory, parameters,
                      connection, __LINE__);
        check_directories(test2_1_directory, test2_1_return_directory,
                          parameters, __LINE__);
        change_directory("../../test1", parameters, connection, __LINE__);
        get_directory(&test1_return_directory, parameters,
                      connection, __LINE__);
        check_directories(test1_directory, test1_return_directory,
                          parameters, __LINE__);

        // Now test some limits
        change_directory("~", parameters, connection, __LINE__);
        change_directory(".", parameters, connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);

        change_directory("./test1/./../test2/./..", parameters,
                         connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);

        // Make sure we can remove our directories.
        change_directory("~", parameters, connection, __LINE__);
        remove_directory("test1", parameters, connection, __LINE__);
        remove_directory("test2/test2.1", parameters, connection, __LINE__);
        remove_directory("test2", parameters, connection, __LINE__);


        // Do some tests with file names and directories having
        // . and .. in their names
        get_directory(&initial_directory, parameters, connection, __LINE__);
        make_directory(".t..est.1..", parameters, connection, __LINE__);
        change_directory(".t..est.1..", parameters, connection, __LINE__);
        make_directory("..t.est..2.", parameters, connection, __LINE__);
        change_directory("..t.est..2.", parameters, connection, __LINE__);
        change_directory("../..", parameters, connection, __LINE__);
        get_directory(&return_directory, parameters, connection, __LINE__);
        check_directories(initial_directory, return_directory,
                          parameters, __LINE__);
        change_directory("../../", parameters, connection, __LINE__);
        remove_directory("./../.././.t..est.1.././..t.est..2.",
                          parameters, connection, __LINE__);
        remove_directory(".t..est.1..", parameters, connection, __LINE__);


    }
    return;
}

/******************************************************************************
 *
 * Function: get_directory
 * Purpose:  Finds out the current working directory.
 *
 *****************************************************************************/
static bool
get_directory(
    char              **name,
    struct parameters *parameters,
    NestConnection    connection,
    int               line)
{
    bool             succeeded;
    NestReplyStatus  status;

    status = NestGetPwd(name, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestGetPwd (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Got current directory: %s in line %d\n",
               *name, line);
        succeeded = true;
    }

    return succeeded;
}

/******************************************************************************
 *
 * Function: make_directory
 * Purpose:  Checks to make sure we can make a directory.
 *
 *****************************************************************************/
static bool
make_directory(
    char              *name,
    struct parameters *parameters,
    NestConnection    connection,
    int               line)
{
    bool             succeeded;
    NestReplyStatus  status;

    status = NestMkDir(name, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestMkDir (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = true;
    } else {
        printf("OK: Made directory: %s in line %d\n", name, line);
        succeeded = false;
    }
    return succeeded;
}

/******************************************************************************
 *
 * Function: change_directory
 * Purpose:  Checks to make sure we can change directories.
 *
 *****************************************************************************/
static bool
change_directory(
    char              *name,
    struct parameters *parameters,
    NestConnection    connection,
    int               line)
{
    bool             succeeded;
    NestReplyStatus  status;

    status = NestChangeDir(name, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestChangeDir (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Changed into %s in %d.\n", name,  line);
        succeeded = true;
    }

    return succeeded;
}

/******************************************************************************
 *
 * Function: remove_directory
 * Purpose:  Checks to make sure we can remove a directory.
 *
 *****************************************************************************/
static bool
remove_directory(
    char              *name,
    struct parameters *parameters,
    NestConnection    connection,
    int               line)
{
    bool             succeeded;
    NestReplyStatus  status;

    status = NestRmDir(name, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestRmDir (%s) in line %d.\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Removed %s in line %d.\n", name, line);
        succeeded = true;
    }

    return succeeded;
}

/******************************************************************************
 *
 * Function: check_directories
 * Purpose:  This is a wrapper for strcmp: The strings contain
 *           two directory names gotten from GetPwd().
 *
 *****************************************************************************/
static bool
check_directories(
    char              *name1,
    char              *name2,
    struct parameters *parameters,
    int               line)
{
    bool  succeeded;

    if (strcmp(name1, name2) == 0) {
        printf("OK: check_directories in line %d.\n", line);
        succeeded = true;
    } else {
        printf("FAILED: check_directories (%s, %s) in line %d.\n",
               name1, name2, line);
        number_of_errors++;
        succeeded = false;
    }
    return succeeded;
}


/*-----------------------------------------------------------------------------
 *
 * Functions to twiddle with sending and receiving files.
 *
 *---------------------------------------------------------------------------*/

static void
test_transfers(
    struct parameters *parameters,
    NestConnection     connection)
{
    test_send_file(parameters, connection, __LINE__);
    // We should test sending/receiving a 0K file
    // We should test sending/receiving with progress indication.
}


/******************************************************************************
 *
 * Function: test_send_file
 * Purpose:  This sends and receives a single file. It should be broken
 *           into multiple functions. If the should_fail
 *           flag is set, sending the file is expected to fail (due to
 *           space restrictions, non-reservation of lots, etc.).
 *
 *****************************************************************************/
static void
test_send_file(
    struct parameters *parameters,
    NestConnection     connection,
    int line,
    bool should_fail)
{

    if (make_test_file("blah")) {
        if( put_file("blah", "blah", connection, line, should_fail)
            && !should_fail){
            if( get_file("blah", "blah2", NULL, connection, line) ){
                if( !compare_files("blah2", "blah") ){
                    printf("FAILED: Test file \"%s\" was corrupted in line "
                           "%d.\n", "blah", line);
                    number_of_errors++;
                } else{
                    printf("OK: NestReceiveFile did not get corrupted file "
                           "in line %d.\n", line);
                }
                unlink("blah2");
            }
            remove_file("blah", connection, line);
        }
        unlink("blah");
    }

    return;
}

/******************************************************************************
 *
 * Function: test_send_large_file
 * Purpose:  This sends and receives a single file of size greater
 *           than one MB. The file that is used is /bin/rpm and is
 *           assumed to be present. If the should_fail
 *           flag is set, sending the file is expected to fail (due to
 *           space restrictions, non-reservation of lots, etc.).
 *
 *****************************************************************************/
static void
test_send_large_file(
    struct parameters *parameters,
    NestConnection     connection,
    int line,
    bool should_fail)
{
    char    *testfile1;
    char    *testfile2;

    testfile1 = strdup("/bin/rpm");
    testfile2 = NestExtractFilename(testfile1);

    if( put_file(testfile1, testfile2, connection, line, should_fail)
        && !should_fail){
        if( get_file(testfile2, testfile2, NULL, connection, line) ){
            if( !compare_files(testfile2, testfile1) ){
                printf("FAILED: Test file \"%s\" was corrupted in line "
                       "%d.\n", testfile1, line);
                number_of_errors++;
            } else{
                printf("OK: NestReceiveFile did not get corrupted file "
                       "in line %d.\n", line);
            }
            unlink(testfile2);
        }
        remove_file(testfile2, connection, line);
    }

    return;
}



/******************************************************************************
 *
 * Funtion: put_file
 * Purpose: Checks to make sure we can send files. If the should_fail
 *          flag is set, sending the file is expected to fail (due to
 *          space restrictions, non-reservation of lots, etc.).
 *          Returns true if file was sent successfully.
 *
 *****************************************************************************/
static bool
put_file(
    char           *localfile,
    char           *remotefile,
    NestConnection connection,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestSendFile(localfile, remotefile, connection);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestSendFile did not fail as expected in line %d.\n",
                line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestSendFile (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestSendFile (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestSendFile succeeded in line %d.\n",
                    line);
        succeeded = true;
    }

    return succeeded;
}


/******************************************************************************
 *
 * Funtion: get_file
 * Purpose: Checks to make sure we can get files.
 *          Note that NEST_TOO_MANY_LINKS is treated as a success condition
 *          rather than a failure. This is because it is known that cyclic
 *          links are not introduced in the tests expect where they are
 *          explicitly created and that test makes sure that cyclic links
 *          are handled correctly
 *
 *****************************************************************************/
static bool
get_file(
    char            *remotefile,
    char            *localfile,
    NestReplyStatus *status,
    NestConnection  connection,
    int             line)
{
    bool            succeeded;
    if(status == NULL)
        status = new NestReplyStatus;

    *status = NestReceiveFile(remotefile, localfile, connection);
    if (*status != NEST_SUCCESS && *status != NEST_TOO_MANY_LINKS) {
        printf("FAILED: NestReceiveFile (%s) in line %d.\n",
                    NestErrorString(*status), line);
        number_of_errors++;
        succeeded = false;
    }
    else {
        printf("OK: NestReceiveFile succeeded in line %d.\n", line);
        succeeded = true;
    }

    return succeeded;
}



/******************************************************************************
 *
 * Funtion: remove_file
 * Purpose: Checks to make sure we can remove files
 *
 *
 *****************************************************************************/
static bool
remove_file(
    char           *filename,
    NestConnection connection,
    int            line)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestRemoveFile( filename , connection );
    if ( status != NEST_SUCCESS ) {
        printf( "FAILED: RemoveNestFile (%s) in line %d.\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        succeeded = true;
    }

    return succeeded;

}



/*-----------------------------------------------------------------------------
 *
 * Functions to twiddle with lots
 *
 *---------------------------------------------------------------------------*/

/******************************************************************************
 *
 * Function: test_lots
 * Purpose:  Does a bunch of operations on lots.
 * Note:     This needs to be generalized. It should call subfunctions
 *           that test everything we expect to work.
 *           We could make a lot with a small size, try to transfer a
 *           file, hopefully fail, upgrade the lot size, try again, etc...
 *
 *****************************************************************************/
static void
test_lots(
    struct parameters *parameters,
    NestConnection    connection)
{

    NestReplyStatus status1;
    NestReplyStatus status2;
    NestLot         nest_lot;
    NestLotHandle   handle1;
    NestLotHandle   handle2;
    double          size_in_MB;
    double          minutes;
    double          avail_size;


    /* Return all existing lots */
    return_all_lots(connection, __LINE__);

    /* Reserve lot of 1 MB and again list the  Lots */
    size_in_MB = 1;
    minutes = 120;
    get_lot(connection, &handle1, size_in_MB, minutes, __LINE__);
    list_user_lots(connection, __LINE__);

    /* Transfer a small file (less than 1 MB and test lot allocation */
    test_send_file( parameters, connection, __LINE__);

    /* Transfer a large file (greater than 1 MB and test lot allocation
     * Should hopefully fail
     */
    printf("Transfer of large file greater than 1 MB should fail now\n");
    test_send_large_file( parameters, connection, __LINE__, true );

    /* Now reserve another lot of 1 MB and again list the Lots */
     size_in_MB = 1;
    get_lot(connection, &handle2, size_in_MB, minutes, __LINE__);
    list_user_lots(connection, __LINE__);

    /* Now transfer the large file again and test the lot allocation */
    test_send_large_file( parameters, connection, __LINE__);

    return_lot(connection, &handle1, __LINE__);
    return_lot(connection, &handle2, __LINE__);

    /* Now reserve lots for short duration and check whether lot
     * expiration is handled correctly
     */

    avail_size = get_avail_nest_space(connection, __LINE__);

    size_in_MB = 5;
    minutes = 1;
    printf("OK: Trying to get a lot of 5 MB for 1 minute duration\n");
    get_lot(connection, &handle1, size_in_MB, minutes, __LINE__);
    list_user_lots(connection, __LINE__);
    test_send_large_file( parameters, connection, __LINE__);

    printf("OK: Waiting for lot to expire (about 2 minutes)...\n");
    /* Now Wait for the lot to expire and test sending file again
     * Should fail now
     */
    sleep(120);
    list_user_lots(connection, __LINE__);
    test_send_large_file( parameters, connection, __LINE__);

    /* Now try to get all the NeST space so that recovery of expired
     * lots is correctly handled
     */

    size_in_MB = avail_size;
    minutes = 1;
    printf("OK: Trying to get a lot of %f MB for 1 minute duration. "
           " This should clean up the expired lots.\n", size_in_MB );
    get_lot(connection, &handle2, size_in_MB, minutes, __LINE__);
    status1 = NestGetLotStat(&nest_lot, connection, &handle1);
    status2 = NestReceiveFile("rpm", "rpm", connection);
    if( (status1 == NEST_LOT_NOT_FOUND ||
         status1 == NEST_NONPRIVELEDGED_USER) &&
         status2 == NEST_FILE_NOT_FOUND )
        printf("OK: Expired Lot was successfully cleaned.\n");
    else{
        printf("FAILED: Expired Lot was not successfully cleaned.\n");
        number_of_errors++;
    }

    list_user_lots(connection, __LINE__);

    return_lot(connection, &handle2, __LINE__);
    if( !(status1 == NEST_LOT_NOT_FOUND ||
          status1 == NEST_NONPRIVELEDGED_USER) )
        return_lot(connection, &handle1, __LINE__);

    return;
}

/******************************************************************************
 *
 * Function: check_lots_enabled
 * Purpose: Checks if the lots are enabled at the server
 *
 *
 *****************************************************************************/
static bool
check_lots_enabled(
    NestConnection connection)
{
    NestLotList     *list;
    NestReplyStatus status;

    status = NestListUserLots(&list, connection);

    if( status == NEST_LOT_DISABLED ){
        printf("OK: Lots are disabled.\n");
        return false;
    }
    else{
        printf("OK: Lots are enabled. \n");
        return true;
    }

}



/******************************************************************************
 *
 * Function: Reserve_Lot
 * Purpose:  Reserves lot of specified size and duration and
 *             prints out suitable diagnostic messages
 *
 *****************************************************************************/
bool
get_lot(
    NestConnection connection,
    NestLotHandle  *handle,
    double         size_in_MB,
    double         minutes,
    int            line)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestRequestUserLot(handle, connection, size_in_MB, minutes);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestRequestUserLot (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Reserved lot of size %f\n", size_in_MB);
        succeeded = true;
    }
    return succeeded;
}


/******************************************************************************
 *
 * Funtion: return_lot
 * Purpose: Terminates the specified lot and prints out suitable
 *            diagnostic messages
 *
 *****************************************************************************/
bool
return_lot(
    NestConnection connection,
    NestLotHandle  *handle,
    int            line)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestTerminateLot(connection, handle);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestTerminateLot (%s) in line %d\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Terminated lot\n");
        succeeded = true;
    }

    return succeeded;
}



/******************************************************************************
 *
 * Funtion: return_all_lots
 * Purpose: Terminates all the lots owned by the current user
 *
 *
 *****************************************************************************/
static bool
return_all_lots(
    NestConnection connection,
    int            line)
{
    NestLotList     *list;
    NestReplyStatus status;
    NestLotHandle   handle;
    bool            succeeded;

    status = NestListUserLots( &list, connection );

    if ( status != NEST_SUCCESS && status != NEST_LOT_NOT_FOUND ) {
        printf("FAILED: NestListUserLots (%s) in line %d\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        if(status != NEST_LOT_NOT_FOUND){
            while ( list !=NULL ){
                NestString_to_handle( &handle, list->stat.lot_handle );
                return_lot( connection, &handle, __LINE__);
                list = list->next;
            }
        }
        succeeded = true;
    }
    return succeeded;
}


/******************************************************************************
 *
 * Function: list_user_lots
 * Purpose : Lists down all the lots held by the current user.
 *           Returns true if lots are enabled else returns false
 *
 *****************************************************************************/
static bool
list_user_lots(
    NestConnection connection,
    int            line)
{
    NestLotList     *list;
    NestReplyStatus status;
    bool            succeeded;

    status = NestListUserLots( &list, connection );
    if ( status != NEST_SUCCESS ) {
            printf("FAILED: NestListUserLots (%s) in line %d\n",
                        NestErrorString(status), line);
            number_of_errors++;
            succeeded = false;
    } else {
        printf("OK:\tLot Handle\tSize\t\tSpace-used\tTime Left\n");
        while ( list !=NULL ){
            printf("   \t%s\t\t%f\t%f\t%f\n",
                   list->stat.lot_handle,
                   list->stat.size,
                   list->stat.used,
                   list->stat.time_left );

            list = list->next;
        }
        succeeded = true;
    }
    return succeeded;

}


/******************************************************************************
 *
 * Function: test_misc
 * Purpose:  Does a bunch of miscellaneous NeST API call tests
 *           If the all_cases flag is set then all tests irrespective of
 *           whether the function is bug free/implemented will be tested.
 *
 *****************************************************************************/
static void
test_misc(
    struct parameters *parameters,
    NestConnection    connection,
    bool              all_cases)
{
    sync_nest(connection, __LINE__);
    get_nest_server_time(connection, __LINE__);
    test_get_physical_name(parameters, connection);
    test_links(parameters, connection, all_cases);
}

/******************************************************************************
 *
 * Function: sync_nest
 * Purpose:  Checks the sync call
 *
 *
 *****************************************************************************/
static bool
sync_nest(
    NestConnection connection,
    int            line)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestSync(connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestSync (%s) in line %d\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Synchronized data on disk with "
               "memory files succesfully.\n");
        succeeded = true;
    }
    return succeeded;
}


/******************************************************************************
 *
 * Function: get_nest_server_time
 * Purpose:  Prints out the nest server timestamp and returns the timestamp
 *           value
 *
 *****************************************************************************/
static bool
get_nest_server_time(
    NestConnection connection,
    int line)
{
    NestReplyStatus status;
    long            timestamp;
    char            formatted_time[MAXPATHLEN];
    bool            succeeded = true;

    status = NestGetTimestamp(timestamp, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestTimestamp (%s) in line %d\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        strftime( formatted_time, MAXPATHLEN, "%A %B %d %H:%M:%S %Z %G",
                  localtime(&timestamp) );
        printf("OK: Time at the nest server: %s.\n", formatted_time);
    }

    return succeeded;

}


/******************************************************************************
 *
 * Function: test_get_physical_name
 * Purpose:  Checks whether NeST returns the correct physical file names of
 *           the logical NeST filenames.
 *           This function uses the file /usr/include/stdio.h as the test
 *           file. The logical name of the file is ~system/usr/include/stdio.h
 *
*****************************************************************************/
static void
test_get_physical_name(
    struct parameters *parameters,
    NestConnection    connection)
{
    char            *current_dir,
                    *logical_test_dir,
                    *physical_test_file,
                    *test_file,
                    *physicalname;

    logical_test_dir        = strdup("~system/usr/include/");
    physical_test_file      = strdup("/usr/include/stdio.h");
    test_file               = strdup("stdio.h");

    if( get_directory(&current_dir, parameters, connection, __LINE__ ) ){
        change_directory( logical_test_dir, parameters, connection, __LINE__);
        get_physical_name(connection, &physicalname, test_file, __LINE__);
        if( !strcmp(physicalname, physical_test_file) ){
            printf("OK: NestGetPhysicalName returned the correct physical "
                   "name %s of the file %s.\n", physicalname, test_file);
        } else{
            printf("FAILED: NestGetPhysicalName returned the wrong physical "
                   "name %s of the file %s.\n", physicalname, test_file);
            number_of_errors++;
        }
        change_directory( current_dir, parameters, connection, __LINE__ );
    } else{
        printf("FAILED: Could not get the physical name of "
               "the current working directory.\n");
        number_of_errors++;
    }

}

/******************************************************************************
 *
 * Function: get_physical_name
 * Purpose:  Prints out the physical file name of the given logical file name
 *           Added this function anticipating its reuse in other tests
 *
 *****************************************************************************/
static bool
get_physical_name(
    NestConnection connection,
    char           **physicalname,
    char           *logicalname,
    int            line)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestGetPhysicalName(physicalname, logicalname, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestGetPhysicalName (%s) in line %d\n",
                    NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Phyical name of file %s : %s.\n",
                    logicalname, *physicalname );
        succeeded = true;
    }

    return succeeded;
}


/******************************************************************************
 *
 * Function: test_links
 * Purpose:  Checks the correct functioning of create link call. Creates links
 *           from nest fs to system fs and vice versa.
 *           If all_cases flag is set, all combinations are checked; else
 *           the link from ~system fs to NeST fs is not checked as it is not
 *           implemented yet.
 *
 *****************************************************************************/
static void
test_links(
    struct parameters *parameters,
    NestConnection    connection,
    bool              all_cases)
{

    NestLotHandle   handle;
    NestReplyStatus status;
    double          size_in_MB,
                    minutes;
    char            *system_file,
                    *system_link,
                    *nest_file,
                    *nest_link,
                    *nest_link_a,
                    *nest_link_b;
    bool            lot_used;

    system_file = strdup("~system/tmp/system_file");
    system_link = strdup("~system/tmp/link");
    nest_file   = strdup("nest_file");
    nest_link   = strdup("nest_link");
    nest_link_a = strdup("nest_link_a");
    nest_link_b = strdup("nest_link_b");


    // Create a file locally and put it to the remote nest fs and ~system fs
    make_test_file("blah");
    if( check_lots_enabled(connection) ){
        size_in_MB = 1;
        minutes = 10;
        if( get_lot(connection, &handle, size_in_MB, minutes, __LINE__) )
            lot_used = true;
        else
            lot_used = false;
    } else
        lot_used = false;
    put_file("blah", nest_file, connection, __LINE__);
    put_file("blah", system_file, connection, __LINE__);

    // Checks for NeST fs ---> NeST fs links
    if( create_link( nest_file, nest_link, connection, __LINE__) )
        remove_file(nest_link, connection, __LINE__);

    // Checks for NeSt fs ---> ~system fs links
    if( create_link(system_file, nest_link, connection, __LINE__ ) )
        remove_file(nest_link, connection, __LINE__);

    // Checks for ~system fs ---> ~system fs links
    if( create_link(system_file, system_link, connection, __LINE__) )
        remove_file(system_link, connection, __LINE__);

    // Checks for ~system fs ---> NeST fs links
    if(all_cases)
        if( create_link(nest_file, system_link, connection, __LINE__) )
            remove_file(system_link, connection, __LINE__);

    remove_file(nest_file, connection, __LINE__);
    remove_file(system_file, connection, __LINE__);

    // Check for cyclic link detection
    put_file("blah", nest_link_a, connection, __LINE__);
    create_link(nest_link_a, nest_link_b, connection, __LINE__);
    remove_file(nest_link_a, connection, __LINE__);
    create_link(nest_link_b, nest_link_a, connection, __LINE__);
    if( get_file(nest_link_a, "new_blah", &status, connection, __LINE__) ){
        if (status != NEST_TOO_MANY_LINKS) {
            printf("FAILED: NestLink (%s) in line %d\n",
                    NestErrorString(status), __LINE__);
            printf("FAILED: Cyclic links are not handled correctly.\n");
            number_of_errors++;
        } else
            printf("OK: Cyclic links are handled correctly.\n");
    }

    remove_file(nest_link_a, connection, __LINE__);
    remove_file(nest_link_b, connection, __LINE__);
    if(lot_used)
        return_lot(connection, &handle, __LINE__);
    unlink("new_blah");
    unlink("blah");

}


/******************************************************************************
 *
 * Function: create_link
 * Purpose:  Creates a link to the specified file
 *
 *****************************************************************************/
static bool
create_link(
    char           *oldPath,
    char           *newPath,
    NestConnection connection,
    int            line)
{
    NestReplyStatus status;
    // This flag will be made a parameter when hard links are implemented in
    // NeST
    int             isSymbolic = true;
    int             succeeded;


    status = NestLink(oldPath, newPath, isSymbolic, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestLink (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Link %s created pointing to %s.\n", newPath, oldPath);
        succeeded = true;
    }

    return succeeded;
}





/******************************************************************************
 *
 * Function: test_user_mgmt
 * Purpose:  Tests the user management APIs
 *
 *****************************************************************************/
static void
test_user_mgmt(
    struct parameters *parameters,
    NestConnection    connection)
{
    char    *user1,
            *user2,
            *group1,
            *group2;
    bool    is_admin;

    user1  = strdup("nest_user1");
    user2  = strdup("nest_user2");
    group1 = strdup("nest_group1");
    group2 = strdup("nest_group2");


    /************************************************************************
     *   Does a buch of operations with the user management APIs
     *   Certain conditions which should fail are checked to make sure that
     *   they fail; but the exact error returned on failure is not checked.
     *   This should be done to make sure that the failure is because of
     *   the expected reason and not something else
     ************************************************************************/

    is_admin = check_admin_user(connection);

    clean_test_users(connection);
    if(is_admin){
        // Add group1
        add_group(connection, group1, __LINE__);
        // Add user1 to group1
        add_user(connection, user1, group1, __LINE__);


        // This call should fail as group1 is added again
        add_group(connection, group1, __LINE__, true);
        // This call should fail as user1 is added again to group1
        // This is done thro user mgmt calls
        add_user(connection, user1, group1, __LINE__, true);
        // This call should fail as user1 is added again to group1
        // This is done thro group mgmt calls
        add_user_to_group(connection, user1, group1, __LINE__, true);
        // This call should fail as user1 is added again.
        // Even though user1 is added to group2, this should be done using
        // add_user_to_group and not using add_user
        add_user(connection, user1, group2, __LINE__, true);
        // This call should fail as a non existing user is added to a group
        add_user_to_group(connection, user2, group1, __LINE__, true);


        // Add group2
        add_group(connection, group2, __LINE__);
        // Add user2 to group1
        add_user(connection, user2, group1, __LINE__);
        // Add user2 to group2
        add_user_to_group(connection, user2, group2, __LINE__);
        // Add user1 to group2
        add_user_to_group(connection, user1, group2, __LINE__);
        // Delete user1. Implicitly deletes user1 from group1 and group2
        del_user(connection, user1, __LINE__);


        // Should fail as user1 is already deleted
        del_user(connection, user1, __LINE__, true);


        // Delete user2 from group1
        del_user_from_group(connection, user2, group1, __LINE__);
        // Delete group1
        del_group(connection, group1, __LINE__);


        // Should fail as user2 is still in group2
        del_group(connection, group2, __LINE__, true);


        // Delete user2 from group2
        del_user_from_group(connection, user2, group2, __LINE__);
        // Delete group2
        del_group(connection, group2, __LINE__);
        // Delete user2
        del_user(connection, user2, __LINE__);


        //Just another round of add group/user and delete group/user
        add_group(connection, group1, __LINE__);
        add_user(connection, user1, group1, __LINE__);
        del_user(connection, user1, __LINE__);
        del_group(connection, group1, __LINE__);

        // These calls should fail as the user "nest_everybody" and 
		// "nest_no_group" are internal to NeST and should not be deleted
        del_user(connection, "nest_everybody", __LINE__, true);
        del_group(connection, "no_nest_group", __LINE__, true);
        // These calls should fail as no such user or group exists
        del_user(connection, "blah", __LINE__, true);
        del_group(connection, "blah", __LINE__, true);

        add_user_to_group(connection, "nest_everybody", "no_nest_group",
                          __LINE__, true);
        del_user_from_group(connection, "nest_everybody", "no_nest_group",
                            __LINE__);
    } else {
        printf("OK: Should be admin user to run user management tests.\n");
        // Add group1
        add_group(connection, group1, __LINE__);
        add_user_to_group(connection, "nest_everybody", group1,
                          __LINE__);


        // This call should fail as group1 is added again
        add_group(connection, group1, __LINE__, true);
        // This call should fail as a non existing user is added to a group
        add_user_to_group(connection, user2, group1, __LINE__, true);


        // Add group2
        add_group(connection, group2, __LINE__);

        // Should fail as nest_everybody is still in group1
        del_group(connection, group1, __LINE__, true);


        // Delete nest_everybody from group1
        del_user_from_group(connection, "nest_everybody", group1, __LINE__);


        // Delete group1
        del_group(connection, group1, __LINE__);
        // Delete group2
        del_group(connection, group2, __LINE__);


    }

}

static void
clean_test_users(
    NestConnection connection
)
{
    NestDelUser(connection, "nest_user1");
    NestDelUser(connection, "nest_user2");
    NestDelGroup(connection, "nest_group1");
    NestDelGroup(connection, "nest_group2");
}

/******************************************************************************
 *
 * Function: check_admin_user
 * Purpose:  Check if the user runnning the tests is admin user
 *
 *****************************************************************************/
static bool
check_admin_user(
    NestConnection connection)
{
    NestReplyStatus status;
    bool            is_admin;

    status = NestAddUser(connection, "test_user", "no_nest_group", "");
    if(status == NEST_SUCCESS){
        is_admin = true;
        NestDelUser(connection, "test_user");
    }
    else
        is_admin = false;

    return is_admin;
}


/******************************************************************************
 *
 * Function: add_group
 * Purpose: Adds a new group
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
add_group(
    NestConnection connection,
    char           *group,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;


    status = NestAddGroup(connection, group);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestAddGroup did not fail as expected in line %d.\n",
                line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestAddGroup (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestAddGroup (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestAddGroup succeeded in line %d.\n",
                    line);
        printf("OK: Created the group %s.\n", group);
        succeeded = true;
    }

    return succeeded;

}


/******************************************************************************
 *
 * Function: del_group
 * Purpose: Deletes a group
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
del_group(
    NestConnection connection,
    char           *group,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestDelGroup(connection, group);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestDelGroup did not fail as expected in line %d.\n",
                line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestDelGroup (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestDelGroup (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestDelGroup succeeded in line %d.\n",
                    line);
        printf("OK: Deleted the group %s.\n", group);
        succeeded = true;
    }

    return succeeded;

}



/******************************************************************************
 *
 * Function: add_user_to_group
 * Purpose: Adds an existing user to a group
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
add_user_to_group(
    NestConnection connection,
    char           *user,
    char           *group,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestAddUserToGroup(connection, user, group);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestAddUserToGroup did not fail as expected "
               "in line %d.\n", line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestAddUserToGroup (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestAddUserToGroup (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestAddUserToGroup succeeded in line %d.\n",
                    line);
        printf("OK: Added User %s to group %s.\n", user, group);
        succeeded = true;
    }

    return succeeded;


}


/******************************************************************************
 *
 * Function: del_user_from_group
 * Purpose: Deletes a user from a group
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
del_user_from_group(
    NestConnection connection,
    char           *user,
    char           *group,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestDelUserFromGroup(connection, user, group);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestDelFromGroup did not fail as expected "
               "in line %d.\n", line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestDelFromGroup (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestDelFromGroup (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestDelFromGroup succeeded in line %d.\n",
                    line);
        printf("OK: Deleted User %s from group %s.\n", user, group);
        succeeded = true;
    }

    return succeeded;
}


/******************************************************************************
 *
 * Function: add_user
 * Purpose: Adds an new user to a group
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
add_user(
    NestConnection  connection,
    char            *user,
    char            *group,
    int             line,
    bool            should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestAddUser(connection, user, group, "");
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestAddUser did not fail as expected "
               "in line %d.\n", line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestAddUser (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestAddUser (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestAddUser succeeded in line %d.\n",
                    line);
        printf("OK: Added User %s to group %s.\n", user, group);
        succeeded = true;
    }

    return succeeded;

}


/******************************************************************************
 *
 * Function: del_user
 * Purpose: Deletes an existing user
 *          If should_fail flag is set this function is supposed to fail due
 *          to the conditions created in the testing process
 *
 *****************************************************************************/
static bool
del_user(
    NestConnection connection,
    char           *user,
    int            line,
    bool           should_fail)
{
    NestReplyStatus status;
    bool            succeeded;

    status = NestDelUser(connection, user);
    if(should_fail && status == NEST_SUCCESS){
        printf("FAILED: NestDelUser did not fail as expected "
               "in line %d.\n", line);
        number_of_errors++;
        succeeded = false;
    } else if (!should_fail && status != NEST_SUCCESS) {
        printf("FAILED: NestDelUser (%s) in line %d.\n",
                NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else if (should_fail && status != NEST_SUCCESS){
        printf("OK: NestDelUser (%s) failed as expected in line %d.\n",
                NestErrorString(status), line);
        succeeded = true;
    } else if (!should_fail && status == NEST_SUCCESS){
        printf("OK: NestDelUser succeeded in line %d.\n",
                    line);
        printf("OK: Deleted User %s.\n", user);
        succeeded = true;
    }

    return succeeded;
}

/******************************************************************************
 *
 * Function: test_robustness
 * Purpose: Sends garbled packets to nest using different protocols.
 *          These tests are used to check the robustness of nest server.
 *          If loop is set packets are blasted till the nest server fails.
 *
 *****************************************************************************/
static void
test_robustness(
    struct parameters *parameters,
    NestConnection    connection,
    bool              loop)
{
    bool succeeded = true;

    // After each of the fuzz tests the nest server is queried to find its time
    // stamp to make sure that the server is still functioning correctly

    do{
        if(succeeded){
            test_protocol_robustness(parameters, connection, "chirp");
            succeeded = succeeded && check_nest_sanity(parameters, __LINE__);
        }
        if(succeeded){
            test_protocol_robustness(parameters, connection, "http");
            succeeded = succeeded && check_nest_sanity(parameters, __LINE__);
        }
        if(succeeded){
            test_protocol_robustness(parameters, connection, "ftp");
            succeeded = succeeded && check_nest_sanity(parameters, __LINE__);
        }
        if(succeeded){
            test_protocol_robustness(parameters, connection, "gftp");
            succeeded = succeeded && check_nest_sanity(parameters, __LINE__);
        }
    }while(succeeded && loop);
}

/******************************************************************************
 *
 * Function: test_protocol_robustness
 * Purpose: Opens a connection to the nest server using the specifies protocol
 *          and sends packets with random data till send fails.
 *
 *****************************************************************************/
static bool
test_protocol_robustness(
    struct parameters *parameters,
    NestConnection    connection,
    char              *protocol )
{
    int                 sockfd;
    int                 portnumber;
    struct hostent      *nest_hostent;
    struct sockaddr_in  nest_serv_addr;
    char                buffer[MAXPATHLEN];
    bool                succeeded = true;

    nest_hostent = gethostbyname(parameters->server);
    if( !strcmp(protocol, "chirp") )
        portnumber = atoi(parameters->chirp_port);
    else if( !strcmp(protocol, "http") )
        portnumber = atoi(parameters->http_port);
    else if( !strcmp(protocol, "ftp") )
        portnumber = atoi(parameters->ftp_port);
    else if( !strcmp(protocol, "gftp") )
        portnumber = atoi(parameters->gftp_port);

    memset( &nest_serv_addr, 0, sizeof( sockaddr_in ) );
    nest_serv_addr.sin_family = AF_INET;
    memcpy( &nest_serv_addr.sin_addr, nest_hostent->h_addr,
            nest_hostent->h_length );
    nest_serv_addr.sin_port = htons(portnumber);

    if( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
        printf("FAILED: Socket resource is unavailable.\n");
        number_of_errors++;
        succeeded = false;
    } else{
        printf("OK: Socket created.\n");
    }

    if( connect(sockfd, (struct sockaddr*) &nest_serv_addr,
                sizeof(nest_serv_addr))< 0){
        printf("FAILED: Unable to connect to the nest server.\n");
        number_of_errors++;
        succeeded = false;
    } else{
        printf("OK: Connected to the nest server using %s port.\n", protocol);
    }

    int count = 0;
    while(1){
        fill_buffer(buffer, MAXPATHLEN);

        // send sometimes fails because either the server resets the connection
        // or the send queue fills up.
        // the total number of packets send upto that point is printed out.
        if( send(sockfd, buffer, MAXPATHLEN,
                 MSG_DONTWAIT|MSG_OOB|MSG_NOSIGNAL) > 0){
            count++;
        } else {
            break;
        }
    }

    printf("OK: Sent %d garbled packets to nest server.\n", count);
    if( close(sockfd) ){
        printf("FAILED: Unable to close the %s connection to nest.\n",
                protocol);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: Closed the %s connection to nest server.\n", protocol);
    }

    return succeeded;

}


/******************************************************************************
 *
 * Function: fill_buffer
 * Purpose: Fills the given buffer with random data.
 *
 *****************************************************************************/
static void
fill_buffer(
    char *buffer,
    int  len)
{
    int count = 0;

    srandom(time(NULL));
    while(count < len)
        buffer[count++] = random();
}


/******************************************************************************
 *
 * Function: copy_file
 * Purpose: Copies the specified oldfile to the newfile
 *
 *****************************************************************************/
static bool 
copy_file(
    char *oldPath,
    char *newPath,
    NestConnection connection,
    int line)
{
    NestReplyStatus status;
    int succeeded;

    status = NestCopy(oldPath, newPath, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestCopy (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: File %s copied to file %s.\n", oldPath, newPath);
        succeeded = true;
    }
     
    return succeeded;
}

/******************************************************************************
 *
 * Function: move_file
 * Purpose: Moves the specified oldfile to the newfile
 *
 *****************************************************************************/
static bool 
move_file(
    char *oldPath,
    char *newPath,
    NestConnection connection,
    int line)
{
    NestReplyStatus status;
    int succeeded;

    status = NestRenameFile(oldPath, newPath, connection);
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestMove (%s) in line %d\n",
               NestErrorString(status), line);
        number_of_errors++;
        succeeded = false;
    } else {
        printf("OK: File %s moved to file %s.\n", oldPath, newPath);
        succeeded = true;
    }
     
    return succeeded;
}

/******************************************************************************
 *
 * Function: test_file_operations
 * Purpose:  Checks the correct functioning of basic file operations - copy,
 *           move.
 *           If all_cases flag is set, copy and move across ~system fs and 
 *           NeST fs is checked; else these operations involving ~system fs 
 *           is not checked as it is not yet implemented.
 *
 *****************************************************************************/
static void
test_file_operations(
    struct parameters *parameters,
    NestConnection    connection,
    bool              all_cases)
{

    NestLotHandle   handle;
    double          size_in_MB,
                    minutes;
    char            *system_file1,
                    *system_file2,
                    *nest_file1,
                    *nest_file2;
    bool            lot_used;

    system_file1 = strdup("~system/tmp/system_file1");
    system_file2 = strdup("~system/tmp/system_file2");
    nest_file1   = strdup("nest_file1");
    nest_file2   = strdup("nest_file2");

    // Create a file locally and put it to the remote nest fs and ~system fs
    make_test_file("blah");
    if( check_lots_enabled(connection) ){
        size_in_MB = 1;
        minutes = 10;
        if( get_lot(connection, &handle, size_in_MB, minutes, __LINE__) )
            lot_used = true;
        else
            lot_used = false;
    } else
        lot_used = false;
    put_file("blah", nest_file1, connection, __LINE__);
    put_file("blah", system_file1, connection, __LINE__);

    //Check copy operations
    
    //NeST fs ---> NeST fs copy
    if( copy_file(nest_file1, nest_file2, connection, __LINE__) )
        remove_file(nest_file2, connection, __LINE__);

    if(all_cases){
        //NeST fs ---> ~system fs copy
        if( copy_file(nest_file1, system_file2, connection, __LINE__) )
            remove_file(system_file2, connection, __LINE__);
        
        //~system fs ---> ~system fs copy
        if( copy_file(system_file1, system_file2, connection, __LINE__) )
            remove_file(system_file2, connection, __LINE__);

        //~system fs ---> NeST fs copy
        if( copy_file(system_file1, nest_file2, connection, __LINE__) )
            remove_file(nest_file2, connection, __LINE__);
    }

    //Check move operations

    //NeST fs ---> NeST fs move
    if( move_file(nest_file1, nest_file2, connection, __LINE__) )
        move_file(nest_file2, nest_file1, connection, __LINE__);

    if(all_cases){
        //NeST fs ---> ~system fs move
        if( move_file(nest_file1, system_file2, connection, __LINE__) )
            move_file(system_file2, nest_file1, connection, __LINE__);
        
        //~system fs ---> ~system fs move
        if( move_file(system_file1, system_file2, connection, __LINE__) )
            move_file(system_file2, system_file1, connection, __LINE__);

        //~system fs ---> NeST fs move
        if( move_file(system_file1, nest_file2, connection, __LINE__) )
            move_file(nest_file2, system_file1, connection, __LINE__);
    }

    remove_file(nest_file1, connection, __LINE__);
    remove_file(system_file1, connection, __LINE__);

    if(lot_used)
        return_lot(connection, &handle, __LINE__);
    unlink("blah");

}
