/******************************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <pthread.h>
#include <semaphore.h>

#include "nest_speak.h"

/* For platforms that don't have this bit, ignore it. */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

/*-----------------------------------------------------------------------------
 *
 * Private data structures
 *
 *---------------------------------------------------------------------------*/
typedef struct
{
	pthread_t		Thread;
	int				ThreadId;
	class NestDrm	*Drm;
} DrmThreadInfo;

class Thread
{
public:
	Thread( class NestDrm *drm, int id );
	~Thread( void );
	int	getId( void ) { return Info.ThreadId; };

private:
	DrmThreadInfo	Info;
};

class NestDrm
{
public:
	NestDrm( const char *server );
	~NestDrm( void );

	int Open( void );
	int RunTests( void );
	int Close( void );

	void SetLotSize( int size ) { master_lot_mb = size; };
	void SetLotTime( int time ) { master_lot_time = time; };
	void SetChildLotSize( int size ) { child_lot_mb = size; };
	void SetNumThreads( int num ) { num_threads = num; };
	NestLotHandle* GetLot( void ) { return &master_lot; };
	void RunThread( int id );

// Internal functions
private:
	int CreateParentLot( void );
	int DestroyParentLot( void );
	int NestDrm::CreateChildLot( int size );

	int SemInit( void );
	int SemWait( void );
	int SemPost( void );

// Private data
private:
    NestConnection	connection;
	const char		*server;
	NestLotHandle	master_lot;
	bool			master_lot_valid;
	int				master_lot_mb;
	int				master_lot_time;
	int				child_lot_mb;
	int				num_threads;
	sem_t			semaphore;
	
};


/*-----------------------------------------------------------------------------
 *
 * Private global variables
 *
 *---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------------
 *
 * Private functions
 *
 *---------------------------------------------------------------------------*/


int
main(int argc, char *argv[])
{
	NestDrm		*drm;

	if ( argc < 2 )
	{
		fprintf( stderr, "usage: %s server\n", argv[0] );
		exit( 1 );
	}
	char *server = argv[1];

    // Connect to the Nest server
	drm = new NestDrm( server );

	// Configure the tests
	drm->SetLotSize( 2000 );
	drm->SetLotTime( 30 );
	drm->SetNumThreads( 10 );
	drm->SetChildLotSize( 125 );
	drm->Open( );
	drm->RunTests( );
	drm->Close( );

	delete drm;

	return 0;
}


NestDrm::NestDrm( const char *server )
{
	if ( ! server ) {
		fprintf( stderr, "No server specified!\n" );
		exit( 1 );
	}
	char *tmp = strdup( server );
	this->server = tmp;
	connection = -1;
	master_lot_valid = false;
}

NestDrm::~NestDrm(void)
{
	Close( );
	free( (void*)server );
}

int NestDrm::Open( void )
{
	NestReplyStatus status;

    printf("Connecting to %s...\n\n", server);
    status = NestOpenConnection(connection, server);

    if(status != NEST_SUCCESS) {
        printf("FAILED: Couldn't connect to %s: %s in line %d\n",
                server, NestErrorString(status), __LINE__);
		exit( 1 );
	}

	printf("OK: Connected to %s in line %d.\n", server, __LINE__);

	return 0;
}

int NestDrm::Close( void )
{
	NestReplyStatus status;

	if ( connection < 0 ) {
		return 0;
	}

	// Cleanup
	(void) DestroyParentLot( );

	// Detroy our connection
	status = NestCloseConnection(connection);

	if(status != NEST_SUCCESS) {
		printf("FAILED: Couldn't close connection to %s: %s in line %d\n",
			   server, NestErrorString(status), __LINE__);
	} else {
		printf("OK: Connection closed in line %d.\n", __LINE__);
	}

	connection = -1;
	return 0;
}

int NestDrm::RunTests( void )
{
	// Create a "parent" lot to work from
	if ( CreateParentLot( ) ) {
		return -1;
	}

	// Threading logic
	int			threadno;
	Thread		*threads[num_threads];

	if ( SemInit( ) ) {
		printf( "Semaphore init failed!\n" );
		return -1;
	}

	// Create the threads, let them work
	for( threadno = 0;  threadno < num_threads;  threadno++ ) {
		threads[threadno] = new Thread( this, threadno );
	}

	// Wait for the threads to finish
	for( threadno = 0;  threadno < num_threads;  threadno++ ) {
		delete( threads[threadno] );
	}
	char	buf[10];
	printf( "Press any key: " ); fgets( buf, sizeof( buf ), stdin );

	// Cleanup
	if ( DestroyParentLot( ) ) {
		return -1;
	}
	return 0;
}

int NestDrm::CreateParentLot( void )
{
	NestReplyStatus	status;
    status = NestRequestUserLot( &master_lot,
								 connection,
								 master_lot_mb,
								 master_lot_time );

	if (status != NEST_SUCCESS) {
		printf("FAILED: Couldn't create master lot: %s in line %d\n",
			   NestErrorString(status), __LINE__);
		return -1;
	} else {
		printf("OK: Created master lot (%d M) in line %d.\n", master_lot_mb, __LINE__);
	}

	const char *user =
		"_C_US_O_Globus_O_University_of_Wisconsin_OU_Computer_Sciences_Department_CN_Nick_LeRoy";
	status = NestAddUserToLot( connection,
							   &master_lot,
							   user );
	if (status != NEST_SUCCESS) {
		printf("FAILED: Couldn't add user '%s' to master lot: %s in line %d\n",
			   user, NestErrorString(status), __LINE__);
		return -1;
	} else {
		printf("OK: Add user to master lot in line %d.\n", __LINE__);
	}

	master_lot_valid = true;
	return 0;
	
}

int NestDrm::DestroyParentLot( void )
{
	NestReplyStatus	status;

	if ( ! master_lot_valid ) {
		return 0;
	}
    status = NestTerminateLot( connection, &master_lot );
    if (status != NEST_SUCCESS) {
        printf("FAILED: NestTerminateLot (%s) in line %d\n",
                    NestErrorString(status), __LINE__ );
		return -1;
    } else {
        printf("OK: Terminated lot\n");
    }
	master_lot_valid = false;

	return 0;
	
}

int NestDrm::CreateChildLot( int size )
{
	NestReplyStatus	status;
	NestLotHandle	lot;

	if ( ! master_lot_valid ) {
		printf( "Master lot invalid!\n" );
		return -1;
	}
	if ( SemWait( ) ) {
		printf( "Failed waiting on semaphore!\n" );
		return -1;
	}
	status = NestRequestUserLot( &lot,
								 connection,
								 size,
								 1,
								 NULL,
								 NULL,
								 NULL,
								 &master_lot,
								 G_LOT );

	(void) SemPost( );
	if (status != NEST_SUCCESS) {
		printf("FAILED: Couldn't create child lot: %s in line %d\n",
			   NestErrorString(status), __LINE__);
		return -1;
	} else {
		printf("OK: Created child lot (%d M) in line %d = %lu\n", size, __LINE__,
			   lot.rsv_id);
	}
	return 0;
	
}

int NestDrm::SemInit( void )
{
	int	status = sem_init( &semaphore, 0, 1 );
	if ( status != 0 ) {
		printf( "FAILED: sem_init returned %d; errno %d\n", status, errno );
	} else {
		printf( "ok: sem_init success\n" );
	}

	return status;
}

int NestDrm::SemWait( void )
{
	int	status = sem_wait( &semaphore );
	if ( status != 0 ) {
		printf( "FAILED: sem_wait returned %d; errno %d\n", status, errno );
	} else {
		//printf( "ok: sem_wait success\n" );
	}

	return status;
}

int NestDrm::SemPost( void )
{
	int	status = sem_post( &semaphore );
	if ( status != 0 ) {
		printf( "FAILED: sem_post returned %d; errno %d\n", status, errno );
	} else {
		//printf( "ok: sem_post success\n" );
	}

	return status;
}

void NestDrm::RunThread( int id )
{
	//printf( "C++: Thead %d starting\n", id );
	CreateChildLot( child_lot_mb );

	// Copy file to the NeST
	char	srcUrl[256];
	char	dstUrl[256];
	char	wd[256];
	char	cmd[512];
	getcwd( wd, sizeof (wd ) );
	sprintf( srcUrl, "file:%s/%s", wd, "8m" );
	sprintf( dstUrl, "gsiftp://%s/%03d.dat", server, id );
	sprintf( cmd, "globus-url-copy %s %s", srcUrl, dstUrl );
	printf( "%03d: %s\n", id, cmd );
	system( cmd );

	//printf( "C++: Thead %d Done\n", id );
};

// C helper function
void* DrmStartThread( void * arg )
{
	DrmThreadInfo *Info = (DrmThreadInfo *) arg;
	Info->Drm->RunThread( Info->ThreadId );
	return NULL;
}

Thread::Thread( NestDrm *drm, int id )
{
	Info.ThreadId = id;
	Info.Drm = drm;

	int		status;

	status = pthread_create( &(Info.Thread), NULL, DrmStartThread, &Info );
	if ( status ) {
		printf( "Thread creation failed: %d, %s\n", status, strerror( status ) );
	}
}

Thread::~Thread( void )
{
	int		status;

	status = pthread_join( Info.Thread, NULL );
	if ( status ) {
		printf( "Thread join failed: %d, %s\n", status, strerror( status ) );
	}
}
